/*
 * IdentifierManager.java
 * Copyright 2004, Gerald R Duprey Jr
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy of
 * the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0 
 *
 * Unless required by applicable law or agreed to in writing, software distributed
 * under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
 * CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License. 
 */

package org.cdp1802.xpl;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import org.cdp1802.xpl.ethernet.BindableAddress;


/**
 * Handles creation, lookup and management of xPL identifiers
 *
 * @author Gerald Duprey Jr
 */
public class IdentifierManager {
  private static IdentifierManager identiferManager = null;
  
  /**
   * Return the instance of the IdentifierManager
   * 
   * @return instance of the IdentifierManager to use
   */
  static IdentifierManager getManager() {
    if (identiferManager == null) identiferManager = new IdentifierManager();
    return identiferManager;
  }
  
  // Instance data
  xPL_IdentifierI broadcastIdentifier = null;
  
  long lastInstanceID = -1;
  String baseInstanceID = null;
  
  HashMap<String, xPL_IdentifierI> identiferList = new HashMap<String, xPL_IdentifierI>();
  
  // Only we can create an instance of ourselves
  private IdentifierManager() { }

  void debug(String theMessage) {
    System.out.println("IDENT_MGR:: " + theMessage);
  }
  
  // Always returns a two digit hex value (even if value is 0-f)
  private String getHexPair(int theValue) {
    if (theValue < 0) return "00";
    if (theValue > 255) return "ff";
    if (theValue < 15) return "0" + Integer.toString(theValue, 16);
    return Integer.toString(theValue, 16);
  }
  
  /**
   * Get a unique identifier that can be used as an instance ID
   *
   * NOTE: Because of the limitation on xPL identifier length,
   * this will never be a truly unique identifier, but in a given 
   * JVM it'll be unique and should be pretty unique on a given 
   * network.
   *
   * @return a fairly unique identifier
   */
  public synchronized String getFairlyUniqueIdentifer() {
    // See if we need to create a seed
    if (baseInstanceID == null) {
      // See if we have a network we and try this on
      Collection<BindableAddress> theAddresses = null;
      try {
        theAddresses = BindableAddress.getBindableAddresses();
      } catch (Exception anyError) { theAddresses = null; }
      
      // Attempt to determine a local net
      if (theAddresses != null) {
        for (BindableAddress theAddress: theAddresses) {
          if (theAddress.isLoopback()) continue;
          
          byte netAddr[] = theAddress.getInetAddress().getAddress();
          baseInstanceID = getHexPair((int) netAddr[0])
                         + getHexPair((int) netAddr[1])
                         + getHexPair((int) netAddr[2])
                         + getHexPair((int) netAddr[3]);
          break;
        }
      }
      
      // If we still don't have a base value (i.e. no network), create
      // one
      if (baseInstanceID == null) {
        baseInstanceID = Long.toString(System.currentTimeMillis(), 36);
      }
      
      // See the instance ID
      lastInstanceID = (long) (Math.random() * 100000000.0);
    }
    
    // Construct the FUID and if its more than 16 chars long, trim it
    String FUID = baseInstanceID + Long.toString(lastInstanceID++, 36);
    while (FUID.length() > 16) FUID = FUID.substring(1);
    return FUID;
  }
 
  /**
   * Return a shared instance to a broadcast identifier.
   *
   * @return a shared, read-only instance of a broadcast identifier
   */
  public xPL_IdentifierI getBroadcastIdentifier() {
    if (broadcastIdentifier == null) {
      broadcastIdentifier = new xPL_Identifier(true);
      identiferList.put("*", broadcastIdentifier);
    }
    return broadcastIdentifier;
  }
  
  private String getIdentifierKey(String vendorID, String deviceID, String instanceID) {
    return vendorID.toLowerCase() + "-" + deviceID.toLowerCase() + "." + instanceID.toLowerCase();
  }

  /**
   * See if the named, passed identifier exists as a known identifier.
   *
   * @param vendorID vendorID to use for lookup
   * @param deviceID device ID to use for lookup
   * @param instanceID instance ID to use for lookup
   * @return true if the identifier exists, false if not, bad identifers, group or broadcast identifier, etc
   */
  public boolean isExistingIdentifier(String vendorID, String deviceID, String instanceID) {
    // Insure we have something worth looking up
    if ((vendorID == null) || (vendorID.length() == 0)) return false;
    if ((deviceID == null) || (deviceID.length() == 0)) return false;
    if ((instanceID == null) || (instanceID.length() == 0)) return false;
    
    // Build a key
    String identifierKey = getIdentifierKey(vendorID, deviceID, instanceID);
    xPL_IdentifierI theIdentifier = identiferList.get(identifierKey);
    if ((theIdentifier == null) || theIdentifier.isBroadcastIdentifier() || theIdentifier.isGroupIdentifier()) return false;
    return true;
  }
  
  /**
   * Get a named identifier.   If there is no such identifier, then we attempt to create
   * one (if the passed identifiers are valid).
   * 
   * @param vendorID vendorID to use for lookup
   * @param deviceID device ID to use for lookup
   * @param instanceID instance ID to use for lookup
   * @return instance of a xplDevice matching this identifier or null if name is invalid
   */
  public xPL_IdentifierI getInstance(String vendorID, String deviceID, String instanceID) {
    // Insure we have something worth looking up
    if ((vendorID == null) || (vendorID.length() == 0)) return null;
    if ((deviceID == null) || (deviceID.length() == 0)) return null;
    if ((instanceID == null) || (instanceID.length() == 0)) return null;
    
    // Build a key
    String identifierKey = getIdentifierKey(vendorID, deviceID, instanceID);

    // Attempt to look this up.  If not found, create it and register it
    synchronized(identiferList) {
      xPL_IdentifierI theIdentifier = identiferList.get(identifierKey);
      if (theIdentifier == null) {
        theIdentifier = new xPL_Identifier(vendorID, deviceID, instanceID);
        identiferList.put(identifierKey, theIdentifier);
      } else {
        if (theIdentifier.isBroadcastIdentifier() || theIdentifier.isGroupIdentifier()) return null;
      }
      return theIdentifier;
    }
  }
  
  /**
   * See if the named, passed group exists as a known identifier.
   *
   * @param groupName group name (final component only) to lookup
   * @return true if the group exists, false if not, bad identifers, not a group or is broadcast identifier, etc
   */
  public boolean isExistingGroup(String groupName) {
    // Insure we have something worth looking up
    if ((groupName == null) || (groupName.length() == 0)) return false;
    
    // Build a key
    String identifierKey = getIdentifierKey("xpl", "group", groupName);
    xPL_IdentifierI theIdentifier = identiferList.get(identifierKey);
    if ((theIdentifier == null) || theIdentifier.isBroadcastIdentifier() || !theIdentifier.isGroupIdentifier()) return false;
    return true;
  }

  /**
   * Get a group identifier.   If there is no such group, we attempt to create one
   * and return it.
   *
   * @param groupName simple name of the group
   * @return instance of a xplIdentifier matching this group or null if name is invalid
   */
  public xPL_IdentifierI getGroupIdentifier(String groupName) {
    // Insure we have something worth looking up
    if ((groupName == null) || (groupName.length() == 0)) { 
      return null;
    }
    
    // Build a key
    String identifierKey = getIdentifierKey("xpl", "group", groupName);

    // Attempt to look this up.  If not found, create it and register it
    synchronized(identiferList) {
      xPL_IdentifierI theIdentifier = identiferList.get(identifierKey);
      if (theIdentifier == null) {
        theIdentifier = new xPL_Identifier(groupName);
        identiferList.put(identifierKey, theIdentifier);
      } else {
        if (!theIdentifier.isGroupIdentifier()) {
          return null;
        }
      }
      return theIdentifier;
    }
  }
  
  /**
   * Get a named identifier, parsing a name spec to do it.  If the name
   * spec refers to a broadcast identifier, then an instance of the broadcast
   * identifier is returned.  Otherwise, the name is parsed.  If the name is
   * not in the correct form (vendor.device.instance), then null is
   * returned.  Otherwise, a shared instance to the identifier is returned
   *
   * @param identifierName name of the identifier
   * @return shared instance of identifier or null if identifier name is invalid
   */
  public xPL_IdentifierI parseNamedIdentifier(String identifierName) {
    // Handle null or empty string
    if ((identifierName == null) || (identifierName.length() == 0)) {
      return null;
    }
    
    // Handle a broadcast identifier
    if (identifierName.equals("*")) return getBroadcastIdentifier();
    
    // Parse the name out
    String vendorID = "";
    String deviceID = "";
    String instanceID = "";
    int firstDelim = -1, secondDelim = -1;
    
    // Parse first period
    if ((firstDelim = identifierName.indexOf('-')) == -1) { 
      return null;
    }
    if ((secondDelim = identifierName.indexOf('.', firstDelim + 1)) == -1) { 
      return null;
    }
    
    // Extract the parts
    vendorID = identifierName.substring(0, firstDelim);
    deviceID = identifierName.substring(firstDelim + 1, secondDelim);
    instanceID = identifierName.substring(secondDelim + 1);

    // See if this is a group and look it up if so
    if (vendorID.equalsIgnoreCase("xpl") && deviceID.equalsIgnoreCase("group")) {
    return getGroupIdentifier(instanceID);
    }
    
    // Attempt to lookup/create the name
    return getInstance(vendorID, deviceID, instanceID);
  }
  
  /**
   * Removed the passed identifier from the internal list of identifiers.  This
   * should only be done when we know all use of the identifier is gone.  While
   * any remaining xplIdentifierI objects will remain in a valid state, they
   * will not be in the cache and future attempts to lookup identifiers of the
   * same name will return a different, new instance of the xplIdentifier.
   * 
   * Note: any attempt to remove the broadcast identifier will fail
   * 
   * Note: Once a identifier is removed the passed xPL_IdentifierI instance is
   * no longer valid (it has been wiped)
   */
  public void removeIdentifier(xPL_IdentifierI theIdentifier) {
    // Skip attempts to whack null identifiers
    if (theIdentifier == null) return;
    
    // Skip attempts to whack the broadcast identifier
    if (theIdentifier.isBroadcastIdentifier()) return;
    
    // If we found it, whack it
    synchronized(identiferList) {
      identiferList.remove(theIdentifier.toString());
    }
    
    // Free resources the identifier consumed.
    theIdentifier.releaseResources();
  }
  
  /**
   * Remove the named identifier from the internal list of identifiers.  This
   * should only be done when we know all use of the identifier is gone.  While
   * any remaining xplIdentifierI objects will remain in a valid state, they
   * will not be in the cache and future attempts to lookup identifiers of the
   * same name will return a different, new instance of the xplIdentifier.
   */
  public void removeIdentifier(String vendorID, String deviceID, String instanceID) {
    removeIdentifier(getInstance(vendorID, deviceID, instanceID));
  }
  
  /**
   * Removed the named identifier from the internal list of identifiers.  This
   * should only be done when we know all use of the identifier is gone.  While
   * any remaining xplIdentifierI objects will remain in a valid state, they
   * will not be in the cache and future attempts to lookup identifiers of the
   * same name will return a different, new instance of the xplIdentifier.
   *
   * Note: any attempt to remove the brodcast identifier will fail
   */
  public void removeIdentifier(String namedIdentifier) {
    removeIdentifier(parseNamedIdentifier(namedIdentifier));
  }
  
  /**
   * Return a all identifiers
   *
   * @return iterable list of all identifiers or null if no identifiers
   */
  public Collection<xPL_IdentifierI> getAllIdentifiers() {
    synchronized(identiferList) {
      // Skip if nothing to go through
      if (identiferList.isEmpty()) return null;
      return identiferList.values();
    }
  }
  
  /**
   * Return all identifiers for a given vendor
   *
   * @param matchingVendorID vendor ID of identifiers to match
   * @return all identifiers matching the passed vendor or null if no match
   */
  public Collection<xPL_IdentifierI> getIdentifiers(String matchingVendorID) {
    synchronized(identiferList) {
      // Skip if nothing to go through
      if (identiferList.isEmpty()) return null;

      ArrayList<xPL_IdentifierI> matchingIdentifiers = new ArrayList<xPL_IdentifierI>();

      // Filter the identifiers
      for (xPL_IdentifierI theIdentifier: identiferList.values()) {
        if (!theIdentifier.getVendorID().equals(matchingVendorID)) continue;
        matchingIdentifiers.add(theIdentifier);
      }

      // Reutrn the matching list, if any
      if (matchingIdentifiers.isEmpty()) return null;
      return matchingIdentifiers;
    }
  }
  
  /**
   * Return all identifiers for a given vendor/device
   *
   * @param matchingVendorID vendor ID of identifiers to match
   * @param matchingDeviceID device ID of identifiers to match
   * @return all identifiers matching the passed vendor/device or null if no matches
   */
  public Collection<xPL_IdentifierI> getIdentifiers(String matchingVendorID, String matchingDeviceID) {
    synchronized(identiferList) {
      // Skip if nothing to go through
      if (identiferList.isEmpty()) return null;

      ArrayList<xPL_IdentifierI> matchingIdentifiers = new ArrayList<xPL_IdentifierI>();

      // Filter the identifiers
      for (xPL_IdentifierI theIdentifier: identiferList.values()) {
        if (!theIdentifier.getVendorID().equals(matchingVendorID)) continue;
        if (!theIdentifier.getDeviceID().equals(matchingDeviceID)) continue;
        matchingIdentifiers.add(theIdentifier);
      }

      // Reutrn the matching list, if any
      if (matchingIdentifiers.isEmpty()) return null;
      return matchingIdentifiers;
    }
  }
  
  /**
   * Return all group identifiers
   *
   * @return all group identifiers
   */
  public Collection<xPL_IdentifierI> getGroupIdentifiers() {
    synchronized(identiferList) {
      // Skip if nothing to go through
      if (identiferList.isEmpty()) return null;

      ArrayList<xPL_IdentifierI> matchingIdentifiers = new ArrayList<xPL_IdentifierI>();

      // Filter the identifiers
      for(xPL_IdentifierI theIdentifier: identiferList.values()) {
        if (!theIdentifier.isGroupIdentifier()) continue;
        matchingIdentifiers.add(theIdentifier);
      }

      // Reutrn the matching list, if any
      if (matchingIdentifiers.isEmpty()) return null;
      return matchingIdentifiers;
    }
  }
  
  /**
   * Change a identifiers instance ID.  This is generally only needed after a identifier
   * has been reconfigured and as a result, a default instance ID has been changed
   * to a specified one.  This is normally all handled in the xPL_Device, so 
   * it would be rare to never that anything outside of the xPL_Device implementation
   * would need it.  
   * 
   * In short, if you are not 100% you need this, you probably don't.
   * 
   * If the resulting new identifier would be non-unique, the attempt to change the
   * instance ID will fail.
   * 
   * The reconfigured identifier is returned (if there is no error).  This may or may
   * not be the same xPL_IdentifierI passed in.  If this comes back non-null, you have
   * to be able to reinstall the returned identifier as your current identifier from then
   * on.  It's possible the passed xPL_IdentifierI may be "delisted" as a result of
   * this reconfig and as such, you should stop using it if the returned xPL_IdentifierI
   * does not match it.
   * 
   * @param theIdentifier identifier to change instance ID on
   * @param newInstanceID new instance ID to assign to identifier
   * @return instance of the reconfigured identifier or null
   */
  public xPL_IdentifierI reconfigureInstanceID(xPL_IdentifierI theIdentifier, String newInstanceID) {
    // Fail if we cannot modify this
    if ((theIdentifier == null) || theIdentifier.isBroadcastIdentifier() || theIdentifier.isGroupIdentifier() || !(theIdentifier instanceof xPL_Identifier)) return null;
    
    synchronized(identiferList) {
      // Make sure this is not any real change
      if (theIdentifier.getInstanceID().equalsIgnoreCase(newInstanceID)) return null;
      
      // See if we are already running under this identifier ID (it's an async world -- these things happen)
      String identifierKey = getIdentifierKey(theIdentifier.getVendorID(), theIdentifier.getDeviceID(), newInstanceID);
      xPL_IdentifierI existingIdentifier = identiferList.get(identifierKey);
      if (existingIdentifier != null) {
        identiferList.remove(getIdentifierKey(theIdentifier.getVendorID(), theIdentifier.getDeviceID(), theIdentifier.getInstanceID()));
        return existingIdentifier;
      }

      // Remove old keyed reference, install instance ID and install new keyed reference
      if (identiferList.remove(getIdentifierKey(theIdentifier.getVendorID(), theIdentifier.getDeviceID(), theIdentifier.getInstanceID())) == null) return null;

      // Install new instance ID and update mapping
      ((xPL_Identifier) theIdentifier).setInstanceID(newInstanceID);
      identiferList.put(identifierKey, theIdentifier);
      return theIdentifier;
    }
  }
}
