/*
 * MessageFormatter.java
 * Copyright 2004, Gerald R Duprey Jr
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy of
 * the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0 
 *
 * Unless required by applicable law or agreed to in writing, software distributed
 * under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
 * CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License. 
 */

package org.cdp1802.xpl;



/**
 * Formats an XPL message into text/wire format
 *
 * @author  gerry
 */
class MessageFormatter {
  private static final char LF = '\n';
  
  MessageFormatter() { }
  
  /**
   * Given a passed xplMessage instance, format it into a wire-format 
   * version in text, suitable for transmission.
   *
   * @param theMessage message instance to format into text
   * @param forMediaHandler mediaHandler this message is destined to be written to (can be null)
   * @param bumpHopCount if true, hop count is written as actual hop count + 1
   * @return formatted text version of the message or null if message was incomplete/invalid
   */
  String formatMessage(xPL_MessageI theMessage, xPL_MediaHandlerI forMediaHandler, boolean bumpHopCount) {
    // Make sure the message is minimally valid
    if ((theMessage.getSource() == null) || (theMessage.getTarget() == null)) return null;
    if ((theMessage.getSchemaClass().length() == 0) || (theMessage.getSchemaType().length() == 0)) return null;
    
    StringBuffer msgText = new StringBuffer(64);
    
    // Format message type
    switch(theMessage.getType()) {
      case COMMAND:
        msgText.append("xpl-cmnd");
        break;
        
      case TRIGGER:
        msgText.append("xpl-trig");
        break;
        
      case STATUS:
        msgText.append("xpl-stat");
        break;
        
      default:
        return null;
    }
    msgText.append(LF);
    
    // Start message header
    msgText.append('{');
    msgText.append(LF);
    
    // Format hop count
    msgText.append("hop=");
    msgText.append(bumpHopCount ? theMessage.getHopCount() + 1 : theMessage.getHopCount());
    msgText.append(LF);
    
    // Format source
    msgText.append("source=");
    msgText.append(theMessage.getSource().toString().toLowerCase());
    msgText.append(LF);
    
    // Format target
    msgText.append("target=");
    msgText.append(theMessage.getTarget().toString().toLowerCase());
    msgText.append(LF);
    
    // End the header
    msgText.append('}');
    msgText.append(LF);
    
    // Format the schema class/type
    msgText.append(theMessage.getSchemaClass().toLowerCase());
    msgText.append('.');
    msgText.append(theMessage.getSchemaType().toLowerCase());
    msgText.append(LF);
    
    // Start of body
    msgText.append('{');
    msgText.append(LF);
    
    // Format name/value pairs
    for (NamedValueI aValue: theMessage.getMessageBody().getAllNamedValues()) {
      msgText.append(aValue.getName().toLowerCase());
      msgText.append('=');
      msgText.append(aValue.getValue());
      msgText.append(LF);
    }
    
    // End of body
    msgText.append('}');
    msgText.append(LF);
    
    // And we are done
    return msgText.toString();
  }
}
