/*
 * MessageParser.java
 * Copyright 2004, Gerald R Duprey Jr
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy of
 * the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0 
 *
 * Unless required by applicable law or agreed to in writing, software distributed
 * under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
 * CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License. 
 */

package org.cdp1802.xpl;

import java.util.NoSuchElementException;
import java.util.StringTokenizer;

/**
 * Parses a passed message into an immutable xplMessage
 *
 * @author  gerry
 */
class MessageParser {
  IdentifierManager identifierManager = xPL_Manager.getManager().getIdentifierManager();
  
  MessageParser() { }
  
  // Parse the passed string for the named value.  The name is expected to
  // start the string.  If the name does not start the string or does not
  // match, null is returned.  The name is case insensitive.  If there is
  // a match, the value is returned.  In the event there is no value 
  // (i.e. name=), an empty string (not null) is returned
  private String parseNamedValue(String theText, String theName) {
    // Get name/value delimiter
    int delimPtr = theText.indexOf('=');
    if (delimPtr == -1) return null;
    
    // Extract the string for direct testing
    String testName = theText.substring(0, delimPtr);
      
    // If there is no match, ignore this
    if (!testName.equalsIgnoreCase(theName)) return null;

    // Extract the value
    return theText.substring(delimPtr + 1);
  }
  
  private int parseNameValuePairs(StringTokenizer tokenizer, int lineCount, xPL_Message theMessage) throws xPL_ParseException, NoSuchElementException {
    String theToken = null, theName = null, theValue = null;
    int delimPtr = 0;
    
    while (tokenizer.hasMoreTokens()) {
      theToken = tokenizer.nextToken();  lineCount++;
      
      // Handle a terminal token
      if (theToken.equals("}")) return lineCount;
        
      // Look for delimiter
      if ((delimPtr = theToken.indexOf('=')) == -1) throw new xPL_ParseException("Empty/Missing name/value pair in message body on line " + lineCount);

      // Break up parts
      if ((theName = theToken.substring(0, delimPtr)).length() == 0) throw new xPL_ParseException("Missing name in name/value pair in message body on line " + lineCount);
      theValue = theToken.substring(delimPtr + 1);

      // Add this to the list
      ((NamedValues) theMessage.namedValues).namedValues.add(new NamedValue(theName, theValue));
    }
    
    return lineCount;
  }
  
  /**
   * Given a string that represents a validly formatted textual representation 
   * of an XPL message, parse it into an xplMessage object, doing validation
   * along the way.
   *
   * @param textMessage text formatted version of the XPL message
   * @param receivedFrom media handler the message was received from
   * @return a xplMessage isntance representing the passed text message
   * @throws xplParseException if there are any problems found in parsing
   */
  xPL_MessageI parseMessage(String textMessage, xPL_MediaHandlerI receivedFrom) throws xPL_ParseException {
    // Make sure the message is even vaugley parsable
    if ((textMessage == null) || (textMessage.length() < 13)) throw new xPL_ParseException("Message is empty or too short");
    
    String theToken = null;
    String theName = null, theValue = null;
    int lineCount = 0;
    
    xPL_Message theMessage = null;
    int hopCount = 0;
    xPL_MessageI.MessageType messageType = xPL_MessageI.MessageType.UNKNOWN;
    xPL_IdentifierI sourceDevice = null, targetDevice = null;
    String schemaClass = null, schemaType = null;
    
    // Create line based tokenizer -- this makes our lives a bit easier
    try {
      StringTokenizer tokenizer = new StringTokenizer(textMessage, "\n");
    
      // See what kind of message this is
      theToken = tokenizer.nextToken(); lineCount++;
      if (theToken.equalsIgnoreCase("xpl-cmnd")) 
        messageType = xPL_MessageI.MessageType.COMMAND;
      else if (theToken.equalsIgnoreCase("xpl-trig"))
        messageType = xPL_MessageI.MessageType.TRIGGER;
      else if (theToken.equalsIgnoreCase("xpl-stat"))
        messageType = xPL_MessageI.MessageType.STATUS;
      else 
        throw new xPL_ParseException("Unknown message header/type: " + theToken);
      
      // Next should be a header intro
      theToken = tokenizer.nextToken(); lineCount++;
      if (!theToken.equals("{")) throw new xPL_ParseException("Missing header { after message type -- improperly formatted message");
      
      // Read in hop count
      theToken = tokenizer.nextToken(); lineCount++;
      if (((theValue = parseNamedValue(theToken, "hop")) == null) || (theValue.length() == 0))
        throw new xPL_ParseException("Missing/invalid hop= in message header");
      
      try {
        hopCount = Integer.parseInt(theValue);
      } catch (Exception anyError) {
        throw new xPL_ParseException("Invalid/non-numeric hop count in message header");
      }
      
      // Read in the source to the message 
      theToken = tokenizer.nextToken(); lineCount++;
      if (((theValue = parseNamedValue(theToken, "source")) == null) || (theValue.length() == 0))
        throw new xPL_ParseException("Missing/invalid source= in message header");
      
      // Lookup the source
      if ((sourceDevice = identifierManager.parseNamedIdentifier(theValue)) == null) {
        throw new xPL_ParseException("Invalid/improperly formatted source in message header");
      }

      // Read in the target to the message 
      theToken = tokenizer.nextToken(); lineCount++;
      if (((theValue = parseNamedValue(theToken, "target")) == null) || (theValue.length() == 0))
        throw new xPL_ParseException("Missing/invalid target= in message header");
      
      // Lookup the source
      if ((targetDevice = identifierManager.parseNamedIdentifier(theValue)) == null) {
        throw new xPL_ParseException("Invalid/improperly formatted target in message header");
      }
      
      // Next should be a header close
      theToken = tokenizer.nextToken(); lineCount++;
      if (!theToken.equals("}")) throw new xPL_ParseException("Missing header closing of } after header data -- improperly formatted message");

      // Next should be the schema class/type
      theToken = tokenizer.nextToken(); lineCount++;
      int delimPtr = theToken.indexOf('.');
      if (delimPtr == -1) throw new xPL_ParseException("Missing/improperly formatted message schema class/type in message");
      if ((schemaClass = theToken.substring(0, delimPtr)).length() == 0) throw new xPL_ParseException("Empty/missing schema class in message");
      if ((schemaType = theToken.substring(delimPtr + 1)).length() == 0) throw new xPL_ParseException("Empty/missing schema type in message");

      // Next should be the message body start
      theToken = tokenizer.nextToken(); lineCount++;
      if (!theToken.equals("{")) throw new xPL_ParseException("Missing body section open { after header -- improperly formatted message");
      
      // Time to create a message
      theMessage = new xPL_Message();
      theMessage.hopCount = hopCount;
      theMessage.msgType = messageType;
      theMessage.msgSource = sourceDevice;
      theMessage.msgTarget = targetDevice;
      theMessage.schemaClass = schemaClass;
      theMessage.schemaType = schemaType;
      theMessage.receivedFrom = receivedFrom;
      theMessage.namedValues = new NamedValues();
      
      // Hand Name/Value parsing off 
      lineCount = parseNameValuePairs(tokenizer, lineCount, theMessage);
    } catch (NoSuchElementException noMoreTokens) {
      throw new xPL_ParseException("Prematurely ran out of message tokens/lines -- short/invalid message, " + lineCount + " lines read OK");
    } catch (xPL_ParseException parseError) {
      throw parseError;
    } catch (Exception anyOtherException) {
      throw new xPL_ParseException("Unexpected error parsing message -- " + anyOtherException.getMessage(), anyOtherException);
    }
    
    // And we are done
    return theMessage;
  }
}
