/*
 * NamedValues.java
 * Copyright 2004, Gerald R Duprey Jr
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy of
 * the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0 
 *
 * Unless required by applicable law or agreed to in writing, software distributed
 * under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
 * CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License. 
 */

package org.cdp1802.xpl;

import java.util.ArrayList;

import java.util.Collection;

/**
 *
 * @author  gerry
 */
public class NamedValues implements NamedValuesI {
  ArrayList<NamedValueI> namedValues = new ArrayList<NamedValueI>();
  
  NamedValues() { }

  public void releaseResources() {
    if (namedValues != null) {
      for (NamedValueI namedValue: namedValues) {
        namedValue.releaseResources();
      }
      namedValues.clear();
      namedValues = null;
    }
  }
  
  // Return the first occurance of the named name/value.  If not
  // found, return -1
  int getFirstIndexOfNamedValue(String theName) {
    NamedValueI namedValue = null;

    int nvCount = namedValues.size();
    for (int nvIndex = 0; nvIndex < nvCount; nvIndex++) {
      if ((namedValue = namedValues.get(nvIndex)).getName().equalsIgnoreCase(theName)) return nvIndex;
    }
    
    return -1;
  }

  // Return first name/value for the passed name
  NamedValueI getFirstNamedValue(String theName) {
    int nvIndex = getFirstIndexOfNamedValue(theName);
    if (nvIndex == -1) return null;
    return namedValues.get(nvIndex);
  }
  
  public boolean isEmpty() {
    return ((namedValues == null) || namedValues.isEmpty());
  }
  
  public boolean hasNamedValue(String theName) {
    synchronized(namedValues) {
      return (getFirstIndexOfNamedValue(theName) != -1);
    }
  }

  public boolean hasNamedValue(String theName, String theValue) {
    synchronized(namedValues) {
      NamedValueI namedValue = null;

      int nvCount = namedValues.size();
      for (int nvIndex = 0; nvIndex < nvCount; nvIndex++) {
        if ((namedValue = namedValues.get(nvIndex)).getName().equalsIgnoreCase(theName)) {
          if (namedValue.getValue().equalsIgnoreCase(theValue)) return true;
        }
      }
      
      return false;
    }
  }
  
  public boolean hasNamedValue(String theName, int theValue) {
    return hasNamedValue(theName, String.valueOf(theValue));
  }
  
  public boolean hasNamedValue(String theName, boolean theValue) {
    return hasNamedValue(theName, (theValue ? "1" : "0"));
  }
  
  public String getNamedValue(String theName) {
    NamedValueI theNameValue = getFirstNamedValue(theName);
    if (theNameValue == null) return null;
    return theNameValue.getValue();
  }

  public String getNamedValue(String theName, String defaultValue) {
    NamedValueI theNameValue = getFirstNamedValue(theName);
    if (theNameValue == null) return defaultValue;
    return theNameValue.getValue();
  }
  
  public int getIntNamedValue(String theName, int defaultValue) {
    NamedValueI theNameValue = getFirstNamedValue(theName);
    if (theNameValue == null) return defaultValue;
    return theNameValue.getIntValue();
  }
  
  public int getIntNamedValue(String theName) {
    return getIntNamedValue(theName, -1);
  }
  
  public boolean getBoolNamedValue(String theName, boolean defaultValue) {
    NamedValueI theNameValue = getFirstNamedValue(theName);
    if (theNameValue == null) return defaultValue;
    return theNameValue.getBoolValue();
  }
  public boolean getBoolNamedValue(String theName) {
    return getBoolNamedValue(theName, false);
  }

  public int getNamedValuesCount() { return namedValues.size(); }

  public int getNamedValueCount(String theName) {
    NamedValueI namedValue = null;
    int valueCount = 0;

    synchronized(namedValues) {
      int nvCount = namedValues.size();
      for (int nvIndex = 0; nvIndex < nvCount; nvIndex++) {
        if ((namedValue = namedValues.get(nvIndex)).getName().equalsIgnoreCase(theName)) valueCount++;
      }
    }
    
    return valueCount;
  }

  public java.util.List<String> getNamedValues(String theName) {
    ArrayList<String> valueList = new ArrayList<String>();

    synchronized(namedValues) {
      NamedValueI namedValue = null;
      int nvCount = namedValues.size();

      for (int nvIndex = 0; nvIndex < nvCount; nvIndex++) {
        if ((namedValue = namedValues.get(nvIndex)).getName().equalsIgnoreCase(theName)) {
          valueList.add(namedValue.getValue());
        }
      }
    }
    
    if (valueList.size() == 0) return null;
    return valueList;
  }
  
  public Collection<NamedValueI> getAllNamedValues() {
    return namedValues;
  }

  /**
   * Create a mutable copy of this list of named values.  Even if this is itself
   * mutable, we still make a copy
   *
   * @return a mutable copy of this named/value list
   */
  public MutableNamedValuesI getMutableCopy() {
    MutableNamedValuesI theCopy = new MutableNamedValues();
    for (NamedValueI origValue: namedValues) {
      theCopy.addNamedValue(origValue.getName(), origValue.getValue());
    }
    return theCopy;
  }
  
  
  public String toString() {
    StringBuffer nvText = new StringBuffer();
    nvText.append(namedValues.size());
    nvText.append(" N/V Pairs: ");

    NamedValueI namedValue = null;
    int nvCount = namedValues.size();
    for (int nvIndex = 0; nvIndex < nvCount; nvIndex++) {
      if (nvIndex > 0) nvText.append(", ");
      nvText.append('[');
      nvText.append(namedValues.get(nvIndex).toString());
      nvText.append(']');
    }

    return nvText.toString();
  }

}
