/*
 * NameValueListI.java
 * Copyright 2004, Gerald R Duprey Jr
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy of
 * the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0 
 *
 * Unless required by applicable law or agreed to in writing, software distributed
 * under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
 * CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License. 
 */

package org.cdp1802.xpl;

import java.util.Collection;
import java.util.List;

/**
 * NameValues is a list of non-unqiue name/value pairs.  Non-unique is
 * important to note:  You can have multiple occurances of a single
 * name with different values.
 *
 * @author  gerry
 */
public interface NamedValuesI {
  /**
   * Return true if there are no named values defined
   *
   * @return true of no named values present
   */
  public boolean isEmpty();
  
  /**
   * Returns the total number of named values contained
   *
   * @return count of named values
   */
  public int getNamedValuesCount();

  /**
   * Returns true if there is at least one value with the
   * passed name.
   *
   * @param theName the name to look for in this list of named values
   * @return true if there is 1 (or more) values with the passed name
   */
  public boolean hasNamedValue(String theName);

  
  /**
   * Returns true if there is at least one named value by the 
   * passed name that has the passed value.
   *
   * @param theName the name to look for
   * @param theValue the value to match
   * @return true if there is at least named value with the passed name and value
   */
  public boolean hasNamedValue(String theName, String theValue);
  
  /**
   * Returns true if there is at least one named value by the 
   * passed name that has the passed value.
   *
   * @param theName the name to look for
   * @param theValue the value to match
   * @return true if there is at least named value with the passed name and value
   */
  public boolean hasNamedValue(String theName, int theValue);
  
  /**
   * Returns true if there is at least one named value by the 
   * passed name that has the passed value.
   *
   * @param theName the name to look for
   * @param theValue the value to match
   * @return true if there is at least named value with the passed name and value
   */
  public boolean hasNamedValue(String theName, boolean theValue);
  
  
  /**
   * Returns the total number of values with the passed named
   *
   * @return number of values with the passed name
   */
  public int getNamedValueCount(String theName);

  /**
   * Return the value associated with the passed name.
   * 
   * If there is more than one value for this name, only the first
   * set/installed value is returned.
   *
   * If there is no instance of a value with this name, null is returned.
   *
   * If there value for the name is empty, an empty (not-null) string
   * is returned
   *
   * @param theName name of the value to fetch
   * @return value of the first matching named value or null if not found
   */
  public String getNamedValue(String theName);

  /**
   * Return the value associated with the passed name.
   * 
   * If there is more than one value for this name, only the first
   * set/installed value is returned.
   *
   * If there is no instance of a value with this name, the passed
   * default value is returned.
   *
   * If there value for the name is empty, an empty (not-null) string
   * is returned
   *
   * @param theName name of the value to fetch
   * @param defaultValue value to return if there is no matching value with this name
   * @return value of the first matching named value or the default value
   */
  public String getNamedValue(String theName, String defaultValue);
  
  /**
   * Return an integer value associated with the passed name.
   * 
   * If there is more than one value for this name, only the first
   * set/installed value is returned.
   *
   * If there is no instance of a value with this name, -1 is returned.
   *
   * If the value of is not numeric, -1 is returned.
   *
   * If there value for the name is empty, 0 (zero) is returned
   *
   * @param theName name of the value to fetch
   * @return value of the first matching named value or -1 if not found
   */
  public int getIntNamedValue(String theName);

  /**
   * Return an integer value associated with the passed name.
   * 
   * If there is more than one value for this name, only the first
   * set/installed value is returned.
   *
   * If there is no instance of a value with this name, the defaultValue is returned.
   *
   * If the value of is not numeric, -1 is returned.
   *
   * If there value for the name is empty, 0 (zero) is returned
   *
   * @param theName name of the value to fetch
   * @param defaultValue value to return if name is not found
   * @return value of the first matching named value or -1 if invalid integer found
   */
  public int getIntNamedValue(String theName, int defaultValue);
  
  /**
   * Return a boolean value associated with the passed name.
   * 
   * If there is more than one value for this name, only the first
   * set/installed value is returned.
   *
   * If there is no instance of a value with this name, false is returned.
   *
   * The following values are considered true: 'true', 'yes', 't', 'y' or '1'
   * The check is case-insensitive
   *
   * If there value for the name is empty, false is returned
   *
   * @param theName name of the value to fetch
   * @return value of the first matching named value or false if not found
   */
  public boolean getBoolNamedValue(String theName);

  /**
   * Return a boolean value associated with the passed name.
   * 
   * If there is more than one value for this name, only the first
   * set/installed value is returned.
   *
   * If there is no instance of a value with this name, defaultValue is returned.
   *
   * The following values are considered true: 'true', 'yes', 't', 'y' or '1'
   * The check is case-insensitive
   *
   * If there value for the name is empty, false is returned
   *
   * @param theName name of the value to fetch
   * @param defaultValue value to return if no matching name is found
   * @return value of the first matching named value or false if not found
   */
  public boolean getBoolNamedValue(String theName, boolean defaultValue);
  
  /**
   * Return a list of all values that match the passed name.  The list is 
   * in order the values are listed.  If there are no matching values for
   * the passed name, null is returned
   *
   * @param theName name to find matching values for
   * @return list of matching values for name or null if name is not found
   */
  public List<String> getNamedValues(String theName);
  
  /**
   * Return the list of all Name/Value pairs
   *
   * @return the name/value list
   */
  public Collection<NamedValueI> getAllNamedValues();
  
  /**
   * Create a mutable copy of this list of named values.  Even if this is itself
   * mutable, we still make a copy
   *
   * @return a mutable copy of this named/value list
   */
  public MutableNamedValuesI getMutableCopy();
  
  
  /**
   * Release all resources this set of Named Values is using.  After this
   * call, this instance cannot be reused.
   */
  public void releaseResources();
}
