/*
 * DeviceManager.java
 * Copyright 2004, Gerald R Duprey Jr
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy of
 * the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0 
 *
 * Unless required by applicable law or agreed to in writing, software distributed
 * under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
 * CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License. 
 */

package org.cdp1802.xpl.device;

import java.io.File;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.ArrayList;
import org.cdp1802.xpl.xPL_IdentifierI;
import org.cdp1802.xpl.xPL_Manager;

/**
 * Handles the creation, desctruction and management of devices
 * @author  gerry
 */
public class DeviceManager {
  private static DeviceManager deviceManager = new DeviceManager();  

  /**
   * Return the DeviceManager instance
   * @return the device manager instance
   */
  public static DeviceManager getManager() { return deviceManager; }
  
  // Instance variables
  HeartbeatManager heartbeatManager = null;
  DevicePersister devicePersister = null;
  ArrayList<xPL_DeviceI> managedDevices = new ArrayList<xPL_DeviceI>();
  
  private DeviceManager() {  }

  void debug(String theMessage) {
    System.out.println("SVC_MGR:: " + theMessage);
  }
  
  private void startHeartbeatManager() {
    if (heartbeatManager == null) heartbeatManager = new HeartbeatManager(managedDevices);
    heartbeatManager.start();
  }
  
  private void stopHeartbeatManager() {
    if (heartbeatManager == null) return;
    heartbeatManager.stop();
  }
  
  /**
   * Add a device to the manager.  The manager will take care of asking the
   * device to send out heartbeats and other mainteance issues.
   *
   * You normally do not need to use this device as when you enable a
   * device, it will be automatically added.  
   *
   * @param theDevice device to manage
   */
  public void addManagedDevice(xPL_DeviceI theDevice) {
    synchronized(managedDevices) {
      // Skip if already managed
      if (managedDevices.contains(theDevice)) return;

      boolean heartbeatManagerNeeded = (managedDevices.size() == 0);
      managedDevices.add(theDevice);

      // Alert sleeper so initial heartbeat can go out.
      if (heartbeatManagerNeeded) 
        startHeartbeatManager();
      else
        managedDevices.notifyAll();
    }
  }
  
  /**
   * Remove a device from the manager.
   * 
   * You normally do not need to use this serviec as when you disable a
   * device, it will be automatically removed.
   * 
   * @param theDevice device to unmanage
   */
  public void removeManagedDevice(xPL_DeviceI theDevice) {
    synchronized(managedDevices) {
      // Remove device.  If not found, bomb out
      if (!managedDevices.remove(theDevice)) return;
      
      // If found and there are now no devices, stop the heartbeat manager
      if (managedDevices.size() == 0) stopHeartbeatManager();
    }
  }

  /**
   * Disable/shutdown all currently managed devices
   *
   * This is really only meant to be used once, when the JVM is being shutdown
   */
  public void shutdownAllManagedDevices() {
    synchronized(managedDevices) {
      int deviceCount = managedDevices.size();
      for (int deviceIndex = deviceCount - 1; deviceIndex >= 0; deviceIndex--) {
        managedDevices.get(deviceIndex).setEnabled(false);
      }
    }
  }
  
  /** 
   * Return the device persister.  This is generally NOT needed outside the 
   * core xPL package itself. Beware!
   *
   * @return The current device persister
   */
  public DevicePersister getDevicePersisterr() {
    if (devicePersister != null) return devicePersister;
    
    synchronized(this) {
      if (devicePersister == null) devicePersister = new DevicePersister();
    }
    return devicePersister;
  }
  
  /** 
   * Create a new device.  You will need to set a device ident in 
   * place before you can enable it.
   *
   * @return new instance of a device
   */
  public xPL_DeviceI createDevice() {
    return new xPL_Device();
  }
  
  /**
   * Create a new device with the passed identifier.
   *
   * If the instance ID is set to *, a unique instance ID will be created and used 
   *
   * @param theIdent identifier for this device
   * @return new instance of this device or null for a bad device
   */
  public xPL_DeviceI createDevice(xPL_IdentifierI theIdent) {
    if (theIdent == null) return null;

    // Get the instance ID and see if we need to change anything
    String instanceID = theIdent.getInstanceID();
    if ((instanceID == null) || (instanceID.length() == 0) || instanceID.equals("*"))
      instanceID = xPL_Manager.getManager().getIdentifierManager().getFairlyUniqueIdentifer();
    
    xPL_DeviceI theDevice = new xPL_Device();
    theDevice.setIdent(theIdent.getVendorID(), theIdent.getDeviceID(), instanceID);
    return theDevice;
  }
  
  /**
   * Create a new device with the passed identifer
   *
   * If the instance ID is set to *, a unique instance ID will be created and used 
   *
   * @param vendorID vendorID for the device
   * @param deviceID deviceID for the device
   * @param instanceID instanceID for the device or null to auto-create an instance ID
   * @return new instance of this device or null if missing/invalid data passed
   */
  public xPL_DeviceI createDevice(String vendorID, String deviceID, String instanceID) {
    // Make sure we have a valid vendor, device and create a instance ID if needed
    if ((vendorID == null) || (vendorID.length() == 0)) return null;
    if ((deviceID == null) || (deviceID.length() == 0)) return null;
    if ((instanceID == null) || (instanceID.length() == 0)) {
      instanceID = xPL_Manager.getManager().getIdentifierManager().getFairlyUniqueIdentifer();
    }
    
    xPL_DeviceI theDevice = new xPL_Device();
    theDevice.setIdent(vendorID, deviceID, instanceID);
    return theDevice;
  }
  
  /**
   * Create a configurable device and attempt to restore the values of the
   * device from it's configuration value.  If there is no configuration 
   * file to read from, a default instance ID is generated.  The config file
   * is also installed as the auto-preserve file for the device as well.
   *
   * @param vendorID vendorID for the device
   * @param deviceID deviceID for the device
   * @param configFile config file to use for this device
   * @return new instance of this device or null if missing/invalid data passed
   */
  public xPL_DeviceI createConfigurableDevice(String vendorID, String deviceID, File configFile) {
    // Make sure we have a valid vendor, device and create a instance ID if needed
    if ((vendorID == null) || (vendorID.length() == 0)) return null;
    if ((deviceID == null) || (deviceID.length() == 0)) return null;
    if (configFile == null) return null;
    
    xPL_DeviceI theDevice = new xPL_Device();
    theDevice.setIdent(vendorID, deviceID, xPL_Manager.getManager().getIdentifierManager().getFairlyUniqueIdentifer());
    theDevice.setConfigurable(true);
    theDevice.restoreDeviceConfig(configFile);
    theDevice.setAutoPreserveConfigFile(configFile);
    return theDevice;
  }

  /**
   * Create a configurable device and attempt to restore the values of the
   * device from it's configuration value.  If there is no configuration 
   * file to read from, a default instance ID is generated.  The config file
   * is also installed as the auto-preserve file for the device as well.
   *
   * If the passed device identifier is a broadcase ident or a group ident, 
   * this call will fail and return null
   *
   * If the instance ID is set to *, a unique instance ID will be created and used 
   *
   * @param theIdent identifier which contains the vendor/device/instance ID
   * @param configFile config file to use for this device
   * @return new instance of this device or null if missing/invalid data passed
   */
  public xPL_DeviceI createConfigurableDevice(xPL_IdentifierI theIdent, File configFile) {
    // Make sure we have a valid vendor, device and create a instance ID if needed
    if ((theIdent == null) || theIdent.isBroadcastIdentifier() || theIdent.isGroupIdentifier()) return null;
    if (configFile == null) return null;

    // Get the instance ID and see if we need to change anything
    String instanceID = theIdent.getInstanceID();
    if ((instanceID == null) || (instanceID.length() == 0) || instanceID.equals("*"))
      instanceID = xPL_Manager.getManager().getIdentifierManager().getFairlyUniqueIdentifer();
    
    xPL_DeviceI theDevice = new xPL_Device();
    theDevice.setIdent(theIdent.getVendorID(), theIdent.getDeviceID(), instanceID);
    theDevice.setConfigurable(true);
    theDevice.restoreDeviceConfig(configFile);
    theDevice.setAutoPreserveConfigFile(configFile);
    return theDevice;
  }

  /**
   * Restore the passed device from the passed input stream.  
   * 
   * If the device has a vendor/device in it, then the stream will be read until
   * a matching device/device is found.  If the device has no vendor/device
   * in it, then the first device found in the stream is used and installed.
   * 
   * @param theDevice device to restore state into
   * @param fromStream stream to read state from
   * @return true if any state was restored, false if state not changed in device
   */
  public boolean restoreDevice(xPL_DeviceI theDevice, InputStream fromStream) {
    // Insure we have a device persister
    if (devicePersister == null) synchronized(this) {
      if (devicePersister == null) devicePersister = new DevicePersister();
    }

    return devicePersister.restoreDevice(theDevice, fromStream);
  }
  
  /**
   * Restore the passed device from the passed input file
   * 
   * If the device has a vendor/device in it, then the file will be read until
   * a matching device/device is found.  If the device has no vendor/device
   * in it, then the first deice found in the file is used and installed.
   * 
   * @param theDevice device to restore state into
   * @param fromFile file to read state from
   * @return true if any state was restored, false if state not changed in device
   */
  public boolean restoreDevice(xPL_DeviceI theDevice, File fromFile) {
    // Insure we have a device persister
    if (devicePersister == null) synchronized(this) {
      if (devicePersister == null) devicePersister = new DevicePersister();
    }

    return devicePersister.restoreDevice(theDevice, fromFile);
  }
  
  /**
   * Save the state of the the passed device to the passed stream
   * 
   * @param theDevice device to save state from
   * @param toStream stream to write state to
   * @return true if state was saved, false if state was not preservable or an error occured
   */
  public boolean preserveDevice(xPL_DeviceI theDevice, OutputStream toStream) {
    // Insure we have a device persister
    if (devicePersister == null) synchronized(this) {
      if (devicePersister == null) devicePersister = new DevicePersister();
    }
    
    // Do persistence
    return devicePersister.preserveDevice(theDevice, toStream);
  }
  
  /**
   * Save the state of the the passed device to the passed file
   * 
   * @param theDevice device to save state from
   * @param toFile file to write state to
   * @return true if state was saved, false if state was not preservable or an error occured
   */
  public boolean preserveDevice(xPL_DeviceI theDevice, File toFile) {
    // Insure we have a device persister
    if (devicePersister == null) synchronized(this) {
      if (devicePersister == null) devicePersister = new DevicePersister();
    }
    
    // Do persistence
    return devicePersister.preserveDevice(theDevice, toFile);
  }
  
  /**
   * Release the devices resources.  This can only be done if the device is
   * disabled.  Once done, the instance of the device is invalid and cannot
   * be reused.  
   * 
   * NOTE: This only works on devices returned from createDevice.  If you
   * implement your own xPL_DeviceI class, you're responsible for releasing
   * it's resources on your own.
   * 
   * @param theDevice device to release resources for
   */
  public void releaseDeviceResources(xPL_DeviceI theDevice) {
    if (theDevice.isEnabled()) return;
    if (!(theDevice instanceof xPL_Device)) return;
    ((xPL_Device) theDevice).releaseResources();
  }
}
