/*
 * DevicePersister.java
 * Copyright 2004, Gerald R Duprey Jr
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy of
 * the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0 
 *
 * Unless required by applicable law or agreed to in writing, software distributed
 * under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
 * CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License. 
 */

package org.cdp1802.xpl.device;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.PrintWriter;
import java.util.Iterator;
import org.cdp1802.xpl.MutableNamedValuesI;
import org.cdp1802.xpl.NamedValuesI;
import org.cdp1802.xpl.NamedValueI;
import org.cdp1802.xpl.xPL_IdentifierI;
import org.cdp1802.xpl.xPL_Manager;
import org.cdp1802.xpl.xPL_Util;
import org.cdp1802.xpl.xPL_FilterI;

/**
 * Support for saving and restoring a devices attributes 
 *
 * @author gduprey
 */
public class DevicePersister {
  
  /**
   * Creates a new instance of DevicePersister 
   */
  DevicePersister() { }

  void debug(String theMessage) {
    System.out.println("DEV_PRST:: " + theMessage);
  }
  
  MutableNamedValuesI parseNamedValues(BufferedReader from) {
    MutableNamedValuesI namedValues = xPL_Util.createMutableNamedValues();
    String theLine = null, theName = null, theValue = null;
    int delimPtr = -1;
    
    
    for(;;) {
      try {
        if ((theLine = from.readLine()) == null) break;
      } catch (IOException ioError) { return null; }
      
      // Find the delimiter -- skip the line if none
      if ((delimPtr = theLine.indexOf("=")) == -1) continue;
      theName = theLine.substring(0, delimPtr);
      theValue = theLine.substring(delimPtr + 1);
      
      // Validate the name
      if (theName.length() == 0) continue;
      
      // Add this in
      namedValues.addNamedValue(theName, theValue);
    }

    // Return the named values
    return namedValues;
  }
  
  // Parse the named/values for option/config/reconfig items, define
  // the configuration items in the device approrpriatly and remove
  // those items from the named values set.  If there are any fatal
  // errors, return false
  public boolean parseConfigDefinitions(xPL_DeviceI theDevice, NamedValuesI namedValues) {
    // Now we pick the config item definitions off
    String theName = null, theValue = null, theParam = null;
    int maxValues = -1, startDelim = -1, endDelim = -1;
    boolean isMandatory = false, isReconfigurable = false;
    xPL_DeviceConfigItemI configItem = null;

    Iterator<NamedValueI> allNamedValues = namedValues.getAllNamedValues().iterator();
    NamedValueI theNamedValue = null;
    while(allNamedValues.hasNext()) {
      theNamedValue = allNamedValues.next();
      theName = theNamedValue.getName();
      theValue = theNamedValue.getValue();
      
      // See if this is the device is configurable flag
      if (theName.equalsIgnoreCase("deviceIsConfiguable")) {
        theDevice.setConfigurable(theValue.equals("1"));
        allNamedValues.remove();
        continue;
      }
      
      // See if these are configuration item definitions and skip if not
      if (theName.equalsIgnoreCase("config")) {
        isMandatory = true;
        isReconfigurable = false;
      } else if (theName.equalsIgnoreCase("reconf")) {
        isMandatory = true;
        isReconfigurable = true;
      } else if (theName.equalsIgnoreCase("option")) {
        isMandatory = false;
        isReconfigurable = true;
      } else {
        // Not one of our tags
        continue;
      }
        
      // Take our tag out of the list of tags
      allNamedValues.remove();
      
      // Get the config item name and parse for a max values indicator
      maxValues = 1;
      if ((startDelim = theValue.indexOf("[")) != -1) {
        // Skip if no end delimiter
        if ((endDelim = theValue.indexOf("]", startDelim + 1)) == -1) continue;
        theParam = theValue.substring(startDelim + 1, endDelim);
        theValue = theValue.substring(0, startDelim);
        if (theParam.length() != 0) {
          try {
            maxValues = Integer.parseInt(theParam);
          } catch (Exception anError) { continue; }
        } else {
          maxValues = 1;
        }
      }
      
      // Handle filters and groups
      if (theValue.equalsIgnoreCase("filter")) {
        theDevice.setMaxFilterCount(maxValues); 
        continue;
      } else if (theValue.equalsIgnoreCase("group")) {
        theDevice.setMaxGroupCount(maxValues);
        continue;
      }

      // Get existing config item, if any, or create one
      if ((configItem = theDevice.getConfigurable(theValue)) == null) {
        configItem = theDevice.addConfigurable(theValue, maxValues, isMandatory, isReconfigurable);
      } else {
        configItem.setMaxValueCount(maxValues);
        configItem.setMandatory(isMandatory);
        configItem.setReconfigurable(isReconfigurable);
      }
    }  
     
    return true;
  }
  
  
  /**
   * Restore the passed device from the passed input stream.  
   * 
   * If the device has a vendor/device in it, then the stream will be read until
   * a matching device/device is found.  If the device has no vendor/device
   * in it, then the first device found in the stream is used and installed.
   * 
   * @param theDevice device to restore state into
   * @param fromStream stream to read state from
   * @return true if any state was restored, false if state not changed in device
   */
  boolean restoreDevice(xPL_DeviceI theDevice, InputStream fromStream) {
    // Make sure we even have a chance
    if ((theDevice == null) || (fromStream == null)) return false;
    
    xPL_IdentifierI theIdent = theDevice.getIdent();
    BufferedReader deviceReader = new BufferedReader(new InputStreamReader(fromStream));
    String theLine = null;
    
    // Look for a header
    synchronized(theDevice) {
      for(;;) {
        // Read a line, bailing out if we hit end of the stream
        try {
          if ((theLine = deviceReader.readLine()) == null) {
            return false;
          }
        } catch (IOException ioError) { 
          return false;
        }

        // See if this line is a header
        if (!theLine.startsWith("[")) continue;

        // Extract the device name
        xPL_IdentifierI aDevice = xPL_Manager.getManager().getIdentifierManager().parseNamedIdentifier(theLine.substring(1, theLine.indexOf("]")));
        if (aDevice == null) {
          continue;
        }

        // See if this is a match 
        if (theIdent != null) {
          // If this is a match, we have our header
          if (theIdent.getVendorID().equalsIgnoreCase(aDevice.getVendorID()) && theIdent.getDeviceID().equalsIgnoreCase(aDevice.getDeviceID())) break;

          // Otherwise, try again
          continue;
        } else {
          // Install this as the device device
          theDevice.setIdent(aDevice);
          break;
        }
      }
    }
    
    // Parse the named values
    MutableNamedValuesI namedValues = parseNamedValues(deviceReader);
    if (namedValues == null) {
      return false;
    }
    
    // Parse out configuration item definitions
    parseConfigDefinitions(theDevice, namedValues);
    
    // Install the actual configuration values
    ((xPL_Device) theDevice).installConfigValues(namedValues);

    // And we are done
    return true;
  }
  
  /**
   * Restore the passed device from the passed input file
   * 
   * If the device has a vendor/device in it, then the stream will be read until
   * a matching deviceis found.  If the device has no vendor/device
   * in it, then the first device found in the stream is used and installed.
   * 
   * @param theDevice device to restore state into
   * @param fromFile file name to read from
   * @return true if any state was restored, false if state not changed in device
   */
  boolean restoreDevice(xPL_DeviceI theDevice, File fromFile) {
    // Make sure we have a chance
    if (!fromFile.exists() || !fromFile.canRead()) return false;
    
    FileInputStream fromStream = null;
    try {
      fromStream = new FileInputStream(fromFile);
      return restoreDevice(theDevice, fromStream);
    } catch (Exception anyError) { 
      anyError.printStackTrace();
      return false;
    } finally {
      if (fromStream != null) try { fromStream.close(); } catch (IOException ignoredError) { }
    }
  }
  
  /**
   * Write the current state of this device out to the passed stream.
   * 
   * @param theDevice device to write state out to
   * @param toStream stream to write state out to
   * @return true if state was written, false if device was not preservable
   */
  boolean preserveDevice(xPL_DeviceI theDevice, OutputStream toStream) {
    synchronized(theDevice) {
      // Insure we can preserve this
      if ((theDevice == null) || (toStream == null) || (theDevice.getIdent() == null)) return false;

      // Get a writable stream
      PrintWriter deviceWriter = new PrintWriter(toStream);

      // Write the basics for the device
      deviceWriter.print("["); deviceWriter.print(theDevice.getIdent().toString()); deviceWriter.println("]");
      deviceWriter.print("deviceIsConfigurable="); deviceWriter.println(theDevice.isConfigurable() ? "1" : "0");
      deviceWriter.print("newconf=");   deviceWriter.println(theDevice.getIdent().getInstanceID());
      deviceWriter.print("interval="); deviceWriter.println(theDevice.getHeartbeatInterval());

      // Write any filters
      deviceWriter.print("option=filter["); deviceWriter.print(theDevice.getMaxFilterCount()); deviceWriter.println("]");
      if (theDevice.getFilterCount() == 0)
        deviceWriter.println("filter=");
      else {
        for(xPL_FilterI theFilter: theDevice.getFilters()) {
          deviceWriter.print("filter=");
          deviceWriter.println(theFilter.toString());
        }
      }

      // Write any groups
      deviceWriter.print("option=group["); deviceWriter.print(theDevice.getMaxGroupCount()); deviceWriter.println("]");
      if (theDevice.getGroupCount() == 0) 
        deviceWriter.println("group=");
      else {
        for (xPL_IdentifierI theGroup: theDevice.getGroups()) {
          deviceWriter.print("group=");
          deviceWriter.println(theGroup.toString());
        }
      }

      // Write the configurables out
      for (xPL_DeviceConfigItemI configItem: theDevice.getConfigurables()) {
        // Write config item header
        if (configItem.isMandatory()) {
          if (configItem.isReconfigurable())
            deviceWriter.print("reconf=");
          else
            deviceWriter.print("config=");
        } else {
          deviceWriter.print("option=");
        }
        deviceWriter.print(configItem.getName());
        if (configItem.getMaxValueCount() > 1) {
          deviceWriter.print("[");
          deviceWriter.print(configItem.getMaxValueCount());
          deviceWriter.print("]");
        }
        deviceWriter.println();

        // Write current values out
        if (configItem.getValues() != null) {
          for (String itemValue: configItem.getValues()) {
            deviceWriter.print(configItem.getName());
            deviceWriter.print("=");
            deviceWriter.println(itemValue);
          }
        }
      }

      // Flush it and we are done
      deviceWriter.flush();

      // And we are done
      return true;
    }
  }

  /**
   * Preserve the current device
   * 
   * @param theDevice device to preserve
   * @param toFile file name to preserve to
   * @return true if any state was preserved, false if not
   */
  boolean preserveDevice(xPL_DeviceI theDevice, File toFile) {
    FileOutputStream toStream = null;
    try {
      toStream = new FileOutputStream(toFile);
      return preserveDevice(theDevice, toStream);
    } catch (Exception anyError) { 
      return false;
    } finally {
      if (toStream != null) {
        try { toStream.flush(); } catch (IOException ignoredError) { }
        try { toStream.close(); } catch (IOException ignoredError) { }
      }
    }
  }
}
