/*
 * HeartbeatManager.java
 * Copyright 2004, Gerald R Duprey Jr
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy of
 * the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0 
 *
 * Unless required by applicable law or agreed to in writing, software distributed
 * under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
 * CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License. 
 */

package org.cdp1802.xpl.device;

import java.util.ArrayList;

import org.cdp1802.xpl.*;


/**
 * Manage periodic heartbeat sending for devices
 *
 * @author gduprey
 */
class HeartbeatManager implements Runnable {
  xPL_Manager theManager = null;
  ArrayList<xPL_DeviceI> managedDevices = null;
  boolean managerStarted = false;
  boolean hubsConfirmed = false;
  Thread managerThread = null;
  
  HeartbeatManager(ArrayList<xPL_DeviceI> managedDevices) {
    this.managedDevices = managedDevices;
    theManager = xPL_Manager.getManager();
  }

  void debug(String theMessage) {
    System.out.println("HBEAT_MGR:: " + theMessage);
  }
  
  public void start() {
    if (managerStarted) return;

    synchronized(managedDevices) {
      // Create thread and start it
      new Thread(this, "xPL Heartbeat Device Manager").start();

      // Wait for thread to signal it's running
      while(!managerStarted) {
        try {
          managedDevices.wait();
        } catch (Exception anyException) { }
      }
    }
  }

  public void stop() {
    if (!managerStarted) return;

    synchronized(managedDevices) {
      // Stop the devices and awaken the sleepers
      managerStarted = false;
      managedDevices.notifyAll();

      // Wait for thread to exit
      try {
        managedDevices.wait();
      } catch (Exception anyError) { }
    }
  }

  // Give each managed device a chance to
  // send out a heartbeat
  void doHeartbeatUpdates() {
    for (xPL_DeviceI theDevice: managedDevices) {
      // If hubs aren't confirmed, then force the hbeat.  Otherwise, send
      // only if it's time to 
      theDevice.sendTimelyHeartbeat(!hubsConfirmed);
    }
  }

  public void run() {
    long waitTime = 0;
    
    // Track when we started sending heartbeats
    long hubDiscoveryWaitStartTime = System.currentTimeMillis();
    
    synchronized(managedDevices) {
      // Flag we are running and notify start thread
      managerStarted = true;
      managedDevices.notifyAll();
      
      // Send out any pending 
      doHeartbeatUpdates();
      
      // Go into a sleep loop (once a minute we wake up)
      while(managerStarted) {
        // See if we'return confirmed or not
        if (hubsConfirmed != theManager.areHubsConfirmed()) {
          // Cache the status of this
          hubsConfirmed = theManager.areHubsConfirmed();

          // Once confirmed, we settle back to once-a-minute timers
          // Otherwise, reset the timer that tracks how long we've been
          // attempting to confirm the hubs
          if (hubsConfirmed) {
            waitTime = 60000;
          } else {
            hubDiscoveryWaitStartTime = System.currentTimeMillis();
          }
        }

        // If hubs are not yet confirmed, then figure out how long we should wait
        if (!hubsConfirmed)  {
          // If it's been over 2 minutes without sign of a hub, drop down
          // how agressive we are about soliciting one
          waitTime = ((System.currentTimeMillis() - hubDiscoveryWaitStartTime) <= 120000) ? 3000 : 30000;
        }

        // Wait a moment
        try {
          managedDevices.wait(waitTime);
        } catch (Throwable anyError) { }

        // Do Heartbeat checks
        doHeartbeatUpdates();
      }

      // Signal we are stopping
      managedDevices.notifyAll();
    }
  }
}
