/*
 * xPL_Device.java
 * Copyright 2004, Gerald R Duprey Jr
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy of
 * the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0 
 *
 * Unless required by applicable law or agreed to in writing, software distributed
 * under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
 * CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License. 
 */

package org.cdp1802.xpl.device;

import java.io.File;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import org.cdp1802.xpl.IdentifierManager;
import org.cdp1802.xpl.NamedValueI;
import org.cdp1802.xpl.NamedValuesI;
import org.cdp1802.xpl.xPL_IdentifierI;
import org.cdp1802.xpl.xPL_Manager;
import org.cdp1802.xpl.xPL_MessageI;
import org.cdp1802.xpl.xPL_MessageListenerI;
import org.cdp1802.xpl.xPL_MutableMessageI;
import org.cdp1802.xpl.xPL_Util;
import org.cdp1802.xpl.xPL_Filter;
import org.cdp1802.xpl.xPL_FilterI;

/**
 *
 * @author  gerry
 */
public class xPL_Device implements xPL_DeviceI, xPL_MessageListenerI {
  static xPL_Manager xplManager = xPL_Manager.getManager();
  static IdentifierManager identifierManager = xplManager.getIdentifierManager();
  static DeviceManager deviceManager = DeviceManager.getManager();

  class HeartbeatResponder implements Runnable {
    xPL_Device theDevice = null;
    
    HeartbeatResponder(xPL_Device theDevice) {
      this.theDevice = theDevice;
    }

    public void respond() {
      new Thread(this, "Heartbeat Responder").start();
    }
    
    public void run() {
      // Delay from .5 to 2.5 seconds
      long delayTime = (long) (Math.random() * 2000) + 500;
      try {
        Thread.currentThread().sleep(delayTime);
      } catch (Exception anyError) { }
      
      // Send the message
      xplManager.sendHeartbeat(theDevice, true);
      theDevice.heartbeatLastSentAt = System.currentTimeMillis();
    }
    
    public void releaseResources() {
      theDevice = null;
    }
  }
  
  xPL_IdentifierI deviceIdent = null;
  int heartbeatInterval = 5;
  long heartbeatLastSentAt = 0;
  boolean deviceEnabled = false;
  int maxFilterCount = 255;
  int maxGroupCount = 255;
  String deviceVersion = null;
  
  HeartbeatResponder heartbeatResponder = null;
  
  boolean deviceIsConfigurable = false;
  boolean deviceIsConfigured = false;
  boolean deviceIsBeingConfigured = false;
  boolean deviceRestored = false;
  boolean passAllMessages = false;
  File autoPreserveFile = null;
  
  ArrayList<xPL_DeviceMessageListenerI> messageListeners = new ArrayList<xPL_DeviceMessageListenerI>();
  ArrayList<xPL_DeviceChangeListenerI> changeListeners = new ArrayList<xPL_DeviceChangeListenerI>();
  ArrayList<xPL_FilterI> deviceFilters = new ArrayList<xPL_FilterI>();
  HashMap<String, xPL_IdentifierI> deviceGroups = new HashMap<String, xPL_IdentifierI>();
  HashMap<String, xPL_DeviceConfigItemI> configItems = new HashMap<String, xPL_DeviceConfigItemI>();
  
  xPL_Device() { }
  
  xPL_Device(xPL_IdentifierI deviceIdent) {
    setIdent(deviceIdent);
  }
  
  xPL_Device(xPL_IdentifierI devuceIdent, boolean configurableDevice) {
    setIdent(deviceIdent);
    setConfigurable(configurableDevice);
  }

  void debug(String theMessage) {
    System.out.println("DEVICE[" + (deviceIdent == null ? "<<UNKNOWN>>" : deviceIdent.toString()) + "]:: " + theMessage);
  }
  
  public xPL_IdentifierI getIdent() { return deviceIdent;  }
  public void setIdent(xPL_IdentifierI theIdentifier) { 
    if (theIdentifier != null) deviceIdent = theIdentifier;
  }
  public void setIdent(String vendorID, String deviceID, String instanceID) { 
    xPL_IdentifierI theIdentifier = xplManager.getIdentifierManager().getInstance(vendorID, deviceID, instanceID);
    if (theIdentifier != null) deviceIdent = theIdentifier;
  }
  
  public int getHeartbeatInterval() { return heartbeatInterval; }
  public void setHeartbeatInterval(int newHeartbeatInterval) { 
  if (newHeartbeatInterval < 1)
    heartbeatInterval = 5;
  else if (newHeartbeatInterval > 30)
    heartbeatInterval = 30;
  else
    heartbeatInterval = newHeartbeatInterval;
  }

  int minutesSinceLastHeartbeat() {
    if (heartbeatLastSentAt == 0) return 0;
    return (int) ((System.currentTimeMillis() - heartbeatLastSentAt) / 60000);
  }
  
  public void sendTimelyHeartbeat(boolean forceHeartbeat) {
    // See if we should send a heartbeat
    int minutesElapsed = minutesSinceLastHeartbeat();
    boolean timeToSendHeartbeat = (heartbeatLastSentAt == 0) || (deviceIsBeingConfigured && (minutesElapsed >= 1)) || (minutesElapsed >= heartbeatInterval);
    if (forceHeartbeat || timeToSendHeartbeat) {
      // Send a heartbeat
      xplManager.sendHeartbeat(this, timeToSendHeartbeat);
      
      // Mark heartbeat last sent at
      if (timeToSendHeartbeat) heartbeatLastSentAt = System.currentTimeMillis();
    }
  }
  public void sendTimelyHeartbeat() { sendTimelyHeartbeat(false); }
  
  public boolean isEnabled() { return deviceEnabled; }
  public void setEnabled(boolean enableDevice) { 
    // Skip if no change
    if (enableDevice == deviceEnabled) return;
  
    // Enable the device
    if (enableDevice) {
      // Insure we can be started
      if (deviceIdent == null) return;
      
      // If this is a configurable device, flag it for configuration
      if (deviceIsConfigurable && !deviceIsConfigured) {
        deviceIsBeingConfigured = true;
      }
      
      // Add oursevles as message listeners
      xplManager.addMessageListener(this);
      heartbeatLastSentAt = 0;
      deviceEnabled = true;
      deviceManager.addManagedDevice(this);
      fireDeviceChanged(xPL_DeviceChangeEventI.ChangeReason.DEVICE_ENABLED);
      return;
    }
    
    // Disable the device
    xplManager.removeMessageListener(this);
    deviceManager.removeManagedDevice(this);
    
    // Mark us disabled, send the "end" heartbeat and clear any transient state
    deviceEnabled = false;
    xplManager.sendHeartbeat(this, true);
    deviceIsBeingConfigured = false;
    fireDeviceChanged(xPL_DeviceChangeEventI.ChangeReason.DEVICE_DISABLED);
  }

  public void addMessageListener(xPL_DeviceMessageListenerI theListener) { 
    synchronized(this) {
      if (messageListeners.contains(theListener)) return;
      messageListeners.add(theListener);
    }
  }

  public void removeMessageListener(xPL_DeviceMessageListenerI theListener) { 
    synchronized(this) {
      messageListeners.remove(theListener);
    }
  }

  public void setPassAllMessages(boolean passAll) {
    passAllMessages = passAll;
  }
  public boolean getPassAllMessages() { return passAllMessages; }
  
  
  
  /******** Device Change Support **********/
  
  public void addDeviceChangeListener(xPL_DeviceChangeListenerI theListener) {
    synchronized(this) {
      if (changeListeners.contains(theListener)) return;
      changeListeners.add(theListener);
    }
  }
  
  public void removeDeviceChangeListener(xPL_DeviceChangeListenerI theListener) { 
    synchronized(this) {
      changeListeners.remove(theListener);
    }
  }
  
  protected void fireDeviceChanged(xPL_DeviceChangeEventI.ChangeReason theReason, Object extraInfo) {
    xPL_DeviceChangeEvent changeEvent = null;
    
    if (extraInfo instanceof xPL_IdentifierI)
      changeEvent = new xPL_DeviceChangeEvent(this, theReason, (xPL_IdentifierI) extraInfo);
    else if (extraInfo instanceof xPL_FilterI)
      changeEvent = new xPL_DeviceChangeEvent(this, theReason, (xPL_FilterI) extraInfo);
    else if (extraInfo instanceof xPL_DeviceConfigItemI)
      changeEvent = new xPL_DeviceChangeEvent(this, theReason, (xPL_DeviceConfigItemI) extraInfo);
    else
      changeEvent = new xPL_DeviceChangeEvent(this, theReason);
    
    for (xPL_DeviceChangeListenerI theListener: changeListeners) {
      try {
        theListener.handleXPLDeviceChange(changeEvent);
      } catch (Throwable anyError) { }
    }
    changeEvent.releaseResources();
  }
  
  protected void fireDeviceChanged(xPL_DeviceChangeEventI.ChangeReason theReason) {
    fireDeviceChanged(theReason, null);
  }

  
  /****** FILTER Support ********/
  
  public void clearFilters() {
    synchronized(this) {
      deviceFilters.clear();
    }
    fireDeviceChanged(xPL_DeviceChangeEventI.ChangeReason.FILTERS_CLEARED);
  }

  public void setMaxFilterCount(int maxAllowedFilters) {
    synchronized(this) {
      if ((maxAllowedFilters < 0) || (maxAllowedFilters == maxFilterCount)) return;

      maxFilterCount = maxAllowedFilters;
      while (deviceFilters.size() > maxFilterCount) deviceFilters.remove(deviceFilters.size() - 1);
    }
  }
  
  public int getMaxFilterCount() { return maxFilterCount; }
  
  public int getFilterCount() { return deviceFilters.size(); }
  
  public xPL_FilterI addFilter(xPL_FilterI aFilter) {
    if (aFilter == null) return null;
    synchronized(this) {
      if (deviceFilters.size() >= maxFilterCount) return null;
      deviceFilters.add(aFilter);
    }
    fireDeviceChanged(xPL_DeviceChangeEventI.ChangeReason.FILTER_ADDED, aFilter);
    return aFilter;
  }
  
  public xPL_FilterI addFilter(String filterSpec) { 
    if ((filterSpec == null) || (filterSpec.length() == 0)) return null;
    return addFilter(new xPL_Filter(filterSpec));
  }

  public Collection<xPL_FilterI> getFilters() { 
    return deviceFilters;
  }

  // Return true if at least one filter matches.  If there
  // are no filters, then we always return true
  protected boolean doesMessageMatchAFilter(xPL_MessageI theMessage) {
    // If there are no filters, we match
    if ((deviceFilters == null) || (deviceFilters.size() == 0)) return true;
    
    // Check each filter
    for (xPL_FilterI theFilter: deviceFilters) {
      if (theFilter.doesMessageMatchFilter(theMessage)) return true;
    }
    
    // No match
    return false;
  }
  
  
  /******** GROUP Support *********/
  
  public void setMaxGroupCount(int maxAllowedGroups) {
    synchronized(this) {
      if ((maxAllowedGroups < 0) || (maxAllowedGroups == maxGroupCount)) return;

      maxGroupCount = maxAllowedGroups;
    }
  }
  
  public int getMaxGroupCount() { return maxGroupCount; }
  
  public int getGroupCount() { return deviceGroups.size(); }

  public void clearGroups() { 
    synchronized(this) {
      deviceGroups.clear();
    }
    fireDeviceChanged(xPL_DeviceChangeEventI.ChangeReason.GROUPS_CLEARED);
  }
  
  public xPL_IdentifierI addGroup(xPL_IdentifierI theGroup) {
    if ((theGroup == null) || !theGroup.isGroupIdentifier()) return null;
    synchronized(this) {
      if (deviceGroups.size() >= maxGroupCount) return null;
      deviceGroups.put(theGroup.getGroupName().toLowerCase(), theGroup);
    }
    fireDeviceChanged(xPL_DeviceChangeEventI.ChangeReason.GROUP_ADDED, theGroup);
    return theGroup;
  }

  public xPL_IdentifierI addGroup(String groupName) {
    if ((groupName == null) || (groupName.length() == 0)) return null;
    
    xPL_IdentifierI theGroup = xPL_Manager.getManager().getIdentifierManager().getGroupIdentifier(groupName);
    if (theGroup == null) return null;
    return addGroup(theGroup);
  }

  public boolean isAssignedToGroup(xPL_IdentifierI theGroup) {
    if ((theGroup == null) || !theGroup.isGroupIdentifier()) return false;
    return deviceGroups.containsKey(theGroup.getGroupName().toLowerCase());
  }
  
  public boolean isAssignedToGroup(String theGroupName) {
    if ((theGroupName == null) || (theGroupName.length() == 0)) return false;
    xPL_IdentifierI theGroup = xPL_Manager.getManager().getIdentifierManager().getGroupIdentifier(theGroupName);
    if (theGroup == null) return false;
    return isAssignedToGroup(theGroup);
  }
  
  public Collection<xPL_IdentifierI> getGroups() {  
    return deviceGroups.values();
  }

  // Return true if message matches any group.  If the message is
  // not a group message, return false.  If there are no groups, 
  // always return true
  protected boolean doesMessageMatchAGroup(xPL_MessageI theMessage) {
    if (deviceGroups.size() == 0) return true;
    return isAssignedToGroup(theMessage.getTarget());
  }
  
  /******* Version Support **********/

  /**
   * Install a version number for this device.  This can be any string or by null to 
   * clear out any version number.  The version number, if non-null, will be included
   * with each hearbeat sent out
   *
   * @param theVersion version string for this devices device
   */
  public void setVersion(String theVersion) {
    deviceVersion = theVersion;
  }
  
  /**
   * Return the most recently set version or null if there is no version string
   *
   * @return version string for this device or null
   */
  public String getVersion() { return deviceVersion; }

  
  /******* Configuration Support **********/
  
  public void clearConfigurables() {
    synchronized(this) {
      configItems.clear();
    }
    fireDeviceChanged(xPL_DeviceChangeEventI.ChangeReason.CONFIGURATION_ITEMS_CLEARED);
  }

  public int getConfigurableCount() {
    return configItems.size();
  }
  
  public xPL_DeviceConfigItemI getConfigurable(String configName) {
    if ((configName == null) || (configName.length() == 0)) return null;
    synchronized(this) {
      return configItems.get(configName.toLowerCase());
    }
  }
  
  /**
   * Read a simple configuration element and return it's value.  If 
   * the config element doesn't exist or has no value, return the 
   * defaultValue.
   *
   * If the configuration item can have more than one value, only the
   * very first one is returned (used getConfigurable() to access all aspects
   * of configurable items).
   *
   * @param configName name of the configuration element to return
   * @param defaultValue value to return if config element doesn't exist or has no value assigned
   * @return the current value of the configuration element or the default value
   */
  public String getConfigValue(String configName, String defaultValue) {
    xPL_DeviceConfigItemI theItem = getConfigurable(configName);
    if (theItem == null) return defaultValue;
    if (theItem.getValueCount() == 0) return defaultValue;
    return theItem.getValue();
  }

  /**
   * Read a simple configuration element and return it's value.  If 
   * the config element doesn't exist or has no value, return null.
   *
   * If the configuration item can have more than one value, only the
   * very first one is returned (used getConfigurable() to access all aspects
   * of configurable items).
   *
   * @param configName name of the configuration element to return
   * @return the current value of the configuration element or null
   */
  public String getConfigValue(String configName) {
    return getConfigValue(configName, null);
  }
  
  /**
   * Read a simple configuration element and return it's value.  If 
   * the config element doesn't exist or has no value, return the 
   * defaultValue.
   *
   * If the configuration item can have more than one value, only the
   * very first one is returned (used getConfigurable() to access all aspects
   * of configurable items).
   *
   * @param configName name of the configuration element to return
   * @param defaultValue value to return if config element doesn't exist or has no value assigned
   * @return the current value of the configuration element or the default value
   */
  public int getIntConfigValue(String configName, int defaultValue) {
    xPL_DeviceConfigItemI theItem = getConfigurable(configName);
    if (theItem == null) return defaultValue;
    if (theItem.getValueCount() == 0) return defaultValue;
    return theItem.getIntValue();
  }
  
  /**
   * Read a simple configuration element and return it's value.  If 
   * the config element doesn't exist or has no value, return -1
   *
   * If the configuration item can have more than one value, only the
   * very first one is returned (used getConfigurable() to access all aspects
   * of configurable items).
   *
   * @param configName name of the configuration element to return
   * @return the current value of the configuration element or -1
   */
  public int getIntConfigValue(String configName) {
    return getIntConfigValue(configName, -1);
  }

  /**
   * Read a simple configuration element and return it's value.  If 
   * the config element doesn't exist or has no value, return the 
   * defaultValue.
   *
   * If the configuration item can have more than one value, only the
   * very first one is returned (used getConfigurable() to access all aspects
   * of configurable items).
   *
   * @param configName name of the configuration element to return
   * @param defaultValue value to return if config element doesn't exist or has no value assigned
   * @return the current value of the configuration element or the default value
   */
  public boolean getBoolConfigValue(String configName, boolean defaultValue) {
    xPL_DeviceConfigItemI theItem = getConfigurable(configName);
    if (theItem == null) return defaultValue;
    if (theItem.getValueCount() == 0) return defaultValue;
    return theItem.getBoolValue();
  }

  /**
   * Read a simple configuration element and return it's value.  If 
   * the config element doesn't exist or has no value, return false
   *
   * If the configuration item can have more than one value, only the
   * very first one is returned (used getConfigurable() to access all aspects
   * of configurable items).
   *
   * @param configName name of the configuration element to return
   * @return the current value of the configuration element or false
   */
  public boolean getBoolConfigValue(String configName) {
    return getBoolConfigValue(configName, false);
  }
  
  public xPL_DeviceConfigItemI addConfigurable(xPL_DeviceConfigItemI configItem) {
    if (configItem == null) return null;
    synchronized(this) {
      configItems.put(configItem.getName().toLowerCase(), configItem);
    }
    fireDeviceChanged(xPL_DeviceChangeEventI.ChangeReason.CONFIGURATION_ITEM_ADDED, configItem);
    return configItem;
  }
  
  public xPL_DeviceConfigItemI addConfigurable(String itemName, int maxValueCount, boolean isMandatory, boolean isReconfigurable) {
    if ((itemName == null) || (itemName.length() == 0)) return null;
    return addConfigurable(new xPL_DeviceConfigItem(itemName, maxValueCount, isMandatory, isReconfigurable));
  }

  public Collection<xPL_DeviceConfigItemI> getConfigurables() {
    synchronized(this) {
      return configItems.values();
    }
  }

  public void setConfigurable(boolean deviceConfigurable) { 
    // Skip if no change
    if (deviceConfigurable == deviceIsConfigurable) return;
    
    deviceIsConfigurable = deviceConfigurable;
  }
  public boolean isConfigurable() { return deviceIsConfigurable;  }
  public boolean isBeingConfigured() { return deviceIsBeingConfigured;  }
  
  
  public boolean isConfigured() { return deviceIsConfigured; }
  public void setConfigured(boolean deviceConfigured) {
    // Skip if no change
    if (deviceIsConfigured == deviceConfigured) return;
    
    // Handle a change when the device is inactive
    if (!deviceEnabled || !deviceConfigured) {
      deviceIsConfigured = deviceConfigured;
      deviceIsBeingConfigured = false;
      return;
    }
    
    // If we were being configured, we no longer are, so handle the transition
    // to a configured state.
    if (deviceIsBeingConfigured || !deviceIsConfigured) {
      deviceIsBeingConfigured = false;
      deviceIsConfigured = true;

      fireDeviceChanged(xPL_DeviceChangeEventI.ChangeReason.CONFIGURATION_CHANGED);
      fireDeviceChanged(xPL_DeviceChangeEventI.ChangeReason.DEVICE_CONFIGURED);

      heartbeatLastSentAt = 0;
      sendTimelyHeartbeat();
    }
  }

  /**
   * Return true if the device has been successfully restored from a configuration
   * file.  In cases, of new devices or non-configurable devices, this will be
   * false.  If there was any error restoring the device, it will also be false.
   *
   * Usually, if this is true, isConfigured() will also be true, but there isn't
   * a direct connection, so donb't make assumptions.
   *
   * @return true if the device was successfully restored from a configuration file
   */
  public boolean wasRestored() { return deviceRestored; }  
  
  // Send the list of configurable items for this device
  protected void sendConfigItemsList() {
    xPL_MutableMessageI responseMessage = createBroadcastMessage(xPL_MessageI.MessageType.STATUS, "config", "list");

    synchronized(this) {
      // Add stock configurables
      responseMessage.addNamedValue("reconf", "newconf");
      responseMessage.addNamedValue("option", "interval");

      if (maxGroupCount > 0) responseMessage.addNamedValue("option", "group[" + maxGroupCount + "]");
      if (maxFilterCount > 0) responseMessage.addNamedValue("option", "filter[" + maxFilterCount + "]");

      // Add in additional device configurables
      String confName, confFlag;
      for (xPL_DeviceConfigItemI theItem: configItems.values()) {
        // Extract the item name
        if (theItem.getMaxValueCount() > 1)
          confName = theItem.getName() + "[" + theItem.getMaxValueCount() + "]";
        else
          confName = theItem.getName();

        // Determine what sort of item it is
        if (theItem.isMandatory()) {
          if (theItem.isReconfigurable())
            confFlag = "reconf";
          else
            confFlag = "config";
        } else
          confFlag = "option";

        // Add it in
        responseMessage.addNamedValue(confFlag, confName);
      }
    }
    
    // Send the message
    sendMessage(responseMessage);
    responseMessage.releaseResources();
  }

  // Send the state of the current configuration
  protected void sendConfigItemValues() {
    xPL_MutableMessageI responseMessage = createBroadcastMessage(xPL_MessageI.MessageType.STATUS, "config", "current");

    synchronized(this) {
      // Add stock configurables
      responseMessage.addNamedValue("newconf", deviceIdent.getInstanceID());
      responseMessage.addNamedValue("interval", heartbeatInterval);

      // Send current filters
      if (maxFilterCount > 0) {
        if (deviceFilters.size() == 0)
          responseMessage.addNamedValue("filter", "");
        else {
          for (xPL_FilterI theFilter: deviceFilters) {
            responseMessage.addNamedValue("filter", theFilter.toString());
          }
        }
      }

      // Send current groups
      if (maxGroupCount > 0) {
        if (deviceGroups.size() == 0)
          responseMessage.addNamedValue("group", "");
        else {
          for (xPL_IdentifierI theGroup: deviceGroups.values()) {
            responseMessage.addNamedValue("group", theGroup.toString());
          }
        }
      }

      // Add in additional device configurables
      String confName, confFlag;
      for (xPL_DeviceConfigItemI theItem: configItems.values()) {
        if (theItem.getValueCount() == 0) {
          responseMessage.addNamedValue(theItem.getName(), "");
        } else {
          for (String theValue: theItem.getValues()) {
            responseMessage.addNamedValue(theItem.getName(), theValue);
          }
        }
      }
    }
    
    // Send the message
    sendMessage(responseMessage);
    responseMessage.releaseResources();
  }

  // Install configuration from the passed config values
  // configValues is the list to extract from
  // settingDefaults true suppresses the installation marking the device as configured
  // updateOnly true disables setting instance ID, interface, filters and groups
  public boolean installConfigValues(NamedValuesI configValues, boolean settingDefaults, boolean updateOnly) {
    boolean heartbeatNeeded = false;
    boolean deviceJustConfigured = false;
    
    synchronized(this) {
      if (!updateOnly) {      
        // See if there is a new instance name
        String instanceID = configValues.getNamedValue("newconf");
        if ((instanceID != null) && !instanceID.equalsIgnoreCase(deviceIdent.getInstanceID())) {
          // Logoff with the old instance ID
          if (deviceEnabled) {
            deviceEnabled = false;
            xplManager.sendHeartbeat(this, true);
            deviceEnabled = true;
            heartbeatNeeded = true;
          }

          // Attempt to update/reconfigure the device and if we do, flag us needing a new heartbeat
          xPL_IdentifierI reconfigDevice = identifierManager.reconfigureInstanceID(deviceIdent, instanceID);
          if (reconfigDevice != null) {
            deviceIdent = reconfigDevice;
            heartbeatNeeded = true;
          }
        }

        // See if there is a new interval
        int newInterval = configValues.getIntNamedValue("interval", -1);
        if (newInterval != -1) setHeartbeatInterval(newInterval);

        // Check for filters
        if (configValues.hasNamedValue("filter")) {
          // Clear any current filters
          clearFilters();

          // Install new filters
          for (String theFilter: configValues.getNamedValues("filter")) {
            if ((theFilter == null) || (theFilter.length() == 0)) continue;
            addFilter(theFilter);
          }
        }

        // Check for groups
        if (configValues.hasNamedValue("group")) {
          // Clear any current groups
          clearGroups();

          // Install new groups
          for (String theGroup: configValues.getNamedValues("group")) {
            if ((theGroup == null) || (theGroup.length() == 0)) continue;
            addGroup(identifierManager.parseNamedIdentifier(theGroup));
          }
        }
      }

      // Iterate over all the remaining values and install them as config
      // items, if they in fact match a config item
      xPL_DeviceConfigItemI theConfigItem = null;
      String theName = null;
      HashMap<String, xPL_DeviceConfigItemI>configuredList = null;
      
      for(NamedValueI namedValue: configValues.getAllNamedValues()) {
        // Get the config item and cache the name
        theName = namedValue.getName();
        
        // Lookup the config item.  If it's unknown, skip it
        if ((theConfigItem = configItems.get(theName.toLowerCase())) == null) continue;

        // See if this is a reconfiguration and the item is not reconfigurable.
        // If it's not reconfigurable in that case, then skip it.
        if (!theConfigItem.isReconfigurable() && !deviceIsBeingConfigured) continue;

        // See if we have our list of configured items and init it if not
        if (configuredList == null) configuredList = new HashMap<String, xPL_DeviceConfigItemI>();

        // If this is the first instance of this config item, clear any values
        if (!configuredList.containsKey(theName)) {
          theConfigItem.clearValues();
          configuredList.put(theName, theConfigItem);
        }

        // Add the value on
        theConfigItem.addValue(namedValue.getValue());
      }

      // If we were being configured, we no longer are
      if (!settingDefaults && !updateOnly) {
        if (deviceIsBeingConfigured || !deviceIsConfigured) {
          heartbeatNeeded = deviceIsBeingConfigured;
          deviceIsBeingConfigured = false;
          deviceIsConfigured = true;
          deviceJustConfigured = true;
        }
      }
    }
      
    // Fire message configuration was installed
    fireDeviceChanged(xPL_DeviceChangeEventI.ChangeReason.CONFIGURATION_CHANGED);
    
    // If this configured the device, send a message
    if (deviceJustConfigured) fireDeviceChanged(xPL_DeviceChangeEventI.ChangeReason.DEVICE_CONFIGURED);

    // Send out a heartbeat if we need one
    if (heartbeatNeeded && deviceEnabled) {
      heartbeatLastSentAt = 0;
      sendTimelyHeartbeat();
    }
    
    return true;
  }
  public boolean installConfigValues(NamedValuesI configValues) {
    return installConfigValues(configValues, false, false);
  }
  
  // Handle receiving a configuration message
  // Note: We know the schema class is "config" and this message is
  // meant for us bythe time this is called
  protected void processConfigMessage(xPL_MessageI configMessage) {
    // Grab the schema type
    String schemaType = configMessage.getSchemaType();
    
    // Handle list type requests
    if (schemaType.equalsIgnoreCase("list")) {
      // See if this is request
      if (configMessage.hasNamedValue("command", "request")) {
        sendConfigItemsList();
        return;
      }
      
      // Unknown list command
      debug("Unknown config.list message -- ignored");
      return;
    }
    
    // Handle current values requests
    if (schemaType.equalsIgnoreCase("current")) {
      // See if this is request
      if (configMessage.hasNamedValue("command", "request")) {
        sendConfigItemValues();
        return;
      }
      
      // Unknown current command
      debug("Unknown config.current message -- ignored");
      return;
    }
    
    // Handle setting config values
    if (schemaType.equalsIgnoreCase("response")) {
      if (installConfigValues(configMessage.getMessageBody())) {
        doAutoPreserveConfig();
      }

      return;
    }
    
    // Anything else is a mystery
    debug("Unknown config." + schemaType + " -- message ignored");
  }
  
  public void handleXPLMessage(xPL_MessageI theMessage) {
    // Skip any message originating from us (echo)
    if (theMessage.getSource() == deviceIdent) return;

    // For broadcasts, insure message passes the filters
    // For group messages, insure this device is associated with the group
    // For all other messages, insure the target is in fact us
    if (theMessage.isBroadcast()) {
      // Handle heartbeat messages and requests, filtering them out (unless they are wanted)
      if (theMessage.isHeartbeatRequest()) {
        if (heartbeatResponder == null) heartbeatResponder = new HeartbeatResponder(this);
        heartbeatResponder.respond();
        if (!passAllMessages) return;
      } else if (theMessage.isHeartbeat()) {
        if (!passAllMessages) return;
      }

      // Insure this broadcast matches at least one filter (if we have any) or dump it
      if (!doesMessageMatchAFilter(theMessage)) return;
      
    } else if (theMessage.isGroupMessage()) {
      // Insure this group is one we are attached to or dump the message
      if (!doesMessageMatchAGroup(theMessage)) return;
    } else {
      // Insure this targeted message is in fact for us
      if (theMessage.getTarget() != deviceIdent) return;
      
      // Since it is specifically targetted at us, check for a config
      // message and revector it as needed
      if (deviceIsConfigurable && theMessage.isConfigurationMessage()) {
        processConfigMessage(theMessage);
        return;
      }
    }
    
    // Send this message to all device listeners
    if (messageListeners != null) {
      for (xPL_DeviceMessageListenerI theListener: messageListeners) {
        try {
          theListener.handleXPLDeviceMessage(this, theMessage);
        } catch (Throwable ignoredError) { }
      }
    }
  }

  public xPL_MutableMessageI createMessage(xPL_MessageI.MessageType type, String schemaClass, String schemaType) {
    if (deviceIdent == null) return null;
    
    xPL_MutableMessageI newMessage = xPL_Util.createMessage();
    newMessage.setSource(deviceIdent);
    newMessage.setType(type);
    newMessage.setSchema(schemaClass, schemaType);
    return newMessage;
  }

  public xPL_MutableMessageI createMessage(xPL_IdentifierI to, xPL_MessageI.MessageType type, String schemaClass, String schemaType) {
    if (deviceIdent == null) return null;
    return xPL_Util.createMessage(deviceIdent, to, type, schemaClass, schemaType);
  }

  public xPL_MutableMessageI createMessage(String vendorID, String deviceID, String instanceID, xPL_MessageI.MessageType type, String schemaClass, String schemaType) {
    if (deviceIdent == null) return null;
    xPL_IdentifierI toDevice = identifierManager.getInstance(vendorID, deviceID, instanceID);
    if (toDevice == null) return null;
    return xPL_Util.createMessage(deviceIdent, toDevice, type, schemaClass, schemaType);
  }

  public xPL_MutableMessageI createMessage(String deviceSpec, xPL_MessageI.MessageType type, String schemaClass, String schemaType) {
    if (deviceIdent == null) {
      System.out.println("No device ident -- message not created");
      return null;
    }
    xPL_IdentifierI toDevice = identifierManager.parseNamedIdentifier(deviceSpec);
    if (toDevice == null) {
      System.out.println("Unparsable target [" + deviceSpec + "] -- message not created");
      return null;
    }
    return xPL_Util.createMessage(deviceIdent, toDevice, type, schemaClass, schemaType);
  }

  public xPL_MutableMessageI createMessage(String toDeviceSpec, String type, String schemaClass, String schemaType) {
    xPL_MessageI.MessageType theType;
    if (type.equalsIgnoreCase("xpl-stat")) {
      theType = xPL_MessageI.MessageType.STATUS;
    } else if (type.equalsIgnoreCase("xpl-cmnd")) {
      theType = xPL_MessageI.MessageType.COMMAND;
    } else if (type.equalsIgnoreCase("xpl-trig")) {
      theType = xPL_MessageI.MessageType.TRIGGER;
    } else {
      System.out.println("Unknown message type [" + type + "] -- message not created");
      return null;
    }
    
    return createMessage(toDeviceSpec, theType, schemaClass, schemaType);
  }
  
  public xPL_MutableMessageI createGroupMessage(xPL_IdentifierI toGroup, xPL_MessageI.MessageType type, String schemaClass, String schemaType) {
    if (deviceIdent == null) return null;
    return xPL_Util.createGroupMessage(deviceIdent, toGroup, type, schemaClass, schemaType);
  }

  public xPL_MutableMessageI createGroupMessage(String toGroupName, xPL_MessageI.MessageType type, String schemaClass, String schemaType) {
    if (deviceIdent == null) return null;
    xPL_IdentifierI toGroup = identifierManager.getGroupIdentifier(toGroupName);
    if (toGroup == null) return null;
    return xPL_Util.createGroupMessage(deviceIdent, toGroup, type, schemaClass, schemaType);
  }

  public xPL_MutableMessageI createBroadcastMessage(xPL_MessageI.MessageType type, String schemaClass, String schemaType) {
    if (deviceIdent == null) return null;
    return xPL_Util.createBroadcastMessage(deviceIdent, type, schemaClass, schemaType);
  }

  public boolean preserveDeviceConfig(OutputStream toStream) {
    return deviceManager.preserveDevice(this, toStream);
  }

  public boolean preserveDeviceConfig(File toFile) {
    return deviceManager.preserveDevice(this, toFile);
  }

  public boolean restoreDeviceConfig(InputStream fromStream) {
    return (deviceRestored = deviceManager.restoreDevice(this, fromStream));
  }

  public boolean restoreDeviceConfig(File fromFile) {
    return (deviceRestored = deviceManager.restoreDevice(this, fromFile));
  }

  public File getAutoPreserveConfigFile() { return autoPreserveFile; }
  public void setAutoPreserveConfigFile(File toFile) {
    autoPreserveFile = toFile;
  }
  
  public boolean doAutoPreserveConfig() {
    if (autoPreserveFile == null) return false;
    return preserveDeviceConfig(autoPreserveFile);
  }

  public void sendMessage(xPL_MessageI theMessage) {
    if (theMessage.getSource() != deviceIdent) {
      if (theMessage instanceof xPL_MutableMessageI) {
        ((xPL_MutableMessageI) theMessage).setSource(deviceIdent);
      }
    }
    xplManager.sendMessage(theMessage);
  }

  public void releaseResources() {
    deviceIdent = null;
    deviceVersion = null;
    autoPreserveFile = null;
    if (heartbeatResponder != null) heartbeatResponder.releaseResources();
    heartbeatResponder = null;
    
    if (messageListeners != null) messageListeners.clear();
    messageListeners = null;
    
    if (changeListeners != null) changeListeners.clear();
    changeListeners = null;
    
    if (deviceFilters != null) {
      for (xPL_FilterI theFilter: deviceFilters) {
        theFilter.resetFilters();
      }
      deviceFilters.clear();
    }
    deviceFilters = null;
    
    if (deviceGroups != null) {
      IdentifierManager devManager = xPL_Manager.getManager().getIdentifierManager();
      for (xPL_IdentifierI theGroup: deviceGroups.values()) {
        devManager.removeIdentifier(theGroup);
      }
      
      deviceGroups.clear();
    }
    deviceGroups = null;
    
    if (configItems != null) {
      for (xPL_DeviceConfigItemI configItem: configItems.values()) {
        configItem.releaseResources();
      }
      
      configItems.clear();
    }
    configItems = null;
  }
  
  public String toString() { return (deviceIdent == null ? "UNKNOWN SVC" : deviceIdent.toString()); }
  
  public boolean equals(xPL_IdentifierI otherDevice) {
    // Insure they are either both null or neither null
    if ((otherDevice == null) != (deviceIdent == null)) return false;
    if (otherDevice == null) return false;
    
    // See if they describe the same thing
    return deviceIdent.equals(otherDevice);
  }
  
  public boolean equals(xPL_DeviceI otherDevice) {
    if (otherDevice == null) return false;
    if (otherDevice == this) return true;
    
    if (deviceIdent == null) return false;
    return deviceIdent.equals(otherDevice.getIdent());
  }
}
