/*
 * xPL_DeviceConfigItem.java
 * Copyright 2004, Gerald R Duprey Jr
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy of
 * the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0 
 *
 * Unless required by applicable law or agreed to in writing, software distributed
 * under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
 * CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License. 
 */

package org.cdp1802.xpl.device;

import java.util.ArrayList;

import java.util.Collection;

/**
 *
 * @author gduprey
 */
public class xPL_DeviceConfigItem implements xPL_DeviceConfigItemI {
  String itemName = "";
  int maxValues = 1;
  boolean isMandatory = false;
  boolean isReconfigurable = true;
  ArrayList<String> itemValues = new ArrayList<String>();

  public xPL_DeviceConfigItem(String itemName) {
    setName(itemName);
  }
  
  public xPL_DeviceConfigItem(String itemName, int maxValueCount, boolean isMandatory, boolean isReconfigurable) {
    setName(itemName);
    setMaxValueCount(maxValueCount);
    setMandatory(isMandatory);
    setReconfigurable(isReconfigurable);
  }

  public void setName(String newName) {
    if ((newName == null) && (newName.length() == 0)) return;
    
    // Handle a name in the form of name[maxValues]
    int newMaxValues = -1, startDelim = -1, endDelim = -1;
    if (((startDelim = newName.indexOf("[")) != -1) && ((endDelim = newName.indexOf("]", startDelim + 1)) != -1)) {
      try {
        newMaxValues = Integer.parseInt(newName.substring(startDelim + 1, endDelim));
        newName = newName.substring(0, startDelim);
      } catch (Exception anyError) { return; }
      
      if ((newMaxValues < 1) || (newName.length() == 0)) return;
      setMaxValueCount(newMaxValues);
    }
    
    itemName = newName;
  }
  public String getName() { return itemName; }

  public void setMandatory(boolean itemIsMandatory) {
    isMandatory = itemIsMandatory;
  }
  public boolean isMandatory() { return isMandatory; }

  public void setReconfigurable(boolean itemIsReconfigurable) {
    isReconfigurable = itemIsReconfigurable;
  }
  public boolean isReconfigurable() { return isReconfigurable; }

  public void setMaxValueCount(int maxValueCount) {
    if ((maxValueCount == maxValues) || (maxValueCount < 1)) return;
    
    synchronized(itemValues) {
      maxValues = maxValueCount;
      while(itemValues.size() > maxValues) itemValues.remove(itemValues.size() - 1);
    }
  }
  public int getMaxValueCount() { return maxValues; }

  public int getValueCount() { return itemValues.size(); }

  private boolean areValuesChangable() {
    if (isReconfigurable) return true;
    
    // If there are no values, we can set one
    if (itemValues.size() == 0) return true;
    return false;
  }
  
  public boolean clearValues() {
    synchronized(itemValues) {
      if (!areValuesChangable()) return false;
      itemValues.clear();
      return true;
    }
  }
  
  public boolean isValueSet() {
    return (itemValues.size() != 0);
  }

  public boolean isValueSetAt(int valueIndex) {
    synchronized(itemValues) {
      if ((valueIndex < 0) || (valueIndex >= itemValues.size())) return false;
      return true;
    }
  }

  public boolean setValue(String theValue) {
    if ((theValue == null) || !areValuesChangable()) return false;
            
    synchronized(itemValues) {
      if ((theValue == null) || (itemValues.size() >= maxValues)) return false;
      
      if (itemValues.size() == 0)
        itemValues.add(theValue);
      else
        itemValues.set(0, theValue);
      
      return true;
    }
  }

  public boolean setValue(int theValue) {
    return setValue(String.valueOf(theValue));
  }

  public boolean setValue(boolean theValue) {
    return setValue(theValue ? "1" : "0");
  }

  public boolean setValueAt(int valueIndex, String theValue) {
    synchronized(itemValues) {
      if (!areValuesChangable() || (theValue == null) || (valueIndex < 0) || (valueIndex >= maxValues) || (valueIndex > itemValues.size()) ) return false;
      if (valueIndex >= itemValues.size())
        itemValues.add(theValue);
      else
        itemValues.set(valueIndex, theValue);
      
      return true;
    }
  }

  public boolean setValueAt(int valueIndex, int theValue) {
    return setValueAt(valueIndex, String.valueOf(theValue));
  }
  
  public boolean setValueAt(int valueIndex, boolean theValue) {
    return setValueAt(valueIndex, (theValue ? "1" : "0"));
  }
  
  /**
   * Remove a value at a given index.
   *
   * If yopu provide an invalid value index, the attempt is silently ignore
   * 
   * @param valueIndex Index of value to remove
   * @return true if the value was removed, false if not
   */
  public boolean removeValueAt(int valueIndex) {
    synchronized(itemValues) {
      if (!areValuesChangable() || (valueIndex < 0) || (valueIndex >= itemValues.size())) return false;
      itemValues.remove(valueIndex);
      return true;
    }
  }
  
  public boolean addValue(String theValue) {
    if (!areValuesChangable() || (theValue == null) || (itemValues.size() >= maxValues)) return false;
    itemValues.add(theValue);
    return true;
  }

  public boolean addValue(int theValue) {
    return addValue(String.valueOf(theValue));
  }

  public boolean addValue(boolean theValue) {
    return addValue(theValue ? "1" : "0");
  }
  
  public String getValue() {
    synchronized(itemValues) {
      if (itemValues.size() == 0) return null;
      return itemValues.get(0);
    }
  }

  public int getIntValue() {
    String theValue = getValue();
    if (theValue == null) return -1;
    if (theValue.length() == 0) return 0;
    try {
      return Integer.parseInt(theValue);
    } catch (Exception anyError) { }
    
    return -1;
  }

  public boolean getBoolValue() {
    String theValue = getValue();
    if ((theValue == null) || (theValue.length() == 0)) return false;
    
    if (theValue.length() == 1) {
      char theChar = theValue.charAt(0);
      if ((theChar == 't') || (theChar == 'T') || (theChar == 'y') || (theChar == 'Y') || (theChar == '1')) return true;
      return false;
    }
    
    if (theValue.equalsIgnoreCase("true") || theValue.equalsIgnoreCase("yes")) return true;
    return false;
  }

  public String getValueAt(int valueIndex) {
    synchronized(itemValues) {
      if ((valueIndex < 0) || (valueIndex >= itemValues.size())) return null;
      return itemValues.get(valueIndex);
    }
  }
  
  public int getIntValueAt(int valueIndex) {
    String theValue = getValueAt(valueIndex);
    if (theValue == null) return -1;
    if (theValue.length() == 0) return 0;
    try {
      return Integer.parseInt(theValue);
    } catch (Exception anyError) { }
    
    return -1;
  }

  public boolean getBoolValueAt(int valueIndex) {
    String theValue = getValueAt(valueIndex);
    if ((theValue == null) || (theValue.length() == 0)) return false;
    
    if (theValue.length() == 1) {
      char theChar = theValue.charAt(0);
      if ((theChar == 't') || (theChar == 'T') || (theChar == 'y') || (theChar == 'Y') || (theChar == '1')) return true;
      return false;
    }
    
    if (theValue.equalsIgnoreCase("true") || theValue.equalsIgnoreCase("yes")) return true;
    return false;
  }

  public Collection<String> getValues() {
    synchronized(itemValues) {
      return itemValues;
    }
  }
  
  public void releaseResources() {
    itemName = null;
    if (itemValues != null) itemValues.clear();
    itemValues = null;
  }
}
