/*
 * xPL_DeviceConfigItemI.java
 * Copyright 2004, Gerald R Duprey Jr
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy of
 * the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0 
 *
 * Unless required by applicable law or agreed to in writing, software distributed
 * under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
 * CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License. 
 */

package org.cdp1802.xpl.device;

import java.util.Collection;

/**
 * Description of a configurable element of a xPL device
 *
 * @author  gerry
 */
public interface xPL_DeviceConfigItemI {
  /**
   * Set a new name for this configuation item
   *
   * NOTE: Name can be a simple name (i.e. "myItemName") or a name that also
   * specifies the max # of values the item can contain (i.e. "myItemName[10]"). If
   * the later form is used, the # of items must be a valid, non-zero number or
   * neither the max values or name is set.  This is meant primarily to be used
   * with parsing xPL configuration messages, but can be used for other good purposes too.
   * 
   * NOTE: Once a device is enabled, you should not change the name, max value
   * count, mandatory or reconfigurable status of any configuration item/element
   * for that device.  Doing so may lead to an inconsistant internal and external
   * state for this device.
   *
   * @param newName new, non-null, non-empty name to set for this item
   */
  public void setName(String newName);
  
  /**
   * Get the items name
   * @return Current name of this item or an empty (non-null) string if non was ever set
   */
  public String getName();
  
  /**
   * Indicate if this is a mandatory item that must have a value set before
   * the device can be used.
   * 
   * NOTE: Once a device is enable, you should not change the name, max value
   * count, mandatory or reconfigurable status of any configuration item/element
   * for that device.  Doing so may lead to an inconsistant internal and external
   * state for this device.
   * @param itemIsMandatory true if this item is mandatorily configured/required by the device
   */
  public void setMandatory(boolean itemIsMandatory);

  /**
   * Is this item require mandatory setting for the device?
   * @return true if the device requires this item to be set
   */
  public boolean isMandatory();
  
  /**
   * Set if this items values can be reconfigured remotely after initial configuration
   * or not.  Even if this is false, you program can still alter the values, but 
   * attempts by remote xPL apps to change them after initial settings will be 
   * rejected.
   * 
   * NOTE: Once a device is enable, you should not change the name, max value
   * count, mandatory or reconfigurable status of any configuration item/element
   * for that device.  Doing so may lead to an inconsistant internal and external
   * state for this device.
   * @param itemIsReconfigurable true if the items values can be changed after initial configuration
   */
  public void setReconfigurable(boolean itemIsReconfigurable);

  /**
   * Is this items values reconfigurable
   * @return true if this items values can be changed remotely after initial reconfiguration.
   */
  public boolean isReconfigurable();

  /**
   * Set the maximum number of values for this item.  If the number is lower
   * than the number of values already set, the addition values are lost.
   * 
   * NOTE: Once a device is enable, you should not change the name, max value
   * count, mandatory or reconfigurable status of any configuration item/element
   * for that device.  Doing so may lead to an inconsistant internal and external
   * state for this device.
   * @param maxValueCount Maximum number of values for this configuration item.  Numbers between 1 and 16 are considered
   * valid, though you can go higher (other xPL apps may not handle it as well)
   */
  public void setMaxValueCount(int maxValueCount);
  
  /**
   * Get the current maximum number of values for this item
   * @return Maximum number of values this item can contain
   */
  public int getMaxValueCount();
  
  
  /**
   * Clear all values for this item.  If the items cannot be cleared (because
   * the item is not reconfigurable and has values set), this will fail
   *
   * @return true if values were cleared, false if not.
   */
  public boolean clearValues();
  
  /**
   * Get the number of values actually set into this item.  This can range 
   * from 0 to the result of getMaxValueCount()
   * @return Number of actual values set into this item (0-n)
   */
  public int getValueCount();

  /**
   * Set the first value for this item.  If the item supports more than one
   * item, this sets only the first items value.  If there is no first item, 
   * then this adds it.
   *
   * @param theValue Value to be installed as the first value for this item
   * @return true if item added, false if value is null or too many values in item or is not configurable
   */
  public boolean setValue(String theValue);
  
  /**
   * Set the first value for this item.  If the item supports more than one
   * item, this sets only the first items value.  If there is no first item, 
   * then this adds it.
   *
   * @param theValue Value to be installed as the first value for this item
   * @return true if item added, false if value is null or too many values in item or is not configurable
   */
  public boolean setValue(int theValue);
  
  /**
   * Set the first value for this item.  If the item supports more than one
   * item, this sets only the first items value.  If there is no first item, 
   * then this adds it.
   *
   * @param theValue Value to be installed as the first value for this item
   * @return true if item added, false if value is null or too many values in item or is not configurable
   *
   */
  public boolean setValue(boolean theValue);

  /**
   * Set a value at a given index.
   * 
   * NOTE: You cannot set a value in an index higher than the number of values
   * already in the item (+1).  So, if there are no items, the only valid index
   * to set a value is 0.  If there are 3 items, you can use any index from 0-3 
   * (where 3 would be adding a new item to the list).
   * 
   * If the index does not meet this, the attempt will be silently rejected/ignored
   *
   * @param valueIndex Index to set value at
   * @param theValue The value to install.  This cannot be null or the attempt is silently rejected/ignored
   * @return true if item added, false if value is null or too many values in item or is not configurable
   */
  public boolean setValueAt(int valueIndex, String theValue);

  /**
   * Set a value at a given index.
   * 
   * NOTE: You cannot set a value in an index higher than the number of values
   * already in the item (+1).  So, if there are no items, the only valid index
   * to set a value is 0.  If there are 3 items, you can use any index from 0-3 
   * (where 3 would be adding a new item to the list).
   * 
   * If the index does not meet this, the attempt will be silently rejected/ignored
   *
   * @param valueIndex Index to set value at
   * @param theValue The value to install.  This cannot be null or the attempt is silently rejected/ignored
   * @return true if item added, false if value is null or too many values in item or is not configurable
   */
  public boolean setValueAt(int valueIndex, int theValue);

  /**
   * Set a value at a given index.
   * 
   * NOTE: You cannot set a value in an index higher than the number of values
   * already in the item (+1).  So, if there are no items, the only valid index
   * to set a value is 0.  If there are 3 items, you can use any index from 0-3 
   * (where 3 would be adding a new item to the list).
   * 
   * If the index does not meet this, the attempt will be silently rejected/ignored
   *
   * @param valueIndex Index to set value at
   * @param theValue The value to install.  This cannot be null or the attempt is silently rejected/ignored
   * @return true if item added, false if value is null or too many values in item or is not configurable
   */
  public boolean setValueAt(int valueIndex, boolean theValue);

  /**
   * Attempt to add an additional value to the item.  If there are already
   * as many values as the item allows, this is silently rejected/ignored
   *
   * @param theValue value to be added to the item
   * @return true if item added, false if value is null or too many values in item or is not configurable
   */
  public boolean addValue(String theValue);
  
  /**
   * Attempt to add an additional value to the item.  If there are already
   * as many values as the item allows, this is silently rejected/ignored
   *
   * @param theValue value to be added to the item
   * @return true if item added, false if value is null or too many values in item or is not configurable
   */
  public boolean addValue(int theValue);

  /**
   * Attempt to add an additional value to the item.  If there are already
   * as many values as the item allows, this is silently rejected/ignored
   *
   * @param theValue value to be added to the item
   * @return true if item added, false if value is null or too many values in item or is not configurable
   */
  public boolean addValue(boolean theValue);

  /**
   * See if there is at least one value set for this item
   * @return true if at least one value is set, false if there are no values
   */
  public boolean isValueSet();

  /**
   * Remove a value at a given index.
   *
   * If yopu provide an invalid value index, the attempt is silently ignore
   * 
   * @param valueIndex Index of value to remove
   * @return true if item was removed, false if not 
   */
  public boolean removeValueAt(int valueIndex);

  /**
   * Determine if the nth value for this item is set or not.
   * @param valueIndex index of the value to test for (0 based index)
   * @return true if there is an item set there, false if no item is set or index is invalid
   */
  public boolean isValueSetAt(int valueIndex);

  /**
   * Get the first/only value for this item.  If the item has more than one
   * value, only the first one is returned.
   * @return First value for this item or null if no items are set
   */
  public String getValue();
  
  /**
   * Get the first/only value for this item.  If the item has more than one
   * value, only the first one is returned.
   *
   * @return First value for this item or -1if no items are set
   */
  public int getIntValue();

  /**
   * Get the first/only value for this item.  If the item has more than one
   * value, only the first one is returned.
   *
   * @return First value for this item or false if no items are set
   */
  public boolean getBoolValue();

  /**
   * Get the value, if any, at the passed index for this item
   * @param valueIndex index of the value to retrieve
   * @return The value at the passed index or null if the value is not set or index is invalid
   */
  public String getValueAt(int valueIndex);
  
  /**
   * Get the value, if any, at the passed index for this item
   *
   * @param valueIndex index of the value to retrieve
   * @return The value at the passed index or -1 if the value is not set or index is invalid
   */
  public int getIntValueAt(int valueIndex);

  /**
   * Get the value, if any, at the passed index for this item
   *
   * @param valueIndex index of the value to retrieve
   * @return The value at the passed index or false if the value is not set or index is invalid
   */
  public boolean getBoolValueAt(int valueIndex);
  
  /** 
   * Return all values for this item.  If there are no items, an empty list is returned
   *
   * @return list of values
   */
  public Collection<String> getValues();
  
  /** 
   * Release all resources for a given configuration item.  Once called, this config
   * item is completly wiped and invalid.  You must not wipe a config item while it is
   * still associated with a device
   */
  public void releaseResources();
}
