/*
 * xPL_DeviceI.java
 * Copyright 2004, Gerald R Duprey Jr
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy of
 * the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0 
 *
 * Unless required by applicable law or agreed to in writing, software distributed
 * under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
 * CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License. 
 */

package org.cdp1802.xpl.device;

import java.io.File;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.Collection;
import org.cdp1802.xpl.xPL_IdentifierI;
import org.cdp1802.xpl.xPL_MessageI;
import org.cdp1802.xpl.xPL_MutableMessageI;
import org.cdp1802.xpl.xPL_FilterI;

/**
 * Describes an xPL device
 * @author gerry
 */
public interface xPL_DeviceI {
  /**
   * Get the identifier for this device.
   *
   * NOTE: The returned value can change over time, especially if the device is
   * reconfigurable.  Because of this, it's important that all outgoing messages
   * for this device use the devices sendMessage() method.
   *
   * @return identifier for this device
   */
  public xPL_IdentifierI getIdent();
  
  /**
   * Set the identifier for this device.  Any attempt to change the identifier
   * while the device is enabled is ignored.
   * @param theIdentifier new identifier to use as this devices identifier.
   */
  public void setIdent(xPL_IdentifierI theIdentifier);
  
  /**
   * Set the identifier for this device.  Any attempt to change the identifier
   * while the device is enabled is ignored.
   * @param vendorID vendor ID for this devices identifier
   * @param deviceID device ID for this devices identifier
   * @param instanceID instance ID for this devices identifier
   */
  public void setIdent(String vendorID, String deviceID, String instanceID);
  
  /**
   * Get current device enabled status.
   * @return true if device is enabled, receiving messages and sending out heartbeats
   */
  public boolean isEnabled();
  
  /**
   * Enable/disable the device.  
   * 
   * This will not work if there is no valid device identifier installed.  In that
   * case, any attempt to enable/disable the device will be ignored
   * 
   * Enabling the device will start sending out heartbeats, start
   * dispatching device messages and, if the device is configurable, start the
   * configuration process for the device.
   * 
   * Disabling the device will stop heartbeats, send out a "end" message for the
   * device and stop dispatching device related messages to listeners.
   * 
   * A device may be enabled and disabled multiple times.  However, re-enabling
   * a previously enabled configurable device that had complete device configuration
   * while last enabled will not result in the device starting a new configuration
   * cycle.
   * @param enabledevice true to enable the device, false to disable it
   */
  public void setEnabled(boolean enabledevice);
  
  /**
   * Add a device message listener.  Only xPL certain messages are delivered.  
   * An xPL message is considered deliverable if
   * 
   * 1) the message specific targets this device
   * 2) the message targets a group this device is part of
   * 3) the message is a broadcast message and either
   *    a) there are no filters for the device 
   *        -OR-
   *    b) the message matches at least one of the filters for this device
   *    
   * If a message listener was already added, a second attempt to add it will be
   * silently ignored.
   * @param theListener message listener to add
   */
  public void addMessageListener(xPL_DeviceMessageListenerI theListener);
  
  /**
   * Remove a previously added message listener
   * @param theListener listener to remove
   */
  public void removeMessageListener(xPL_DeviceMessageListenerI theListener);
  
  /**
   * Normally a device will filter out messages that deal with xPL bookkeeping
   * such as hbeat and config messages.  In some special occasions, you may 
   * need to hear even those messages and you can enable the device to pass
   * all messages to the listeners.
   *
   * <P>NOTE: Messages specifically targeted for another device or those not
   * matching the devices filters (if any) are never delivered, regardless of
   * this setting.  Only broadcast messages, group messages to a group this
   * device is part of or targeted messages to this device are candidates for
   * being passed to a listener and then only if they meet any filter criteria.
   *
   * @param passAll true to pass all device messages (heartbeats/config), false to filter extraneous messages out
   */
  public void setPassAllMessages(boolean passAll);
  
  /**
   * Determine if this device is passing all messages to a device message 
   * listener or if some messages such as heartbeats and configuration
   * messages are being left out.
   *
   * @return true if all messages are being passed and delivered
   */
  public boolean getPassAllMessages();
  
  /**
   * Get the current heartbeat interval
   * @return interval, in minutes, for this devices heartbeats to be sent
   */
  public int getHeartbeatInterval();
  
  /**
   * Set the heartbeat interval for this device. 
   * 
   * The valid range is 1 to 30 minutes with a default of 5 minutes.  Values less
   * than 1 are rounded up to 5 minutes (the recommended value) and values over
   * 30 are capped at 30 minutes.
   * @param newHeartbeatInterval new heartbeat interval, in minutes, to install
   */
  public void setHeartbeatInterval(int newHeartbeatInterval);

  /**
   * Determines if it's time to send a heartbeat for this device (based on when
   * the last heartbeat was sent and the heartbeat interval) and if it's time,
   * initiates a hearetbeat transmission.
   * 
   * This is not normally needed to be called as any device that is being managed
   * (and those using setEnable() ghenerally are) will have this called periodically
   * automatically.
   *
   * @param forceHeartbeat if true, the heartbeat is always sent.  If false, it's only sent if it's time to send it
   */
  public void sendTimelyHeartbeat(boolean forceHeartbeat);

  /**
   * Determines if it's time to send a heartbeat for this device (based on when
   * the last heartbeat was sent and the heartbeat interval) and if it's time,
   * initiates a hearetbeat transmission.
   * 
   * This is not normally needed to be called as any device that is being managed
   * (and those using setEnable() ghenerally are) will have this called periodically
   * automatically.
   */
  public void sendTimelyHeartbeat();
  
  /**
   * Sets the upper limit on the number of groups this device should advertise
   * allowing.  The default is 255 and the minimum is 0.  The maximum is 
   * in the billions.
   *
   * devices can actually belong to an unlimited number of groups, but the 
   * xPL configuration requires a number. 
   *
   * @param maxAllowedGroups max # of allowed groups.
   */
  public void setMaxGroupCount(int maxAllowedGroups);
  
  /**
   * Return the maximum number of groups that this device will allow.
   *
   * @return maximum # of groups allowed for this device
   */
  public int getMaxGroupCount();
  
  /**
   * Get the number of groups currently assigned to this device
   *
   * @return number of groups assigned to the device
   */
  public int getGroupCount();
  
  
  /**
   * Clear any existing group assignments for this device
   */
  public void clearGroups();
  
  /**
   * Add a group assignment to this device. A device many have 
   * many group names (by convention, this is limited to 16, though you can
   * have as many as free memory permits).  An attempt to add the same group
   * to a device more than once is silently ignored.
   * @param theGroup identifier describing group to add to this device
   * @return instance of the group added or null if group could not be added
   */
  public xPL_IdentifierI addGroup(xPL_IdentifierI theGroup);
  
  /**
   * Add a group assignment to this device. A device many have 
   * many group names (by convention, this is limited to 16, though you can
   * have as many as free memory permits).  An attempt to add the same group
   * to a device more than once is silently ignored.
   * @param groupName group name to assign to this device
   * @return instance of the group added or null if group could not be added
   */
  public xPL_IdentifierI addGroup(String groupName);
  
  /**
   * See if the device is a member of the passed group
   *
   * @param theGroup group device to check
   * @return true if this device is associated with this group, false if not
   */
  public boolean isAssignedToGroup(xPL_IdentifierI theGroup);
  
  /**
   * See if the device is a member of the passed group
   *
   * @param theGroupName name of group to check
   * @return true if this device is associated with this group, false if not
   */
  public boolean isAssignedToGroup(String theGroupName);
  
  /**
   * Get the current list of groups assigned to this device
   * @return list of groups assigned to this device or null if there are no groups assigned
   */
  public Collection<xPL_IdentifierI> getGroups();
  
  /**
   * Sets the upper limit on the number of filters this device should advertise
   * allowing.  The default is 255 and the minimum is 0.  The maximum is 
   * in the billions.
   *
   * devices can actually belong to an unlimited number of filters, but the 
   * xPL configuration requires a number. 
   *
   * @param maxAllowedFilters max # of allowed filters.
   */
  public void setMaxFilterCount(int maxAllowedFilters);
  
  /**
   * Return the maximum number of filters that this device will allow.
   *
   * @return maximum # of filters allowed for this device
   */
  public int getMaxFilterCount();
  
  /**
   * Get the number of filters currently installed for this device
   *
   * @return number of filters installed into this device
   */
  public int getFilterCount();

  /**
   * Clear out all filters for this device
   */
  public void clearFilters();
  
  /**
   * Add the passed filter to the list of filters applied to incoming broadcast messages
   * for this device.  If there is one or more filters installed, a broadcast must 
   * match at least one filter or it's discarded and not delivered to the device
   * message listener.
   * @param aFilter Filter to add to the list of filters
   * @return Instance of the filter added or null if filter could not be added
   */
  public xPL_FilterI addFilter(xPL_FilterI aFilter);
  
  /**
   * Attempt to add the passed filter spec to the list of filters applied to incoming broadcast messages
   * for this device.  If there is one or more filters installed, a broadcast must 
   * match at least one filter or it's discarded and not delivered to the device
   * message listener.
   * 
   * The passed filter spec will be parsed into a filter using the syntax of a filter
   * as description in the xPL protocol docs or the xPL_FilterI.setFilter(String) 
   * method and if it's valid, a new xPL_FilterI is created and installed.
   * @param filterSpec filter spec using the sytax described in the xPL protocol docs and/or xPL_FilterI.setFilter(String) method
   * @return Instance of the filter added or null if filter could not be added
   */
  public xPL_FilterI addFilter(String filterSpec);
  
  /**
   * Get the list of currently installed filters
   * @return List of currentlu installed filters. If there are no filters, null is returned.
   */
  public Collection<xPL_FilterI> getFilters();
  
  /**
   * Install a version number for this device.  This can be any string or by null to 
   * clear out any version number.  The version number, if non-null, will be included
   * with each hearbeat sent out
   *
   * @param theVersion version string for this devices device
   */
  public void setVersion(String theVersion);
  
  /**
   * Return the most recently set version or null if there is no version string
   *
   * @return version string for this device or null
   */
  public String getVersion();
  
  /**
   * Determine if this device is configurable or not
   * @return true if the device is marked as configurable.
   */
  public boolean isConfigurable();
  
  /**
   * Set the configurable status for this devices.
   * 
   * NOTE: You must set the configurable status BEFORE you enable the device.  If
   * you change it afterwards, it may have no effect unless the device is then 
   * disabled and re-enabled.
   * @param deviceConfigurable true to mark this device configurable
   */
  public void setConfigurable(boolean deviceConfigurable);

  /**
   * Determine if the device is awaiting configuration from a configuration 
   * manager.  If the device is configurable and has not been able to retain
   * it's state, it'll initiate a request to be configured from a xPL Configuration
   * Manager.  Until it receives that configuration, it'll remove in a "Being Configured"
   * state.
   * @return true if this device is currently awaiting configuration
   */
  public boolean isBeingConfigured();
  
  /**
   * Determine if this device is configurable and has a valid configuration
   * available.
   *
   * @return true if device is configuable and is currently considered configured
   */
  public boolean isConfigured();
  
  /**
   * Return true if the device has been successfully restored from a configuration
   * file.  In cases, of new devices or non-configurable devices, this will be
   * false.  If there was any error restoring the device, it will also be false.
   *
   * Usually, if this is true, isConfigured() will also be true, but there isn't
   * a direct connection, so donb't make assumptions.
   *
   * @return true if the device was successfully restored from a configuration file
   */
  public boolean wasRestored();
  
  /**
   * Mark a device as being configured, even if it has not.  If not otherwise
   * set, this also marks the device as configurable.
   *
   * You should only do this if you have a device that is configurable and you
   * do not want to load a configuration (from a file or manually) or advertise
   * yourself as needing configuration.  Needless to say, this is very, very
   * rarely needed and if you think it is, you may want to think again.
   *
   * NOTE: This should be done before the device is enabled.  If done after the
   * fact, it will not have an immediate impact on configuration messages being
   * sent/handled by the device.
   *
   * @param deviceConfigured true to mark the device as configurable and configured
   */
  public void setConfigured(boolean deviceConfigured);
  
  /**
   * Clear all configurable items out
   */
  public void clearConfigurables();
  
  /**
   * Return the number of currently defined configuable items
   *
   * @return # of configurable items
   */
  public int getConfigurableCount();
  
  /**
   * Lookup an existing configurable item and return it or null if it
   * does not exist
   *
   * @param configName name of configuration item to lookup
   * @return the matching config item or null if no matching config item found
   */
  public xPL_DeviceConfigItemI getConfigurable(String configName);
  
  /**
   * Read a simple configuration element and return it's value.  If 
   * the config element doesn't exist or has no value, return the 
   * defaultValue.
   *
   * If the configuration item can have more than one value, only the
   * very first one is returned (used getConfigurable() to access all aspects
   * of configurable items).
   *
   * @param configName name of the configuration element to return
   * @param defaultValue value to return if config element doesn't exist or has no value assigned
   * @return the current value of the configuration element or the default value
   */
  public String getConfigValue(String configName, String defaultValue);

  /**
   * Read a simple configuration element and return it's value.  If 
   * the config element doesn't exist or has no value, return null.
   *
   * If the configuration item can have more than one value, only the
   * very first one is returned (used getConfigurable() to access all aspects
   * of configurable items).
   *
   * @param configName name of the configuration element to return
   * @return the current value of the configuration element or null
   */
  public String getConfigValue(String configName);
  
  /**
   * Read a simple configuration element and return it's value.  If 
   * the config element doesn't exist or has no value, return the 
   * defaultValue.
   *
   * If the configuration item can have more than one value, only the
   * very first one is returned (used getConfigurable() to access all aspects
   * of configurable items).
   *
   * @param configName name of the configuration element to return
   * @param defaultValue value to return if config element doesn't exist or has no value assigned
   * @return the current value of the configuration element or the default value
   */
  public int getIntConfigValue(String configName, int defaultValue);
  
  /**
   * Read a simple configuration element and return it's value.  If 
   * the config element doesn't exist or has no value, return -1
   *
   * If the configuration item can have more than one value, only the
   * very first one is returned (used getConfigurable() to access all aspects
   * of configurable items).
   *
   * @param configName name of the configuration element to return
   * @return the current value of the configuration element or -1
   */
  public int getIntConfigValue(String configName);

  /**
   * Read a simple configuration element and return it's value.  If 
   * the config element doesn't exist or has no value, return the 
   * defaultValue.
   *
   * If the configuration item can have more than one value, only the
   * very first one is returned (used getConfigurable() to access all aspects
   * of configurable items).
   *
   * @param configName name of the configuration element to return
   * @param defaultValue value to return if config element doesn't exist or has no value assigned
   * @return the current value of the configuration element or the default value
   */
  public boolean getBoolConfigValue(String configName, boolean defaultValue);

  /**
   * Read a simple configuration element and return it's value.  If 
   * the config element doesn't exist or has no value, return false
   *
   * If the configuration item can have more than one value, only the
   * very first one is returned (used getConfigurable() to access all aspects
   * of configurable items).
   *
   * @param configName name of the configuration element to return
   * @return the current value of the configuration element or false
   */
  public boolean getBoolConfigValue(String configName);
  
  /**
   * Install a new configurable item.  The item name must be unique or the item
   * will not be installed.
   * 
   * You should only install new configurable items before enabing the device.  If
   * you install them afterwards, they may not be picked up by the xPL Configuration
   * Manager.
   * @param itemName unique name of this configuration item
   * @param maxValueCount maximum number of values this config item can have (1-n)
   * @param isMandatory true if the config item mandatory/required before device can function
   * @param isReconfigurable true if the item can have it values changed by a Configuration Manager at any time.  False if it can only be set once
   * @return instance of config item added or null if config item could not be added
   */
  public xPL_DeviceConfigItemI addConfigurable(String itemName, int maxValueCount, boolean isMandatory, boolean isReconfigurable);

  /**
   * Install a new configurable item.  The item name must be unique or the item
   * will not be installed.
   * 
   * You should only install new configurable items before enabing the device.  If
   * you install them afterwards, they may not be picked up by the xPL Configuration
   * Manager.
   * @param configItem config item to add to the device
   * @return instance of config item added or null if config item could not be added
   */
  public xPL_DeviceConfigItemI addConfigurable(xPL_DeviceConfigItemI configItem);
  /**
   * Get the list of all currently configurable items
   * @return list of configrable items or null if there are no items
   */
  public Collection<xPL_DeviceConfigItemI> getConfigurables();
  
  /**
   * Add a device change listener.  The listener will receive updates
   * whenever critical aspects of the device is changed (such as being
   * configured, being enabled/disabled, etc).
   * 
   * An attempt to add the same listener more than once for a device will be
   * silently ignored.
   * @param theListener device change listener to add
   */
  public void addDeviceChangeListener(xPL_DeviceChangeListenerI theListener);
  /**
   * Remove the passed device change listener from this device
   * @param theListener change listener to remove from this device
   */
  public void removeDeviceChangeListener(xPL_DeviceChangeListenerI theListener);
  
  /**
   * Create a message from this device.  The message is not directed at anything
   * 
   * NOTE: For configurable devices, the device ident installed into the 
   * created message may change over time.  This is not a problem as long as
   * you ALWAYS use the xPL_DeviceI.sendMessage() method.  That method will
   * insure the source is correct (and if not, will attempt to correct the 
   * message to match the sending device). 
   * 
   * NOTE: You must direct the message to a target before sending
   * 
   * @param type type of new message
   * @param schemaClass schema class for this message
   * @param schemaType scheme type for this message
   * @return new instance of a mutable xPL message or null if bad parameters
   */
  public xPL_MutableMessageI createMessage(xPL_MessageI.MessageType type, String schemaClass, String schemaType);
  
  /**
   * Create a message from this device directed at a specific xPL Device.
   * 
   * NOTE: For configurable devices, the device ident installed into the 
   * created message may change over time.  This is not a problem as long as
   * you ALWAYS use the xPL_DeviceI.sendMessage() method.  That method will
   * insure the source is correct (and if not, will attempt to correct the 
   * message to match the sending device). 
   * 
   * @param to xPL Device to direct message at
   * @param type type of new message
   * @param schemaClass schema class for this message
   * @param schemaType scheme type for this message
   * @return new instance of a mutable xPL message or null if bad parameters
   */
  public xPL_MutableMessageI createMessage(xPL_IdentifierI to, xPL_MessageI.MessageType type, String schemaClass, String schemaType);
  
  /**
   * Create a message from this device directed at a specific xPL Device.
   * 
   * NOTE: For configurable devices, the device ident installed into the 
   * created message may change over time.  This is not a problem as long as
   * you ALWAYS use the xPL_DeviceI.sendMessage() method.  That method will
   * insure the source is correct (and if not, will attempt to correct the 
   * message to match the sending device). 
   * 
   * @param vendorID vendor of device to direct message at
   * @param deviceID device ID to direct message at
   * @param instanceID instance ID to direct message at
   * @param type type of new message
   * @param schemaClass schema class for this message
   * @param schemaType scheme type for this message
   * @return new instance of a mutable xPL message or null if bad parameters
   */
  public xPL_MutableMessageI createMessage(String vendorID, String deviceID, String instanceID, xPL_MessageI.MessageType type, String schemaClass, String schemaType);
  
  /**
   * Create a message from this device directed at a specific xPL Device.
   * 
   * NOTE: For configurable devices, the device ident installed into the 
   * created message may change over time.  This is not a problem as long as
   * you ALWAYS use the xPL_DeviceI.sendMessage() method.  That method will
   * insure the source is correct (and if not, will attempt to correct the 
   * message to match the sending device).
   * 
   * @param toDeviceSpec identifier spec to target message toward
   * @param type type of new message
   * @param schemaClass schema class for this message
   * @param schemaType scheme type for this message
   * @return new instance of a mutable xPL message or null if bad parameters
   */
  public xPL_MutableMessageI createMessage(String toDeviceSpec, xPL_MessageI.MessageType type, String schemaClass, String schemaType);
  

  /**
   * Create a message from this device directed at a specific xPL Device.
   * 
   * NOTE: For configurable devices, the device ident installed into the 
   * created message may change over time.  This is not a problem as long as
   * you ALWAYS use the xPL_DeviceI.sendMessage() method.  That method will
   * insure the source is correct (and if not, will attempt to correct the 
   * message to match the sending device).
   * 
   * @param toDeviceSpec device spec to target message toward
   * @param type type of new message.  Mut be xpl-trig, xpl-stat or xpl-cmnd
   * @param schemaClass schema class for this message
   * @param schemaType scheme type for this message
   * @return new instance of a mutable xPL message or null if bad parameters
   */
  public xPL_MutableMessageI createMessage(String toDeviceSpec, String type, String schemaClass, String schemaType);
  
  /**
   * Create a message from this device directed at an xPL Group
   * 
   * NOTE: For configurable devices, the device ident installed into the 
   * created message may change over time.  This is not a problem as long as
   * you ALWAYS use the xPL_DeviceI.sendMessage() method.  That method will
   * insure the source is correct (and if not, will attempt to correct the 
   * message to match the sending device). 
   * 
   * @param toGroupName group name to direct message at
   * @param type type of new message
   * @param schemaClass schema class for this message
   * @param schemaType scheme type for this message
   * @return new instance of a mutable xPL message or null if bad parameters
   */
  public xPL_MutableMessageI createGroupMessage(String toGroupName, xPL_MessageI.MessageType type, String schemaClass, String schemaType);
  
  /**
   * Create a message from this device directed at an xPL Group
   * 
   * NOTE: For configurable devices, the device ident installed into the 
   * created message may change over time.  This is not a problem as long as
   * you ALWAYS use the xPL_DeviceI.sendMessage() method.  That method will
   * insure the source is correct (and if not, will attempt to correct the 
   * message to match the sending device). 
   * 
   * @param toGroup xPL_IdentifierI representation of a group to direct message at (this MUST be a group identifier)
   * @param type type of new messsage
   * @param schemaClass schema class for this message
   * @param schemaType scheme type for this message
   * @return new instance of a mutable xPL message or null if bad parameters
   */
  public xPL_MutableMessageI createGroupMessage(xPL_IdentifierI toGroup, xPL_MessageI.MessageType type, String schemaClass, String schemaType);
  
  /**
   * Create a broadcast message from this device
   * 
   * NOTE: For configurable devices, the device ident installed into the 
   * created message may change over time.  This is not a problem as long as
   * you ALWAYS use the xPL_DeviceI.sendMessage() method.  That method will
   * insure the source is correct (and if not, will attempt to correct the 
   * message to match the sending device). 
   * 
   * @param type type of new message
   * @param schemaClass schema class of new message
   * @param schemaType schema type of new message
   * @return new instance of a mutable xPL message or null if bad parameters
   */
  public xPL_MutableMessageI createBroadcastMessage(xPL_MessageI.MessageType type, String schemaClass, String schemaType);
  
  /**
   * Preserve the current state of this device to the specified stream.  All
   * configuration about the device (instance name, heartbeat interval, if
   * it's configurable, filters, groups and configuration item definitions and
   * current values) are saved.
   *
   * Note: The device must have a valid identifier installed or this will
   * not actually do anything.
   *
   * @param toStream stream to save this device to
   * @return true if device was successfully saved to the stream, false if device not savable or an error occured.
   */
  public boolean preserveDeviceConfig(OutputStream toStream);
  
  /**
   * Preserve the current state of this device to the specified file.  All
   * configuration about the device (instance name, heartbeat interval, if
   * it's configurable, filters, groups and configuration item definitions and
   * current values) are saved.
   *
   * Note: The device must have a valid identifier installed or this will
   * not actually do anything.
   *
   * @param toFile file to save this device to
   * @return true if device was successfully saved to the file, false if device not savable or an error occured.
   */
  public boolean preserveDeviceConfig(File toFile);
  
  /**
   * Restore this devices configuration from the passed input stream.  
   *
   * If the device has a vendor/device in it, then the stream will be read until
   * a matching vendor/device is found (instance is not used).  If the device has 
   * no vendor/device in it, then the first device found in the stream is used
   * and installed.
   *
   * This allows a completely blank device to be installed from a stream as 
   * long as the contents of the stream define only one device.  It also allows
   * multiple devices to be configured from the same stream, as long as each
   * device has a vendor and device installed and the devices are restored 
   * in the same order the definitions occur in the stream.
   *
   * @param fromStream stream to restore device settings from
   * @return true if a device configuration was found and installed
   */
  public boolean restoreDeviceConfig(InputStream fromStream);
  
  /**
   * Restore this devices configuration from the passed file.  
   *
   * If the device has a vendor/device in it, then the file will be read until
   * a matching vendor/device is found (instance is not used).  If the device has 
   * no vendor/device in it, then the first device found in the file is used
   * and installed.
   *
   * This allows a completely blank device to be installed from a file as 
   * long as the contents of the stream define only one device.  It also allows
   * multiple devices to be configured from the same file.
   *
   * @param fromFile file to restore device settings from
   * @return true if a device configuration was found and installed
   */
  public boolean restoreDeviceConfig(File fromFile);

  /**
   * Install a specific file to which the device will automatically write it's
   * values/configuration to whenver the device is remotely reconfigured.
   * 
   * Passing null clears any previously installed auto-preserve file
   * 
   * The preserve will only occur when the device receives an xPL configuration
   * message and has successfully installed it.  Any other need to preserve it
   * (such as after you programtically change the configuration) will require 
   * manual preservation, though using the doAutoPreserveConfig() method will
   * slightly simplify that (you don't have to lookup the auto-config file and
   * invoke the preserveDeviceConfig method in two steps).
   * 
   * @param toFile file to auto-preserve to or null to disable auto-preservation
   */
  public void setAutoPreserveConfigFile(File toFile);
  
  /**
   * Get the current auto-preserve file, if any
   *
   * @return auto-preserve file or null if there is no active auto-preserve file
   */
  public File getAutoPreserveConfigFile();
  
  /**
   * Perform a preservation of this device to the file last set in the
   * setAutoPreserveConfigFile() method.  This is just a convience as you
   * can accomplish the same thing like this:
   * 
   *  preserveDeviceConfig(getAutoPreserveConfigFile())
   * 
   * @return true if device preserved, false if no auto-preserve file, an error, etc
   */
  public boolean doAutoPreserveConfig();
  
  
  /**
   * Send an xPL message.   
   
   * It's important all outgoing messages for a device to use this, especially
   * if you are re-using a message instance over and over and the device is
   * configurable.  
   *
   * In that case (common), it's possible the "from" identifier you stamped
   * into the message at creation no longer represents the devices message
   * identifier.  This method will check for that and correct it, if necessary
   * and possible (i.e. the message is at some point an instance of 
   * xPL_MutableMessageI -- otherwise, the message cannot be changed and
   * will not be sent).
   *
   * @param theMessage message to send
   */
  public void sendMessage(xPL_MessageI theMessage);
  
  
  public boolean equals(xPL_IdentifierI otherDevice);
  public boolean equals(xPL_DeviceI otherDevice);
}
