/*
 * BindableAddress.java
 * Copyright 2004, Gerald R Duprey Jr
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy of
 * the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0 
 *
 * Unless required by applicable law or agreed to in writing, software distributed
 * under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
 * CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License. 
 */

package org.cdp1802.xpl.ethernet;

     
import java.net.*;
import java.util.*;
import org.cdp1802.xpl.xPL_MediaHandlerException;

/**
 * Represents a single, bindable address on the local system and
 * services for discovering them
 *
 * @author  gerry
 */
public class BindableAddress {
   /**
    * Given a passed address, dervice a netmask to use for that address.
    * @param fromAddress address to compute a default netmask from
    * @return a default netmask based on the class of the passed address
    */
   public static InetAddress getNetMask(InetAddress fromAddress) {
    // Create a container for the netemask
    int netNumber = (int) fromAddress.getAddress()[0] & 0xff;
    byte netMask[] = new byte[] { 0, 0, 0, 0 };

    // Create a netmask
    if (netNumber <= 127)
      netMask[0] = (byte) 0xff;
    else if (netNumber < 192) {
      netMask[0] = (byte) 0xff;
      netMask[1] = (byte) 0xff;
    } else if (netNumber < 224) {
      netMask[0] = (byte) 0xff;
      netMask[1] = (byte) 0xff;
      netMask[2] = (byte) 0xff;
    }

    // Create an address for it and return it
    try {
      return InetAddress.getByAddress(netMask);
    } catch (Exception anyError) {
      System.err.println("Could not create NetMask for IP address " + fromAddress.getHostAddress() + " -- " + anyError.getMessage());
      return null;
    }
  }

   /**
    * Create a broadcast address for the passed address
    * @param fromAddress address to create broadcast address from
    * @return broadcast address compatible with the passed address
    */
   public static InetAddress getBroadcastAddress(InetAddress fromAddress) {
    // Create a container for the netemask
    byte origAddr[] = fromAddress.getAddress();
    byte netMask[] = new byte[] { (byte) 0xff, (byte) 0xff, (byte) 0xff, (byte) 0xff };
    int netNumber = (int) origAddr[0] & 0xff;

    // Create a netmask
    if (netNumber <= 127)
      netMask[0] = origAddr[0];
    else if (netNumber < 192) {
      netMask[0] = origAddr[0];
      netMask[1] = origAddr[1];
    } else if (netNumber < 224) {
      netMask[0] = origAddr[0];
      netMask[1] = origAddr[1];
      netMask[2] = origAddr[2];
    }

    // Create an address for it and return it
    try {
      return InetAddress.getByAddress(netMask);
    } catch (Exception anyError) {
      System.err.println("Could not create Broadcast for IP address " + fromAddress.getHostAddress() + " -- " + anyError.getMessage());
      return null;
    }
  }

  /**
   * Parse a string representing a dotted quad, numeric IPV4 address and
   * return the InetAddress version of the string.
   * @param dottedQuad textual version of the address (i.e. 10.0.2.3)
   * @return InetAddress version of the passed string or null if the passed string format is not valid
   */
  public static InetAddress parseInetAddress(String dottedQuad) {
    byte addrParts[] = new byte[4];
    int addrPartCount = 0;

    // Break this up and parse it
    StringTokenizer quadParser = new StringTokenizer(dottedQuad, ".");
    while (quadParser.hasMoreTokens()) {
      try {
        addrParts[addrPartCount++] = (byte) Integer.parseInt(quadParser.nextToken());
      } catch (Exception anyError) {
        System.err.println("Unable to parse passed dotted quad IPV4 address [" + dottedQuad + "] -- " + anyError.getMessage());
        return null;
      }
    }
    
    // Insure we got 4 parts
    if (addrPartCount != 4) {
      System.err.println("Unable to parse passed dotted quad IPV4 address [" + dottedQuad + "] -- Wrong number of quads!");
      return null;
    }
    
    // Turn it into an inetaddr
    try {
      return InetAddress.getByAddress(addrParts);
    } catch (Exception anyError) {
      System.err.println("Unable to convert parsed dotted quad IPV4 address [" + dottedQuad + "] into an InetAddress -- " + anyError.getMessage());
      return null;
    }
  }

   /**
    * Enumerate all the network interfaces on the local machine and return them.
    * @throws org.cdp1802.xpl.xPL_MediaHandlerException Thrown if there are any errors or no network interfaces on the system
    * @return HashMap of network interfaces.  The HashMap is keyed by IP address
    */
   public static Collection<BindableAddress> getBindableAddresses() throws xPL_MediaHandlerException {
    ArrayList<BindableAddress> servableList = new ArrayList<BindableAddress>();
    boolean nonLoopbackNetworkFound = false;
    NetworkInterface theInterface = null;
    InetAddress theAddress = null;
    Enumeration<NetworkInterface> interfaceList = null;

    try {
      interfaceList = NetworkInterface.getNetworkInterfaces();
      if (!interfaceList.hasMoreElements()) throw new xPL_MediaHandlerException("There are no network interfaces on this machine");
    } catch (Exception anyError) {
      throw new xPL_MediaHandlerException("Unable to fetch list of network interfaces -- " + anyError.getMessage());
    }
    
    while (interfaceList.hasMoreElements()) {
      theInterface = interfaceList.nextElement();
      
      Enumeration<InetAddress> boundIPAddrList = theInterface.getInetAddresses();
      while (boundIPAddrList.hasMoreElements()) {
        theAddress = boundIPAddrList.nextElement();
        if (!theAddress.isLoopbackAddress()) nonLoopbackNetworkFound = true;
        BindableAddress bindableAddress = new BindableAddress(theInterface, theAddress);
        servableList.add(bindableAddress);
      }
    }
    
    // Handle no itnerfaces at all
    if (!nonLoopbackNetworkFound || (servableList.size() == 0)) throw new xPL_MediaHandlerException("There are no network interfaces on this machine");
    
    // Copy into an array
    return servableList;
  }
  
  NetworkInterface theInterface = null;
  InetAddress theBoundIPAddress = null;
  InetAddress theNetMask = null;
  InetAddress theBroadcastAddress = null;
  boolean isLoopback = false;
 
  BindableAddress(NetworkInterface theInterface, InetAddress theBoundIPAddress) {
    this.theInterface = theInterface;
    this.theBoundIPAddress = theBoundIPAddress;
    isLoopback = theBoundIPAddress.isLoopbackAddress();
  }

  /**
   * Is this address in fact a loopback on the system (machine local pseudo network)
   * @return true if this is a local loopback network
   */
  public boolean isLoopback() { return isLoopback; }
  
  /**
   * Get the network interface associated with this address.  Since a single network
   * interface can have >1 IP address bound to it, the relationship between network interfaces
   * and IP addresses is not always 1:1
   * @return network interface associated with this ip address
   */
  public NetworkInterface getNetworkInterface() { return theInterface; }
  /**
   * Get the IP address for this network address
   * @return address of this network address
   */
  public InetAddress getInetAddress() { return theBoundIPAddress; }

  /**
   * Get the related, default netmask for this network
   * @return default netmask for this network
   */
  public InetAddress getNetMask() {
    // If we already have a netmask, return it
    if (theNetMask != null) return theNetMask;
    return (theNetMask = getNetMask(theBoundIPAddress));
  }
  
  /**
   * Get the default broadcast address for this network
   * @return Broadcast address for this network
   */
  public InetAddress getBroadcastAddress() {
    if (theBroadcastAddress != null) return theBroadcastAddress;
    return (theBroadcastAddress = getBroadcastAddress(theBoundIPAddress));
  }

  /**
   * Get a string description of this network.  It includes the IP address and
   * actual network interface
   * @return Description of this network
   */
  public String toString() {
    return theBoundIPAddress.getHostAddress() + " on " + theInterface.getDisplayName();
  }
 
}
