/*
 * EthernetHandler.java
 * Copyright 2004, Gerald R Duprey Jr
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy of
 * the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0 
 *
 * Unless required by applicable law or agreed to in writing, software distributed
 * under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
 * CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License. 
 */

package org.cdp1802.xpl.ethernet;

import java.io.*;
import java.net.*;
import java.util.*;
import org.cdp1802.xpl.*;
import org.cdp1802.xpl.device.xPL_DeviceI;

/**
 * Implementation of a media handler for UDP/Ethernet
 *
 * @author  gerry
 */
public class EthernetHandler extends xPL_MediaHandler {
  /**
   * Describes methods of conection that a handler will support
   */
  public enum ConnectMode { 
    /**
     * Undefined connection method.  This is a placeholder and if actually
     * used as a connection type, the attempt to start the handler will fail.
     */
    UNKNOWN, 
            
    /**
     * Make the connection in standalone mode  -- it'll bind directly to the XPL_PORT.  
     * This assumes there is no hub on the network interface the handler will use and
     * never will be,
     * 
     * In general, the only need for this is setting up a Hub (which should be the only
     * service that ever starts in standalone mode).  You should not use this for 
     * general xPL apps.
     *
     * NOTE: xPL doesn't technically support this mode.  This only works if there is
     * no hub, but then it uses the hubs port.  That would break any other xPL app
     * that tried to run, so use this wisely or not at all.
     */
    STANDALONE, 
            
    /**
     * Attempt to connect to the network via an existing xPL hub.  A test is made to
     * insure there is a hub running on the network interface and if not, the attempt
     * to start the handler will fail.  Otherwise, the connection will be made to an
     * anonymous port in anticipation of hub participation in event propogation.
     * 
     * This is the method most commonly used for writing xPL apps/services
     */
    VIA_HUB, 
            
    /**
     * Try to determine if there is a hub running.  If not, then try to start up in
     * STANDALONE mode.  If there appears to be a hub, then try to start up in VIA_HUB
     * mode.  You normally would not use this connect type.
     */
    AUTO };
    
  /**
   * Assigned Port # for XPL services
   */
  public static final int XPL_PORT = 3865;
  
  class MessageDispatcher implements Runnable {
    boolean dispatcherRunning = false;
    EthernetHandler theHandler = null;
    
    MessageDispatcher(EthernetHandler theHandler) {
      this.theHandler = theHandler;
    }
    
    public void run() {
      xPL_Manager xplManager = xPL_Manager.getManager();
      xPL_MessageI receivedMessage = null;
      EthernetReceiver theReceiver = networkReceiver;
      
      dispatcherRunning = true;
      
      // Flag that we are running
      synchronized(this) { this.notifyAll(); }
      
      try {
        while(dispatcherRunning) {
          // Get next message.  If we get null, reader is shutting down
          if ((receivedMessage = theReceiver.getNextQueuedMessage()) == null) break;

          // See if this is an echo
          if (!hubConfirmed && isHubEcho(receivedMessage)) {
            hubConfirmed = true;

            // Alert listeners
            fireHandlerEvent(new xPL_MediaHandlerEvent(theHandler, xPL_MediaHandlerEvent.EventCode.HANDLER_CONFIRMED_HUB));
          }

          // Distribute the message to all listeners
          xplManager.distributeReceivedMessage(receivedMessage);
          receivedMessage.releaseResources();
        }
      } finally {
        xplManager = null;
        receivedMessage = null;
        theReceiver = null;
        dispatcherRunning = false;
        theHandler = null;
      }
    }
  }

  ConnectMode connectMode = ConnectMode.VIA_HUB;
  ConnectMode actualConnectMode = ConnectMode.UNKNOWN;

  InetAddress networkInterface = null;
  InetAddress connectedNetworkInterface = null;

  InetAddress advertisedInterface = null;
  InetAddress currentAdvertisedInterface = null;

  InetAddress broadcastAddr = null;
  DatagramSocket broadcastSocket = null;
  DatagramPacket broadcastPacket = null;
  Collection<BindableAddress> bindableAddresses = null;
  
  int connectedPort = 0;

  EthernetReceiver networkReceiver = null;
  MessageDispatcher messageDispatcher = null;
  
  /**
   * Create a new instance of an ethernet handler
   */
  public EthernetHandler() {
    setDescription("Ethernet");
  }

  private void debug(String debugMessage) {
    System.err.println("ENET_HANDLER:: " + debugMessage);
  }
  
  /**
   * Select the network interface to bind to.
   *
   * Passing null means bind to the all network interfaces and 
   * use the first non-loopback interface as our "source" address.
   *
   * @param networkInterface address of the network interface to use
   */
  public void setNetworkInterface(InetAddress networkInterface) {
    // Detect and xlate wildcard network
    if (networkInterface != null) {
      byte theBytes[] = networkInterface.getAddress();
      if ((theBytes[0] == 0) && (theBytes[1] == 0) && (theBytes[2] == 0) && (theBytes[3] == 0))
        networkInterface = null;
    }
    
    this.networkInterface = networkInterface;
  }

  /**
   * Return the network interface last set.  If the network interface
   * has been changed since the handler started, this may not represent
   * the currently in use interface, but will describe the interface 
   * to be used next time the handler is started.  If no specific
   * interface has been selected, null is returned.
   *
   * @return address to use when handler is started or null for any
   */
  public InetAddress getNetworkInterface() { return networkInterface; }
  
  /**
   * Return the actual local address the network is connected to 
   * or null if there is no local address or the handler is not
   * working.
   *
   * @return network handler is listening on or null
   */
  public InetAddress getConnectedNetworkInterface() {
    return connectedNetworkInterface;
  }
  
  /**
   * Set the advertised address for packets coming from this media handler.
   * Notmally, this is the same as the value used in setNetworkInterface, but
   * if you are binding to all network interfaces, you need to pick one as
   * primary.  If you pass null here, the first non-loopback address/network
   * on this machine will be used by default.
   *
   * Note: Changes only take affect when the handler is started.
   *
   * @param advertiseAs address of network to advertise xPL packets as coming from
   */
  public void setAdvertisedInterface(InetAddress advertiseAs) {
    advertisedInterface = advertiseAs;
  }
  
  /**
   * Return the last installed interface to advertise packets from.  If
   * there was no installed advertised interface, return null
   *
   * Note: If the adververtised interface ws changed after the handler was
   * started, this may no longer match what the handler is currently using
   *
   * @return network address packets are set to be advertised from
   */
  public InetAddress getAdvertisedInterface() { return advertisedInterface; }
  
  /**
   * Return the currently in use network interface/address that packets
   * are being advertised as from.  If the handler is not running, this
   * will be null.
   *
   * @return current address packets are advertised from
   */
  public InetAddress getCurrentAdvertisedInterface() { return currentAdvertisedInterface; }

  /**
   * Return true if the passed address is one of the local interfaces.
   *
   * NOTE: This will only work while the handler is running.  Used any
   * other time and it will return false.
   *
   * @param theAddress text address to check.  
   * @return true if address represents a local network interface
   */
  public boolean isLocalNetworkInterface(String theAddress) {
    if (bindableAddresses == null) return false;
    for (BindableAddress boundAddress: bindableAddresses) {
      if (boundAddress.getInetAddress().getHostAddress().equals(theAddress)) return true;
    }
    
    return false;
  }
  
  /**
   * Set type of connection to attempt.  Changing this will
   * not change any current status (if started -- you'd need to stop
   * and restart to make it take effect).  Options are:
   *
   * STANDALONE - An attempt is made to use the xPL port 3865.  If
   *   anything else is using it (a hub or other standalone XPL app), this
   *   will fail.
   *
   * VIA_HUB - An attempt ot made to insure that something is in fact listening
   *   on port 3865 (presumably, a hub).  If nothing is, this fails.  If something
   *   is, then we open a system assigned port #. VIA_HUB is the default connect 
   *   mode.
   *
   * AUTO - We first attempt to start in HUB mode.  If there is something listening
   *   on port 3865, we then connect as if it were VIA_HUB mode.  If there is nothing
   *   listening on port 3865, then we start in STANDALONE mode. 
   *
   * @param newConnectMode connection mode to use next time handler is started
   */
  public void setConnectMode(ConnectMode newConnectMode) {
    connectMode = newConnectMode;
  }
  
  /**
   * Return type of connection to attempt when handler is next started.  This
   * may not match the actual/eventual/current connect mode.
   *
   * @return mode to use for next connection
   */
  public ConnectMode getConnectMode() { return connectMode; }

  /**
   * Return the actual connection mode used for the current or last 
   * connection attempt.  If the handler has never been started, then UNKNOWN
   * is returned.  If the handler has been successfully started (even if
   * it is currently stopped), then this will be VIA_HUB or STANDALONE.  It 
   * will never be AUTO (AUTO controls how we pick the right connect mode, not
   * the actual connection mode used).
   *
   * @return last actual connection mode or UNKNOWN if never connected/started
   */
  public ConnectMode getActualConnectMode() { return actualConnectMode; }
  
  /**
   * Return the actual port # that we are attached/listening on.
   * This is only valid when the handler is started.  When stopped,
   * this returns the last successfully connected port (or 0 if called
   * before the first time the handler successfully connects)
   *
   * @return port # handler is listening on
   */
  public int getConnectedPort() {
    return connectedPort;
  }

  void fireErrorEvent(Throwable theError) {
    fireHandlerEvent(new xPL_MediaHandlerEvent(this, xPL_MediaHandlerEvent.EventCode.HANDLER_ERROR, theError));
  }
  
  boolean sendNetworkMessage(xPL_MessageI theMessage, InetAddress destAddr, int destPort) {
    // Bail out if handler is not running
    if (!handlerStarted) return false;
    
    // validate and format the message for sending
    String textMessage = prepareMessage(theMessage);
    if (textMessage == null) return false;
    
    synchronized(broadcastPacket) {
      // Install the message and send it
      broadcastPacket.setData(textMessage.getBytes());
      broadcastPacket.setAddress(destAddr);
      broadcastPacket.setPort(destPort);
      try {
        broadcastSocket.send(broadcastPacket);
      } catch (IOException ioError) { 
        // Alert listeners
        fireErrorEvent(ioError);
        return false;
      }
    }
    
    // Message sent!
    return true;
  }

  /**
   * Send the passed message via this network
   * @param theMessage message to send
   * @return true if message sent, false if handler not started or there was an error
   */
  public boolean sendMessage(xPL_MessageI theMessage) {
    return sendNetworkMessage(theMessage, broadcastAddr, XPL_PORT);
  }

  private boolean isHubEcho(xPL_MessageI theMessage) {
    if (theMessage == null) return false;

    // We only look at heartbeat messages
    if (!theMessage.isHeartbeat()) return false;
    
    // Parse the identifying marks
    String theIPAddr = theMessage.getNamedValue("remote-ip");
    int thePort = theMessage.getIntNamedValue("port");
    if ((theIPAddr == null) || (theIPAddr.length() == 0) || (thePort == -1)) return false;

    // See if the heartbeat is from us
    if (!theIPAddr.equals(currentAdvertisedInterface.getHostAddress())) return false;
    if (thePort != connectedPort) return false;
    
    // If it is, it's an echo
    return true;
  }
  
  
  /**
   * Send a heartbeat message for the passed device
   * 
   * NOTE: You would not normally use this device as xPL_DeviceI have a method
   * better suited toward sending heartbeats (and in fact, will generally manage
   * the process on their own).
   * 
   * NOTE: out of cadence (forced) heartbeats will only be sent out if the handler
   * has not yet confirmed a hub.  Non-forced heartbeats are always sent.
   * 
   * @param forDevice device to send heartbeat for
   * @param timelyHeartbeat true for normal heartbeats, false for forced, out-of-cadence heartbeats
   * @return true if heartbeat was sent, false if handler is not started or there was an error
   */
  public boolean sendHeartbeat(xPL_DeviceI forDevice, boolean timelyHeartbeat) {
    // Bail out if handler is not running
    if (!handlerStarted) return false;
    
    // If this is a forced heartbeat and the hub is already confirmed, then we
    // do not send out the message because it's unneeded.
    // This assumes forced heartbeats are used primarily in hub discovery
    if (!timelyHeartbeat && hubConfirmed) return true;
    
    // Create the message
    xPL_MutableMessageI heartbeatMessage = xPL_Util.createBroadcastMessage(forDevice.getIdent(), xPL_MessageI.MessageType.STATUS, "hbeat", "app");
    
    // Handle being in configuration mode
    if (forDevice.isConfigurable() && forDevice.isBeingConfigured()) heartbeatMessage.setSchemaClass("config");
    
    // If the device is not enabled, this is a shutdown message then
    if (!forDevice.isEnabled()) heartbeatMessage.setSchemaType("end");
    
    // Install message parameters
    heartbeatMessage.addNamedValue("interval", forDevice.getHeartbeatInterval());
    heartbeatMessage.addNamedValue("port", connectedPort);
    heartbeatMessage.addNamedValue("remote-ip", currentAdvertisedInterface.getHostAddress());
    
    // Add a version, if supplied
    if (forDevice.getVersion() != null) heartbeatMessage.addNamedValue("version", forDevice.getVersion());
    
    // Send the message
    boolean messageSent = sendMessage(heartbeatMessage);
    heartbeatMessage.releaseResources();
    return messageSent;
  }
  
  /**
   * Attempt to start the handler.
   * @throws org.cdp1802.xpl.xPL_MediaHandlerException thrown for any number of reasons (bad network, bad settings, failure to bind to the address, etc)
   */
  public void startHandler() throws xPL_MediaHandlerException {
    
    // Skip this if we are already started
    if (handlerStarted) return;

    // Install advertised address
    currentAdvertisedInterface = advertisedInterface;
    
    // Get list of bindable address to validate our connection address
    bindableAddresses = BindableAddress.getBindableAddresses();
    
    // Setup network interface binding, if needed
    boolean usedDefaultNetwork = false;
    if (networkInterface == null) {
      // Get the "default" address

      // If there is no advertised interface, select the first non-loopback 
      // interface we have (if any)
      if (currentAdvertisedInterface == null) {
        for (BindableAddress bindableAddress: bindableAddresses) {
          if (bindableAddress.isLoopback()) continue;
          currentAdvertisedInterface = bindableAddress.getInetAddress();
          networkInterface = bindableAddress.getInetAddress();
          broadcastAddr = bindableAddress.getBroadcastAddress();
          break;
        }

        if (currentAdvertisedInterface == null) {
          stopHandler();
          throw new xPL_MediaHandlerException("Unable to select/find a usable, non-loopback advertisable network interface");
        }
      }

      usedDefaultNetwork = true;
    } else {
      BindableAddress knownAddress = null;
      for (BindableAddress boundAddress: bindableAddresses) {
        if (boundAddress.getInetAddress().getHostAddress().equals(networkInterface.getHostAddress())) {
          knownAddress = boundAddress;
          break;
        }
      }

      if (knownAddress == null) {
        stopHandler();
        throw new xPL_MediaHandlerException("Specified Network to use is not bound to any known network interface");        
      }
      broadcastAddr = knownAddress.getBroadcastAddress();
    }
    if (currentAdvertisedInterface == null) currentAdvertisedInterface = networkInterface;

    // Create a new ethernet receiver
    networkReceiver = new EthernetReceiver(this);
    try {
      networkReceiver.start();
      connectedNetworkInterface = networkReceiver.getReceiverSocket().getLocalAddress();
    } catch (xPL_MediaHandlerException bindError) {
      stopHandler();
      throw bindError;
    }
    
    // Create the broadcast packet for re-use
    broadcastPacket = new DatagramPacket(new byte[16], 16, broadcastAddr, XPL_PORT);

    // Create a braodcast socket
    try {
      broadcastSocket = new DatagramSocket(0, networkInterface);
    } catch (SocketException socketError) {
      stopHandler();
      throw new xPL_MediaHandlerException("Unable to create broadcast socket", socketError);
    }
    
    // Fetch actual details from receiver
    connectedPort = networkReceiver.getBoundPort();
    actualConnectMode = networkReceiver.getActualConnectMode();
    setDescription("Ethernet Handler (Port " + connectedPort + ")");
    
    // Undo temporary wildcard network install
    if (usedDefaultNetwork) networkInterface = null;

    // Flag this handler as running and unconfirmed, hub wise
    setStarted(true);
    hubConfirmed = false;
    
    // Start Message dispatcher thread
    // Create the message dispatcher
    messageDispatcher = new MessageDispatcher(this);
    synchronized(messageDispatcher) {
      // Fire off the thread
      new Thread(messageDispatcher, "Ethernet Received Message Dispatcher").start();
      
      // Wait for it to startup
      for(;;) {
        try {
          messageDispatcher.wait();
          break;
        } catch (Exception anyError) { continue; }
      }
    }
    
    // Alert listeners
    fireHandlerEvent(new xPL_MediaHandlerEvent(this, xPL_MediaHandlerEvent.EventCode.HANDLER_ACTIVE));
  }

  /**
   * Stop the handler, if it is running.
   */
  public void stopHandler() {
    // Stop the receiver 
    if (networkReceiver != null) {
      // Ask the receiver to stop
      networkReceiver.stop();
      networkReceiver = null;
    }

    // Shutdown the broadcaster
    if (broadcastSocket != null) broadcastSocket.close();
    broadcastSocket = null;
    broadcastPacket = null;
    broadcastAddr = null;
    connectedNetworkInterface = null;
    if (bindableAddresses != null) bindableAddresses.clear();
    bindableAddresses = null;

    currentAdvertisedInterface = null;

    messageDispatcher = null;
    
    // And indicate we are stopped
    setStarted(false);

    // Alert listeners
    fireHandlerEvent(new xPL_MediaHandlerEvent(this, xPL_MediaHandlerEvent.EventCode.HANDLER_INACTIVE));
  }
}
