/*
 * EthernetHub.java
 * Copyright 2004, Gerald R Duprey Jr
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy of
 * the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0 
 *
 * Unless required by applicable law or agreed to in writing, software distributed
 * under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
 * CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License. 
 */

package org.cdp1802.xpl.ethernet;

import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.HashMap;
import java.util.Iterator;
import org.cdp1802.xpl.xPL_Manager;
import org.cdp1802.xpl.xPL_MediaHandlerException;
import org.cdp1802.xpl.xPL_MessageI;
import org.cdp1802.xpl.xPL_MessageListenerI;

/**
 * Implements an xPL Hub
 * @author  gerry
 */
public class EthernetHub implements xPL_MessageListenerI {
  class LocalClient {
    InetAddress clientAddr = null;
    int clientPort = 0;
    int hbeatInterval = 0;
    long lastHeardFromAt = 0;
    
    LocalClient(InetAddress theAddress, int thePort, int heartbeatInterval) {
      clientAddr = theAddress;
      clientPort = thePort;
      hbeatInterval = heartbeatInterval;
      heardFromClient();
    }
    
    public boolean sendToClient(xPL_MessageI theMessage) {
      if (hubHandler == null) return false;
      return hubHandler.sendNetworkMessage(theMessage, clientAddr, clientPort);
    }
    
    public void setHeartbeatInterval(int newInterval) {
      hbeatInterval = newInterval;
      heardFromClient();
    }
    
    public void heardFromClient() {
      lastHeardFromAt = System.currentTimeMillis();
    }
    
    public boolean isClientDead() {
      int minutesSinceLastContact = (int) ((System.currentTimeMillis() - lastHeardFromAt) / 60000);
      if (minutesSinceLastContact > (hbeatInterval * 2)) return true;
      return false;
    }
    
    public String toString() {
      return clientAddr.getHostName() + ':' + clientPort;
    }
    
    public void releaseResources() {
      clientAddr = null;
      clientPort = 0;
      hbeatInterval = 0;
      lastHeardFromAt = 0;
    }
  }
  
  class GrimReaper implements Runnable {
    boolean reaperRunning = false;
    
    public void start() {
      if (reaperRunning) return;
      
      synchronized(this) {
        // Start the thread
        new Thread(this, "xPL Hub Dead Client Reaper").start();
        
        // Wait until the reaper starts
        while (!reaperRunning) {
          try {
            this.wait();
            break;
          } catch (Exception anyError) { continue; }
        }
      }
    }
    
    public void stop() {
      synchronized(this) {
        reaperRunning = false;
        this.notifyAll();
        
        try {
          this.wait();
        } catch (Exception ignoredError) { }
      }
    }

    void reapDeadClients() {
      synchronized(clientList) {
        LocalClient localClient = null;
        Iterator<LocalClient> localClients = clientList.values().iterator();
        while (localClients.hasNext()) {
          localClient = localClients.next();
          if (localClient.isClientDead()) {
           localClients.remove();
           localClient.releaseResources();
          }
        }
      }
    }
    
    public void run() {
      synchronized(this) {
        // Indicate we've started and poke folks
        reaperRunning = true;
        this.notifyAll();
        
        while(reaperRunning) {
          // We reap once a minute
          try {
            this.wait(60000);
            if (reaperRunning) reapDeadClients();
          } catch (Exception anyError) { }
        }
        
        // Notify any waiters that we are on our way out
        this.notifyAll();
      }
    }
  }

  /** 
   * This will search the network (as set in setNetworkInterface) and
   * determine if there is another hub running on this machine.  if
   * there is, true is returned.  If not, then false.
   *
   * @param testNetwork network address to do test on or null for all
   * @return true if there is an internal or external (to the JVM) hub on this network
   */
  public static boolean isExternalHubRunning(InetAddress testNetwork) {
    // Get a network handler and configure it for hub use
    EthernetHandler testHandler = xPL_Manager.getManager().createNetworkHandler();
    if (testNetwork != null) testHandler.setNetworkInterface(testNetwork);
    testHandler.setConnectMode(EthernetHandler.ConnectMode.STANDALONE);
    
    // Attempt to start it
    try {
      testHandler.startHandler();
    } catch (xPL_MediaHandlerException handlerError) { return true; }

    // If we were able to start it, stop it now
    testHandler.stopHandler();
    
    // And say no hub running
    return false;
  }
  
  // Instance data
  xPL_Manager xplManager = xPL_Manager.getManager();
  EthernetHandler hubHandler = null;
  InetAddress hubNetwork = null;
  boolean hubRunning = false;
  GrimReaper grimReaper = null;
  HashMap<String, LocalClient> clientList = new HashMap<String, LocalClient>();
  
  public EthernetHub() {  }
    
  void debug(String theMessage) {
    System.out.println("XPL_HUB::" + theMessage);
  }
  
  /**
   * Set a specific network interface to use.  You can als
   * pass null to bind/use all network interfaces
   *
   * @param theNetworkAddr address of the network interface to use or null
   */
  public void setNetworkInterface(InetAddress theNetworkAddr) {
    hubNetwork = theNetworkAddr;
  }
  
  /**
   * Return the last installed network interface.  If interface has been changed
   * since the hub was started (and if the hub is running), this may not be
   * the same interface the hub is currently using, though next time throws hub
   * is stopped and started, it will use this.  null means all networks are
   * being used.
   *
   * @return address of network interface to use next time the hub is started
   */
  public InetAddress getNetworkInterface() { return hubNetwork; }

  /** 
   * Return true of the hub is actually running/started
   *
   * @return true if hub is started and running, false if not
   */
  public boolean isHubRunning() { return hubRunning; }
  
  /**
   * Attempt to start the hub on the installed network interface.
   * If the interface is null, the hub is started for all interfaces
   *
   * @throws xPL_MediaHandlerException if hub could not start
   */
  public void startHub() throws xPL_MediaHandlerException {
    // Don't attempt to double start the hub
    if (hubRunning) return;
    
    // Start a grim reaper, if needed
    if (grimReaper == null) grimReaper = new GrimReaper();
    grimReaper.start();
    
    // Get a network handler and configure it for hub use
    if (hubHandler == null) hubHandler = xplManager.createNetworkHandler();
    if (hubNetwork != null) hubHandler.setNetworkInterface(hubNetwork);
    hubHandler.setConnectMode(EthernetHandler.ConnectMode.STANDALONE);
    
    // Attempt to start it
    try {
      hubHandler.startHandler();
    } catch (xPL_MediaHandlerException handlerError) {
      hubHandler = null;
      throw handlerError;
    }
    
    // Add us as a message handler
    xplManager.addMessageListener(this);
    
    // Add this to list of managed handlers
    xplManager.addMediaHandler(hubHandler);
    
    // Mark the hub running and exit
    hubRunning = true;
  }

  public void stopHub() {
    if (!hubRunning) return;
    
    // Disconnect us from the world
    xplManager.removeMediaHandler(hubHandler);
    xplManager.removeMessageListener(this);
    
    // Shutdown the handlers
    if (hubHandler != null) hubHandler.stopHandler();
    if (grimReaper != null) grimReaper.stop();
    
    // Clear any clients out
    for (LocalClient theClient: clientList.values()) {
      theClient.releaseResources();
    }
    clientList.clear();
    
    // Flag hub stopped
    hubRunning = false;
  }
  
  // Parse the info from the message.  Use that to lookup
  // the local client.  If no client is found, create a new
  // one.  In either case, return that client.  If the message
  // contains invalid data, return null.
  //
  // In the case of an End message, the client is removed and
  // null returned.
  LocalClient updateLocalClient(xPL_MessageI theMessage) {
    String schemaClass = theMessage.getSchemaClass();
    String schemaType = theMessage.getSchemaType();
    
    // Skip unless it's a heartbeat of some sorts
    if (!theMessage.isHeartbeat()) return null;

    // Parse out the identifying bits
    String msgIPAddr = theMessage.getNamedValue("remote-ip");
    int msgIPPort = theMessage.getIntNamedValue("port", -1);
    int msgInterval = theMessage.getIntNamedValue("interval", -1);
    
    // Skip if this message is missing data or has invalid values
    if ((msgIPAddr == null) || (msgIPAddr.length() == 0) || (msgIPPort == -1) || (msgInterval == -1)) return null;
    
    // See if this is a local device and skip if not
    if (!hubHandler.isLocalNetworkInterface(msgIPAddr)) return null;
    
    // See if this is in fact a packet from a hub port (which does not need to be managed)
    if (msgIPPort == EthernetHandler.XPL_PORT) return null;
    
    // See if there is an existing client
    synchronized(clientList) {
      LocalClient localClient = clientList.get(msgIPAddr + ':' + msgIPPort);
      if (localClient != null) {

        // Check for "end" message
        if (theMessage.isHeartbeatEnd()) {
          clientList.remove(msgIPAddr + ':' + msgIPPort);
          localClient.releaseResources();
          localClient = null;
        } else {
          // Update when we last heard from this client
          localClient.heardFromClient();
        }
      } else {
        // Don't do this if the first thing we ever heard was a shutdown
        if (!theMessage.isHeartbeatEnd()) {
          // Convert IP address into internal format
          InetAddress msgInetAddr = null;
          try {
            msgInetAddr = InetAddress.getByName(msgIPAddr);
          } catch (UnknownHostException badIPAddr) { return null; }

          // Create a new LocalClient
          localClient = new LocalClient(msgInetAddr, msgIPPort, msgInterval);
          clientList.put(msgIPAddr + ':' + msgIPPort, localClient);
        }
      }
      
      // Return the client
      return localClient;
    }
  }

  // Send the message to all existing clients
  void rebroadcastToClients(xPL_MessageI theMessage) {
    synchronized(clientList) {
      for (LocalClient localClient: clientList.values()) {
        localClient.sendToClient(theMessage);
      }
    }
  }
  
  // Handle messages
  public void handleXPLMessage(xPL_MessageI theMessage) {
    // Skip any message not received by our handler
    if (theMessage.getReceivedFrom() != hubHandler) return;
    
    // Update local client (if this is a heart beat message)
    updateLocalClient(theMessage);
    
    // Rebroadcast this message to all clients
    rebroadcastToClients(theMessage);
  }
  
  
}
