/*
 * EthernetReceiver.java
 * Copyright 2004, Gerald R Duprey Jr
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy of
 * the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0 
 *
 * Unless required by applicable law or agreed to in writing, software distributed
 * under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
 * CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License. 
 */

package org.cdp1802.xpl.ethernet;

import java.net.*;
import java.util.LinkedList;
import java.util.Queue;
import org.cdp1802.xpl.*;

/**
 *
 * @author  gerry
 */
public class EthernetReceiver implements Runnable {
  Thread receiverThread = null;
  boolean receiverRunning = false;
  EthernetHandler ethernetHandler = null;
  EthernetHandler.ConnectMode actualConnectionMode = EthernetHandler.ConnectMode.UNKNOWN;

  int boundPort = 0;
  byte packetBuff[] = null;
  DatagramSocket receiveSocket = null;
  Queue<xPL_MessageI> messageQueue = new LinkedList<xPL_MessageI>();

  EthernetReceiver(EthernetHandler theEthernetHandler) {
    ethernetHandler = theEthernetHandler;
  }
  
  private void debug(String debugMessage) {
    System.err.println("ENET_RCVR:: " + debugMessage);
  }
  
  void releaseResources() {
    receiverThread = null;
    receiverRunning = false;
    ethernetHandler = null;
    packetBuff = null;
    receiveSocket = null;
    if (messageQueue != null) messageQueue.clear();
  }

  EthernetHandler.ConnectMode getActualConnectMode() {
    return actualConnectionMode;
  }
  
  int getBoundPort() { return boundPort; }
  
  Thread getReceiverThread() { return receiverThread; }
  boolean isRunning() { return receiverRunning; }

  DatagramSocket getReceiverSocket() { return receiveSocket; }
  
  // Waits and returns the next queued message.  If there are 
  // no queued messages, we normally wait indefintely.  However,
  // if there never will be any more messages (i.e. reader is 
  // shutting down), then we return with null
  xPL_MessageI getNextQueuedMessage() {
    // If the receiver is not running, then don't bother trying
    if (!receiverRunning) return null;
    
    // Sync on the message queue
    synchronized(messageQueue) {
      while(receiverRunning) {
        // If there are any messages, return it
        if (!messageQueue.isEmpty()) return messageQueue.remove();          
        
        // Wait for a new message
        try {
          messageQueue.wait();
        } catch (Throwable anyError) { continue; }
      }
    }
    
    // We are out of messages and the receiver is no longer running
    return null;
  }
  
  // Attempt to make a standalone connection
  private boolean attemptStandaloneConnection() {
    try {
      //receiveSocket = new DatagramSocket(EthernetHandler.XPL_PORT, ethernetHandler.getNetworkInterface());
      receiveSocket = new DatagramSocket(EthernetHandler.XPL_PORT);
    } catch (SocketException bindError) {
      receiveSocket = null;
      return false;
    }
    return true;
  }
  
  // Attempt a hub (non-xPL port based) connection
  private boolean attemptHubConnection() {
    try {
      // receiveSocket = new DatagramSocket(0, ethernetHandler.getNetworkInterface());
      receiveSocket = new DatagramSocket();
    } catch (SocketException bindError) {
      receiveSocket = null;
      return false;
    }
    
    return true;
  }
  
  // Attempt to start the media handler.  This will not return
  // until the media handler/receiver are all up and running.
  void start() throws xPL_MediaHandlerException {
    // Attempt binding in the appropriate mode
    switch(ethernetHandler.getConnectMode()) {
      case STANDALONE:
        if (!attemptStandaloneConnection()) {
          throw new xPL_MediaHandlerException("Unable to bind to xPL port in Standalone mode"); 
        }
        
        // Record actual port and connection type
        boundPort = EthernetHandler.XPL_PORT;
        actualConnectionMode = EthernetHandler.ConnectMode.STANDALONE;
        break;
        
      case VIA_HUB:
        // Make a hub based connection
        if (!attemptHubConnection()) {
          throw new xPL_MediaHandlerException("Unable to bind to any port in Via Hub mode");
        }
        
        // Record actual port and connect mode
        boundPort = receiveSocket.getLocalPort();
        actualConnectionMode = EthernetHandler.ConnectMode.VIA_HUB;
        break;

      case AUTO:
        // See if we can make a standalone connection.  If so, there is no hub, so we start
        // in standalone mode.  If there is a hub, we start in HUB mode.
        if (attemptStandaloneConnection()) {
          boundPort = EthernetHandler.XPL_PORT;
          actualConnectionMode = EthernetHandler.ConnectMode.STANDALONE;
        } else {
          // Make a hub based connection
          if (!attemptHubConnection()) {
            throw new xPL_MediaHandlerException("Unable to bind to any port in Auto mode");            
          }
          
          // Record actual port and connect mode
          boundPort = receiveSocket.getLocalPort();
          actualConnectionMode = EthernetHandler.ConnectMode.VIA_HUB;
        }
        break;
    }

    // Allocate a packet buffer, if needed
    if (packetBuff == null) packetBuff = new byte[1500];

    // Attempt to maximize the receive buffer
    int bufferSize = -1;
    for(;;) {
      if (bufferSize == -1) {
        try {
          bufferSize = receiveSocket.getReceiveBufferSize();
        } catch (Throwable anyError) { break; }
      }

      // Try a new buffer size
      bufferSize *= 2;
      
      // First try to install the current buffer size
      try {
        receiveSocket.setReceiveBufferSize(bufferSize);
      } catch (Exception anyError) { break; }
      
      // If the actual size is lower than the attempted install
      // size, we hit the O/S limit on this and we are done
      try {
        if (receiveSocket.getReceiveBufferSize() < bufferSize) {
          bufferSize = receiveSocket.getReceiveBufferSize();
          break;
        }
      } catch (Exception anyError) { break; }
    }
    
    // We are ready to start our receiver thread.  We start this
    // and once started, we wait until it is up and going.
    receiverThread = new Thread(this, "xPL Receiver: Ethernet (Port " + boundPort + ")");
    receiverThread.setPriority(Thread.MAX_PRIORITY);
    synchronized(messageQueue) {
      // Start the thread
      receiverThread.start();
      
      // Wait until something interesting happens (either we are running
      // or the thread dies because of an error/failure)
      while (!receiverRunning && receiverThread.isAlive()) {
        try {
          messageQueue.wait();
        } catch (Throwable anyError) { }
      }
    }
    
    // See if we are OK
    if (!receiverRunning || !receiverThread.isAlive()) {
      receiveSocket.close();
      releaseResources();
      throw new xPL_MediaHandlerException("Unable to start xPL receiver");
    }
  }
  
  // Stop the receiver.  Once stopped, this instances resources are
  // released and it cannot be used again.
  void stop() {
    // Bail out if we are not running
    if (!receiverRunning) return;
    
    // Clear the running flag and close the socket
    receiverRunning = false;
    receiveSocket.close();
    
    // Wait for the reader thread to stop
    for(;;) {
      try {
        receiverThread.join();
        releaseResources();
        return;
      } catch (Throwable anyError) { }
    }
  }
  
  // Fetch new messages and distribute them.  The messages are simply queued
  // so that the receiver can keep pace with the incoming message flow
  public void run() {
    // Create the receive packet
    DatagramPacket receivedPacket = new DatagramPacket(packetBuff, packetBuff.length);
    xPL_Manager xplManager = xPL_Manager.getManager();
    String textMessage = null;
    xPL_MessageI receivedMessage = null;
    int bytesRead = 0;
    boolean rapidReadMode = false;
    
    // Actually start reading packets
    synchronized(messageQueue) {
      receiverRunning = true;
      messageQueue.notifyAll();
    }
    
    try {
      while(receiverRunning) {
        // Wait here until I we get a packet
        try {
          // Wait for eternity for a new packet, then insure we got at least
          // a vaugely plausible packet of data
          receiveSocket.setSoTimeout(0);
          receiveSocket.receive(receivedPacket);
          if ((bytesRead = receivedPacket.getLength()) < 20) continue;
        } catch (Throwable variousError) { 
          ethernetHandler.fireErrorEvent(variousError);
          continue;
        }

        synchronized(messageQueue) {
          // We now loop trying to read packets in as quickly as possible
          // until we exhaust any pending packets
          rapidReadMode = true;
          
          while(rapidReadMode && receiverRunning) {
            // Decode the last received packet
            try {
              textMessage = new String(packetBuff, 0, bytesRead);
              receivedMessage = xPL_Util.parseMessage(textMessage, ethernetHandler);

              // Queue it
              messageQueue.add(receivedMessage);
              messageQueue.notifyAll();
            } catch (xPL_ParseException parseError) {
              debug("Received bad xPL message from network -- " + parseError.getMessage());
              ethernetHandler.fireErrorEvent(parseError);
            } catch (Throwable anyOtherError) {
              debug("Something bad happened in receiver -- " + anyOtherError.getMessage());
              ethernetHandler.fireErrorEvent(anyOtherError);
              anyOtherError.printStackTrace();
            }
            
            // Wait a bit for another packet.  If none is found in
            // quick order, we're done and we'll go back to the
            // longer term sleep.
            while(rapidReadMode && receiverRunning) {
              try {
                receiveSocket.setSoTimeout(50);
                receiveSocket.receive(receivedPacket);
                if ((bytesRead = receivedPacket.getLength()) < 20) continue;
                break;
              } catch (SocketTimeoutException readTimeout) {   
                rapidReadMode = false;
                break;
              } catch (Throwable anyOtherException) {
                ethernetHandler.fireErrorEvent(anyOtherException);
                continue;
              }
            }
          }
        }
      }
    } finally {
      receiverRunning = false;
      
      synchronized(messageQueue) {
        messageQueue.notifyAll();
      }
    }
  }
  
}
