/*
 * LightingDevice.java
 *
 * Copyright (c) 2005, Gerald R Duprey Jr
 * More info at http://www.xPL4Java.org
 */

package org.cdp1802.xpl.lighting;

import java.util.*;

/**
 * Describes a single controllable lighting device.
 *
 * <P>NOTES ON CHANNELS
 *
 * <P>Most devices have only a single channel (that is, in most cases, a channel is
 * a controllable load).  Some have more than one and a very few have none at all.
 * Channels are numbered from 1 up.  Some commands will accept a channel #0 which
 * is a wildcard and applies to all channels on the device.
 *
 * <P>For devices with no channels, the channel count is 0 and the primary channel
 * will be returned as INVALID_CHANNEL.
 *
 * <P>Devices that have more than channel tend to have a primary channel which is
 * often, though not necessarily, channel 1.  
 *
 * @author gerry
 */
public class LightingDevice {
  /** Used to indicate an invalid channel.  Only valid for devices that do not have any channels */
  public static final int INVALID_CHANNEL = -1;

  /** Used to indicate all channels for a device */
  public static final int ALL_CHANNELS = 0;
  
  /** Used to indicate the device should use it's default fade rate */
  public static final float DEFAULT_FADE_RATE = 99999;
  
  /** Used to indicate invalid/unassigned device state */
  public static final int UNASSIGNED_DEVICE_STATE = -1;

  /** Special value for commands telling devices to go to their last dim level (if supported, 100% otherwise) */
  public static final int LAST_DIM_LEVEL = 255;

  /** Special value for commands telling devices to go to their defaultdim level (if supported, 100% otherwise) */
  public static final int DEFAULT_DIM_LEVEL = 256;
  
  LightingNetwork deviceNetwork = null;
  String deviceID = null;
  
  String firmwareVersion = "";
  String manufacturerID = "";
  String manufacturerName = "";
  String productID = "";
  String productName = "";
  
  String deviceComment = "";
  
  int channelCount = ALL_CHANNELS;
  int primaryChannel = INVALID_CHANNEL;
  
  boolean reportsManualChanges = false;
  
  String deviceName = "";
  LightingRoom deviceRoom = null;
  LightingFloor deviceFloor = null;
  
  ArrayList<LightingSceneDevice> deviceScenes = new ArrayList<LightingSceneDevice>();
  
  boolean isDimmable[] = null;;
  int deviceState[] = null;
  float defaultFadeRate[] = null;
  
  LightingDevice() { }
  
  LightingDevice(LightingNetwork theNetwork, String deviceID, int channelCount, int primaryChannel) {
    setDeviceInfo(theNetwork, deviceID, channelCount, primaryChannel);
  }
  
  void setDeviceInfo(LightingNetwork theNetwork, String deviceID, int channelCount, int primaryChannel) {
    this.deviceNetwork = theNetwork;
    this.deviceID = deviceID;
    this.channelCount = channelCount;
    this.primaryChannel = primaryChannel;
    
    // Handle channel stuff
    if ((this.channelCount) < 1) {
      this.channelCount = 0;
      this.primaryChannel = LightingDevice.INVALID_CHANNEL;
    } else {
      if (primaryChannel < 1) primaryChannel = 1;
      if (primaryChannel > this.channelCount) primaryChannel = 1;
      this.primaryChannel = primaryChannel;
    }
    
    // Init space for device channel level tracking
    deviceState = new int[(this.channelCount < 1) ? 1 : this.channelCount];
    isDimmable = new boolean[(this.channelCount < 1) ? 1 : this.channelCount];
    defaultFadeRate = new float[(this.channelCount < 1) ? 1 : this.channelCount];
    for (int chanIndex = 0; chanIndex < deviceState.length; chanIndex++) {
      deviceState[chanIndex] = UNASSIGNED_DEVICE_STATE;
      isDimmable[chanIndex] = false;
      defaultFadeRate[chanIndex] = 0;
    }
  }
  
  void debug(String theMessage) {
    deviceNetwork.getGateway().intDebug("DEVICE[" + deviceID + "]:: " + theMessage);
  }
  
  void error(String theMessage) {
    deviceNetwork.getGateway().intError("DEVICE[" + deviceID + "]:: " + theMessage);
  }
  
  /**
   * Return the network associated with this device.  All devices are associated
   * with a single network
   *
   * @return network associated with this device
   */
  public LightingNetwork getNetwork() { return deviceNetwork; }
  
  /**
   * Return this devices ID (or device number).  Each device in a network has
   * a unique ID.
   *
   * @return device ID for this device
   */
  public String getDeviceID() { return deviceID; }

  /**
   * Get the textual name for this device
   *
   * @return name of this device
   */
  public String getDeviceName() { return deviceName; }
  void setDeviceName(String newName) {
    if ((newName == null) || (newName.length() == 0) || newName.equals(deviceName)) return;
    
    deviceName = newName;
  }

  /**
   * Get the comment, if any, for this device
   */
  public String getComment() { return deviceComment; }
  
  /**
   * Get the devices firmware version, if any
   */
  public String getFirmwareVersion() { return firmwareVersion; }

  /**
   * Get the manufacturer ID for this device, if any
   */
  public String getManufacturerID() { return manufacturerID; }

  /**
   * Get the manufacturers name of this device, if any
   */
  public String getManufacturerName() { return manufacturerName; }
  
  /**
   * Get the product ID for this device, if any
   */
  public String getProductID() { return productID; }
  
  /**
   * Get the product name for this device, if any
   */
  public String getProductName() { return productName; }
  
  /**
   * Get the room this device is in.
   *
   * @return instance of the room this device is in/assigned to
   */
  public LightingRoom getRoom() { return deviceRoom; }
  void setRoom(LightingRoom theRoom) {
    // Handle a bad or non-update
    if ((theRoom == null) || (theRoom == deviceRoom)) return;
    
    // Set room for device
    if (deviceRoom != null) deviceRoom.removeDevice(this);
    deviceRoom = theRoom;
    if (deviceRoom != null) deviceRoom.addDevice(this);
  }
  
  /**
   * Get the floor this device is in.
   *
   * @return instance of the floor this device is in/assigned to
   */
  public LightingFloor getFloor() { return deviceFloor; }
  void setFloor(LightingFloor theFloor) {
    // Handle a bad or non-update
    if ((theFloor == null) || (theFloor == deviceFloor)) return;
    
    // Set floor for device
    if (deviceFloor != null) deviceFloor.removeDevice(this);
    deviceFloor = theFloor;
    if (deviceFloor != null) deviceFloor.addDevice(this);
  }
  
  /**
   * Determine if this device will report it's state when manually
   * changed.
   */
  public boolean getReportsManualChanges() { return reportsManualChanges; }
  
  /**
   * Return the number of channels this device has
   *
   * @return number of channels or 0 if device has none
   */
  public int getChannelCount() { return channelCount; }
  
  /**
   * Return primary channel for the device (or INVALID_CHANNEL if no channels)
   *
   * @return primary channel for device or INVALID_CHANNEL
   */
  public int getPrimaryChannel() { return primaryChannel; }
  
  /**
   * Get the default fade rate for this device.  All dimmable devices have
   * a default fade rate -- the rate that the device takes to transition from
   * one level to another.  On most devices, this rate is configuable.
   *
   * <P>When a request is made to change the level of a device and that request
   * is not specific about how fast, this rate is used.
   *
   * @param forChannel channel to get default fade rate for
   * @return default fade rate for this device in seconds
   */
  public float getDefaultFadeRate(int forChannel) { 
    return defaultFadeRate[forChannel - 1];
  }

  /**
   * Get the default fade rate for this device.  All dimmable devices have
   * a default fade rate -- the rate that the device takes to transition from
   * one level to another.  On most devices, this rate is configuable.
   *
   * <P>When a request is made to change the level of a device and that request
   * is not specific about how fast, this rate is used.
   *
   * @return default fade rate for this device in seconds
   */
  public float getDefaultFadeRate() { 
    return getDefaultFadeRate(primaryChannel);
  }

  /**
   * Is this device dimmable or not.  Being dimmable is a combination of whether
   * the underlying device is actually capable of dimming and if the device has
   * been configured to dim or not (you can configure dimming modules to not
   * dim and they will only have an on and off status).
   *
   * @param theChannel channel to check to see if dimmable or not
   * @return is this device capable of being dimmed or not
   */
  public boolean isDimmable(int theChannel) {
    return isDimmable[theChannel - 1];
  }

  /**
   * Is this device dimmable or not.  Being dimmable is a combination of whether
   * the underlying device is actually capable of dimming and if the device has
   * been configured to dim or not (you can configure dimming modules to not
   * dim and they will only have an on and off status).
   *
   * @return is this device capable of being dimmed or not
   */
  public boolean isDimmable() { 
    return isDimmable(primaryChannel);
  }

  /**
   * Add a listener that will receives events generated by this device
   * 
   * <P>NOTE: This is just shorthand for LightingGateway.getManager().addDeviceListener(theListener, this);
   * 
   * @param theListener listener to receive this devices events
   */
  public void addListener(LightingDeviceListenerI theListener) {
    deviceNetwork.getGateway().addDeviceListener(theListener, this);
  }
  
  /**
   * Remove a listener for this device
   * 
   * <P>NOTE: This is just shorthand for LightingGateway.getManager().removeDeviceListener(theListener, this);
   * 
   * @param theListener the listener to remove
   */
  public void removeListener(LightingDeviceListenerI theListener) {
    deviceNetwork.getGateway().removeDeviceListener(theListener, this);
  }
  
  /**
   * Request this device send us it's current status
   *
   * <P>NOTE: if there is a request to send this devices status already
   * in the queue, this will be skipped
   */
  public void requestStateFromDevice() {
    deviceNetwork.getGateway().sendStateRequest(this);
  }

  
  /********* Scene Related Services ***********/
  
  /**
   * Get the number of scenes this device is associated with.
   *
   * @return number of scenes device is currently associated with
   */
  public int getSceneCount() { return deviceScenes.size(); }
  
  /**
   * Get a scene associated with this device.  If the passed index is
   * invalid (too high/low), null is returned.  Indicies start at 0.
   *
   * @param sceneIndex index into the scenes to return
   * @return instance of the scene/device record or null if bad index
   */
  public LightingSceneDevice getSceneAt(int sceneIndex) {
    if ((sceneIndex < 0) || (sceneIndex >= deviceScenes.size())) return null;
    return deviceScenes.get(sceneIndex);
  }
  
  /**
   * Given a specific scene ID, see if this device is associated with it
   * and if so, return the scene/device record describing that association
   *
   * @param sceneID scene ID to check device for
   * @return instance of the scene/device record or null if device is not associated with the passed scene ID
   */
  public LightingSceneDevice getSceneByID(String sceneID) {
    synchronized(deviceScenes) {
      for (LightingSceneDevice devScene: deviceScenes) {
        if (devScene.getScene().getSceneID().equals(sceneID)) return devScene;
      }
    }
    return null;
  }
  
  /**
   * Return a collection of scenes for this device, useful for
   * iterating over all the scenes.
   *
   * @return collection of the scenes this device is associated with
   */
  public Collection<LightingSceneDevice> getScenes() { return deviceScenes; }
  
  
  /******** Device State Related Services *************/
  
  /**
   * Is this device On?  
   *
   * @param forChannel check the state for the channel (channels are 1 based (i.e. channel #1 is passed as 1)
   * @return true if device is on (not at level 0) or false if off
   */
  public boolean isDeviceOn(int forChannel) {
    if ((forChannel < 1) || (forChannel > deviceState.length)) return false;
    return deviceState[forChannel - 1] > 0;
  }

  /**
   * Is this device On?
   *
   * <P>NOTE: This checks the primary channel for the device which is almost
   * always what you want (most devices have only a single channel);
   *
   * @return true if device is on (not at level 0) or false if off
   */
  public boolean isDeviceOn() {
    return isDeviceOn(primaryChannel);
  }
  
  /**
   * Get the current level for given channel of this device. This will return a value
   * between 0 and 100 (percent).
   *
   * @param forChannel channel to check level for
   * @return level device is currently at, 0 to 100
   */
  public int getDeviceLevel(int forChannel) { return deviceState[forChannel - 1];  }

  /**
   * Get the current level of the primary channel for this device. This will return a value
   * between 0 and 100 (percent).
   *
   * @return level device is currently at, 0 to 100
   */
  public int getDeviceLevel() { return getDeviceLevel(primaryChannel); }
  
  /**
   * Tell the device to change the level of the passed channel to the passed
   * level at the passed rate.
   *
   * @param toLevel new level (0-100) for the device to goto or DEFAULT_DIM_LEVEL or LAST_DIM_LEVEL
   * @param fadeRate rate/speed device to goto new level at in seconds
   * @param channel channel that device should change to reflect this
   */
  public void setDeviceLevel(int toLevel, float fadeRate, int channel) {
    transmitNewDeviceLevel(toLevel, fadeRate, channel);
  }
  
  /**
   * Tell the device to change it's level to the passed level at the 
   * passed rate.
   *
   * @param toLevel new level (0-100) for the device to goto or DEFAULT_DIM_LEVEL or LAST_DIM_LEVEL
   * @param fadeRate rate/speed device to goto new level at in seconds
   */
  public void setDeviceLevel(int toLevel, float fadeRate) {
    setDeviceLevel(toLevel, fadeRate, ALL_CHANNELS);
  }
  
  /**
   * Tell the device to change it's level to the passed level.  The passed
   * level must be between 0 and 100 or DEFAULT_DIM_LEVEL or LAST_DIM_LEVEL
   *
   * <P>In this method, the switch will use it's default fade rate to effect
   * it's change
   *
   * @param theLevel level to set device to (0-100) or DEFAULT_DIM_LEVEL or LAST_DIM_LEVEL
   */
  public void setDeviceLevel(int theLevel) {
    setDeviceLevel(theLevel, DEFAULT_FADE_RATE, ALL_CHANNELS);
  }

  // Invoked whenever the a devices state has changed.
  //
  // Should take care of any house keeping and then fire a state changed
  // message for those who care to hear it.
  void deviceStateChanged(int forChannel) {
    deviceNetwork.getGateway().fireDeviceEvent(new LightingDeviceEvent(this, LightingDeviceEvent.EventCode.DEVICE_STATE_CHANGED, forChannel));
  }
  
  /**
   * Update the internal state of this devices level, but do not send any
   * commands to the gateway.  
   *
   * <P>If the new level is not a change, we discard and return
   *
   * <P>If it's a change to a new, invalid level, we do not apply it, but we do
   * queue a request to the device to ask what it's real level is
   *
   * <P>Otherwise, install the new level and fire an event to tell folks
   * about the new level
   *
   * @param newLevel new level to apply
   * @param atFadeRate fade rate to apply at
   * @param forChannel channel to apply level change to
   */
  void updateInternalDeviceLevel(int newLevel, float atFadeRate, int forChannel) {
    // Check for a "last level" value.  Since we don't know what that
    // was, we'll just ask the device it's current setting
    if ((newLevel < 0) || (newLevel > 100)) {
      if (LightingGateway.DEBUG_MODE) debug("Got request to set light to non-normative level -- queueing status request to confirm");
      deviceNetwork.getGateway().sendStateRequest(this);
      return;
    }

    // Handle a potential range of updates
    int lowChannel = 0, highChannel = deviceState.length - 1;
    if (forChannel > 0) {
      lowChannel = forChannel - 1;
      highChannel = forChannel - 1;
    }
    
    // Update all affected channels
    for (int chanIndex = lowChannel; chanIndex <= highChannel; chanIndex++) {
      if (newLevel == deviceState[chanIndex]) continue;

      // Otherwise, install the new setting
      if (LightingGateway.DEBUG_MODE) debug("Level changing on channel " + chanIndex + " from " + deviceState[chanIndex] + " to " + newLevel);
      deviceState[chanIndex] = newLevel;
      deviceStateChanged(chanIndex + 1);
    }
  }

  /**
   * Given the passed parameters, send a command out telling the device
   * to go to this level at this rate on this channel.  If the command
   * was successful, then return true.  If the command was not confirmed
   * or timed out, return fase
   *
   * <P>This should NEVER be called directly -- it is a helper/intercept point 
   * used exclusively by transmitNewDeviceLevel
   *
   * <P>Descendent classes can override this to differntly format the message
   *
   * @param theLevel level to set device to (0-100 or DEFAULT_DIM_LEVEL OR LAST_DIM_LEVEL
   * @param theFadeRate rate level should fade to at in seconds
   * @param theChannel channel to set level for
   * @return true of the device level was confirmed received by the device, false if not or timedout
   */
  boolean sendNewDeviceLevelMessage(int theLevel, float theFadeRate, int theChannel) {
    return deviceNetwork.getGateway().sendMessage(new LightingMessage(this, LightingMessage.MsgType.GOTO, theLevel, theFadeRate, theChannel));
  }

  /** 
   * Invoked to post an update to the devices internal state after a command
   * to a device has been successfully sent.
   *
   * <P>This should NEVER be called directly -- it is a helper/intercept point 
   * used exclusively by transmitNewDeviceLevel
   *
   * <P>Descendent classes can override this to differntly handle applying/interpretting
   * updates in their own way
   * 
   * @param theLevel new level to install for device
   * @param theFadeRate new fade rate level was applied at in seconds
   * @param theChannel channel level/fade rate were applied to
   */
  void installNewDeviceLevel(int theLevel, float theFadeRate, int theChannel) {
    updateInternalDeviceLevel(theLevel, theFadeRate, theChannel);
  }
  
  /**
   * Given the passed candidate device level, fade rate and channel, determine
   * if this represents a real change to the device that should result in a 
   * command being sent to the device.  If true is retuned, this is new and requires
   * a command be sent to the device.  If false, it's not new and no command 
   * should be sent.
   *
   * <P>This should NEVER be called directly -- it is a helper/intercept point 
   * used exclusively by transmitNewDeviceLevel
   *
   * <P>Descendent classes can override this to allow more complex determinations
   * about whether the new level represents a real change or not.
   *
   * @param newDeviceLevel candidate new level for device
   * @param atFadeRate candidate new fade rate in seconds
   * @param forChannel candidate channel for new level
   * @param true if these values are a change and require a command be sent to the device to effect the new level
   */
  boolean isReallyNewDeviceLevel(int newDeviceLevel, float atFadeRate, int forChannel) {
    if ((newDeviceLevel < 0) || (newDeviceLevel > 100)) return true;
    
    // Handle eitehr a specific channel or a wildcard channel
    int lowChannel = 0, highChannel = deviceState.length - 1;
    if (forChannel > 0) {
      lowChannel = forChannel - 1;
      highChannel = forChannel - 1;
    }
    
    // Look for a mismatch and return true as soon as there is one
    for (int chanIndex = lowChannel; chanIndex <= highChannel; chanIndex++) {
      if (newDeviceLevel != deviceState[chanIndex]) return true;
    }
    
    // No mismatches -- not really a new level
    return false;
  }
  
  /**
   * Given the passed level values, determine if they are a real change for the
   * device or not.  If not, return true and leavel.  If they are a real change,
   * attempt to send a command to the device and if that command is successful,
   * install the values as the devices current state
   * 
   * <P>Descendent classes can override this to differntly control and effect
   * changes to a devices level
   *
   * @param newDeviceLevel new level to conditionally set device to (0-100, DEFAULT_DIM_LEVEL or LAST_DIM_LEVEL)
   * @param atFadeRate fade rate to set device level in seconds
   * @param toChannel channel to apply level to
   * @return true if the device is updated (either already updated or successfully commanded the device), false if there was an error and the update was not applied
   */
  boolean transmitNewDeviceLevel(int newDeviceLevel, float atFadeRate, int toChannel) {
    // See if there is any reason to even send this
    if (!isReallyNewDeviceLevel(newDeviceLevel, atFadeRate, toChannel)) return true;
    
    // Send the request out
    if (!sendNewDeviceLevelMessage(newDeviceLevel, atFadeRate, toChannel)) {
      if (LightingGateway.DEBUG_MODE) debug("Request to set device level failed -- Asking device for current state");
      deviceNetwork.getGateway().sendStateRequest(this);
      return false;
    }
    
    // Install the new device levels, possibly firing off an event
    installNewDeviceLevel(newDeviceLevel, atFadeRate, toChannel);
    if (LightingGateway.DEBUG_MODE) debug("Confirmed request to set device to level succeedded -- updating internal state");
    return true;
  }
  
  /**
   * Parse and install a device state report.  Most times, this means just installing
   * the devices and firing a message
   *
   * <P>Descendent classes can override this to different parse status reports
   *
   * @param theMessage status report message
   */
  void receiveDeviceStateReport(LightingMessage theMessage) {
    // Update our internal level 
    updateInternalDeviceLevel(theMessage.getLevel(), theMessage.getFadeRate(), theMessage.getChannel());
  }
  
  /**
   * Receive a GOTO type message -- a command to direct the device to 
   * a new level (presumably sent by someone else -- not us
   *
   * @param theMessage message received meant to change a devices level
   */
  void receiveDeviceGotoCommand(LightingMessage theMessage) {
    updateInternalDeviceLevel(theMessage.getLevel(), theMessage.getFadeRate(), theMessage.getChannel());
  }
  
  
  /**
   * Process a newly received message for this device.  
   *
   * <P>NOTE: In the case of wildcard commands, the device in the passed message 
   * may not match this device.  But in any other case, it will match this 
   * device
   *
   * @param theMessage message to be received by this device
   */
  void receiveMessage(LightingMessage theMessage) {
    switch(theMessage.getMessageType()) {
      case GOTO:
        receiveDeviceGotoCommand(theMessage);
        break;
        
      case STATE_REPORT:
        receiveDeviceStateReport(theMessage);
        break;
        
    }
  }
  
  // Hook for descendent device classes to copy data after all other device
  // data is copied from the parent
  void customCopyFrom(LightingDevice parentDevice) { }
  
  // Copy our settings from another device.  This is pretty much only of value
  // after device discovery.
  // This device (the target) must NOT be a listed, known device
  void copyFrom(LightingDevice parentDevice) {
    deviceNetwork = parentDevice.deviceNetwork;
    deviceID = parentDevice.deviceID;
    
    firmwareVersion = parentDevice.firmwareVersion;
    
    deviceName = parentDevice.deviceName;
    setRoom(parentDevice.deviceRoom);
    setFloor(parentDevice.deviceFloor);
    
    deviceState = parentDevice.deviceState;
    
    // Copy scenes and fix their references
    deviceScenes.clear();
    deviceScenes.addAll(parentDevice.deviceScenes);
    for (LightingSceneDevice theDeviceScene: deviceScenes) {
      theDeviceScene.theDevice = this;
    }
    
    // Allow descendent classes easy copying
    customCopyFrom(parentDevice);
  }
  
  /**
   * Get a printable summary of this device composed of it's name and ID
   *
   * @return printable id of this device including it's name and ID
   */
  public String toString() { return deviceName + " (" + deviceID + ")"; }
  
  /**
   * Release resources when the device is no longer needed
   * 
   * <P>NOTE: Only the LightingGateway should ever invoke this.  Anyone else
   * doing so will surely break things badly
   */
  void releaseResources() {
    // Insure nothing is left handing around
    if (deviceNetwork != null) deviceNetwork.getGateway().removeAllRelatedDeviceListeners(this);
    
    // Clear up device scenes
    if (deviceScenes != null) {
      int sceneCount = deviceScenes.size();
      for (int sceneIndex = sceneCount - 1; sceneIndex >= 0; sceneIndex--) {
        deviceScenes.get(sceneIndex).getScene().removeDevice(deviceScenes.get(sceneIndex));
      }
      deviceScenes.clear();
    }

    isDimmable = null;
    deviceState = null;
    defaultFadeRate = null;
    deviceNetwork = null;
    deviceID = null;
    deviceName = null;
    deviceScenes = null;
    deviceRoom = null;
    deviceFloor = null;
  }
}
