/*
 * LightingDiscovery.java
 *
 * Copyright (c) 2005, Gerald R Duprey Jr
 * More info at http://www.xPL4Java.org
 */

package org.cdp1802.xpl.lighting;

import java.util.*;

import org.cdp1802.xpl.*;
import org.cdp1802.xpl.device.*;
import org.cdp1802.xpl.xPL_MessageI.MessageType;

/**
 * Will use the passed adapter to learn everything possible about
 * the underlying lighting gateway
 *
 * @author gerry
 */
class LightingDiscovery implements Runnable {
  static final int ERROR_VALUE = -1;
  
  /**
   * Discovery state
   * 0 - idle
   * 1 - waiting for gateway info
   * 2 - waiting for network list
   * 3 - waiting for network info
   * 4 - waiting for scene list
   * 5 - waiting for scene info
   * 6 - waiting for device list
   * 7 - waiting for device info
   */
  private int discoveryState = 0;
  
  private LightingXPLAdapter theAdapter = null;
  private xPL_MutableMessageI requestMessage = null;
  
  private LinkedList<String> NetworkIDList = new LinkedList<String>();
  private LinkedList<String> SceneIDList = new LinkedList<String>();
  private LinkedList<String> DeviceIDList = new LinkedList<String>();
  
  protected LightingGateway ourGateway = null;
  protected LightingNetwork importNetwork = null;
  
  LightingDiscovery(LightingGateway ourGateway, LightingXPLAdapter theAdapter) { 
    this.ourGateway = ourGateway;
    this.theAdapter = theAdapter;
    requestMessage = theAdapter.getAdapterDevice().createBroadcastMessage(xPL_MessageI.MessageType.COMMAND, LightingXPLAdapter.SCHEMA_CLASS, "request");
  }

  void setRequestTarget(xPL_IdentifierI newDestination) {
    synchronized(requestMessage) {
      requestMessage.setTarget(newDestination);
    }
  }
  
  void sendMessage(xPL_MessageI theMessage) {
    theAdapter.getAdapterDevice().sendMessage(theMessage);
  }
  
  
  protected void debug(String theMessage) {
    ourGateway.intDebug("LIGHTING_DISCOVERY:: " + theMessage);
  }

  protected void warn(String theMessage) {
    ourGateway.intWarn("LIGHTING_DISCOVERY:: " + theMessage);
  }

  protected void error(String theMessage) {
    ourGateway.intError("LIGHTING_DISCOVERY:: " + theMessage);
  }
  
  protected boolean importNetwork(String networkID, String networkName, String networkPassword) {
    // Create new network, if needed
    if ((importNetwork = ourGateway.getNetworkByID(networkID)) == null) {
      importNetwork = new LightingNetwork(ourGateway, networkID);
      ourGateway.networkList.add(importNetwork);
    }
    
    importNetwork.networkName = (networkName == null ? "" : networkName);
    importNetwork.networkPassword = (networkPassword == null ? "" : networkPassword);

    // Handle installing the preferred network for the gateway.
    if (ourGateway.preferredNetworkID == null) ourGateway.preferredNetworkID = networkID;
    if (ourGateway.preferredNetworkID.equals(networkID)) ourGateway.preferredNetwork = importNetwork;

    // And we are ready to import
    return true;
  }
  
  protected boolean importScene(String sceneID, String sceneName) {
    boolean sceneNew = false;
    LightingScene importScene = null;
    
    // Fetch/create scene
    if ((importScene = importNetwork.getSceneByID(sceneID)) == null) {
      importScene = new LightingScene(importNetwork, sceneID);
      sceneNew = true;
    }
    
    // Set attributes
    if ((sceneName == null) || (sceneName.length() == 0)) {
      importScene.sceneName = "";
    } else {
      importScene.sceneName = sceneName;
    }
    
    // Install it
    if (sceneNew) importNetwork.addScene(importScene);
    
    // Purge all scene devices (only way to insure upcoming scene devices will import right)
    importScene.removeAllDevices();
    return true;
  }

  protected boolean importDevice(String deviceID, String deviceName, String manufacturerInfo, String productInfo, String firmwareVersion, int channelCount, int primaryChannel, String roomName, String floorName, String comment, boolean reportsManualChanges) {
    boolean deviceNew = false;
    LightingDevice importDevice = null;
    
    // Fetch/create device
    if ((importDevice = importNetwork.getDeviceByID(deviceID)) == null) {
      importDevice = new LightingDevice(importNetwork, deviceID, channelCount, primaryChannel);
      deviceNew = true;
    } else {
      if ((importDevice.channelCount != channelCount) || (importDevice.primaryChannel != primaryChannel)) {
        importDevice.setDeviceInfo(importNetwork, deviceID, channelCount, primaryChannel);
      }
    }

    // Check for name change
    importDevice.deviceName = (deviceName == null ? "" : deviceName);

    // Do simple stuff
    importDevice.deviceComment = (comment == null ? "" : comment);
    importDevice.reportsManualChanges = reportsManualChanges;

    // Get the room & install it
    LightingRoom importRoom = null;
    if ((roomName == null) || (roomName.length() == 0)) {
      importRoom = importNetwork.getDefaultRoom();
    } else {
      if ((importRoom = importNetwork.getRoomNamed(roomName)) == null) {
        importRoom = new LightingRoom(importNetwork, roomName);
        importNetwork.addRoom(importRoom);
      }
    }
    importDevice.setRoom(importRoom);

    // Get the floor & install it
    LightingFloor importFloor = null;
    if ((floorName == null) || (floorName.length() == 0)) {
      importFloor = importNetwork.getDefaultFloor();
    } else {
      if ((importFloor = importNetwork.getFloorNamed(floorName)) == null) {
        importFloor = new LightingFloor(importNetwork, floorName);
        importNetwork.addFloor(importFloor);
      }
    }
    importDevice.setFloor(importFloor);

    // Parse manufacturer data
    if ((manufacturerInfo == null) || (manufacturerInfo.length() == 0)) {
      importDevice.manufacturerID = "";
      importDevice.manufacturerName = "";
    } else {
      int delimPtr = manufacturerInfo.indexOf(',');
      if (delimPtr != -1) {
        if (delimPtr == 0) {
          importDevice.manufacturerID = "";
          importDevice.manufacturerName = manufacturerInfo;
        } else {
          importDevice.manufacturerID = manufacturerInfo.substring(0, delimPtr);
          importDevice.manufacturerName = manufacturerInfo.substring(delimPtr + 1);
        }
      } else {
        importDevice.manufacturerID = "";
        importDevice.manufacturerName = "";
      }
    }

    // Parse product data
    if ((productInfo == null) || (productInfo.length() == 0)) {
      importDevice.productID = "";
      importDevice.productName = "";
    } else {
      int delimPtr = productInfo.indexOf(',');
      if (delimPtr != -1) {
        if (delimPtr == 0) {
          importDevice.productID = "";
          importDevice.productName = productInfo;
        } else {
          importDevice.productID = productInfo.substring(0, delimPtr);
          importDevice.productName = productInfo.substring(delimPtr + 1);
        }
      } else {
        importDevice.productID = "";
        importDevice.productName = "";
      }
    }

    if (deviceNew) importNetwork.addDevice(importDevice);
    return true;
  }
  
  protected boolean importChannelInfo(String deviceID, int theChannel, boolean isDimmable, float defaultFadeRate, int curLevel) {
    boolean deviceChanged = false;
    
    LightingDevice importDevice = importNetwork.getDeviceByID(deviceID);
    if (importDevice == null) {
      error("Channel record references device ID " + deviceID + " which is an unknown device");
      discoveryFailed();
      return false;
    }
    
    importDevice.isDimmable[theChannel - 1] = isDimmable;
    importDevice.defaultFadeRate[theChannel - 1] = defaultFadeRate;
    importDevice.deviceState[theChannel - 1] = curLevel;
    return true;
  }
  
  protected boolean importDeviceScene(String deviceID, String sceneID, int theChannel, int theLevel, float fadeRate) {
    LightingDevice importDevice = importNetwork.getDeviceByID(deviceID);
    LightingScene importScene = importNetwork.getSceneByID(sceneID);

    // Verify scene and device
    if (importDevice == null) {
      error("Scene record references device ID " + deviceID + " which is an unknown device");
      discoveryFailed();
      return false;
    }
    if (importScene == null) {
      error("Scene record references scene ID " + sceneID + " which is an unknown scene");
      discoveryFailed();
      return false;
    }
    
    // Handle normalizing fade rate
    if (theChannel > 0) {
      if (fadeRate == importDevice.defaultFadeRate[theChannel - 1]) fadeRate = LightingDevice.DEFAULT_FADE_RATE;
    }
    
    // See if this device exists in the scene or not
    LightingSceneDevice sceneDevice = importScene.getSceneDeviceById(importDevice.getDeviceID(), theChannel);
    if (sceneDevice != null) {
      // See if there are any changes and if not, skip out
      if ((sceneDevice.getLevel() == theLevel) && (sceneDevice.getFadeRate() == fadeRate)) return true;
      
      // Apply the changes
      sceneDevice.theLevel = theLevel;
      sceneDevice.theFadeRate = fadeRate;
      return true;
    }

    // Add in a new device to the scene
    importScene.addDevice(importDevice, theLevel, fadeRate, theChannel);
    return true;
  }
  
  protected void markDiscoveryStarting() {
    if (ourGateway.discoveryInProgress) return;

    // Set flags and fire off events
    ourGateway.discoveryInProgress = true;
    ourGateway.fireGatewayEvent(new LightingGatewayEvent(LightingGatewayEvent.EventCode.DISCOVERY_STARTED));
  }

  protected void markDiscoveryFailed() {
    if (!ourGateway.discoveryInProgress) return;

    ourGateway.discoveryInProgress = false;
    ourGateway.fireGatewayEvent(new LightingGatewayEvent(LightingGatewayEvent.EventCode.DISCOVERY_FAILED));
  }
  
  protected void markDiscoveryComplete() {
    if (!ourGateway.discoveryInProgress) return;

    // End the import
    ourGateway.discoveryInProgress = false;
    ourGateway.discoveryComplete = true;
    ourGateway.fireGatewayEvent(new LightingGatewayEvent(LightingGatewayEvent.EventCode.DISCOVERY_COMPLETE));
  }
  
  public boolean startDiscovery() {
    discoveryState = 0;
    
    // Mark the import as starting
    markDiscoveryStarting();
    
    // Send network info request
    if (LightingGateway.DEBUG_MODE) debug("Lighting Discovery starting");
    sendGatewayInfoRequest();

    // Start the thread
    ourGateway.queueTask(this, "Lighting Discovery Watchdog");
    
    // And we're on our way
    return true;
  }

  void discoveryFailed() {
    markDiscoveryFailed();
    discoveryState = 0;
    synchronized(this) { this.notifyAll(); }
  }
  
  void discoveryComplete() {
    markDiscoveryComplete();
    discoveryState = 0;
    synchronized(this) { this.notifyAll(); }
  }
  
  // Return true if we recognized the message and handled it
  boolean processInfoMessage(xPL_MessageI theMessage) {
    // Skip everything unless we are expecting something
    if (discoveryState == 0) return false;

    // Check on status
    if (!theMessage.getNamedValue("status", "unknown").equalsIgnoreCase("ok")) {
      error("Request for lighting info failed -- got [" + theMessage.getNamedValue("status", "unknown") + "] -- bailing out");
      discoveryFailed();
      return true;
    }
    
    if (theMessage.getSchemaType().equalsIgnoreCase("gateinfo")) {
      receivedGatewayInfo(theMessage);
      return true;
    }

    if (theMessage.getSchemaType().equalsIgnoreCase("netlist")) {
      receivedNetworkList(theMessage);
      return true;
    }

    if (theMessage.getSchemaType().equalsIgnoreCase("netinfo")) {
      receivedNetworkInfo(theMessage);
      return true;
    }

    if (theMessage.getSchemaType().equalsIgnoreCase("scnlist")) {
      receivedSceneList(theMessage);
      return true;
    }

    if (theMessage.getSchemaType().equalsIgnoreCase("scninfo")) {
      receivedSceneInfo(theMessage);
      return true;
    }
  
    if (theMessage.getSchemaType().equalsIgnoreCase("devlist")) {
      receivedDeviceList(theMessage);
      return true;
    }

    if (theMessage.getSchemaType().equalsIgnoreCase("devinfo")) {
      receivedDeviceInfo(theMessage);
      return true;
    }
    
    return false;
  }
  
  // Given the passed list of identifiers, 
  // populates the passed LinkedList with the IDs
  void expandToIDs(List<String> theDefs, LinkedList<String> IDList) {
    // Clear ID list
    IDList.clear();
    
    // Handle there being nothing to parse
    if ((theDefs == null) || (theDefs.size() == 0)) return;
    
    for (String aDef: theDefs) {
      if (aDef.length() == 0) continue;
      
      for (String aPart: aDef.split(",")) {
        if (aPart.length() == 0) continue;
        IDList.add(aPart);
      }
    }
  }
  
  private void sendGatewayInfoRequest() {
    discoveryState = 1;
    
    // Send request for all networks
    synchronized(requestMessage) {
      requestMessage.clearMessageBody();
      requestMessage.addNamedValue("request", "gateinfo");
      sendMessage(requestMessage);
    }
  }
  
  
  private void receivedGatewayInfo(xPL_MessageI theMessage) {
    if (discoveryState != 1) return;

    // Parse gateway info
    ourGateway.gatewayProtocol = theMessage.getNamedValue("protocol", "");
    ourGateway.gatewayDescription = theMessage.getNamedValue("description", "");
    ourGateway.gatewayVersion = theMessage.getNamedValue("version", "");
    ourGateway.gatewayAuthor = theMessage.getNamedValue("author", "");
    ourGateway.gatewayURL = theMessage.getNamedValue("info-url", "");
    ourGateway.preferredNetworkID = theMessage.getNamedValue("preferred-net");
    
    ourGateway.areChannelsSupported = theMessage.getBoolNamedValue("channels-ok");
    ourGateway.areFadeRatesSupported = theMessage.getBoolNamedValue("fade-rate-ok");
    ourGateway.areScenesSupported = theMessage.getBoolNamedValue("scenes-ok");
    
    // See if there is a fade rate list (optional)
    ourGateway.hasFadeRateList = false;
    ourGateway.fadeRateList = null;
    if (theMessage.hasNamedValue("fade-rate-list")) {
      LinkedList<String> fadeRates = new LinkedList<String>();
      expandToIDs(theMessage.getMessageBody().getNamedValues("fade-rate-list"), fadeRates);
      if (fadeRates.size() > 0) {
        int rateIndex = 0;
        double theRates[] = new double[fadeRates.size()];
        
        for (String aRate: fadeRates) {
          try {
            theRates[rateIndex++] = Double.parseDouble(aRate);
          } catch (Exception anyError) {
            theRates = null;
            break;
          }
        }

        if (theRates != null) {
          ourGateway.hasFadeRateList = true;
          ourGateway.fadeRateList = theRates;
        }
      }

      fadeRates.clear();
    }
    
    // Advance to next step
    markItemDiscovered();
    discoveryState = 2;

    // Send request for all networks
    synchronized(requestMessage) {
      requestMessage.clearMessageBody();
      requestMessage.addNamedValue("request", "netlist");
      sendMessage(requestMessage);
    }
  }

  private void receivedNetworkList(xPL_MessageI theMessage) {
    if (discoveryState != 2) return;
    
    // Parse network IDs
    expandToIDs(theMessage.getMessageBody().getNamedValues("network"), NetworkIDList);
    if (NetworkIDList.isEmpty()) {
      error("No networks available in gateway");
      markDiscoveryFailed();
      return;
    }
    
    // Mark an item discovered
    markItemDiscovered();

    // Advance to next step
    discoveryState = 3;
    synchronized(requestMessage) {
      requestMessage.clearMessageBody();
      requestMessage.addNamedValue("request", "netinfo");
      requestMessage.addNamedValue("network", NetworkIDList.removeFirst());
      sendMessage(requestMessage);
    }
  }
  
  void receivedNetworkInfo(xPL_MessageI theMessage) {
    // Make sure things are happening as expected
    if (discoveryState != 3) return;

    if (theMessage != null) {
      // Parse the network info
      String networkID = theMessage.getNamedValue("network");
      String netPassword = theMessage.getNamedValue("password");
      String netName = theMessage.getNamedValue("name");
      if (networkID == null) {
        error("Missing items from lighting.netinfo -- network info cannot be parsed");
        discoveryFailed();
        return;
      }

      // Attempt to install the data
      if (!importNetwork(networkID, netName, netPassword)) {
        discoveryFailed();
        return;
      }
    }

    // Flag this item done
    markItemDiscovered();
    
    // Done with networks -- advance to scenes
    discoveryState = 4;
    synchronized(requestMessage) {
      requestMessage.clearMessageBody();
      requestMessage.addNamedValue("request", "scnlist");
      requestMessage.addNamedValue("network", importNetwork.getNetworkID());
      sendMessage(requestMessage);
    }
  }

  void receivedSceneList(xPL_MessageI theMessage) {
    // Make sure things are happening as expected
    if (discoveryState != 4) return;
    
    if ((importNetwork = ourGateway.getNetworkByID(theMessage.getNamedValue("network", ""))) == null) {
      error("Unable to lookup related network for scnlist -- failed");
      discoveryFailed();
      return;
    }

    // parse the scene list
    expandToIDs(theMessage.getMessageBody().getNamedValues("scene"), SceneIDList);
    markItemDiscovered();
    
    // Kick of scene discovery, if any
    discoveryState = 5;
    receivedSceneInfo(null);
  }
  
  void receivedSceneInfo(xPL_MessageI theMessage) {
    LightingScene theScene = null;
    
    // Make sure things are happening as expected
    if (discoveryState != 5) return;

    // Parse scene (if any)
    if (theMessage != null) {
      if ((importNetwork = ourGateway.getNetworkByID(theMessage.getNamedValue("network", ""))) == null) {
        error("Unable to lookup related network for scninfo -- failed");
        discoveryFailed();
        return;
      }
        
      String sceneID = theMessage.getNamedValue("scene");
      String sceneName = theMessage.getNamedValue("name");
      if ((sceneID != null) && (sceneID.length() > 0) && (sceneName != null) && (sceneName.length() > 0)) {
        this.importScene(sceneID, sceneName);
      }
    }
    
    markItemDiscovered();

    // See if we have another scene to get or not
    if (!SceneIDList.isEmpty()) {
      synchronized(requestMessage) {
        requestMessage.clearMessageBody();
        requestMessage.addNamedValue("request", "scninfo");
        requestMessage.addNamedValue("network", importNetwork.getNetworkID());
        requestMessage.addNamedValue("scene", SceneIDList.removeFirst());
        sendMessage(requestMessage);
      }
      
      return;
    }
    
    // Done with scenes -- advance
    discoveryState = 6;
    synchronized(requestMessage) {
      requestMessage.clearMessageBody();
      requestMessage.addNamedValue("request", "devlist");
      requestMessage.addNamedValue("network", importNetwork.getNetworkID());
      sendMessage(requestMessage);
    }
  }
  
  void receivedDeviceList(xPL_MessageI theMessage) {
    // Make sure things are happening as expected
    if (discoveryState != 6) return;
    
    if ((importNetwork = ourGateway.getNetworkByID(theMessage.getNamedValue("network", ""))) == null) {
      error("Unable to lookup related network for devicelist -- failed");
      discoveryFailed();
      return;
    }
    
    // parse the device list
    expandToIDs(theMessage.getMessageBody().getNamedValues("device"), DeviceIDList);
    markItemDiscovered();

    // Request first devices info
    discoveryState = 7;
    receivedDeviceInfo(null);
  }
  
  private int decodeLevel(String theLevel) {
    if (theLevel.equalsIgnoreCase("default")) return LightingDevice.DEFAULT_DIM_LEVEL;
    if (theLevel.equalsIgnoreCase("last")) return LightingDevice.LAST_DIM_LEVEL;
    
    try {
      int newLevel = Integer.parseInt(theLevel);
      if ((newLevel >= 0) && (newLevel <= 100)) return newLevel;
    } catch (Exception anything) { }
    
    return LightingDevice.LAST_DIM_LEVEL;
  }
  
  void receivedDeviceInfo(xPL_MessageI theMessage) {
    LightingDevice theDevice = null;
    
    // Make sure things are happening as expected
    if (discoveryState != 7) return;

    // Parse message, if any
    if (theMessage != null) {
      if ((importNetwork = ourGateway.getNetworkByID(theMessage.getNamedValue("network", ""))) == null) {
        error("nable to lookup related network for deviceinfo -- failed");
        discoveryFailed();
        return;
      }

      String deviceID = theMessage.getNamedValue("device");
      String deviceName = theMessage.getNamedValue("name");
      String roomName = theMessage.getNamedValue("room");
      String floorName = theMessage.getNamedValue("floor");
      String comment = theMessage.getNamedValue("comment");
      String manufacturerInfo = theMessage.getNamedValue("manufacturer");
      String productInfo = theMessage.getNamedValue("product");
      String firmwareVersion = theMessage.getNamedValue("firmware-version");
      boolean reportsManualChanges = theMessage.getBoolNamedValue("report-on-manual");
      int channelCount = theMessage.getIntNamedValue("channel-count");
      int primaryChannel = theMessage.getIntNamedValue("primary-channel");
      
      // Discovery the device
      if (!importDevice(deviceID, deviceName, manufacturerInfo, productInfo, firmwareVersion, channelCount, primaryChannel, roomName, floorName, comment, reportsManualChanges)) {
        error("Unable to import new device");
        discoveryFailed();
        return;
      }
      
      // Process channels now
      List<String> channels = theMessage.getMessageBody().getNamedValues("channel");
      if (channels != null) {
        String channelParts[];
        int channelNum = 0;
        float fadeRate = 0;
        int curLevel = 0;
        boolean isDimmable = false;
        
        for (String aChannel: channels) {
          channelParts = aChannel.split(",");
          try {
            channelNum = Integer.parseInt(channelParts[0]);
          } catch (Exception error) {
            error("Parsing channel number of device channel info");
            discoveryFailed();
            return;
          }
          
          isDimmable = channelParts[1].equalsIgnoreCase("true");
          fadeRate = LightingGateway.parseFadeRate(channelParts[2]);
          
          try {
            curLevel = Integer.parseInt(channelParts[3]);
            if (curLevel < 0) curLevel = 0;
            if (curLevel > 100) curLevel = 100;
          } catch (Exception anyError) { curLevel = 0; }
          
          
          if (!importChannelInfo(deviceID, channelNum, isDimmable, fadeRate, curLevel)) {
            error("Unable to import channel for device");
            discoveryFailed();
            return;
          }
        }
      }
      
      List<String> scenes = theMessage.getMessageBody().getNamedValues("scene");
      if (scenes != null) {
        String sceneID = null;
        int channel = 0, level = 0;
        float fadeRate = 0;
        String sceneParts[];
        
        for (String aScene: scenes) {
          sceneParts = aScene.split(",");
          sceneID = sceneParts[0];
          try {
            channel = Integer.parseInt(sceneParts[1]);
          } catch (Exception error) {
            error("Parsing channel of device/scene info");
            discoveryFailed();
            return;
          }

          // Parse level & faderate
          level = decodeLevel(sceneParts[2]);
          fadeRate = LightingGateway.parseFadeRate(sceneParts[3]);

          if (!importDeviceScene(deviceID, sceneID, channel, level, fadeRate)) {
            error("Unable to import device/scene association");
            discoveryFailed();
            return;
          }
        }
      }
    }
    
    markItemDiscovered();

    // See if we have another device to get or not
    if (!DeviceIDList.isEmpty()) {
      synchronized(requestMessage) {
        requestMessage.clearMessageBody();
        requestMessage.addNamedValue("request", "devinfo");
        requestMessage.addNamedValue("network", importNetwork.getNetworkID());
        requestMessage.addNamedValue("device", DeviceIDList.removeFirst());
        sendMessage(requestMessage);
      }
      
      return;
    }

    // See if there is another network to advance to
    if (!NetworkIDList.isEmpty()) {
      discoveryState = 3;  // Change back to network info state
      synchronized(requestMessage) {
        requestMessage.clearMessageBody();
        requestMessage.addNamedValue("request", "netinfo");
        requestMessage.addNamedValue("network", NetworkIDList.removeFirst());
        sendMessage(requestMessage);
      }
      
      return;
    }
    
    // Refresh complete
    discoveryComplete();
  }

  boolean itemDiscoveryed = false;
  
  void markItemDiscovered() {
    synchronized(this) {
      itemDiscoveryed = true;
      this.notifyAll();
    }
  }

  // Wait for the import to complete.  We get "pinged" everytime something
  // is imported and that resets our timer.  If more than 15 seconds go 
  // by without a ping, we have a problem and we fail out the import
  public void run() {
    int retryCount = 0;
    
    while(ourGateway.isDiscoveryInProgress()) {
      // Wait 15 seconds
      synchronized(this) {
        try {
          if (LightingGateway.DEBUG_MODE) debug("Wathdog about to sleep for 15 seconds..");
          this.wait(15000);
        } catch (Exception anyError) { }

        if (!ourGateway.isDiscoveryInProgress()) break;
        
        // If we woke up because an item was imported, clear the flag and
        // restart out sleep
        if (itemDiscoveryed) {
          itemDiscoveryed = false;
          if (LightingGateway.DEBUG_MODE) debug("Watchdog got notice of imported item -- resetting");
          continue;
        }
      }
      
      // If we got a timeout but the state is only at 1, it's possible we
      // send the request out before the receiver was ready to serve it.  So
      // we'll retry that
      if (discoveryState == 1) {
        // After a while, give up (150 seconds, 2 minutes+)
        if (retryCount >= 10) {
          error("Unable to request remote gateway info -- ignored for 150+ seconds");
          discoveryFailed();
          break;
        }
      
        // Send it out and go back to sleep
        if (LightingGateway.DEBUG_MODE) debug("Watchdog timedout -- resending request as retry " + retryCount);
        sendGatewayInfoRequest();
        continue;
      }
      
      // If refresher state is 0, the import stopped behind our back.  Shouldn't
      // happen, but handle it gracefully and exit
      if (discoveryState == 0) break;

      // If we got here, we timed out while in the middle of the import
      // process -- fail it!
      error("Watchdog timedout midway through status messages -- failed");
      discoveryFailed();
      break;
    }
  }
}
