/*
 * LightingFloor.java
 *
 * Copyright (c) 2005, Gerald R Duprey Jr
 * More info at http://www.xPL4Java.org
 */

package org.cdp1802.xpl.lighting;

import java.util.*;

/**
 * Describes a floor (basement, attic, 1st, etc) in the network
 *
 * @author gerry
 */
public class LightingFloor {
  LightingNetwork floorNetwork = null;
  String floorName = "";
  ArrayList<LightingDevice> floorDevices = new ArrayList<LightingDevice>();
  
  LightingFloor(LightingNetwork theNetwork, String floorName) {
    this.floorNetwork = theNetwork;
    this.floorName = floorName;
  }

  /**
   * Get the network associated with this floor
   *
   * @return network associated with this floor
   */
  LightingNetwork getNetwork() { return floorNetwork; }

  /**
   * Name of this floor
   *
   * @return name of this floor
   */
  public String getFloorName() { return floorName; }
  void setFloorName(String newName) {
    // Make sure it's valid and different
    if ((newName == null) || (newName.length() == 0)) return;
    if (newName.equals(floorName)) return;
    
    // Assign it
    floorName = newName;
    
    // If there are no devices, we are done
    if (floorDevices.size() == 0) return;
  }

  /**
   * Get the number of devices currently associated with this floor
   *
   * @return number of devices in the floor
   */
  public int getDeviceCount() { return floorDevices.size(); }
  
  /**
   * Given an index between 0 and deviceCount - 1, return the device at 
   * that index for this floor
   *
   * @param deviceIndex index of device in this floor to return
   * @return device instance of device at the index or null for invalid index
   */
  public LightingDevice getDeviceAt(int deviceIndex) {
    if ((deviceIndex < 0) || (deviceIndex >= floorDevices.size())) return null;
    return floorDevices.get(deviceIndex);
  }
  
  /**
   * Given a device ID, check this floor to see if the device is 
   * associated with the floor and if so, return the device
   *
   * @param deviceID device ID to look for in this floor
   * @return instance of matching device if assigned to this floor or null if not found
   */
  public LightingDevice getDeviceByID(String deviceID) {
    // This is faster than scanning floor devices because it's
    // all direct access lookups (no loops)
    LightingDevice floorDevice = floorNetwork.getDeviceByID(deviceID);
    if ((floorDevice == null) || (floorDevice.getFloor() != this)) return null;
    return floorDevice;
  }
  
  /**
   * Get a collection of devices assigned to this floor
   *
   * <P>NOTE: Devices are returned in device ID order
   *
   * @return collection of devices assigned to this floor
   */
  public Collection<LightingDevice> getDevices() { return floorDevices; }
  
  boolean addDevice(LightingDevice theDevice) {
    synchronized(floorDevices) {
      if (getDeviceByID(theDevice.getDeviceID()) != null) return false;
      floorDevices.add(theDevice);
    }
    
    return true;
  }
  
  boolean removeDevice(LightingDevice theDevice) {
    synchronized(floorDevices) {
      return floorDevices.remove(theDevice);
    }
  }

  /**
   * Get a printable description of this floor
   *
   * @return printable description of this floor
   */
  public String toString() { return floorName; }
  
  void releaseResources() {
    floorNetwork = null;
    floorName = null;
    if (floorDevices != null) {
      floorDevices.clear();
    }
    floorDevices = null;
  }
}