/*
 * LightingGateway.java
 *
 * Copyright (c) 2005, Gerald R Duprey Jr
 * More info at http://www.xPL4Java.org
 */

package org.cdp1802.xpl.lighting;

import java.io.*;
import java.util.*;
import java.text.*;

import org.cdp1802.xpl.*;
import org.cdp1802.xpl.device.*;

/**
 * LightingGateway is a reflection of a remote lighting gateway.  It can
 * speak to a specific remote lighting gateway or be allowed to "find" one on 
 * it own. 
 * 
 * xPL4Java must be up and running (i.e. connected to the network) before creating
 * any lighting gateways.
 *
 * Once setup, the gateway should be allowed to discover all it can about the remote
 * side of the gateway using the startGatewayDiscovery() call.  Once that is complete, 
 * you can lookup devices, scenes and networks, check current state of these devices
 * and register to hear when that state changes as well as command the devices and
 * scenes.
 *
 * There can be multiple instances of LightingGateway in an app, though it makes
 * most sense if each one is only associated with one remote gateway.
 * 
 * @author gerry
 */
public class LightingGateway {
  public static final String LIGHTING_VERSION = xPL_Manager.XPL4JAVA_VERSION;
  
  /* Set true and recompile to enable copious debugging output */
  public static final boolean DEBUG_MODE = false;
  
  /* Used for logging */
  static DateFormat dateTimeFormatter = new SimpleDateFormat("MM/dd HH:mm:ss ");

  public static String formatLevel(int theLevel) {
    if (theLevel == LightingDevice.DEFAULT_DIM_LEVEL) return "default";
    if (theLevel == LightingDevice.LAST_DIM_LEVEL) return "last";
    if (theLevel > 100) return "last";
    if (theLevel < 0) return "last";
    return String.valueOf(theLevel);
  }
  
  public static int parseLevel(String theLevel) {
    if (theLevel.equalsIgnoreCase("default")) return LightingDevice.DEFAULT_DIM_LEVEL;
    if (theLevel.equalsIgnoreCase("last")) return LightingDevice.LAST_DIM_LEVEL;
    
    try {
      int newLevel = Integer.parseInt(theLevel);
      if ((newLevel >= 0) && (newLevel <= 100)) return newLevel;
    } catch (Exception anything) { }
    
    return LightingDevice.LAST_DIM_LEVEL;
  }
  
  public static String formatFadeRate(float theRate) {
    if (theRate == LightingDevice.DEFAULT_FADE_RATE) return "default";
    return String.valueOf(theRate);
  }

  public static float parseFadeRate(String theRate) {
    if (theRate.equalsIgnoreCase("default")) return LightingDevice.DEFAULT_FADE_RATE;
    try {
      return Float.parseFloat(theRate);
    } catch (Exception anyError) { return LightingDevice.DEFAULT_FADE_RATE; }
  }

  
  private static ThreadPool threadPool = new ThreadPool();

  /**
   * Queue a task to run in one of the pooled threads.  This is a convience 
   * feature that helps minimize excessive thread creation by re-cycling 
   * threads and using them over and over.
   *
   * <P>This is best for shorter-lived threads.  For longer lived threads 
   * (i.e. that run for most of the applications lifetime), create normal 
   * threads for them.
   *
   * <P>While meant primarily for xPLLighting and related utilities to use, it
   * is perfectly OK to use for simple task queuing on your application.
   *
   * @param theTask task to queue up
   * @param taskDescription short summary of what this task is going to do
   */
  public static void queueTask(Runnable theTask, String taskDescription) {
    threadPool.addTask(theTask, taskDescription);
  }
  
  
  class QualifiedDeviceListener {
    LightingDeviceListenerI theListener = null;
    LightingDevice theDevice = null;
    
    QualifiedDeviceListener(LightingDeviceListenerI theListener, LightingDevice forDevice) {
      this.theListener = theListener;
      this.theDevice = forDevice;
    }
    
    QualifiedDeviceListener(LightingDeviceListenerI theListener) {
      this.theListener = theListener;
    }
    
    boolean sendToListener(LightingDevice toDevice) {
      if (toDevice == null) return true;
      if (theDevice == null) return true;
      
      return (theDevice == toDevice);
    }
    
    void releaseResources() {
      theListener = null;
      theDevice = null;
    }
  }
  
  class QualifiedSceneListener {
    LightingSceneListenerI theListener = null;
    LightingScene theScene = null;
    
    QualifiedSceneListener(LightingSceneListenerI theListener, LightingScene forScene) {
      this.theListener = theListener;
      this.theScene = forScene;
    }
    
    QualifiedSceneListener(LightingSceneListenerI theListener) {
      this.theListener = theListener;
    }
    
    boolean sendToListener(LightingScene toScene) {
      if (toScene == null) return true;
      if (theScene == null) return true;
      
      return (theScene == toScene);
    }
    
    void releaseResources() {
      theListener = null;
      theScene = null;
    }
  }
  
  LightingXPLAdapter xPLAdapter = null;
  
  LightingNetwork preferredNetwork = null;

  boolean discoveryInProgress = false;
  boolean discoveryComplete = false;
  
  ArrayList<LightingGatewayListenerI> managerListenerList = new ArrayList<LightingGatewayListenerI>();
  ArrayList<QualifiedDeviceListener> deviceListenerList = new ArrayList<QualifiedDeviceListener>();
  ArrayList<QualifiedSceneListener> sceneListenerList = new ArrayList<QualifiedSceneListener>();

  ArrayList<LightingNetwork> networkList = new ArrayList<LightingNetwork>();
  
  // For debug/error messages only
  StringBuffer messageBuffer = new StringBuffer();  // Used for debug/log messages
  
  String gatewayProtocol = "";
  String gatewayDescription = "";
  String gatewayVersion = "";
  String gatewayAuthor = "";
  String gatewayURL = "";
  String preferredNetworkID = "";
  
  boolean areScenesSupported = false;
  boolean areChannelsSupported = false;
  boolean areFadeRatesSupported = false;
  
  boolean hasFadeRateList = false;
  double fadeRateList[] = null;
  
  /**
   * Create an instance of a lighting gateway using the passed xPL device to 
   * receive and send messages on and listen only to messages (and only send
   * messages to) the remote gateway.
   *
   * <P>NOTE: Passing null for remote gateway enabled this to listen to any lighting
   * gateway on the xPL network and to broadcast to all gateways.  If there is 
   * only one gateway, this is easy.  If there are multiple gateways though, you 
   * really must specify the gateway.
   *
   * <P>NOTE: If passing anything but a null remoteGateway or a specific remote
   * gateway, the gateway will be ingored and null used.  Specifically, passing
   * a broadcast or group identifier here will result in the same behavior as 
   * passing null (i.e. broadcast behavior).
   *
   * <P>NOTE: If you pass a null/broadcast gateway, this instance will use broadcasts
   * ONLY until he first time it receives a valid lighting message from a gateway.  At
   * that point, that gateway will be "bound" to this gateway and we'll no longer send
   * broadcasts and will no longer listen to messages from anything other than that
   * specific remote gateway.  You can determine this by get the getRemoteGateway()
   * method -- if it return null, we're still broadcasting.  if non-null, we are 
   * bound to a specific remote gateway.
   *
   * @param clientDevice device to use or null to auto-create/enable a device
   * @param remoteGateway identifier of remote gateway or null for any gateway
   */
  public LightingGateway(xPL_DeviceI clientDevice, xPL_IdentifierI remoteGateway) {
    setupAdapter(clientDevice, remoteGateway);
  }

  /**
   * Create an instance of a lighting gateway using the passed xPL device to 
   * receive and send messages on.  Broadcast all messages to all gateways
   * and listen for lighting messages from any gateway.
   *
   * <P>NOTE: This will listen to any lighting gateway on the xPL network and 
   * will broadcast to all gateways.  If there is only one gateway, this is OK.
   * If there are multiple gateways though, you really must specify the gateway.
   *
   * <P>NOTE: This instance will use broadcasts ONLY until he first time it 
   * receives a valid lighting message from a gateway.  At that point, that 
   * gateway will be "bound" to this gateway and we'll no longer send
   * broadcasts and will no longer listen to messages from anything other than that
   * specific remote gateway.  You can determine this by get the getRemoteGateway()
   * method -- if it return null, we're still broadcasting.  if non-null, we are 
   * bound to a specific remote gateway.
   *
   * @param clientDevice device to use or null to auto-create/enable a device
   */
  public LightingGateway(xPL_DeviceI clientDevice) {
    setupAdapter(clientDevice, null);
  }
 
 
  /**
   * Create an instance of a lighting gateway.  We will create our own xPL device
   * to receive and send messages to and will broadcast all messages to all gateways
   * and listen for lighting messages from any gateway.
   *
   * <P>NOTE: This will listen to any lighting gateway on the xPL network and 
   * will broadcast to all gateways.  If there is only one gateway, this is OK.
   * If there are multiple gateways though, you really must specify the gateway.
   *
   * <P>NOTE: This instance will use broadcasts ONLY until he first time it 
   * receives a valid lighting message from a gateway.  At that point, that 
   * gateway will be "bound" to this gateway and we'll no longer send
   * broadcasts and will no longer listen to messages from anything other than that
   * specific remote gateway.  You can determine this by get the getRemoteGateway()
   * method -- if it return null, we're still broadcasting.  if non-null, we are 
   * bound to a specific remote gateway.
   */
  public LightingGateway() {
    setupAdapter(null, null);
  }
  
  private void setupAdapter(xPL_DeviceI clientDevice, xPL_IdentifierI remoteGateway) {
    // Allocate a client device, if needed
    if (clientDevice == null) {
      clientDevice = xPL_Manager.getManager().getDeviceManager().createDevice("cdp1802", "lght-cln", null);
      clientDevice.setVersion(LIGHTING_VERSION);
      clientDevice.setEnabled(true);
    }
    
    // Insure we have a good gateway
    if (remoteGateway != null) {
      if (remoteGateway.isBroadcastIdentifier() || remoteGateway.isGroupIdentifier()) {
        remoteGateway = null;
      }
    }

    // Create the xPL adapter to package and unpackage messages
    xPLAdapter = new LightingXPLAdapter(this, clientDevice, remoteGateway);
  }
  
  void intDebug(String theMessage) {
    synchronized(messageBuffer) {
      messageBuffer.setLength(0);
      messageBuffer.append(dateTimeFormatter.format(new Date()));
      messageBuffer.append(' ');
      messageBuffer.append(theMessage);
      System.out.println(messageBuffer);
    }
  }
  void debug(String theMessage) {
    intDebug("LIGHTING:: " + theMessage);
  }
  
  void intWarn(String theMessage) {
    synchronized(messageBuffer) {
      messageBuffer.setLength(0);
      messageBuffer.append(dateTimeFormatter.format(new Date()));
      messageBuffer.append(" Warning ");
      messageBuffer.append(theMessage);
      System.out.println(messageBuffer);
    }
  }
  void warn(String theMessage) {
    intWarn("LIGHTING:: " + theMessage);
  }

  void intError(String theMessage) {
    synchronized(messageBuffer) {
      messageBuffer.setLength(0);
      messageBuffer.append(dateTimeFormatter.format(new Date()));
      messageBuffer.append(" *ERROR* ");
      messageBuffer.append(theMessage);
      System.err.println(messageBuffer);
    }
  }
  void error(String theMessage) {
    intError("LIGHTING:: " + theMessage);
  }
  
  
  /**
   *  Return the number of available networks
   *
   * @return number of Lighting networks available
   */
  public int getNetworkCount() { return networkList.size(); }
  
  /**
   * Return the network at the passed index
   *
   * @param theIndex index of the network to return
   * @return the network at that index or null for an invalid index
   */
  public LightingNetwork getNetworkAt(int theIndex) {
    if ((theIndex < 0) || (theIndex >= networkList.size())) return null;
    return networkList.get(theIndex);
  }

  /**
   * Get the network matching the specified ID.
   *
   * <P>NOTE: Right now, this either returns the primary network or
   * returns null if no match.  In the future, multiple networks
   * could be handled via this
   *
   * @param networkID network ID to match
   * @return network reference or null if no matching network
   */
  public LightingNetwork getNetworkByID(String networkID) {
    synchronized(networkList) {
      for (LightingNetwork theNetwork: networkList) {
        if (theNetwork.getNetworkID().equals(networkID)) return theNetwork;
      }
    }
    
    return null;
  }

  /**
   * Get the xPL_Device that we are using to talk to the xPL network with
   */
  public xPL_DeviceI getAdapterDevice() { return xPLAdapter.getAdapterDevice(); }
  
  /**
   * Get the xPL_Identifier of the remote gateway. 
   *
   * <P>NOTE: If a broadcast identifier was specified as the remote gateway and
   * there has been no traffic so far, this will return null (null beign the AKA 
   * for a broadcast gateway).  If there is any traffic, the null will be updated
   * with the actual ident of the remote gateway (and that ident is used from
   * then on).
   *
   * <P>NOTE: If this gateway is bound to a remote gateway and that remote
   * gateway announces it's going offline (sends a hbeat.end or config.end 
   * message), this gateway will switch back to being unbound until we again
   * hear from a new gateway.  This will not happen if the remote gateway 
   * just goes away without first announcing itself.
   *
   */
  public xPL_IdentifierI getRemoteGateway() { return xPLAdapter.getRemoteGateway(); }
  
  /**
   * Get the protocol for the remote gateway 
   */
  public String getGatewayProtocol() { return gatewayProtocol; }
  
  /**
   * Get the description of the remote gateway
   */
  public String getGatewayDescription() { return gatewayDescription; }
  
  /**
   * Get the version of the remote gateway
   */
  public String getGatewayVersion() { return gatewayVersion; }
  
  /**
   * Get the author of the remote gateway
   */
  public String getGatewayAuthor() { return gatewayAuthor; }
  
  /**
   * Get the URL for the remote gateways info webpage
   */
  public String getGatewayURL() { return gatewayURL; }
  
  /**
   * Determine if this gateway supports scenes or not
   */
  public boolean areScenesSupported() { return areScenesSupported; }
  
  /**
   * Determine if this gateways devices support channels or not
   */
  public boolean areChannelsSupported() { return areChannelsSupported; }
  
  /**
   * Determine if this gateways scenes and devices support 
   * fade rates or not
   */
  public boolean areFadeRatesSupported() { return areFadeRatesSupported; }

  /**
   * Determine if this gateway has published a list of valid fade rates.
   *
   * @return true if there are defined fade rates, false if not
   */
  public boolean hasFadeRateList() { return hasFadeRateList; }

  /**
   * If this gateway has a published list of fade rates, return the
   * number of them. 
   *
   * @return number of fade rates or 0 if no fade rates available
   */
  public int getFadeRateListCount() { 
    if (!hasFadeRateList || (fadeRateList == null)) return 0;
    return fadeRateList.length;
  }

  /**
   * Given an index, return the fade rate from the published 
   * fade rate list in this gateway.
   *
   * If the index is invalid or the gateway does not have a
   * published list of fade rates, -1 is returned.
   *
   * @param fadeRateIndex index of the fade rate to return, starting at 0
   * @return the fade rate for the passed index or -1 if index is bad or no fade rate list is available
   */
  public double getFadeRateAt(int fadeRateIndex) {
    if (!hasFadeRateList || (fadeRateList == null)) return -1;
    if ((fadeRateIndex < 0) || (fadeRateIndex >= fadeRateList.length)) return -1;
    return fadeRateList[fadeRateIndex];
  } 

  /**
   * Get the preferred network ID for the remote gateway
   */
  public String getPreferredNetworkID () { return preferredNetworkID; }
  
  /**
   * Get the preferred network.
   *
   * @return preferred network instance
   */
  public LightingNetwork getPreferredNetwork() { return preferredNetwork; }
  
  /**
   * Initiate gateway discovery.  This will interrogate the remote gateway
   * and learn/cache everything we can learn about it. After a successful
   * discovery, we'll know the details on this gateway, will have a list
   * of networks, scenes and devices in that gateway as well as current
   * state for all devices.
   *
   * You can either periodically check the isDiscoveryInProgress method or
   * register as a gateway listener to see when the discovery is complete (or
   * if it failed).
   *
   * Discovery can be run at any time and as often as needed.  Subsequent
   * discoveries will only update the new stuff and will not result in 
   * new instances and such for existing networks/devices/scenes.
   *
   * @return true if discovery started, false if not (usually, discovery already in progress)
   *
   */
  public boolean startGatewayDiscovery() {
    return xPLAdapter.startDiscovery();
  }
  
  /**
   * Determine if discovery is in progress or not
   *
   * @return true if in progress, no if not
   */
  public boolean isDiscoveryInProgress() { return discoveryInProgress; }

  /**
   * Determine if we have had a successful discovery or not.  Until discovery
   * is complete, you cannot depend on any values in this gateway instance
   * or any dependent items.
   */
  public boolean isDiscoveryComplete() { return discoveryComplete; }
  
  /**
   * Add a listener that will be invoked for any gateway oriented 
   * events
   *
   * <P>If the listener has already been added, it will not be added again
   *
   * @param theListener listener to add
   */
  public void addGatewayListener(LightingGatewayListenerI theListener) {
    synchronized(managerListenerList) {
      if (managerListenerList.contains(theListener)) return;
      managerListenerList.add(theListener);
    }
  }

  /**
   * Remove a previously added listener
   *
   * @param theListener listener to remove
   */
  public void removeGatewayListener(LightingGatewayListenerI  theListener) {
    synchronized(managerListenerList) {
      managerListenerList.remove(theListener);
    }
  }

  // Fires a manager event to all manager listeners
  void fireGatewayEvent(LightingGatewayEvent theEvent) {
    int listenerCount = managerListenerList.size();
    for (int listenerIndex = listenerCount - 1; listenerIndex >= 0; listenerIndex--) {
      try {
        managerListenerList.get(listenerIndex).handleLightingGatewayEvent(theEvent);
      } catch (Throwable ignoredException) { }
    }
    
    // Free event resources
    theEvent.releaseResources();
  }

  /**
   * Add a Device listener for a specific device.  The listener will only
   * receive events when the event relates to the passed events.
   *
   * <P>An attempt to add a listener for a given device that already has that listener
   * listed is ignored.  However, you can add the same listener more than once as
   * long as each add is for a different device.
   *
   * <P>Passing null for the device means you'll get events for all device
   *
   * @param theListener listener to receive device events
   * @param forDevice device to receive events for or null for all devices
   */
  public void addDeviceListener(LightingDeviceListenerI theListener, LightingDevice forDevice) {
    synchronized(deviceListenerList) {
      for (QualifiedDeviceListener deviceListener: deviceListenerList) {
        if (deviceListener.theListener != theListener) continue;
        if (deviceListener.theDevice != forDevice) continue;
        
        // Listener already exists
        return;
      }
      
      deviceListenerList.add(new QualifiedDeviceListener(theListener, forDevice));
    }
  }
  
  /**
   * Add a Device listener for any device.
   *
   * <P>An attempt to add a listener that already has this listener listed is 
   * ignored.  However, you can add the same listener more than once as
   * long as each add is for a different device.
   *
   * @param theListener listener to receive device events
   */
  public void addDeviceListener(LightingDeviceListenerI theListener) {
    addDeviceListener(theListener, null);
  }

  /**
   * Remove a previously added device listener for a passed device.  Only if
   * there is a listed instance of the passed listener and that has a matching
   * device is it removed.  Other instances of the same listener but for other
   * devices are left alone.
   *
   * <P>If null is passed in forDevice, then an existing listener that was added
   * to listed for any devices is removed.  Other instances of the same listener
   * that are listening to specific devices are not changed.
   *
   * @param theListener listener to remove
   * @param forDevice device that listener was listening to
   */
  public void removeDeviceListener(LightingDeviceListenerI theListener, LightingDevice forDevice) {
    synchronized(deviceListenerList) {
      for (QualifiedDeviceListener deviceListener: deviceListenerList) {
        if (deviceListener.theListener != theListener) continue;
        if (deviceListener.theDevice != forDevice) continue;
        
        // Remove Listener
        deviceListenerList.remove(deviceListener);
        deviceListener.releaseResources();
        return;
      }
    }
  }

  /**
   * Remove a previouly added device listener that was listening for any 
   * devices.  Only the single instance of this listener that was set to 
   * receive any device is removed.  Other instances of this same listener 
   * that are set to receive events for specific devices are left alone.
   *
   * @param theListener listener to remove
   */
  public void removeDeviceListener(LightingDeviceListenerI theListener) {
    removeDeviceListener(theListener, null);
  }

  /**
   * Remove all occurances of the passed listener, regardless of the device
   * they were set to listen for (or if they were set for all devices).
   *
   * @param theListener listener to remove all occurances of
   */
  public void removeAllDeviceListeners(LightingDeviceListenerI theListener) {
    synchronized(deviceListenerList) {
      QualifiedDeviceListener deviceListener = null;
      for (int listenerIndex = deviceListenerList.size() - 1; listenerIndex >= 0; listenerIndex--) {
        deviceListener = deviceListenerList.get(listenerIndex);
        if (deviceListener.theListener != theListener) continue;
        
        // Remove Listener
        deviceListenerList.remove(listenerIndex);
        deviceListener.releaseResources();
      }
    }
  }

  /**
   * Remove all listeners for a specifc device.  This is generally only needed
   * when a device is being removed from the network
   *
   * @param forDevice device to purge listeners for
   */
  public void removeAllRelatedDeviceListeners(LightingDevice forDevice) {
    synchronized(deviceListenerList) {
      QualifiedDeviceListener deviceListener = null;
      for (int listenerIndex = deviceListenerList.size() - 1; listenerIndex >= 0; listenerIndex--) {
        deviceListener = deviceListenerList.get(listenerIndex);
        if (deviceListener.theDevice != forDevice) continue;
        
        // Remove Listener
        deviceListenerList.remove(listenerIndex);
        deviceListener.releaseResources();
      }
    }
  }
  
  // Fires a device event to all appropriate listeners
  void fireDeviceEvent(LightingDeviceEvent theEvent) {
    QualifiedDeviceListener theListener = null;
    LightingDevice eventDevice = theEvent.getDevice();
    
    int listenerCount = deviceListenerList.size();
    for (int listenerIndex = listenerCount - 1; listenerIndex >= 0; listenerIndex--) {
      if (listenerIndex >= listenerCount) continue;
      
      try {
        theListener = deviceListenerList.get(listenerIndex);
        if (theListener.sendToListener(eventDevice)) theListener.theListener.handleLightingDeviceEvent(theEvent);
      } catch (Throwable ignoredException) { }
    }
    
    // Free event resources
    theEvent.releaseResources();
  }

  /**** Scene Listeners ****/
  
  /**
   * Add a scene listener for a specific scene.  The listener will only
   * receive events when the event relates to the passed events.
   *
   * <P>An attempt to add a listener for a given scene that already has that listener
   * listed is ignored.  However, you can add the same listener more than once as
   * long as each add is for a different scene.
   *
   * <P>Passing null for the scene means you'll get events for all scenes
   *
   * @param theListener listener to receive scene events
   * @param forScene scene to receive events for or null for all scenes
   */
  public void addSceneListener(LightingSceneListenerI theListener, LightingScene forScene) {
    synchronized(sceneListenerList) {
      for (QualifiedSceneListener sceneListener: sceneListenerList) {
        if (sceneListener.theListener != theListener) continue;
        if (sceneListener.theScene != forScene) continue;
        
        // Listener already exists
        return;
      }
      
      sceneListenerList.add(new QualifiedSceneListener(theListener, forScene));
    }
  }
  
  /**
   * Add a scene listener for any scene.
   *
   * <P>An attempt to add a listener that already has this listener listed is 
   * ignored.  However, you can add the same listener more than once as
   * long as each add is for a different scene.
   *
   * @param theListener listener to receive scene events
   */
  public void addSceneListener(LightingSceneListenerI theListener) {
    addSceneListener(theListener, null);
  }

  /**
   * Remove a previously added scene listener for a passed scene.  Only if
   * there is a listed instance of the passed listener and that has a matching
   * scene is it removed.  Other instances of the same listener but for other
   * scenes are left alone.
   *
   * <P>If null is passed in forScene, then an existing listener that was added
   * to listed for any scenes is removed.  Other instances of the same listener
   * that are listening to specific scenes are not changed.
   *
   * @param theListener listener to remove
   * @param forScene scene that listener was listening to
   */
  public void removeSceneListener(LightingSceneListenerI theListener, LightingScene forScene) {
    synchronized(sceneListenerList) {
      for (QualifiedSceneListener sceneListener: sceneListenerList) {
        if (sceneListener.theListener != theListener) continue;
        if (sceneListener.theScene != forScene) continue;
        
        // Remove Listener
        sceneListenerList.remove(sceneListener);
        sceneListener.releaseResources();
        return;
      }
    }
  }

  /**
   * Remove a previouly added scene listener that was listening for any 
   * scenes.  Only the single instance of this listener that was set to 
   * receive any scene is removed.  Other instances of this same listener 
   * that are set to receive events for specific scenes are left alone.
   *
   * @param theListener listener to remove
   */
  public void removeSceneListener(LightingSceneListenerI theListener) {
    removeSceneListener(theListener, null);
  }

  /**
   * Remove all occurances of the passed listener, regardless of the scene
   * they were set to listen for (or if they were set for all scenes).
   *
   * @param theListener listener to remove all occurances of
   */
  public void removeAllSceneListeners(LightingSceneListenerI theListener) {
    synchronized(sceneListenerList) {
      QualifiedSceneListener sceneListener = null;
      for (int listenerIndex = sceneListenerList.size() - 1; listenerIndex >= 0; listenerIndex--) {
        sceneListener = sceneListenerList.get(listenerIndex);
        if (sceneListener.theListener != theListener) continue;
        
        // Remove listener
        sceneListenerList.remove(listenerIndex);
        sceneListener.releaseResources();
      }
    }
  }

  /**
   * Remove all listeners for a specifc scene.  This is generally only needed
   * when a scene is being removed from the network
   *
   * @param forScene scene to purge listeners for
   */
  public void removeAllRelatedSceneListeners(LightingScene forScene) {
    synchronized(sceneListenerList) {
      QualifiedSceneListener sceneListener = null;
      for (int listenerIndex = sceneListenerList.size() - 1; listenerIndex >= 0; listenerIndex--) {
        sceneListener = sceneListenerList.get(listenerIndex);
        if (sceneListener.theScene != forScene) continue;
        
        // Remove listener
        sceneListenerList.remove(listenerIndex);
        sceneListener.releaseResources();
      }
    }
  }

  // Fires a scene event to all appropriate listeners
  void fireSceneEvent(LightingSceneEvent theEvent) {
    QualifiedSceneListener theListener = null;
    LightingScene eventScene = theEvent.getScene();
    
    int listenerCount = sceneListenerList.size();
    for (int listenerIndex = listenerCount - 1; listenerIndex >= 0; listenerIndex--) {
      if (listenerIndex >= listenerCount) continue;
      
      try {
        theListener = sceneListenerList.get(listenerIndex);
        if (theListener.sendToListener(eventScene)) theListener.theListener.handleLightingSceneEvent(theEvent);
      } catch (Throwable ignoredException) { }
    }
    
    // Free event resources
    theEvent.releaseResources();
  }

  void sendStateRequest(LightingDevice forDevice) {
    LightingMessage stateMessage = new LightingMessage(forDevice, LightingMessage.MsgType.QUERY_STATE);
    sendMessage(stateMessage);
  }
  
  /**
   * Send a message.  
   *
   * @param theMessage to send
   * @return true if the message was sent, false if there was an error
   */
  boolean sendMessage(LightingMessage theMessage) {
    if (LightingGateway.DEBUG_MODE) debug("passing sendConfirmedMessage for gateway to media-adapter");
    return xPLAdapter.sendMessage(theMessage);
  }
  
  /** 
   * Handle the received message from the media adapter, distribute it to the proper 
   * recipients. This should only be invoked from media adapters!
   *
   * <P>NOTE: This is meant to be called by media adapters only and should not be called
   * by anyone else
   *
   * @param theMessage message to deliver
   */
  void distributeReceivedMessage(LightingMessage theMessage) {
    // Distribute a scene message
    if (theMessage.isSceneMessage()) {
      theMessage.getScene().receiveMessage(theMessage); 
      return;
    }
    
    // Distribute a device message
    if (theMessage.isDeviceMessage()) {
      theMessage.getDevice().receiveMessage(theMessage);
      return;
    }

    // Unhandled message -- ignored
  }
}