/*
 * LightingNetwork.java
 *
 * Copyright (c) 2005, Gerald R Duprey Jr
 * More info at http://www.xPL4Java.org
 */

package org.cdp1802.xpl.lighting;

import java.util.*;

/**
 * Describes all items and resoruces associated with a single
 * lighting network.
 *
 * @author gerry
 */
public class LightingNetwork {
  LightingGateway ourGateway = null;
  String networkID = null;
  String networkPassword = "";
  String networkName = "";
  
  LightingRoom defaultRoom = null;
  LightingFloor defaultFloor = null;
  
  ArrayList<LightingDevice> deviceList = new ArrayList<LightingDevice>();
  ArrayList<LightingScene> sceneList = new ArrayList<LightingScene>();
  ArrayList<LightingRoom> roomList = new ArrayList<LightingRoom>();
  ArrayList<LightingFloor> floorList = new ArrayList<LightingFloor>();
  
  LightingNetwork(LightingGateway ourGateway, String networkID) {
    this.ourGateway = ourGateway;
    this.networkID = networkID;
    
    roomList.add(defaultRoom = new LightingRoom(this, "Default Room"));
    floorList.add(defaultFloor = new LightingFloor(this, "Default Floor"));
  }

  void debug(String theMessage) {
    ourGateway.intDebug("NETWORK[" + networkID + "]:: " + theMessage);
  }
  
  void error(String theMessage) {
    ourGateway.intError("NETWORK[" + networkID + "]:: " + theMessage);
  }

  /**
   * Get the Gateway associated with this network.
   * 
   * @return instance of the gateway associated manages this network
   */
  public LightingGateway getGateway() { return ourGateway; }
  
  /**
   * Get this networks unique ID
   *
   * @return this networks id number
   */
  public String getNetworkID() { return networkID; }
  
  void setNetworkID(String newNetworkID) {
    networkID = newNetworkID;
  }

  LightingRoom getDefaultRoom() { return defaultRoom; }
  LightingFloor getDefaultFloor() { return defaultFloor; }
  
  /**
   * Get the networks password, if any.  
   *
   * @return network password, if there is one and we've learned it yet
   */
  public String getNetworkPassword() { return networkPassword; }
  
  void setNetworkPassword(String newPassword) {
    networkPassword = newPassword;
  }

  /**
   * Get the networks name, if any
   *
   * @return network name, if any
   */
  public String getNetworkName() { return networkName; }
  
  
  /**
   * Get the number of devices associated with this network
   *
   * @return known Devices for this network
   */
  public int getDeviceCount() { return deviceList.size(); }
  
  /**
   * Given an index between 0 and the deviceCount - 1, return the device
   * at that index.    
   * 
   * @param deviceIndex index of device to return
   * @return the device at that index or null if index is invalid
   */
  public LightingDevice getDeviceAt(int deviceIndex) {
    synchronized(deviceList) {
      if ((deviceIndex < 0) || (deviceIndex >= deviceList.size())) return null;
      return deviceList.get(deviceIndex);
    }
  }

  /**
   * Given a device ID, return the instance of that device
   * that matches the ID (if any)
   *
   * @param deviceID device ID to lookup
   * @return instance of that device or null if no matching device
   */ 
  public LightingDevice getDeviceByID(String deviceID) {
    if ((deviceID == null) || (deviceID.length() == 0)) return null;

    synchronized(deviceList) {
      for (LightingDevice theDevice: deviceList) {
        if (theDevice.getDeviceID().equals(deviceID)) return theDevice;
      }
    }

    return null;
  }

  /**
   * Get the index of the passed device in this networks device list
   *
   * @param theDevice device to get index of
   * @return index of device or -1 if not found
   */
  public int getIndexOfDevice(LightingDevice theDevice) {
    int deviceIndex = 0;
    synchronized(deviceList) {
      for (LightingDevice aDevice: deviceList) {
        if (aDevice == theDevice) return deviceIndex;
        deviceIndex++;
      }
    }
    
    return -1;
  }
  
  
  /**
   * Given the name of a device, find the first instance of a device
   * that matches the name and return it.  
   *
   * <P>NOTE: if multiple devices have the same name, the one with the 
   * lowest ID will be returned.
   *
   * @param deviceName name of device to find
   * @return device matching the name or null if no match
   */
  public LightingDevice getDeviceNamed(String deviceName) {
    if ((deviceName == null) || (deviceName.length() == 0)) return null;

    for (LightingDevice theDevice: deviceList) {
      if (theDevice.getDeviceName().equals(deviceName)) return theDevice;
    }
    
    return null;
  }

  /**
   * Get the entire collection of devices for this network.
   *
   * @return collection of devices for this network
   */
  public Collection<LightingDevice> getDevices() { return deviceList; }
  
  boolean addDevice(LightingDevice theDevice) {
    // Don't allow bogus scenes
    if ((theDevice == null) || (theDevice.getNetwork() != this)) return false;
    
    // Insure there are no dups
    synchronized(deviceList) {
      LightingDevice dupDevice = getDeviceByID(theDevice.getDeviceID());
      if (dupDevice != null) {
        if (dupDevice == theDevice) return true;
        return false;
      }

      // Add it in
      deviceList.add(theDevice);
    }
    
    return true;
  }
  
  boolean removeDevice(LightingDevice theDevice) {
    // Don't allow bogus scenes
    if ((theDevice == null) || (theDevice.getNetwork() != this)) return false;

    int deviceIndex = getIndexOfDevice(theDevice);
    if (deviceIndex == -1) return false;
    deviceList.remove(deviceIndex);
    
    return true;
  }

  // Scenes
  
  /**
   * Get the number of scenes defined for this network
   *
   * @return number of scenes defined
   */
  public int getSceneCount() { return sceneList.size(); }

  /**
   * Give an index between 0 and sceneCount - 1, return the scene at that 
   * index.
   *
   * <P>Note that while scenes are stored internally in 
   * a list that has holes in it (which getScenes() sees), this will
   * filter them out and only consideres actually present scenes when
   * determining the one that matches the index.
   *
   * <P>In short, it does what you'd expect, but it's clarified here  to
   * make the difference between it and what getScenes() passes 
   * explicit
   *
   * @param sceneIndex index of scene to return
   * @return the scene at that index or null if index is invalid
   */
  public LightingScene getSceneAt(int sceneIndex) {
    synchronized(sceneList) {
      if ((sceneIndex < 0) || (sceneIndex >= sceneList.size())) return null;
      return sceneList.get(sceneIndex);
    }
  }

  /**
   * Given the passed scene, find the index of it in this networks list 
   * of scenes.
   *
   * @param theScene Scene to get index of
   * @return index of scene or -1 if scene is not in this networks list
   */
  public int getIndexOfScene(LightingScene theScene) {
    synchronized(sceneList) {
      int sceneIndex = 0;
      for (LightingScene aScene: sceneList) {
        if (aScene == theScene) return sceneIndex;
        sceneIndex++;
      }
    }
    
    return -1;
  }

  /**
   * Given the passed scene ID, find the find instance that matches
   * it and return it
   *
   * @param sceneID sceneID to search for
   * @return instance of matching scene or null if nothing matches
   */
  public LightingScene getSceneByID(String sceneID) {
    if ((sceneID == null) || (sceneID.length() == 0)) return null;
    
    synchronized(sceneList) {
      for (LightingScene aScene: sceneList) {
        if (aScene.getSceneID().equals(sceneID)) return aScene;
      }
    }
    
    return null;
  }

  /**
   * Given the name of a scene, find the first instance of a scene
   * that matches the name and return it.  
   *
   * <P>NOTE: if multiple scenes have the same name, the one with the 
   * lowest ID will be returned.
   *
   * @param sceneName name of scene to find
   * @return scene matching the name or null if no match
   */
  public LightingScene getSceneNamed(String sceneName) {
    if ((sceneName == null) || (sceneName.length() == 0)) return null;
    synchronized(sceneList) {
      for (LightingScene theScene: sceneList) {
        if (theScene.getSceneName().equals(sceneName)) return theScene;
      }
    }
    
    return null;
  }

  /**
   * Get the entire collection of scenes for this network.
   *
   * @return collection of scenes for this network
   */
  public Collection<LightingScene> getScenes() { return sceneList; }
  
  boolean addScene(LightingScene theScene) {
    // Don't allow bogus scenes
    if ((theScene == null) || (theScene.getNetwork() != this)) return false;

    synchronized(sceneList) {
      // Insure there are no dups
      LightingScene dupScene = getSceneByID(theScene.getSceneID());
      if (dupScene != null) {
        if (dupScene == theScene) return true;
        return false;
      }

      // Add it in
      sceneList.add(theScene);
    }
    
    if (LightingGateway.DEBUG_MODE) debug("Added Scene [" + theScene + "]");
    return true;
  }

  boolean removeScene(LightingScene theScene) {
    // Don't allow bogus scenes
    if ((theScene == null) || (theScene.getNetwork() != this)) return false;

    synchronized(sceneList) {
      int sceneIndex = getIndexOfScene(theScene);
      if (sceneIndex == -1) return false;
      sceneList.remove(sceneIndex);
    }
    
    return true;
  }
  
  // Rooms
  
  /**
   * Get the number of known rooms for this network
   *
   * @return room count for this network
   */
  public int getRoomCount() { return roomList.size(); }
  
  /**
   * Get the room at the passed index
   *
   * @param roomIndex index of room to fetch
   * @return room at the index or null if invalid index
   */
  public LightingRoom getRoomAt(int roomIndex) {
    if ((roomIndex < 0) || (roomIndex >= roomList.size())) return null;
    return roomList.get(roomIndex);
  }

  /**
   * Given a name, find the first room with that name and return
   * and instance of it.
   *
   * @param theRoomName name of the room to lookup
   * @return instance of a room that matches the name or null if no match is found
   */
  public LightingRoom getRoomNamed(String theRoomName) {
    for (LightingRoom theRoom: roomList) {
      if (theRoom.getRoomName().equals(theRoomName)) return theRoom;
    }
    return null;
  }
  
  /**
   * Get a collection of all known rooms for this network
   *
   * <P>NOTE: Rooms will be returned in alphabetical order
   *
   * @return collection of known rooms for this network
   */
  public Collection<LightingRoom> getRooms() { return roomList; }
  
  boolean addRoom(LightingRoom theRoom) {
    // Don't allow bad rooms in
    if ((theRoom == null) || (theRoom.getNetwork() != this)) return false;
    if ((theRoom.getRoomName() == null) || (theRoom.getRoomName().length() == 0)) return false;
    
    // See if there is a dup room
    LightingRoom dupRoom = getRoomNamed(theRoom.getRoomName());
    if (dupRoom != null) {
      if (dupRoom == theRoom) return true;
      return false;
    }
    
    // Add room in alpha order
    int insertAt = -1;
    for (int roomIndex = 0; roomIndex < roomList.size(); roomIndex++) {
      if (roomList.get(roomIndex).getRoomName().compareTo(theRoom.getRoomName()) > 0) {
        insertAt = roomIndex;
        roomList.add(roomIndex, theRoom);
        break;
      }
    }
    if (insertAt == -1) roomList.add(theRoom);
    
    if (LightingGateway.DEBUG_MODE) debug("Added Room [" + theRoom + "] -- " + roomList.size() + " rooms now known");
    return true;
  }

  boolean removeRoom(LightingRoom theRoom) {
    if ((theRoom == null) || (theRoom.getNetwork() != this)) return false;

    for (int roomIndex = 0; roomIndex < roomList.size(); roomIndex++) {
      if (roomList.get(roomIndex) != theRoom) continue;
      
      roomList.remove(roomIndex);
      return true;
    }
    
    return false;
  }


  // Floors
  
  /**
   * Get the number of known floors for this network
   *
   * @return floor count for this network
   */
  public int getFloorCount() { return floorList.size(); }
  
  /**
   * Get the floor at the passed index
   *
   * @param floorIndex index of floor to fetch
   * @return floor at the index or null if invalid index
   */
  public LightingFloor getFloorAt(int floorIndex) {
    if ((floorIndex < 0) || (floorIndex >= floorList.size())) return null;
    return floorList.get(floorIndex);
  }

  /**
   * Given a name, find the first floor with that name and return
   * and instance of it.
   *
   * @param theFloorName name of the floor to lookup
   * @return instance of a floor that matches the name or null if no match is found
   */
  public LightingFloor getFloorNamed(String theFloorName) {
    for (LightingFloor theFloor: floorList) {
      if (theFloor.getFloorName().equals(theFloorName)) return theFloor;
    }
    return null;
  }
  
  /**
   * Get a collection of all known floors for this network
   *
   * <P>NOTE: Floors will be returned in alphabetical order
   *
   * @return collection of known floors for this network
   */
  public Collection<LightingFloor> getFloors() { return floorList; }
  
  boolean addFloor(LightingFloor theFloor) {
    // Don't allow bad floors in
    if ((theFloor == null) || (theFloor.getNetwork() != this)) return false;
    if ((theFloor.getFloorName() == null) || (theFloor.getFloorName().length() == 0)) return false;
    
    // See if there is a dup floor
    LightingFloor dupFloor = getFloorNamed(theFloor.getFloorName());
    if (dupFloor != null) {
      if (dupFloor == theFloor) return true;
      return false;
    }
    
    // Add floor in alpha order
    int insertAt = -1;
    for (int floorIndex = 0; floorIndex < floorList.size(); floorIndex++) {
      if (floorList.get(floorIndex).getFloorName().compareTo(theFloor.getFloorName()) > 0) {
        insertAt = floorIndex;
        floorList.add(floorIndex, theFloor);
        break;
      }
    }
    if (insertAt == -1) floorList.add(theFloor);
    
    if (LightingGateway.DEBUG_MODE) debug("Added Floor [" + theFloor + "] -- " + floorList.size() + " floors now known");
    return true;
  }

  boolean removeFloor(LightingFloor theFloor) {
    if ((theFloor == null) || (theFloor.getNetwork() != this)) return false;

    for (int floorIndex = 0; floorIndex < floorList.size(); floorIndex++) {
      if (floorList.get(floorIndex) != theFloor) continue;
      
      floorList.remove(floorIndex);
      return true;
    }
    
    return false;
  }
}
