/*
 * LightingRoom.java
 *
 * Copyright (c) 2005, Gerald R Duprey Jr
 * More info at http://www.xPL4Java.org
 */

package org.cdp1802.xpl.lighting;

import java.util.*;

/**
 * Describes a room in a network
 *
 * @author gerry
 */
public class LightingRoom {
  LightingNetwork roomNetwork = null;
  String roomName = "";
  ArrayList<LightingDevice> roomDevices = new ArrayList<LightingDevice>();
  
  LightingRoom(LightingNetwork theNetwork, String roomName) {
    this.roomNetwork = theNetwork;
    this.roomName = roomName;
  }

  /**
   * Get the network associated with this room
   *
   * @return network associated with this room
   */
  LightingNetwork getNetwork() { return roomNetwork; }

  /**
   * Name of this room
   *
   * @return name of this room
   */
  public String getRoomName() { return roomName; }
  void setRoomName(String newName) {
    // Make sure it's valid and different
    if ((newName == null) || (newName.length() == 0)) return;
    if (newName.equals(roomName)) return;
    
    // Assign it
    roomName = newName;
    
    // If there are no devices, we are done
    if (roomDevices.size() == 0) return;
  }

  /**
   * Get the number of devices currently associated with this room
   *
   * @return number of devices in the room
   */
  public int getDeviceCount() { return roomDevices.size(); }
  
  /**
   * Given an index between 0 and deviceCount - 1, return the device at 
   * that index for this room
   *
   * @param deviceIndex index of device in this room to return
   * @return device instance of device at the index or null for invalid index
   */
  public LightingDevice getDeviceAt(int deviceIndex) {
    if ((deviceIndex < 0) || (deviceIndex >= roomDevices.size())) return null;
    return roomDevices.get(deviceIndex);
  }
  
  /**
   * Given a device ID, check this room to see if the device is 
   * associated with the room and if so, return the device
   *
   * @param deviceID device ID to look for in this room
   * @return instance of matching device if assigned to this room or null if not found
   */
  public LightingDevice getDeviceByID(String deviceID) {
    // This is faster than scanning room devices because it's
    // all direct access lookups (no loops)
    LightingDevice roomDevice = roomNetwork.getDeviceByID(deviceID);
    if ((roomDevice == null) || (roomDevice.getRoom() != this)) return null;
    return roomDevice;
  }
  
  /**
   * Get a collection of devices assigned to this room
   *
   * <P>NOTE: Devices are returned in device ID order
   *
   * @return collection of devices assigned to this room
   */
  public Collection<LightingDevice> getDevices() { return roomDevices; }
  
  boolean addDevice(LightingDevice theDevice) {
    synchronized(roomDevices) {
      if (getDeviceByID(theDevice.getDeviceID()) != null) return false;
      roomDevices.add(theDevice);
    }
    
    return true;
  }
  
  boolean removeDevice(LightingDevice theDevice) {
    synchronized(roomDevices) {
      return roomDevices.remove(theDevice);
    }
  }

  /**
   * Get a printable description of this room
   *
   * @return printable description of this room
   */
  public String toString() { return roomName; }
  
  void releaseResources() {
    roomNetwork = null;
    roomName = null;
    if (roomDevices != null) {
      roomDevices.clear();
    }
    roomDevices = null;
  }
}