/*
 * LightingScene.java
 *
 * Copyright (c) 2005, Gerald R Duprey Jr
 * More info at http://www.xPL4Java.org
 */

package org.cdp1802.xpl.lighting;

import java.util.*;

/**
 * A scene is a "group" of a number of lighting devices
 * to be controlled with a single command.  Each scene/device connection (or
 * association) has it's own level that device should go to when the scene
 * is activated as well as a faderate to describe how long that change
 * should take.
 *
 * <P>In Lighting terminology, "activating" a scene tells all devices associated with
 * that scene to go to the level programmed into them for that specific scene.  
 * This often means the many devices in the scene will go to different levels.
 *
 * <P>In addition to activating and deactivating scenes, you can also command all
 * devices in a scene to do the same thing.  For example, you tell all devices
 * in a scene to go to 80% over 10 seconds.  
 *
 * <P>It's important to differentiate commanding all devices in a scene vs. 
 * activating a scene.  When activating, each associated device may go to a
 * different, pre-programmed level at a different pre-programmed rate.  When
 * a scene is commanded, all the devices do the exact same thing (based on
 * the command) and ignore any pre-programmed levels for that scene.
 *
 * @author gerry
 */
public class LightingScene {
  LightingNetwork sceneNetwork = null;
  
  String sceneID = null;
  String sceneName = "";
  
  ArrayList<LightingSceneDevice> sceneDevices = new ArrayList<LightingSceneDevice>();
  
  /**
   * Creates a new instance of LightingScene
   */
  LightingScene(LightingNetwork theNetwork, String theSceneID) {
    this.sceneNetwork = theNetwork;
    this.sceneID = theSceneID;
  }
  
  void debug(String theMessage) {
    sceneNetwork.getGateway().intDebug("SCENE[" + sceneID + "]:: " + theMessage);
  }
  
  void error(String theMessage) {
    sceneNetwork.getGateway().intError("SCENE[" + sceneID + "]:: " + theMessage);
  }
  
  /**
   * Get the network this scene is associated with
   *
   * @return network scene is associated with
   */
  public LightingNetwork getNetwork() { return sceneNetwork; }

  /**
   * Return the scene ID of this scene.  Each scene ID is unique within a network.
   *
   * @return scene ID for this scene
   */
  public String getSceneID() { return sceneID; }

  /**
   * Return the name of this scene.  Some scenes may not have names and that will
   * not affect how they work (though it makes figuring out what is going on
   * a bit harder for people reading the results).
   *
   * @return name of this scene
   */
  public String getSceneName() { return sceneName; }
  void setSceneName(String newName) {
    if ((newName == null) || (newName.length() == 0) || newName.equals(sceneName)) return;
    sceneName = newName;
  }

  /**
   * Activate all devices associated with this scene.
   *
   * <P>Activating means that each device should go to it's own unique pre-programmed
   * level that was set when the device was associated with this scene.
   *
   * <P>The distinction here (vs turning setSceneLevel or turnSceneOn/Off) is that
   * each device in the scene will tend to have differernt levels it'll go 
   * to.  The setSceneLevel/turnSceneOnOff will set ALL devices in the scene to
   * the SAME level (regardless of pre-programmed settings in the scene
   * association).
   */
  public void activate() {
    // Send a message and insure it gets sent
    if (!sceneNetwork.getGateway().sendMessage(new LightingMessage(this, LightingMessage.MsgType.ACTIVATE))) return;


    // Now send an event and tell all devices to assume that state
    sceneNetwork.getGateway().fireSceneEvent(new LightingSceneEvent(this, LightingSceneEvent.EventCode.SCENE_ACTIVATED));
    for (LightingSceneDevice sceneDevice: sceneDevices) {
      sceneDevice.getDevice().updateInternalDeviceLevel(sceneDevice.getLevel(), sceneDevice.getFadeRate(), sceneDevice.getChannel());
    }
  }

  /**
   * Deactivate all devices associated with this scene.
   *
   * <P>Deactivating means that each device in the scene should turn off
   */
  public void deactivate() {
    // Send a message and insure it gets sent
    if (!sceneNetwork.getGateway().sendMessage(new LightingMessage(this, LightingMessage.MsgType.DEACTIVATE))) return;

    // Now send an event and tell all devices to assume that state
    sceneNetwork.getGateway().fireSceneEvent(new LightingSceneEvent(this, LightingSceneEvent.EventCode.SCENE_DEACTIVATED));
    for (LightingSceneDevice sceneDevice: sceneDevices) {
      sceneDevice.getDevice().updateInternalDeviceLevel(0, 0, 0);
    }
  }

  /**
   * Command all devices associated with this scene to all go to the
   * passed level at the passed faderate.
   *
   * <P>This is markedly different than activating a scene (see the notes
   * on the activateScene() method for a more indepth explanation).
   *
   * @param toLevel new level (0-100, DEFAULT_DIM_LEVEL or LAST_DIM_LEVEL) all devices in the scene should go to
   * @param atFadeRate rate devices should use to change level to in seconds
   */
  public void gotoLevel(int toLevel, float atFadeRate) {
    // Send a message and insure it gets sent
    if (!sceneNetwork.getGateway().sendMessage(new LightingMessage(this, LightingMessage.MsgType.GOTO, toLevel, atFadeRate))) return;

    sceneNetwork.getGateway().fireSceneEvent(new LightingSceneEvent(this, LightingSceneEvent.EventCode.SCENE_GOTO, toLevel, atFadeRate));
    for (LightingSceneDevice sceneDevice: sceneDevices) {
      sceneDevice.getDevice().updateInternalDeviceLevel(toLevel, sceneDevice.getFadeRate(), sceneDevice.getChannel());
    }
  }

  /**
   * Command all devices associated with this scene to all go to the
   * passed level at their default fade rates
   *
   * <P>This is markedly different than activating a scene (see the notes
   * on the activateScene() method for a more indepth explanation).
   *
   * @param toLevel new level (0-100, DEFAULT_DIM_LEVEL or LAST_DIM_LEVEL) all devices in the scene should go to
   */
  public void gotoLevel(int toLevel) {
    gotoLevel(toLevel, LightingDevice.DEFAULT_FADE_RATE);
  }

  /**
   * Get the number of devices associated with this scene
   *
   * @return number of devices associated with scene
   */
  public int getSceneDeviceCount() { return sceneDevices.size(); }
  
  /**
   * Get the scene/device record at the passed index
   *
   * @param deviceIndex index of the scene/device to return
   * @return scene/device record at that index or null if invalid index passed
   */
  public LightingSceneDevice getSceneDeviceAt(int deviceIndex) {
    return sceneDevices.get(deviceIndex);
  }

  /**
   * Scan the devices associated with scene and if one of them matches
   * the passed device ID, return it
   *
   * @param deviceID device ID to scan this scene for
   * @param theChannel channel in device to scan for (-1 means find first/any instance of device)
   * @return scene/device record that matches the device ID or null if no match (device not associated with this scene)
   */
  public LightingSceneDevice getSceneDeviceById(String deviceID, int theChannel) {
    for (LightingSceneDevice theSceneDevice: sceneDevices) {
      if (!theSceneDevice.getDevice().getDeviceID().equals(deviceID)) continue;
      if ((theChannel != -1) && (theChannel != theSceneDevice.getChannel())) continue;
      
      return theSceneDevice;
    }
    
    return null;
  }

  /**
   * Return a collection of all scene/device records describing devices
   * associated with this scene.
   *
   * @return collection of all devices associated with this scene
   */
  public Collection<LightingSceneDevice> getSceneDevices() { return sceneDevices; }

  boolean addDevice(LightingSceneDevice theSceneDevice) {
    synchronized(sceneDevices) {
      // Make sure this device isn't already in this scene
      if (getSceneDeviceById(theSceneDevice.getDevice().getDeviceID(), theSceneDevice.getChannel()) != null) return false;
      sceneDevices.add(theSceneDevice);
      
      synchronized(theSceneDevice.getDevice()) {
        theSceneDevice.getDevice().deviceScenes.add(theSceneDevice);
      }
    }
    
    return true;
  }

  
  boolean addDevice(LightingDevice theDevice, int theLevel, float theFadeRate, int forChannel) {
    LightingSceneDevice newScene = new LightingSceneDevice(this, theDevice, theLevel, theFadeRate, forChannel);    
    if (!addDevice(newScene)) {
      newScene.releaseResources();
      return false;
    }

    return true;
  }

  // theChannel == -1 means remove all refs to this device for any channel
  boolean removeDevice(LightingDevice theDevice, int theChannel) {
    LightingSceneDevice sceneDevice = null;
    boolean deviceRemoved = false;
    
    synchronized(sceneDevices) {
      for (int deviceIndex = sceneDevices.size() - 1; deviceIndex >= 0; deviceIndex--) {
        if ((sceneDevice = sceneDevices.get(deviceIndex)).getDevice() != theDevice) continue;
        if ((theChannel != -1) && (theChannel != sceneDevice.getChannel())) continue;

        sceneDevices.remove(deviceIndex);
        theDevice.deviceScenes.remove(sceneDevice);

        sceneDevice.releaseResources();
        deviceRemoved = true;
      }
    }

    return deviceRemoved;
  }
  
  boolean removeDevice(LightingSceneDevice theSceneDevice) {
    return removeDevice(theSceneDevice.getDevice(), theSceneDevice.getChannel());
  }
  
  void removeAllDevices() {
    LightingSceneDevice sceneDevice = null;

    synchronized(sceneDevices) {
      for (int deviceIndex = sceneDevices.size() - 1; deviceIndex >= 0; deviceIndex--) {
        sceneDevice = sceneDevices.get(deviceIndex);

        sceneDevices.remove(deviceIndex);
        sceneDevice.getDevice().deviceScenes.remove(sceneDevice);
        sceneDevice.releaseResources();
      }
    }
  }
  
  /**
   * Add a scene event listener to this scene
   * 
   * <P>NOTE: This is the same as LightingGateway.getManager().addSceneListener(theListener, this);
   * 
   * @param theListener listener to receive events about this scene
   */
  public void addListener(LightingSceneListenerI theListener) {
    sceneNetwork.getGateway().addSceneListener(theListener, this);
  }
  
  /**
   * Remove a scene event listener of this scene
   * 
   * <P>NOTE: This is the same as LightingGateway.getManager().removeSceneListener(theListener, this);
   * 
   * @param theListener listener to remove from list of this scenes event recipients
   */
  public void removeListener(LightingSceneListenerI theListener) {
    sceneNetwork.getGateway().removeSceneListener(theListener, this);
  }

  void receiveMessage(LightingMessage theMessage) {
    int theLevel = 0, theChannel = 0;
    float theFadeRate = 0;
                
    switch(theMessage.getMessageType()) {
      case ACTIVATE:
        if (LightingGateway.DEBUG_MODE) debug("Got Scene ACTIVATE");
        sceneNetwork.getGateway().fireSceneEvent(new LightingSceneEvent(this, LightingSceneEvent.EventCode.SCENE_ACTIVATED, theMessage.getSourceDevice()));
        
        for (LightingSceneDevice sceneDevice: sceneDevices) {
          sceneDevice.getDevice().updateInternalDeviceLevel(sceneDevice.getLevel(), sceneDevice.getFadeRate(), sceneDevice.getChannel());
        }
        
        break;
        
      case DEACTIVATE:
        if (LightingGateway.DEBUG_MODE) debug("Got Scene DEACTIVATE");
        sceneNetwork.getGateway().fireSceneEvent(new LightingSceneEvent(this, LightingSceneEvent.EventCode.SCENE_DEACTIVATED, theMessage.getSourceDevice()));
        
        for (LightingSceneDevice sceneDevice: sceneDevices) {
          sceneDevice.getDevice().updateInternalDeviceLevel(0, 0, 0);
        }
        
        break;
        
      case GOTO:
        if (LightingGateway.DEBUG_MODE) debug("Got Scene GOTO");

        theLevel = theMessage.getLevel();
        theFadeRate = theMessage.getFadeRate();
        theChannel = theMessage.getChannel();
        sceneNetwork.getGateway().fireSceneEvent(new LightingSceneEvent(this, LightingSceneEvent.EventCode.SCENE_GOTO, theLevel, theMessage.getFadeRate(), theMessage.getSourceDevice()));
        
        for (LightingSceneDevice sceneDevice: sceneDevices) {
          sceneDevice.getDevice().updateInternalDeviceLevel(theLevel, theFadeRate, theChannel);
        }
        break;
        
      default:
        if (LightingGateway.DEBUG_MODE) debug("Got an unhandled report code of " + theMessage.getMessageType().toString());
        break;
    }
  }
  
  /**
   * Return a description and ID for this scene
   *
   * @return name and ID of this scene, in printable form
   */
  public String toString() { return sceneName + " (" + sceneID + ")"; }
  
  void releaseResources() {
    // Insure no listeners are left hanging around
    if (sceneNetwork != null) sceneNetwork.getGateway().removeAllRelatedSceneListeners(this);
    
    if (sceneDevices != null) {
      for (LightingSceneDevice theSceneDevice: sceneDevices) {
        theSceneDevice.releaseResources();
      }
      sceneDevices.clear();
    }    
    
    sceneNetwork = null;
    sceneID = null;
    sceneName = null;
    sceneDevices = null;
  }
}
