/*
 * LightingSceneEvent.java
 *
 * Copyright (c) 2005, Gerald R Duprey Jr
 * More info at http://www.xPL4Java.org
 */

package org.cdp1802.xpl.lighting;

/**
 * Describes a scene event on the network.
 *
 * <P>Unlike devices which send events only when the devices state changes, scenes
 * do not have "states", so events describe real time reception of scene
 * based commands.
 *
 * <P>Scene action reports *can* have a "source" associated with them -- that is,
 * the device that initiated the action.  In cases where there is, the getSourceDevice()
 * call will return it.  Scene can have action (activate, goto, etc) taken without
 * direct interference by a device (i.e. from a computer) in which case there may be 
 * no source.  So if you care about the source, you must test for a source device -- never
 * assume there always will be one.
 *
 * @author gerry
 */
public class LightingSceneEvent {
  public enum EventCode { 
    /** This scene has been activated */
    SCENE_ACTIVATED,
    
    /** All devices of this scene are being set to a new, common level */
    SCENE_GOTO,
    
    /** This scene has been deactivated */
    SCENE_DEACTIVATED,
  };
  
  LightingScene eventScene = null;
  LightingDevice eventSourceDevice = null;
  EventCode eventCode = null;
  int eventLevel = LightingDevice.UNASSIGNED_DEVICE_STATE;
  float eventFadeRate = LightingDevice.DEFAULT_FADE_RATE;
  
  LightingSceneEvent(LightingScene theScene, EventCode theEventCode, LightingDevice sourceDevice) {
    this.eventScene = theScene;
    this.eventCode = theEventCode;
    this.eventSourceDevice = sourceDevice;
  }
  
  LightingSceneEvent(LightingScene theScene, EventCode theEventCode) {
    this(theScene, theEventCode, null);
  }

  LightingSceneEvent(LightingScene theScene, EventCode theEventCode, int theLevel, float theFadeRate, LightingDevice sourceDevice) {
    this.eventScene = theScene;
    this.eventCode = theEventCode;
    this.eventLevel = theLevel;
    this.eventFadeRate = theFadeRate;
    this.eventSourceDevice = sourceDevice;
  }
  
  LightingSceneEvent(LightingScene theScene, EventCode theEventCode, int theLevel, float theFadeRate) {
    this(theScene, theEventCode, theLevel, theFadeRate, null);
  }
  
  LightingSceneEvent(LightingScene theScene, EventCode theEventCode, int theLevel) {
    this(theScene, theEventCode, theLevel, LightingDevice.DEFAULT_FADE_RATE);
  }

  /** 
   * Get the scene associated with this event
   *
   * @return scene associated with the event
   */
  public LightingScene getScene() { return eventScene; }
  
  /**
   * Get descriptive event code for this event
   *
   * @return code describing this event
   */
  public EventCode getEventCode() { return eventCode; }
  
  /**
   * For SCENE_GOTO, get the level all devices associated
   * with the scene have been commanded to go to.  For all other events, this
   * is -1
   * 
   * @return level all devices have been commanded to go to or -1
   */
  public int getLevel() { return eventLevel; }

  /**
   * For SCENE_GOTO, get the faderate for all devices associated
   * with the scene that have been commanded to go to a new leel.  For all other events, this
   * is -1
   * 
   * @return fade rate for all devices commanded to go to a new level or -1
   */
  public float getFadeRate() { return eventFadeRate; }
  
  /**
   * For scene activity messages (like activate and goto) this is the originator of 
   * the scene activituy request.  Sometimes the originator cannot be determined in
   * which case this will be null.
   *
   * @return device that originated the sene action, if any
   */
  public LightingDevice getSourceDevice() { return eventSourceDevice; }

  void releaseResources() {
    eventScene = null;
    eventCode = null;
    eventSourceDevice = null;
  }
}
