/*
 * LightingXPLAdapter.java
 *
 * Copyright (c) 2005, Gerald R Duprey Jr
 * More info at http://www.xPL4Java.org
 */
package org.cdp1802.xpl.lighting;

import org.cdp1802.xpl.*;
import org.cdp1802.xpl.device.*;
import org.cdp1802.xpl.lighting.LightingGatewayEvent.EventCode;
import org.cdp1802.xpl.xPL_MessageI.MessageType;

/**
 * The xPL Adapter will talk to a remote lighting client  via xPL.
 * 
 * <P>By default, the adapter will send lighting messages to the entire xPL network (i.e. broadcast).
 * You can, however, restrict the destination to a specific xPL device or xPL group.
 * 
 * @author gerry
 */
class LightingXPLAdapter implements xPL_DeviceMessageListenerI {
  public static final String SCHEMA_CLASS = "lighting";

  xPL_DeviceI adapterDevice = null;
  xPL_IdentifierI specifiedRemoteGateway = null;
  xPL_IdentifierI remoteGateway = null;
  LightingGateway ourGateway = null;
  xPL_MutableMessageI commandMessage = null;
  xPL_MutableMessageI requestMessage = null;
  LightingDiscovery xPLDiscoverer = null;
  
  LightingXPLAdapter(LightingGateway theGateway, xPL_DeviceI adapterDevice, xPL_IdentifierI remoteGateway) {
    this.ourGateway = theGateway;
    this.adapterDevice = adapterDevice;
    this.specifiedRemoteGateway = remoteGateway;
    this.adapterDevice.setPassAllMessages(true);
    
    allocateMessages();
    activateAdapter();
  }
  
  void debug(String theMessage) {
    ourGateway.intDebug("xPLADAPTER:: " + theMessage);
  }

  void warn(String theMessage) {
    ourGateway.intWarn("xPLADAPTER:: " + theMessage);
  }

  void error(String theMessage) {
    ourGateway.intError("xPLADAPTER:: " + theMessage);
  }

  // Common routine to allocate messages to match target
  private void allocateMessages() {
    // We can't do anything until there is a device
    if (adapterDevice == null) return;
    
    // Deallocate old resources, if any
    if (commandMessage != null) commandMessage.releaseResources();
    if (requestMessage != null) requestMessage.releaseResources();
    
    // Allocate message templates with appropriate targets
    commandMessage = adapterDevice.createMessage(remoteGateway, xPL_MessageI.MessageType.COMMAND, SCHEMA_CLASS, "basic");
    requestMessage = adapterDevice.createMessage(remoteGateway, xPL_MessageI.MessageType.COMMAND, SCHEMA_CLASS, "request");
  }

  // Create an xPL device and set us up to hear messages
  boolean activateAdapter() {
    // Enable the listener and device
    adapterDevice.addMessageListener(this);
    adapterDevice.setEnabled(true);
    return true;
  }

  xPL_DeviceI getAdapterDevice() { return adapterDevice; }
  xPL_IdentifierI getRemoteGateway() { return remoteGateway; }
  
  boolean startDiscovery() {
    if (xPLDiscoverer == null) xPLDiscoverer = new LightingDiscovery(ourGateway, this);
    if (ourGateway.isDiscoveryInProgress()) return false;
    return xPLDiscoverer.startDiscovery();
  }
  
  // Send a confirmed message
  boolean sendMessage(LightingMessage theMessage) {
    // Decode type of message
    synchronized(commandMessage) {
      commandMessage.clearMessageBody();
      requestMessage.clearMessageBody();

      switch(theMessage.getMessageType()) {
        case ACTIVATE:
          if (theMessage.isSceneMessage()) {
            commandMessage.addNamedValue("command", "activate");
            commandMessage.addNamedValue("network", theMessage.getScene().getNetwork().getNetworkID());
            commandMessage.addNamedValue("scene", theMessage.getScene().getSceneID());
          }
          break;
          
        case DEACTIVATE:
          if (theMessage.isSceneMessage()) {
            commandMessage.addNamedValue("command", "deactivate");
            commandMessage.addNamedValue("network", theMessage.getScene().getNetwork().getNetworkID());
            commandMessage.addNamedValue("scene", theMessage.getScene().getSceneID());
          }
          break;

        case GOTO:
          if (theMessage.isSceneMessage()) {
            commandMessage.addNamedValue("command", "goto");
            commandMessage.addNamedValue("network", theMessage.getScene().getNetwork().getNetworkID());
            commandMessage.addNamedValue("scene", theMessage.getScene().getSceneID());
            commandMessage.addNamedValue("level", LightingGateway.formatLevel(theMessage.getLevel()));
            commandMessage.addNamedValue("fade-rate", LightingGateway.formatFadeRate(theMessage.getFadeRate()));
          } 
          
          if (theMessage.isDeviceMessage()) {
            commandMessage.addNamedValue("command", "goto");
            commandMessage.addNamedValue("network", theMessage.getDevice().getNetwork().getNetworkID());
            commandMessage.addNamedValue("device", theMessage.getDevice().getDeviceID());
            commandMessage.addNamedValue("channel", theMessage.getChannel());
            commandMessage.addNamedValue("level", LightingGateway.formatLevel(theMessage.getLevel()));
            commandMessage.addNamedValue("fade-rate", LightingGateway.formatFadeRate(theMessage.getFadeRate()));
          }
          
          break;
          
        case QUERY_STATE:
          if (theMessage.isDeviceMessage()) {
            requestMessage.clearMessageBody();
            requestMessage.setNamedValue("network", theMessage.getDevice().getNetwork().getNetworkID());
            requestMessage.setNamedValue("request", "devstate");
            requestMessage.setNamedValue("device", theMessage.getDevice().getDeviceID());
            if (theMessage.getChannel() != LightingDevice.INVALID_CHANNEL) requestMessage.setNamedValue("channel", theMessage.getChannel());
          }
          break;
          
          
        default:
          error("Unable to send lighting command via xPL -- Unknown command " + theMessage.getMessageType());
          return false;
      }

      // Send the command, if any
      if (!commandMessage.getMessageBody().isEmpty()) {
        adapterDevice.sendMessage(commandMessage);
        return true;
      } else if (!requestMessage.getMessageBody().isEmpty()) {
        adapterDevice.sendMessage(requestMessage);
        return true;
      }
      return false;
    }    
  }
  
  private void parseDeviceStateReport(xPL_MessageI theMessage) {
    String networkID = theMessage.getNamedValue("network");
    String deviceID = theMessage.getNamedValue("device");
    int channel = theMessage.getIntNamedValue("channel", -1);
    int level = theMessage.getIntNamedValue("level", -1);
    
    // Make sure we aren't missing anything
    if ((networkID == null) || (deviceID == null) || (channel == -1) || (level == -1)) {
      error("Missing parts of a lighting.device state report -- report ignored");
      return;
    }
    
    // Validate the network
    LightingNetwork theNetwork = ourGateway.getNetworkByID(networkID);
    if (theNetwork == null) {
      if (ourGateway.isDiscoveryComplete()) {
        error("Unknown network ID " + networkID + " specified -- device state report dropped");
      }
      return;
    }
    
    // Get the device
    LightingDevice theDevice = theNetwork.getDeviceByID(deviceID);
    if (theDevice == null) {
      if (ourGateway.isDiscoveryComplete()) {
        error("Unknown device ID " + deviceID + " specified -- device state report dropped");
      }
      return;
    }

    // Encapsulate this into a report and ship it off
    ourGateway.distributeReceivedMessage(new LightingMessage(theDevice, LightingMessage.MsgType.GOTO, level, LightingDevice.DEFAULT_FADE_RATE, channel));
  }
  
  private void parseSceneActivityReport(xPL_MessageI theMessage) {
    String networkID = theMessage.getNamedValue("network");
    String sceneID = theMessage.getNamedValue("scene");
    String theLevel = theMessage.getNamedValue("level", "default");
    String action = theMessage.getNamedValue("action");
    String theFadeRate = theMessage.getNamedValue("fade-rate", "default");
    float fadeRate = LightingDevice.DEFAULT_FADE_RATE;
    
    // Make sure we aren't missing anything
    if ((networkID == null) || (sceneID == null) || (action == null)) {
      error("Missing parts of a lighting scene activity report -- report ignored");
      return;
    }
    
    // Validate the network
    LightingNetwork theNetwork = ourGateway.getNetworkByID(networkID);
    if (theNetwork == null) {
      if (ourGateway.isDiscoveryComplete()) {
        error("Unknown network ID " + networkID + " specified -- scene activity report dropped");
      }
      return;
    }
    
    // Get the scene
    LightingScene theScene = theNetwork.getSceneByID(sceneID);
    if (theScene == null) {
      if (ourGateway.isDiscoveryComplete()) {
        error("Unknown scene ID " + sceneID + " specified -- scene activity report dropped");
      }
      return;
    }

    LightingMessage.MsgType sceneAction = null;
    if (action.equalsIgnoreCase("activate")) sceneAction = LightingMessage.MsgType.ACTIVATE;
    if (action.equalsIgnoreCase("deactivate")) sceneAction = LightingMessage.MsgType.DEACTIVATE;
    if (action.equalsIgnoreCase("goto")) sceneAction = LightingMessage.MsgType.GOTO;
    
    if (sceneAction == LightingMessage.MsgType.GOTO) {
      // Insure we at least have a level
      if ((theLevel == null) || (theLevel.length() == 0)) {
        error("Missing parts of a lighting.scene activity report -- report ignored");
        return;
      }

      // Parse fade rate, if any
      fadeRate = LightingGateway.parseFadeRate(theFadeRate);
    }
    
    // Encapsulate this into a report and ship it off
    int level = LightingGateway.parseLevel(theLevel);
    
    // See if there is a source device
    LightingDevice sourceDevice = null;
    if (theMessage.hasNamedValue("source")) {
      sourceDevice = theNetwork.getDeviceByID(theMessage.getNamedValue("source"));
    }
    
    ourGateway.distributeReceivedMessage(new LightingMessage(theScene, sceneAction, level, fadeRate, sourceDevice));
  }

  /**
   * Receive xPL messages and decode them
   *
   * @exclude not part of the API -- needed to receive xPL messages
   */
  public void handleXPLDeviceMessage(xPL_DeviceI theDevice, xPL_MessageI theMessage) {
    // Insure it's for us
    if (!theMessage.getSchemaClass().equalsIgnoreCase(SCHEMA_CLASS)) {
      // Check for the remote gateway going away
      
      // If we have no remote gateway or this is not an "end" message, skip town
      if ((remoteGateway == null) || !theMessage.isHeartbeatEnd()) return;
      
      // Insure it's from the right gateway
      if (!remoteGateway.equals(theMessage.getSource())) return;
      
      // Okay -- the gateway we were previously bound to is going away -- unbind us!
      if (specifiedRemoteGateway == null) {
        System.out.println("xPLLighting:: Remote gateway " + remoteGateway + " going offline -- unbinding from it");
      } else {
        System.out.println("xPLLighting:: Remote gateway " + remoteGateway + " going offline -- idling til it's back");
      }
      
      // Let listeners know what happened
      ourGateway.fireGatewayEvent(new LightingGatewayEvent(LightingGatewayEvent.EventCode.REMOTE_GATEWAY_UNBOUND));
      remoteGateway = null;
      allocateMessages();
      return;
    }

    // Skip anything but messages from a lighting source
    if (!theMessage.getSchemaClass().equalsIgnoreCase(SCHEMA_CLASS)) return;
    
    // Apply target filter, if any
    if (remoteGateway != null) {
      // Skip messages not from our target
      if (!remoteGateway.equals(theMessage.getSource())) return;
    } else {
      // In broadcast mode, the first time we find a gateway, we bind to it

      // If we have a specific gateway, make sure this matches
      if ((specifiedRemoteGateway != null) && !theMessage.getSource().equals(specifiedRemoteGateway)) return;
      
      // Bind us
      remoteGateway = theMessage.getSource();
      allocateMessages();
      xPLDiscoverer.setRequestTarget(remoteGateway);
      System.out.println("xPLLighting:: Just bound to remote gateway " + remoteGateway + "");
      ourGateway.fireGatewayEvent(new LightingGatewayEvent(LightingGatewayEvent.EventCode.REMOTE_GATEWAY_BOUND));
    }
    
    // Handle trigger based messages
    if (theMessage.getType() == xPL_MessageI.MessageType.TRIGGER) {
      // Handle device updates
      if (theMessage.getSchemaType().equalsIgnoreCase("device")) {
        parseDeviceStateReport(theMessage);
        return;
      }

      // Handle scene updates
      if (theMessage.getSchemaType().equalsIgnoreCase("scene")) {
        parseSceneActivityReport(theMessage);
        return;
      }

      // Handle a gateway update
      if (theMessage.getSchemaType().equalsIgnoreCase("gateway")) {
        if (theMessage.hasNamedValue("report", "gateway-changed")) {
          ourGateway.fireGatewayEvent(new LightingGatewayEvent(LightingGatewayEvent.EventCode.REMOTE_GATEWAY_CHANGED));
          return;
        }

        return;
      }
      
      error("Got unexpected trigger message [" + theMessage.getSchemaType() + "]");
    }
    
    // Try for a discovery update
    if ((xPLDiscoverer != null) && (theMessage.getType() == xPL_MessageI.MessageType.STATUS)) {
      if (xPLDiscoverer.processInfoMessage(theMessage)) return;
    }
  }
}