package org.cdp1802.xpl.lighting;

import java.util.*;

/**
  A ThreadPool manages a small pool of threads, remembering when
  the thread finishes to return it to the pool.  This conserves the
  number of threads being created (which is expensive) at any time.

  <P>A ThreadPool has a maximum number of threads it will spawn
  at any time.  If all threads are busy and an additional request
  comes in, the request is queued until the the first executing
  thread completes and looks for a new job.  If there are multiple
  requests that come in while all threads in the pool are busy, they
  will be queued and used in a FIFO order.

  <P>Having a large number of threads is not wasteful by itself.  Threads
  are not actually created until a request comes in and there are no
  free threads available to execute the request (and the number of executing
  threads is below the thread pool maximum).
 
  Copyright (c) 2005, Gerald R Duprey Jr
  More info at http://www.xPL4Java.org
*/

class ThreadPool {
  private static final int defaultThreadLimit = 60;

  Vector<ThreadPoolRunner> threadList = new Vector<ThreadPoolRunner>(defaultThreadLimit);
  java.util.LinkedList<TaskInfo> requestQueue = new java.util.LinkedList<TaskInfo>();
  ThreadGroup poolGroup = null;
  String threadPoolName = "ThreadPool";
  int threadLimit = defaultThreadLimit;
  int threadCount = 0;
  int idleThreadCount = 0;
  boolean poolOnHold = false;

  class TaskInfo {
    Runnable theTask;
    String taskDescription = "";
    
    TaskInfo(Runnable theTask, String taskDescription) {
      this.theTask = theTask;
      this.taskDescription = taskDescription;
    }
    
    public void setDescription(String theDescription) {
      taskDescription = theDescription;
    }
    public String getDescription() { return taskDescription; }
  }
  
  /**
    Create an unnamed thread pool.
  */
  ThreadPool() { }

  /**
    Create a named thread pool.
  */
  ThreadPool(String poolName) {
    setPoolName(poolName);
  }

  /**
    Create a named thread pool.
  */
  ThreadPool(String poolName, int maxThreads) {
    setPoolName(poolName);
    setMaxThreads(maxThreads);
  }

  /**
    Create a named thread pool.
  */
  ThreadPool(String poolName, int maxThreads, boolean poolOnHold) {
    setPoolName(poolName);
    setMaxThreads(maxThreads);
    setPoolOnHold(poolOnHold);
  }

  /**
    Set the thread pool name.
  */
  void setPoolName(String poolName) {
    // Skip out if this isn't a reasonable name
    if ((poolName == null) || (poolName.length() == 0)) return;

    // Set the name
    threadPoolName = poolName;
  }

  /**
    Return the thread pool name.
  */
  String getPoolName() { return threadPoolName; }

  synchronized void setPoolOnHold(boolean isOnHold) { 
    if (poolOnHold == isOnHold) return;
    
    synchronized(threadList) {
      poolOnHold = isOnHold;
      threadList.notifyAll();
    }
  }
  
  boolean isPoolOnHold() { return poolOnHold; }
  
  /**
    Set the maximum number of threads this pool can spawn.  The value
    can be from 1 to 2 billion (or so) or 0 which implies no limit.

    @param maxThreads Maximum # of threads spawnable by this pool, 0 for unlimited
  */
  void setMaxThreads(int maxThreads) {
    // Don't allow negative maximum threads
    if (maxThreads >= 0) synchronized(threadList) {
      // Set limit and make sure we have room for them all
      threadLimit = maxThreads;
      threadList.ensureCapacity(threadLimit);
    }
  }

  /**
    Return the current maximum number fo threads this pool can
    spawn.  If the pool is set to allow unlimited threads, 0
    is returned.

    @return Number of threads pool can have or 0 for unlimited
  */
  int getMaxThreads() { return threadLimit; }

  /**
    Return the number of threads currently available. (which is to
    say spawned at some point and currently either idle or busy).

    @return number of actual threads in this pool
  */
  int getCurrentThreadCount() { return threadCount; }

  /**
    Return the number of busy threads.  This number can change
    quickly, and should only be used to get a gauge of the thread
    pool usage.

    @return number of busy threads in this pool
  */
  int getBusyThreadCount() {
    synchronized(threadList) { return threadCount - idleThreadCount; }
  }

  /**
    Return the nymber of idle threads.  This number can change
    quickly and should only be used to get a gauge of the
    thread pool usage.

    @return number of idle threads in this pool
  */
  int getIdleThreadCount() { return idleThreadCount; }

  /**
    Purge all idle threads.  All currently idle threads will be destroyed
    and removed from this pool.  You should only use this if you have had
    a dramatic spike in the number of threads for this pool and now beleive
    it's passed and want to reduce resource usage.

    <P> Note: If you expect to need these threads again later, it's probably
    best not to purge them.  The major benefit of a ProThreadPool is to
    provide enconomy with regards to the over head of thread creation.  Forcing
    the threads to be created again later may be expensive - an expense you've
    already incurred.

    <P>It's probably better to just put a ceiling on the number of threads
    that can be created.  Or, if you know you'll need some threads later but
    not as many as you currently have, use purgeIdleThreads(N) to remove excess threads
    but still leave some around for later.
  */
  void purgeAllIdleThreads() { purgeIdleThreads(0); }

  /**
    Purge idle all but N idle threads from the pool.  All idle threads in
    excess of the number you pass are purged from the thread pool. You should
    only use this if you have had a dramatic spike in the number of threads for
    this pool and now beleive it's passed and want to reduce resource usage.

    <P> Note: If you expect to need these threads again later, it's probably
    best not to purge them.  The major benefit of a ProThreadPool is to
    provide enconomy with regards to the over head of thread creation.  Forcing
    the threads to be created again later may be expensive - an expense you've
    already incurred.

    <P>It's probably better to just put a ceiling on the number of threads
    that can be created.

    @param keepIdleCount number of idle threads to keep
  */
   void purgeIdleThreads(int keepIdleCount) {
    ThreadPoolRunner theThread;

    // Skip if there is a negative number
    if (keepIdleCount < 0) return;

    // Purge idle threads
    synchronized(threadList) {
      // Iterate until we have reduced the thread count
      while (idleThreadCount > keepIdleCount) {
        for (int threadPtr = 0; threadPtr < threadList.size();) {
          // See if this thread is busy.  If not, it's idle and we'll
          // kill it mercilessly.  We then retry the next item in the
          // vector (which would be at the same "index" in the vector
          // since everything shifted down one).  If the thread is
          // busy, we go on to the next thread.
          theThread = (ThreadPoolRunner) threadList.elementAt(threadPtr);
          synchronized(theThread) {
            if (!theThread.threadBusy) {
              removeThread(theThread);
              break;
            } else
              threadPtr++;
          }
        }
      }
    }
  }

  // Try to spawn a new thread.  If the number of threads spawned
  // is at or exceeds the thread limit, then we fail.
  private void spawnThread() {
    // Try to add a new thread
    synchronized(threadList) {
      // See if we are maxed out on threads
      if (threadCount >= threadLimit) {
        System.err.println("THREAD:: Needed more threads, but thread limit reached - task queued");
        return;
      }

      // If we don't have a thread group yet, create one
      if (poolGroup == null) poolGroup = new ThreadGroup(threadPoolName);

      // Create a new thread
      ThreadPoolRunner theThread = new ThreadPoolRunner(this, poolGroup, "DVThread");
      theThread.setPriority(Thread.MIN_PRIORITY);

      // Add the thread to our list
      threadCount++;
      threadList.addElement(theThread);
      if (LightingGateway.DEBUG_MODE) System.out.println("Spawned new Thread " + theThread.threadID + ", now " + threadCount + " threads running");

      // Let it rip!
      theThread.start();
    }
  }

  /** Terminate an existing, running thread and remove it from the
      list of known threads.

      @param theThread thread to remove/shutdown
  */
  void removeThread(ThreadPoolRunner theThread) {
    // Stop the thread
    theThread.timeToDie = true;
    theThread.interrupt();

    // Remove the thread
    synchronized(threadList) {
      threadList.removeElement(theThread);
      threadCount--;
      if (LightingGateway.DEBUG_MODE) System.out.println("Shutdown Thread, now " + threadCount + " threads running");
    }
  }

  /**
    Add a request to be executed by a thread from the pool.  If there
    is an idle thread, the thread will immediatly take the request and
    begin executing it.  If there are no idle threads and the total number
    of threads is below the thread maximum, a new thread is started and
    allowed to execute this request.  If there are no idle threads and the
    thread limit is reached, the request is queued and will be run as soon
    as one of the executing threads becomes free.  Requests that are queued
    are processing in a FIFO order.

    <P>Requests must implement the Runnable interface.  There is no need for them
    to extend the Thread class because all that extra baggage won't help
    anything here (though it won't hurt, and if you are trying to make a class
    that can be independently a thread or part of a pool, this isn't such
    a bad thing).

    <P>The request has it's processing in the "run()" method.  You must *NOT*
    execute any thread commands (suspend(), stop(), interrupt(), etc) on the
    thread that is running your request from within your request - you would
    screw up the processing of the thread pool.

    <P>When your request is finished running, it should just exit the run()
    method.  Don't use stop(), join() or any other threading function.  Just
    exit (fall out the bottom, return, etc)

    @param theRequest class that implements Runnable and will have its run() method invoked
  */
  void addTask(Runnable theTask, String taskDescription) {
    synchronized(requestQueue) {
      requestQueue.addFirst(new TaskInfo(theTask, taskDescription));
      if (LightingGateway.DEBUG_MODE) System.out.println("Queued Threaded Task");

      // See if we should try to use an existing thread
      // or create a new one
      synchronized(threadList) {
        if (idleThreadCount == 0) spawnThread();
      }

      // Wake up a listener, if any, letting them know
      // there is a request available
      requestQueue.notify();
    }
  }
  
  // Print status of all currently running threads
  synchronized void dumpThreadStatus() {
    Enumeration theThreads = threadList.elements();
    while (theThreads.hasMoreElements()) {
      ThreadPoolRunner theThread = (ThreadPoolRunner) theThreads.nextElement();
      System.out.println("THREAD_STAT:: " + theThread.toString());
    }
    System.out.println("THREAD_STAT:: " + threadCount + " total threads of which " + idleThreadCount + " are idle");
  }
  
  // Return the description of the current thread (if any)
  String getThreadDescription() {
    return getThreadDescription(null);
  }
  
  // Return the description of the passed thread
  String getThreadDescription(Thread theThread) {
    if (theThread == null) theThread = Thread.currentThread();
    int threadCount = threadList.size();
    ThreadPoolRunner poolThread = null;
    
    for (int threadIndex = 0; threadIndex < threadCount; threadIndex++) {
      poolThread = ((ThreadPoolRunner) threadList.elementAt(threadIndex));
      if (poolThread == theThread) return poolThread.toString();
    }
    
    return "";
  }
}

