/*
 * PluginManager.java
 *
 * Created on October 11, 2005, 8:44 PM
 *
 */

package org.cdp1802.xpl.plugins;

import java.io.*;
import java.net.*;
import java.util.*;

import org.jdom.*;

import org.cdp1802.xpl.*;

/**
 * This class manager fetching, maintaining and providing access to vendor
 * plugin files.  It will conduct network transactions, if allowed to (and
 * you specify a proxy, if needed) and cache the results locally and hand
 * them back to the caller.
 *
 * In most cases, you can simply use
 *
 *  PluginManager.getManager().getPlugin("vendorID");
 *
 * and everything will "just work".  You can disable network access and/or
 * whether local data is cached on disk or just in memory.  You can also 
 * change where the plugin files are stored, but most times, there isn't
 * any need.
 *
 * @author gerry
 */
public class PluginManager {
  private static PluginManager pluginManager = new PluginManager();
  public static PluginManager getManager() { return pluginManager; }
  
  PluginManagerCache pluginCache = null;
  PluginManagerDownloader pluginDownloader = null;
  
  private PluginManager() { 
    pluginCache = new PluginManagerCache();
    pluginDownloader = new PluginManagerDownloader();
  }

  /**
   * Enable or disable network connections.  By default, network connections
   * are enabled.  When enabled, if a request for a plugin is made and that
   * plugin isn't available locally, it will be fetched from the network.  If
   * disabled and the plugin isn't available, the request will return an 
   * unknown plugin
   *
   * Default is network connections are allowed
   *
   * @param allowNetworkConnections true to enable them, false to disable
   */
  public void setNetworkConnectionsAllowed(boolean allowNetworkConnections) {
    pluginDownloader.setNetworkConnectionsAllowed(allowNetworkConnections);
  }

  /**
   * Determine if network connections are currently allowed
   *
   * @return true if network connections are allowed
   */
  public boolean isNetworkConnectionsAllowed() { 
    return pluginDownloader.isNetworkConnectionsAllowed();
  }
            
  /**
   * Install a proxy that the manager should use when making network 
   * connections to fetch plugin files.  If you pass a host of
   * null, it disables the proxy
   *
   * @param proxyHost host to send HTTP requests to, null to disable proxy
   * @param proxyPort port on proxyHost to send HTTP requests to.
   */
  public void setProxy(String proxyHost, int proxyPort) {
    pluginDownloader.setProxy(proxyHost, proxyPort);
  }

  /**
   * Return the current proxy host.  If there is no proxy, null is returned
   *
   * @return proxy host or null if no proxy
   */
  public String getProxyHost() { 
    return pluginDownloader.getProxyHost();
  }
  
  /**
   * Get the current proxy port or -1 if no proxy
   *  
   * @return proxy port or -1 if no proxy
   */
  public int getProxyPort() { 
    return pluginDownloader.getProxyPort();
  }

  
  /**
   * Enable or disable local caching.  If enabled, whenever any plugin information
   * is fetched, it's cached locally (including the plugin index). If false,
   * the cache directory can still be consulted for locally stored files, but
   * no updates/additions are stored there.
   *
   * Default is local caching is allowed
   *
   * @param isAllowed true to allow local caching, false to disable.
   */
  public void setLocalCachingAllowed(boolean isAllowed) {
    pluginCache.setLocalCachingAllowed(isAllowed);
  }
  
  /**
   * Determine if local caching is allowed
   *
   * @return true if local caching is allowed, false if not
   */
  public boolean isLocalCachingAllowed() { 
    return pluginCache.isLocalCachingAllowed();
  }
  
  /**
   * Change the directory that is used for caching data.  Note that doing this
   * will not copy over any existing cached data, so you'll need to do that
   * yourself.
   *
   * @param cacheDir directory to store cache files in
   * @return true if directory is OK, false if it cannot be used
   */
  public boolean setLocalCacheDirectory(File cacheDir) {
    return pluginCache.setLocalCacheDirectory(cacheDir);
  }
   
  /**
   * Return the currently in use caching directory
   *
   * @return current cache directory
   */
  public File getLocalCacheDirectory() {
    return pluginCache.getLocalCacheDirectory();
  }
  
  
  /*** Plugin Index support ****/
  
  
  
  /**
   * Load the plugin index.  If there is no local copy or the
   * local copy is old, we'll fetch it from the network and then
   * attempt to cache it.
   *
   * You normally do NOT need to do this as the various services will
   * implicitly do this as needed.
   *
   * @param forceReload true causes the index to be redownloaded even if cached
   * @return true if the index is now loaded, false if there was an error
   */
  public boolean loadPluginIndex(boolean forceReload) {
    // See if we have one
    if (!forceReload) {
      // If it's already there, we're done
      if (pluginCache.isPluginIndexAvailable()) return true;
      
      // If we could load it, we're done
      if (pluginCache.loadPluginIndex(false)) return true;
    }

    // We do not have a local copy (or do not want to use it)
    
    // Try to fetch the new index
    Document newIndex = pluginDownloader.downloadPluginIndex();
    if (newIndex == null) return false;
      
    // Compare the index and if no change, don't screw with it
    if (pluginCache.isEquivilentIndex(newIndex)) return true;
    
    // Clear any current cache and install this one
    pluginCache.clearCachedIndex();
    if (!pluginCache.installPluginIndex(newIndex)) return false;
    pluginCache.savePluginIndex();
    return true;
  }

  /**
   * Load the plugin index.  If there is no local copy or the
   * local copy is old, we'll fetch it from the network and then
   * attempt to cache it.
   *
   * You normally do NOT need to do this as the various services will
   * implicitly do this as needed.
   *
   * @return true if the index is now loaded, false if there was an error
   */
  public boolean loadPluginIndex() { return loadPluginIndex(false); }
  
  /**
   * Return true of the plugin index is loaded.  If false, it 
   * may still need to be loaded/fetched
   *
   * @return true if plugin index is loaded
   */
  public boolean isPluginIndexAvailble() {
    return pluginCache.isPluginIndexAvailable();  
  }
  
  /**
   * Return a list of vendor IDs that are currently known of 
   *
   * @return collection of vendor IDs or an empty collection if none
   */
  public Collection<String> getKnownVendorIDs() {
    loadPluginIndex();
    return pluginCache.getKnownVendorIDs();
  }
  
  /**
   * Given the passed vendor, get the URL used to fetch the vendors
   * plugin.
   *
   * @param forVendorID vendor ID to get URL for
   * @return URL for vendor plugin or null if unknown
   */
  public String getPluginURL(String forVendorID) {
    String vendorURL = null;
    
    // Insure the index is loaded and load it if not
    boolean indexWasLoaded = pluginCache.isPluginIndexAvailable();
    boolean indexIsStale = pluginCache.isPluginIndexStale();
    if (!loadPluginIndex(indexIsStale)) return null;

    // See if it's known 
    if ((vendorURL = pluginCache.getPluginURL(forVendorID)) != null) return vendorURL;
    
    // If the index was not loaded before the last load or it was stale before, 
    // then it was just fetched and is accurate as possible -- there aren't any options
    if (!indexWasLoaded || indexIsStale) return null;
    
    // Since the index was loaded already, perhaps it's stale.  Fetch a new
    // one and install it and try again
    if (!loadPluginIndex(true)) return null;
    return pluginCache.getPluginURL(forVendorID);
  }
  
  /**
   * Determine if the passed vendor ID is a known one.  That is, 
   * is it an ID who could have plugin information (even if that
   * information is not currently available).
   *
   * If we cannot find the vendor, we'll force a network reload
   * of the master plugin list and if that still comes up missing,
   * we're done.
   *
   * @param forVendorID vendor ID to check on
   * @return true if there should be a plugin available, false if unknown
   */
  public boolean isPluginKnown(String forVendorID) {
    // If it's available locally, fine
    if (pluginCache.isPluginCached(forVendorID)) return true;
    return getPluginURL(forVendorID) != null;
  }
  
  /**
   * Determine if the passed vendor IDs plugin is available locally
   * or not.  If it's loaded or in the cache, we return true
   *
   * @param forVendorID vendor ID to check on
   * @return true if available locally, false if not or not known
   */
  public boolean isPluginAvailableLocally(String forVendorID) {
    return pluginCache.isPluginCached(forVendorID);
  }
  
  /**
   * For the passed plugin vendor, attempt to load the 
   * plugin and return it.  If the vendor does not have a plugin 
   * or the plugin cannot be loaded, null is returned.
   *
   * NOTE: If you force a lookup and the vendor does not have a known
   * website (or there is an error), this will return the locally cached
   * plugin, if any, or null and not change/clear any currently cached entry.
   *
   * @param forVendorID vendor ID to load plugin for
   * @param forceReload true to force lookup of plugin from vendors website
   * @return VendorPlugin instance or null if no plugin available
   */
  public VendorPlugin getPlugin(String forVendorID, boolean forceReload) {
    // Insure it's known 
    if (!isPluginKnown(forVendorID)) return null;

    VendorPlugin localPlugin = null, downloadedPlugin = null;
    
    
    if (!forceReload) {
      // Attempt to get it locally
      if ((localPlugin = pluginCache.getPlugin(forVendorID)) != null) {
        // See if it might be stale and if so, if there is anything
        // we can do about that
        if (!pluginCache.isPluginStale(forVendorID)) return localPlugin;

        // If it's stale but networking is not allowed, there isn't much we can do
        if (!pluginDownloader.isNetworkConnectionsAllowed()) return localPlugin;
      }
    }

    // Get the URL (if any) for this vendor
    String pluginURL = getPluginURL(forVendorID);
    if (pluginURL == null) return localPlugin;
    
    // Attempt to download it
    if ((downloadedPlugin = pluginDownloader.downloadPlugin(pluginURL)) == null) return localPlugin;
    
    // If there is a downloaded and local, see if the local has changed
    if (localPlugin != null) {
      if (pluginCache.isEquivilent(localPlugin, downloadedPlugin)) return localPlugin;
    }
    
    // Update the cache with this
    pluginCache.addPluginToCache(downloadedPlugin);
    
    // And return it
    return downloadedPlugin;
  }
  
  /**
   * For the passed plugin vendor, attempt to load the 
   * plugin and return it.  If the vendor does not have a plugin 
   * or the plugin cannot be loaded, null is returned.
   *
   * @param forVendorID vendor ID to load plugin for
   * @return VendorPlugin instance or null if no plugin available
   */
  public VendorPlugin getPlugin(String forVendorID) {
    return getPlugin(forVendorID, false);
  }
    
  /**
   * This will clear all items from the cache -- the index as well as all
   * plugins that have been locally cached are cleared
   */
  public void clearCache() { pluginCache.clearCache(); }
  
  /**
   * This call will clear the currently loaded plugin index along with
   * any locally cached copies.  It will not reload anything, but any
   * subsequent call to a index related function will.
   */
  public void clearIndexCache() { pluginCache.clearCachedIndex(); }

  /**
   * Clear any cached info for the plugin identified.  
   *
   * @param forVendorID vendor ID of plugin to clear
   */
  public void clearPluginCache(String forVendorID) { pluginCache.clearCachedPlugin(forVendorID); }
   
  /**
    * Clear any cahced info for all plugins (the plugin index is left
    * in place).
    */
  public void clearAllCachedPlugins() { pluginCache.clearAllCachedPlugins(); }
   
  /**
  * This call will check the plugin index and for each listed plugin that
  * is not already locally cached, will download and cache it (assuming 
  * caching is allowed). 
  *
  * If forceRefresh is passed true, then the index and all plugins listed
  * are redownloaded, even if they are already available locally.
  *
  * This can take a while to complete
  *
  * @param forceRefresh true to force everything to be downloaded
  * @return true if everything OK, false is there was an error
  */
  public boolean loadAllPlugins(boolean forceRefresh) {
    // Insure the plugin index is available
    if (!loadPluginIndex(forceRefresh)) return false;
     
    for (String theVendorID: pluginCache.getKnownVendorIDs()) {
      if (getPlugin(theVendorID, forceRefresh) == null) return false;
    }
     
    return true;
  }

  /**
  * This call will check the plugin index and for each listed plugin that
  * is not already locally cached, will download and cache it (assuming 
  * caching is allowed). 
   *
  * This can take a while to complete
  *
  * @return true if everything OK, false is there was an error
  */
  public boolean loadAllPlugins() { return loadAllPlugins(false); }
}
