/*
 * PluginManagerCache.java
 *
 * Created on October 12, 2005, 11:12 AM
 *
 */

package org.cdp1802.xpl.plugins;

import java.io.*;
import java.net.*;
import java.util.*;

import org.jdom.*;
import org.jdom.input.SAXBuilder;
import org.jdom.output.Format;
import org.jdom.output.XMLOutputter;
import org.apache.commons.httpclient.*;
import org.apache.commons.httpclient.methods.*;

import org.cdp1802.xpl.*;

/**
 *
 * @author gerry
 */
public class PluginManagerCache {
  static final String WINDOWS_CACHE = "C:\\Documents and Settings\\All Users\\Application Data\\xPL\\Plugins\\";
  static final String UNIX_CACHE = "~/.xPL/Plugins";
  static final String PLUGIN_INDEX_FILENAME = "plugins.xml";
  static final int MAX_PLUGIN_CACHE_AGE = 30;  // Max plugin cache days

  static XMLOutputter xmlOutputter = null;
  static SAXBuilder xmlParser = null;
  
  static {
    xmlParser = new SAXBuilder();
    xmlOutputter = new XMLOutputter();
    xmlOutputter.setFormat(Format.getPrettyFormat());
  }

  boolean localCachingAllowed = true;
  File pluginCacheDir = null;
  
  Document rootDocument = null;
  HashMap<String, String> vendorPluginIndex = new HashMap<String, String>();
  HashMap<String, VendorPlugin> vendorPlugins = new HashMap<String, VendorPlugin>();
  
  PluginManagerCache() {
    // Figure out where the cache directory should be
    if (File.separatorChar == '/') {
      pluginCacheDir = new File(System.getProperty("user.home"), ".xPL/Plugins");
      pluginCacheDir.mkdirs();
    } else {
      pluginCacheDir = new File(WINDOWS_CACHE);
      pluginCacheDir.mkdirs();
    }
  }
  
  /**
   * Enable or disable local caching.  If enabled, whenever any plugin information
   * is fetched, it's cached locally (including the plugin index). If false,
   * the cache directory can still be consulted for locally stored files, but
   * no updates/additions are stored there.
   *
   * Default is local caching is allowed
   *
   * @param isAllowed true to allow local caching, false to disable.
   */
  public void setLocalCachingAllowed(boolean isAllowed) {
    localCachingAllowed = isAllowed;
  }
  
  /**
   * Determine if local caching is allowed
   *
   * @return true if local caching is allowed, false if not
   */
  public boolean isLocalCachingAllowed() { return localCachingAllowed; }
  
  /**
   * Change the directory that is used for caching data.  Note that doing this
   * will not copy over any existing cached data, so you'll need to do that
   * yourself.
   *
   * @param cacheDir directory to store cache files in
   * @return true if directory is OK, false if it cannot be used
   */
  public boolean setLocalCacheDirectory(File cacheDir) {
    // Make sure it passes the basics
    if ((cacheDir == null) || !cacheDir.exists() || !cacheDir.isDirectory() || !cacheDir.canRead() || !cacheDir.canWrite()) return false;
    pluginCacheDir = cacheDir;
    return true;
  }
   
  /**
   * Return the currently in use caching directory
   *
   * @return current cache directory
   */
  public File getLocalCacheDirectory() {
    return pluginCacheDir;
  }
  
  /**
   * Compute the age of the passed file, in days
   *
   * @param theFile file to compute age on
   * @return age in days or -1 if the file does not exist
   */
  private int computeFileAge(File theFile) {
    if (!theFile.exists()) return -1;
    
    long fileModified = theFile.lastModified();
    long rightNow = System.currentTimeMillis();
    
    // 86,400,000 milli seconds in a day
    return (int) ((rightNow - fileModified) / 86400000);
  }
  
  /**
   * Build a file reference for the passed plugin
   *
   * @param thePlguin plugin to read vendor ID from for file
   * @return cache file reference (or null if vendor ID is bad)
   */
  private File getCacheFileForPlugin(VendorPlugin thePlugin) {
    if ((thePlugin == null) || (thePlugin.getVendorID() == null) || (thePlugin.getVendorID().length() == 0)) return null;
    return new File(pluginCacheDir, thePlugin.getVendorID()  + ".xml");
  }
  
  private File getCacheFileForIndex() {
    return new File(pluginCacheDir, PLUGIN_INDEX_FILENAME);
  }
  
  private boolean nodeEquals(Element nodeA, Element nodeB){
    if (!nodeA.getName().equals(nodeB.getName())) return false;
    
    List nodeAAttrList = nodeA.getAttributes();
    List nodeBAttrList = nodeB.getAttributes();

    Iterator nodeAAttrs = nodeAAttrList.iterator();
    while (nodeAAttrs.hasNext()) {
      Iterator nodeBAttrs = nodeBAttrList.iterator();
      Attribute nodeAAttribute = (Attribute) nodeAAttrs.next();
      Attribute nodeBAttribute = (Attribute) nodeBAttrs.next();
      
      while (!nodeAAttribute.getName().equals(nodeBAttribute.getName()) && nodeBAttrs.hasNext()) {
        nodeBAttribute = (Attribute)nodeBAttrs.next();
      }

      if (!nodeAAttribute.getValue().equals(nodeBAttribute.getValue())) {
         return false;
      }
    }
    
    return true;
  }    

  private boolean xmlEquals(Element nodeA, Element nodeB) {
    // Do these nodes match?
    if(!nodeEquals(nodeA, nodeB)) return false;

    // Check all child nodes 
    Iterator nodeAChildren = nodeA.getChildren().iterator();
    Iterator nodeBChildren = nodeB.getChildren().iterator();
    while (nodeAChildren.hasNext()) {
      if (!xmlEquals((Element) nodeAChildren.next(),  (Element) nodeBChildren.next())) return false;
    }
    
    // Everything matches!
    return true;
  }
  
  /****************************/
  /*** Plugin Index Support ***/
  /****************************/
  
  /**
   * Clear the index cache.  After this call, you need to
   * download a new index as no aspect (loaded or on disk)
   * will be left.
   */
  public void clearCachedIndex() {
    rootDocument = null;
    vendorPluginIndex.clear();

    // Purge file, if appropriate
    if (localCachingAllowed) {
      File saveFile = getCacheFileForIndex();
      if (saveFile.exists()) saveFile.delete();
    }
  }

  /**
   * Write the passed root index out to the cache.  If 
   * local caching is disabled, this returns as if it
   * succeeded without actually doing anything
   *
   * @return true if saved, false if there was an error
   */
  public boolean savePluginIndex() {
    // If we aren't doing local caching, this *always* works
    if (!localCachingAllowed) return true;

    if (rootDocument == null) return false;
    
    try {
      File saveFile = getCacheFileForIndex();
      if (saveFile.exists()) saveFile.delete();
      
      OutputStream outputStream = new FileOutputStream(saveFile);
      BufferedOutputStream bufferedStream = new BufferedOutputStream(outputStream);
      
      xmlOutputter.output(rootDocument, bufferedStream);

      bufferedStream.flush();
      bufferedStream.close();
      
      return true;
    } catch (IOException anyError) {
      return false;
    }
  }

  /**
   * Attempt to load the passed document as the new plugin document.
   * If it's OK, then install it and return true.  If there are any
   * errors, don't install it and return false
   * 
   * @param newDocument document with plugin to install
   * @return true if successfully installed
   */
  public boolean installPluginIndex(Document newDocument) {
    // Parse this out
    HashMap<String, String> newIndex = new HashMap<String, String>();
    List pluginList = newDocument.getRootElement().getChildren("plugin");
    for (Object anObject: pluginList) {
      Element thePlugin = (Element) anObject;
      
      String pluginURL = thePlugin.getAttributeValue("url");
      if (pluginURL == null) continue;
      
      // Insure it's not a "special" URL
      if (pluginURL.indexOf("xpl-schema-collection") != -1) continue;
      
      // Parse a vendor ID out of this
      int slashPtr = pluginURL.lastIndexOf('/');
      if (slashPtr == -1) continue;
      
      String vendorID = pluginURL.substring(slashPtr + 1);
      
      newIndex.put(vendorID.toLowerCase(), pluginURL + ".xml");
    }

    // Make sure we found plugins
    if (newIndex.isEmpty()) return false;
    
    // Install this
    rootDocument = newDocument;
    vendorPluginIndex.clear();
    vendorPluginIndex = newIndex;
    
    // And we are done
    return true;
  }
  
  /**
   * Attempt to load a local cached version of the plugin index and
   * populate the index data from it.  If that cannot be loaded, then
   * attempt to download and index.  If that is successfull, cache it
   * and install it.
   *
   * @param forceReload reload from disk regardless of current status
   * @return true if there is an index loaded, false if not
   */
  public boolean loadPluginIndex(boolean forceReload) {
    boolean rootDocumentLoaded = (rootDocument != null);
    Document newDocument = null;
    
    // If there is no local caching, we either already have it or not
    if (!localCachingAllowed) {
      return rootDocumentLoaded;
    }
    
    // Get the local cache file and see if we can read it
    File indexFile = getCacheFileForIndex();
    if (!indexFile.exists() || !indexFile.canRead() || (indexFile.length() == 0)) return false;
    
    // See if we really *need* to re-read it or not
    if ((rootDocumentLoaded && (computeFileAge(indexFile) <= MAX_PLUGIN_CACHE_AGE)) && !forceReload) return true;
      
    // Attempt to parse this
    try {
      newDocument = xmlParser.build(indexFile);
      return installPluginIndex(newDocument);
    } catch (Exception anyError) {
      return false;
    }
  }
  
  /**
   * Determine if the passed index document is the same as our current document
   * and returns true if so
   *
   * @param compareDoc document to compare index to
   * @return true if the documents are basically the same
   */
  public boolean isEquivilentIndex(Document compareDoc) {
    if ((rootDocument == null) || (compareDoc == null)) return false;
    return xmlEquals(rootDocument.getRootElement(), compareDoc.getRootElement());
  }
  

  /**
   * Determine if there is a plugin index available or not
   *
   * @return true if plugin index is known available, false if not
   */
  public boolean isPluginIndexAvailable() {
    return ((rootDocument != null) && (vendorPluginIndex.size() != 0));
  }

  /**
   * Determine if the plugin index is old and needs to be refreshed.
   *
   * @return true if index is cached and getting stale
   */
  public boolean isPluginIndexStale() {
    File cacheFile = getCacheFileForIndex();
    if (!cacheFile.exists()) return true;
    
    return computeFileAge(cacheFile) > MAX_PLUGIN_CACHE_AGE;
  }
  
  /**
   * For the given vendor ID, attempt to find the URL for the vendor.  If there
   * is no known ID, return null.  
   *
   * @param forVendorID vendor ID to look for URL in
   * @return vendor plugin URL or null if vendor has no URL
   */
  public String getPluginURL(String forVendorID) {
    return vendorPluginIndex.get(forVendorID.toLowerCase());
  }
  
  /** 
   * Return a list of vendor IDs of vendors listed in the master
   * plugin directory. 
   *
   * @return list of known vendor IDs
   */
  public Collection<String> getKnownVendorIDs() {
    return vendorPluginIndex.keySet();
  }

  /*****************************/
  /*** Vendor Plugin Support ***/
  /*****************************/
  
  /**
   * Build a file reference for the passed vendor
   *
   * @param theVendorID vendorID to create file for
   * @return cache file reference (or null if vendor ID is bad)
   */
  private File getCacheFileForVendorID(String vendorID) {
    if ((vendorID == null) || (vendorID.length() == 0)) return null;
    return new File(pluginCacheDir, vendorID + ".xml");
  }

  /**
   * Add the plugin to the cache.  If there is an older version of the 
   * plugin already in the cache, it's replaced with this one.
   *
   * @param thePlugin plugin to add to cache
   * @return true if save was OK, false if there was an error
   */
  public boolean addPluginToCache(VendorPlugin thePlugin) {
    // Make sure we have a chance here
    if ((thePlugin == null) || (thePlugin.getVendorID() == null) || (thePlugin.getVendorID().length() == 0)) return false;

    // Put it in local cache
    vendorPlugins.put(thePlugin.getVendorID().toLowerCase(), thePlugin);
    
    // If we are caching, no biggie
    if (!localCachingAllowed) return true;
    
    // Create a new filename
    File cacheFile = getCacheFileForPlugin(thePlugin);
    if (cacheFile == null) return false;
    if (cacheFile.exists()) cacheFile.delete();
    
    try {
      thePlugin.savePluginToFile(cacheFile);
    } catch (Exception anything) { return false; }
    
    return true;
  }
  
  /**
   * Given a vendor ID, attempt to fetch the vendor plugin info
   * and return it.  
   *
   * @param vendorID vendorID to fetch
   * @return the plugin or null if the vendor ID is unknown
   */
  public VendorPlugin getPlugin(String vendorID) {
    // If already loaded, we're home free
    VendorPlugin thePlugin = vendorPlugins.get(vendorID.toLowerCase());
    if (thePlugin != null) return thePlugin;

    // See if we have it locally
    File localFile = getCacheFileForVendorID(vendorID);
    if (localFile == null) return null;
    if (!localFile.exists() || !localFile.canRead()) return null;
    
    try {
      thePlugin = new VendorPlugin(localFile);
      vendorPlugins.put(thePlugin.getVendorID().toLowerCase(), thePlugin);
      return thePlugin;
    } catch (Exception anyError) { return null; }
  }
  
  /**
   * Determine if the passed vendor IDs plugin information is available
   * locally.  If it is (it's loaded or found in the cache directory),
   * return true.
   *
   * @param vendorID vendor ID to determine if cached or not
   * @return true if vendor plugin is cached locally, false if not
   */
  public boolean isPluginCached(String vendorID) {
    if (vendorPlugins.get(vendorID.toLowerCase()) != null) return true;

    // See if we have it locally
    return getCacheFileForVendorID(vendorID).exists();
  }

  /**
   * Determine if the passed vendor IDs cached data, if any, is old and
   * needs to be refreshed (if no cache file, true is always returned).
   * 
   * @param forVendorID vendor ID to check on 
   * @return true if vendor is cached and the entry is getting stale
   */
  public boolean isPluginStale(String forVendorID) {
    File cacheFile = getCacheFileForVendorID(forVendorID);
    if (!cacheFile.exists()) return true;
    
    // See if cache file is younger than plugin index.  If so, it 
    // can be considered stale
    File cachedIndexFile = getCacheFileForIndex();
    if (cachedIndexFile.exists()) {
      if (cachedIndexFile.lastModified() > cacheFile.lastModified()) return true;
    }

    // Check age
    return computeFileAge(cacheFile) > MAX_PLUGIN_CACHE_AGE;
  }
  
  /**
   * Determine if the passed plugins are basically the same.  If they 
   * are, then return true
   *
   * @param pluginA plugin to check
   * @param pluginB other plugin to check
   * @return true if they are equivilent
   */
  public boolean isEquivilent(VendorPlugin pluginA, VendorPlugin pluginB) {
    if ((pluginA == null) || (pluginB == null)) return false;

    return xmlEquals(pluginA.getNodeElement(), pluginB.getNodeElement());
  }
  
  /**
   * Clear any cached data on the passed vendors plugin
   *
   * @param theVendorID vendor ID of plugins to clear out of cache
   */
  public void clearCachedPlugin(String theVendorID) {
    vendorPlugins.remove(theVendorID.toLowerCase());
    File localFile = getCacheFileForVendorID(theVendorID);
    if (localFile == null) return;
    localFile.delete();
  }

  /**
   * Clear any cached data on the passed vendors plugin
   *
   * @param thePlugin plugin to clear out of cache
   */
  public void clearCachedPlugin(VendorPlugin thePlugin) {
    clearCachedPlugin(thePlugin.getVendorID());
  }
  
  /**
   * Clear all cached plugins 
   */
  public void clearAllCachedPlugins() {
    vendorPlugins.clear();

    // Purge local files
    File cacheFiles[] = pluginCacheDir.listFiles();
    if ((cacheFiles == null) || (cacheFiles.length == 0)) return;
    for (File cacheFile: cacheFiles) {
      cacheFile.delete();
    }
  }

  /**
   * Clear all cached items, locally and on disk
   *
   * After this, nothing will work until you fetch a
   * new plugin index from the network
   */
  public void clearCache() {
    clearCachedIndex();
    clearAllCachedPlugins();
  }
}
