/*
 * PluginManagerDownloader.java
 *
 * Created on October 12, 2005, 11:26 AM
 *
 */

package org.cdp1802.xpl.plugins;

import java.io.*;
import java.net.*;
import java.util.*;

import org.jdom.*;
import org.jdom.input.SAXBuilder;
import org.jdom.output.Format;
import org.jdom.output.XMLOutputter;
import org.apache.commons.httpclient.*;
import org.apache.commons.httpclient.methods.*;

import org.cdp1802.xpl.*;

/**
 *
 * @author gerry
 */
public class PluginManagerDownloader {
  static final int MAX_PLUGIN_CACHE_AGE = 10;  // Max plugin cache days
  static final String SEED_URLS[] = { "http://www.xplproject.org.uk/plugins.xml", "http://www.xpl.myby.co.uk/support/xplhalweb/plugins.xml" };
  
  static final int networkTimeout = 120;

  static XMLOutputter xmlOutputter = null;
  static SAXBuilder xmlParser = null;
  
  boolean networkConnectionAllowed = true;
  String proxyIPAddr = null;
  int proxyIPPort = -1;

  int lastRootServerTried = -1;
  ArrayList<String> rootServers = new ArrayList<String>();
  HttpClient httpClient = new HttpClient();
  
  static {
    xmlParser = new SAXBuilder();
    xmlOutputter = new XMLOutputter();
    xmlOutputter.setFormat(Format.getPrettyFormat());
  }
  
  PluginManagerDownloader() {
    httpClient.getParams().setConnectionManagerTimeout(20000);

    // Pre-load the root servers
    for (String rootServer: SEED_URLS) {
      rootServers.add(rootServer);
    }
  }
  
  /**
   * Enable or disable network connections.  By default, network connections
   * are enabled.  When enabled, if a request for a plugin is made and that
   * plugin isn't available locally, it will be fetched from the network.  If
   * disabled and the plugin isn't available, the request will return an 
   * unknown plugin
   *
   * Default is network connections are allowed
   *
   * @param allowNetworkConnections true to enable them, false to disable
   */
  void setNetworkConnectionsAllowed(boolean allowNetworkConnections) {
    networkConnectionAllowed = allowNetworkConnections;
  }

  /**
   * Determine if network connections are currently allowed
   *
   * @return true if network connections are allowed
   */
  boolean isNetworkConnectionsAllowed() { return networkConnectionAllowed; }
  
  /**
   * Install a proxy that the manager should use when making network 
   * connections to fetch plugin files.  If you pass a host of
   * null, it disables the proxy
   *
   * @param proxyHost host to send HTTP requests to, null to disable proxy
   * @param proxyPort port on proxyHost to send HTTP requests to.
   */
  void setProxy(String proxyHost, int proxyPort) {
    this.proxyIPAddr = proxyHost;
    this.proxyIPPort = proxyPort;
    
    if (proxyIPAddr == null) {
      httpClient.getHostConfiguration().setProxyHost(null);
    } else {
      httpClient.getHostConfiguration().setProxy(proxyIPAddr, proxyIPPort);
    }
  }

  /**
   * Return the current proxy host.  If there is no proxy, null is returned
   *
   * @return proxy host or null if no proxy
   */
  String getProxyHost() { return proxyIPAddr; }
  
  /**
   * Get the current proxy port or -1 if no proxy
   *  
   * @return proxy port or -1 if no proxy
   */
  int getProxyPort() { return proxyIPPort; }
  
  // Cimple convience to create a GetMethod configured as 
  // we want it
  private GetMethod createGetter(String toURL) {
    GetMethod getMethod = new GetMethod(toURL);
    getMethod.setFollowRedirects(true);
    return getMethod;
  }
  
  /**
   * Pick a index server URL to use at random.  This just helps to 
   * mix up the order we hit the server in
   *
   * @return URL of the index server to try
   */
  private String getRandomRootURL() { 
    lastRootServerTried = (int) (Math.random() * (rootServers.size() - 1));
    return rootServers.get(lastRootServerTried);
  }
  
  /**
   * Using one of the plugin servers, fetch the current version of the 
   * plugin index.
   *
   * @return document loaded or null if error loading document
   */
  public Document downloadPluginIndex() {
    String theIndexURL = getRandomRootURL();
    int startingRootServerIndex = lastRootServerTried;
    
    for(;;) {
      // Fetch a getter
      GetMethod getter = createGetter(theIndexURL);

      try {
        int response = httpClient.executeMethod(getter);

        // See if this failed
        if (response != HttpStatus.SC_OK) {
          // Read the response and ditch it
          try { 
            getter.getResponseBody();
          } catch (IOException ioError) { }
        } else {
          // Parse the data into a document
          InputStream responseStream = null;
          try {
            responseStream = getter.getResponseBodyAsStream();
            System.out.println("PLUGIN_MGR:: Downloading Plugin Index from " + theIndexURL);
            return xmlParser.build(responseStream);
          } finally {
            if (responseStream != null) responseStream.close();
          }
        }
      } catch (Exception anyError) {
        // Do nothing
      } finally {
        getter.releaseConnection();
      }
      
      // If we get there, this URL didn't work -- try another
      lastRootServerTried++;
      if (lastRootServerTried >= rootServers.size()) lastRootServerTried = 0;
      if (lastRootServerTried == startingRootServerIndex) {
        // We've tried them all
        return null;
      }
      theIndexURL = rootServers.get(lastRootServerTried);
    }
  }
  
  public VendorPlugin downloadPlugin(String vendorURL) {
    // Fetch a getter
    GetMethod getter = createGetter(vendorURL);

    try {
      int response = httpClient.executeMethod(getter);

      // See if this failed
      if (response != HttpStatus.SC_OK) {
        // Read the response and ditch it
        try { 
          getter.getResponseBody();
        } catch (IOException ioError) { }
        
        System.err.println("PLUGIN_MGR:: Unable to download plugin for " + vendorURL + ", HTTP status " + response);
        return null;
      }
      
      // Parse the data into a document
      InputStream responseStream = null;
      try {
        responseStream = getter.getResponseBodyAsStream();
        System.out.println("PLUGIN_MGR:: Downloaded Vendor Plugin from " + vendorURL);
        return new VendorPlugin(responseStream);
      } finally {
        if (responseStream != null) responseStream.close();
      }
    } catch (Exception anyError) {
      System.err.println("Error loading plugin for vendor " + vendorURL + " -- " + anyError.getMessage());
      anyError.printStackTrace();
      return null;
    } finally {
      getter.releaseConnection();
    }
  }
  
}
