/*
 * VendorPlugin.java
 *
 * Created on September 28, 2005, 6:01 PM
 *
 */

package org.cdp1802.xpl.plugins;

import java.io.*;
import java.util.*;

import org.jdom.*;
import org.jdom.input.SAXBuilder;
import org.jdom.output.Format;
import org.jdom.output.XMLOutputter;

/**
 * Represents a plugin file for a vendor
 *
 * @author gerry
 */
public class VendorPlugin extends BasePluginItem {
  static XMLOutputter xmlOutputter = null;
  static SAXBuilder xmlParser = null;

  Document pluginDocument = null;
  List deviceList = null;
  String vendorName = null;
  ArrayList<PluginDevice> pluginDevices = new ArrayList<PluginDevice>();
  File pluginFile = null;
  
  boolean isNewPlugin = false;
  boolean pluginsModified = false;
  boolean vendorNameMismatchFound = false;
  boolean multipleVendorsFound = false;
  
  static {
    xmlParser = new SAXBuilder();
    xmlOutputter = new XMLOutputter();
    xmlOutputter.setFormat(Format.getPrettyFormat());
  }

  /**
   * Construct a default vendor plugin
   */
  public VendorPlugin() {
    initPlugin();
    createNewDocument();

    // Setup vendor name
    this.vendorName = "default";
    this.pluginFile = new File(this.vendorName + ".xml");
  }
  
  /**
   * Construct a new vendor plugin with the given vendor
   * name.  A new XML document will automatically
   * be created.  A default plugin file will be created from the
   * name.
   *
   * @param vendorName name to associated with this set of plugins
   */
  public VendorPlugin(String vendorName) {
    initPlugin();
    createNewDocument();

    // Setup vendor name
    this.vendorName = vendorName.toLowerCase();
    if (this.vendorName.length() > 8) this.vendorName = this.vendorName.substring(0, 8);
    this.pluginFile = new File(this.vendorName + ".xml");
  }

  /**
   * Construct and read vendor plugin from the passed plugin file and
   * read the contents in.  If the passed file does not exist, it
   * is installed as the default and a new plugin document is
   * created.
   *
   * @param pluginFile file to read plugins from
   */
  public VendorPlugin(File pluginFile) throws Exception {
    initPlugin();

    // Install plugin file. 
    this.pluginFile = pluginFile;
    
    // If the file does not exist, act like a new file will be created
    if (!pluginFile.exists()) {
      createNewDocument();
      vendorName = parseVendorNameFromFile(pluginFile);
      return;
    }

    // Parse the plugin
    readPluginFromFile(pluginFile);
  }

  /**
   * Construct and read vendor plugin from the passed stream and
   * read the contents in.  
   *
   * @param pluginStream stream to read plugins from
   */
  public VendorPlugin(InputStream pluginStream) throws Exception {
    initPlugin();
    readPluginFromStream(pluginStream);
  }
  
  
  // Put plugin into "just created" state
  private void initPlugin() {
    pluginDocument = null;
    nodeElement = null;
    deviceList = null;
    
    if (pluginDevices != null) {
      for (PluginDevice theDevice: pluginDevices) {
        theDevice.releaseResources();
      }
      pluginDevices.clear();
    } else {
      pluginDevices = new ArrayList<PluginDevice>();
    }
  
    isNewPlugin = false;
    pluginsModified = false;
    vendorNameMismatchFound = false;
    multipleVendorsFound = false;
  }
  
  private void installDocument(Document theDocument) {
    // Initialize and install new document
    initPlugin();
    
    // Install the document
    pluginDocument = theDocument;
    nodeElement = pluginDocument.getRootElement();
    deviceList = nodeElement.getChildren("device");
    
    // If we do not have a vendor yet, parse it from the file
    if ((vendorName == null) || (vendorName.length() == 0)) {
      vendorName = parseVendorNameFromFile(pluginFile);
    }
    
    // Proof the plugin file to insure that 1) the vendor in the file matches the
    // files default vendor named (derived from the file name) and 2) that there
    // is only one vendor in the file
    boolean vendorNameConfirmed = false;
    for (Object anObject: deviceList) {
      Element anElement = (Element) anObject;

      String theVendor = anElement.getAttributeValue("id");
      if (theVendor != null) {
        theVendor = theVendor.substring(0, theVendor.indexOf('-')).toLowerCase();
        
        // If the vendor name of the device matches our current vendor name (if any), skip on to the next one
        if ((vendorName != null) && theVendor.equals(vendorName)) {
          vendorNameConfirmed = true;
          continue;
        }

        // They do not match -- if this is the first time, just reset name
        if (!vendorNameConfirmed) {
          // If there was a vendor before, then this is a mismatch
          vendorNameMismatchFound = ((vendorName != null) && (vendorName.length() != 0));
          
          // Set the name and mark it confirmed
          vendorNameConfirmed = true;
          vendorName = theVendor;
          continue;
        }

        // If the name is confirmed but we found a different name, then there
        // are multiple vendors in file
        if (vendorNameConfirmed) {
          multipleVendorsFound = true;
          continue;
        }
      }
    }

    // Copy existing devices in
    if (deviceList.size() != 0) {
      Element theDevice = null;
      String deviceIdent = null;
      String vendorPrefix = vendorName + "-";
      PluginDevice newDevice = null;
      
      try {
        suppressEvents = true;

        for (Object anObject: deviceList) {
          theDevice = (Element) anObject;

          // Skip null, no id attr or mismatched vendor
          if (theDevice == null) continue;
          if ((deviceIdent = theDevice.getAttributeValue("id")) == null) continue;
          if (!deviceIdent.startsWith(vendorPrefix)) continue;

          // Add device
          if (!addDevice(newDevice = new PluginDevice(this, theDevice))) continue;
          newDevice.clearModified();
        }
      } finally {
        suppressEvents = false;
      }
    }
    
    // Migrate root node, if needed
    if (!nodeElement.getName().equalsIgnoreCase("xpl-plugin")) {
      nodeElement.setName("xpl-plugin");
    }

    // Create a file, if needed
    if (pluginFile == null) {
      pluginFile = new File(vendorName + ".xml");
    }
    
    
    // Set flags 
    isNewPlugin = false;
    pluginsModified = false;
  }
  
  private void createNewDocument() {
    this.nodeElement = new Element("xpl-plugin");
    this.pluginDocument = new Document(nodeElement);
    this.deviceList = nodeElement.getChildren("device");
    isNewPlugin = true;
  }

  // Parse the vendor name from the passed pluginfile
  // If there is no plugin file, the vendorName is not changed
  private String parseVendorNameFromFile(File fromFile) {
    if (fromFile == null) return null;
    
    // Parse out the vendor name
    String newVendorName = fromFile.getName().toLowerCase();
    if (newVendorName.endsWith(".xml")) newVendorName = newVendorName.substring(0, newVendorName.indexOf(".xml"));
    
    // Truncate if vendor name is too long
    if (newVendorName.length() > 8) newVendorName = newVendorName.substring(0, 8);

    // And return it
    return newVendorName;
  }
  
  /**
   * Read in plugins from the passed plugin file.  If the file cannot
   * be parsed, no change in this instance is made.  If the file is parsed, any
   * previous settings for this instance are reset and the new ones installed.
   */
  public void readPluginFromFile(File fromFile) throws Exception {
    // Basic sanity checks
    if (fromFile == null) throw new Exception("No vendor plugin file available to be parsed");
    if (!fromFile.exists() || !fromFile.canRead()) throw new Exception("Unable to access vendor plugin file");
    
    // Attempt to parse this
    Document newDocument = null;
    try {
      newDocument = xmlParser.build(fromFile);
    } catch (Exception anyError) {
      throw new Exception("Unable to read vendors plugin file --" + anyError.getMessage(), anyError);
    }

    installDocument(newDocument);
  }
  
  /**
   * Read in plugins from the currently installed plugin file.  If the file cannot
   * be parsed, no change in this instance is made.  If the file is parsed, any
   * previous settings for this instance are reset and the new ones installed.
   */
  public void readPluginFromFile() throws Exception {
    readPluginFromFile(pluginFile);
  }

  /**
   * Read in plugins from the passed stream.  If the stream is parsed, any
   * previous settings for this instance are reset and the new ones installed.
   */
  public void readPluginFromStream(InputStream fromStream) throws Exception {
    // Basic sanity checks
    if (fromStream == null) throw new Exception("No vendor plugin steram available to be parsed");
    
    // Attempt to parse this
    Document newDocument = null;
    try {
      newDocument = xmlParser.build(fromStream);
    } catch (Exception anyError) {
      throw new Exception("Unable to read plugin stream --" + anyError.getMessage(), anyError);
    }

    installDocument(newDocument);
  }

  /**
   * Save the current plugin to the passed file.  If there is no 
   * plugin file, it cannot be written to or the plugin state is invalid,
   * an exception is thrown
   *
   * @throws Exception if there is no plugin file, cannot be written to or plugin in invalid state
   */
  public void savePluginToFile(File toFile) throws Exception {
    if ((pluginDocument == null) || (toFile == null)) throw new Exception("No document and/or vendor plugin file available");
    
    try {
      OutputStream outputStream = new FileOutputStream(toFile);
      BufferedOutputStream bufferedStream = new BufferedOutputStream(outputStream);
      
      xmlOutputter.output(pluginDocument, bufferedStream);

      bufferedStream.flush();
      bufferedStream.close();
      pluginsModified = false;
      
      for (PluginDevice theDevice: pluginDevices) {
        theDevice.clearModified();
        theDevice.clearNewDevice();
        theDevice.clearDiscoveredDevice();
      }
      
      isNewPlugin = false;
    } catch (IOException anyError) {
      throw new Exception("Unable to write plugin file --" + anyError.getMessage(), anyError);
    }
  }
  
  /**
   * Save the current plugin out to the current plugin file.  If there is no 
   * plugin file, it cannot be written to or the plugin state is invalid,
   * an exception is thrown
   *
   * @throws Exception if there is no plugin file, cannot be written to or plugin in invalid state
   */
  public void savePluginToFile() throws Exception {
    savePluginToFile(pluginFile);
  }
  
  /**
   * Return true if, after reading the plugin file, there was a vendor mis-match
   * between the installed vendor name and the vendor name found in the
   * plugin file.  The vendor name in this class has been updated to match
   * the contents of the plugin file.
   *
   * @return true if there was a change in vendor name during the last plugin file reading
   */
  public boolean foundVendorNameMismatch() { return vendorNameMismatchFound; }
  
  /**
   * Returns true if, in the process of reading the plugin file, multiple 
   * vendors were found in the same file.  Only vendors matching the current
   * vendorName are loaded.  The rest are left alone and will be kept in 
   * the plugin file, but cannot be accessed here.
   *
   * @return true if there were two (or more) vendors in the plugin file
   */
  public boolean foundMultipleVendors() { return multipleVendorsFound; }
  
  /**
   * Get the DOM document describing this plugin
   *
   * @return DOM document describing this vendor plugin
   */
  public Document getDocument() { return pluginDocument; }

  /**
   * Get the currently assigned vendor ID
   *
   * @return ID assigned as vendor to this plugin 
   */
  public String getVendorID() { return vendorName; }
  
  /**
   * Install a new vendor ID.  All existing devices are 
   * changed as well to have the new vendor ID.
   *
   * NOTE: If the ID is invalid (too short, too long or no
   * actual change) it's ignored.
   *
   * @param newVendorID new vendor ID to assign
   */
  public void setVendorID(String newVendorID) {
    newVendorID = newVendorID.toLowerCase();
    
    // Skip if no change
    if ((newVendorID == null) || (newVendorID.length() == 0) || (newVendorID.length() > 8) || newVendorID.equals(vendorName)) return;
    
    // Install it
    vendorName = newVendorID;
    
    // Visit each device and set it's vendor
    for (PluginDevice theDevice: pluginDevices) {
      theDevice.setVendorName(newVendorID);
    }
    
    markModified();
    fireEvent(PluginEvent.EventCode.PLUGIN_UPDATED);
  }

  
  public String getVendorName() {
    String theText = nodeElement.getAttributeValue("vendor");
    return (theText == null ? "" : theText);
  }
  
  public void setVendorName(String newValue) {
    String curValue = getVendorName();
    
    if ((newValue == null) || (newValue.length() == 0)) {
      if (curValue.length() != 0) {
        nodeElement.removeAttribute("vendor");
        markModified();
        fireEvent(PluginEvent.EventCode.PLUGIN_UPDATED);
      }
      return;
    }
    
    if (curValue.equals(newValue)) return;
    nodeElement.setAttribute("vendor", newValue);
    markModified();
    fireEvent(PluginEvent.EventCode.PLUGIN_UPDATED);
  }
  
  public String getPluginVersion() {
    String theText = nodeElement.getAttributeValue("version");
    return (theText == null ? "" : theText);
  }
  
  public void setPluginVersion(String newValue) {
    String curValue = getPluginVersion();
    
    if ((newValue == null) || (newValue.length() == 0)) {
      if (curValue.length() != 0) {
        nodeElement.removeAttribute("version");
        markModified();
        fireEvent(PluginEvent.EventCode.PLUGIN_UPDATED);
      }
      return;
    }
    
    if (curValue.equals(newValue)) return;
    nodeElement.setAttribute("version", newValue);
    markModified();
    fireEvent(PluginEvent.EventCode.PLUGIN_UPDATED);
  }
  
  public String getPluginURL() {
    String theText = nodeElement.getAttributeValue("plugin_url");
    return (theText == null ? "" : theText);
  }
  
  public void setPluginURL(String newValue) {
    String curValue = getPluginURL();
    
    if ((newValue == null) || (newValue.length() == 0)) {
      if (curValue.length() != 0) {
        nodeElement.removeAttribute("plugin_url");
        markModified();
        fireEvent(PluginEvent.EventCode.PLUGIN_UPDATED);
      }
      return;
    }
    
    if (curValue.equals(newValue)) return;
    nodeElement.setAttribute("plugin_url", newValue);
    markModified();
    fireEvent(PluginEvent.EventCode.PLUGIN_UPDATED);
  }
  
  public String getInfoURL() {
    String theText = nodeElement.getAttributeValue("info_url");
    return (theText == null ? "" : theText);
  }
  
  public void setInfoURL(String newURL) {
    String curValue = getInfoURL();
    
    if ((newURL == null) || (newURL.length() == 0)) {
      if (curValue.length() != 0) {
        nodeElement.removeAttribute("info_url");
        markModified();
        fireEvent(PluginEvent.EventCode.PLUGIN_UPDATED);
      }
      return;
    }
    
    if (curValue.equals(newURL)) return;
    nodeElement.setAttribute("info_url", newURL);
    markModified();
    fireEvent(PluginEvent.EventCode.PLUGIN_UPDATED);
  }
  
  /**
   * Get the file the plugins XML is to be stored in
   *
   * @return file describing the plugins XML file
   */
  public File getPluginsFile() { return pluginFile; }
  
  /**
   * Set the file used to read/store the XML for 
   * this plugin
   *
   * @param newFile new File to use for storing/reading this plugin
   */
  public void setPluginsFile(File newFile) {
    if ((newFile == null) || ((pluginFile != null) && newFile.equals(pluginFile))) return;
    pluginFile = newFile;
  }
  
  /**
   * See if this set of plugins has been changed since it was 
   * last in sync with the disk XML file
   *
   * @return true if there have been any changes
   */
  public boolean isModified() { return pluginsModified; }

  // Used whenever a change is made to mark it
  protected void markModified() {
    pluginsModified = true;
  }
  
  /**
   * Return true if this plugin is new (not read from a file
   * and never saved).  Once saved, the plugin will never be
   * considered new and plugins read from a plugin file are
   * never considered new either.
   *
   * @return true if new, unsaved plugin
   */
  public boolean isNewPlugin() { return isNewPlugin; }

  /**
   * Get number of devices currently known
   *
   * @return number of known vendor devices
   */
  public int getDeviceCount() { return pluginDevices.size(); }

  /**
   * Get a vendor at the given index.  If the index is
   * invalid, return null
   *
   * @param deviceIndex index of devie to return
   * @return device or null if index is not valid
   */
  public PluginDevice getDeviceAt(int deviceIndex) {
    if ((deviceIndex < 0) || (deviceIndex >= deviceList.size())) return null;
    return pluginDevices.get(deviceIndex);
  }

  /**
   * Get the index of the passed device or -1 if the device
   * is not valid or not in this plugin
   *
   * @param theDevice device to get index of
   * @return index of device in this plugin or -1
   */
  public int getIndexOfDevice(PluginDevice theDevice) {
    return pluginDevices.indexOf(theDevice);
  }
  
  /**
   * Look for a device of the given name and return it or null
   * if not found.
   *
   * NOTE: only pass the deviceID portion of the device -- do
   * not include the vendor name
   *
   * @param deviceName name of the device to look for
   * @return the device matching the name or null if nothing matched
   */
  public PluginDevice getDeviceNamed(String deviceName) {
    String deviceIdent = vendorName + "-" + deviceName.toLowerCase();
    for (PluginDevice aDevice: pluginDevices) {
      if (aDevice.getID().equals(deviceIdent)) return aDevice;
    }
    
    return null;
  }

  /**
   * Creates a new plugin device.  The device is not added to the current plug
   * in -- you need to use addDevice() for that.  If the passed name is 
   * invalid (too short/too long) or there is a device with this name
   * already, null is returned
   *
   * NOTE: only pass the deviceID portion of the device -- do
   * not include the vendor name
   *
   * @param deviceName name of the device to create or null to not set a device name
   * @param flagDiscovered mark new device as discovered or not
   * @return instance of new device or null
   */
  public PluginDevice createDevice(String deviceName, boolean flagDiscovered) {
    // Insure vendor name is valid 
    if ((deviceName != null) && ((deviceName.length() == 0) || (deviceName.length() > 8))) return null;

    // Make sure it's not a dupe
    if (deviceName != null) {
      PluginDevice newDevice = getDeviceNamed(deviceName);
      if (newDevice != null) return null;
    }
    
    // Create a new entry 
    PluginDevice newDevice = new PluginDevice(this, (deviceName == null ? "" : deviceName));
    if (flagDiscovered) newDevice.markDiscoveredDevice();
    return newDevice;
  }
  
  /**
   * Creates a new plugin device.  The device is not added to the current plug
   * in -- you need to use addDevice() for that.  If the passed name is 
   * invalid (too short/too long) or there is a device with this name
   * already, null is returned
   *
   * NOTE: only pass the deviceID portion of the device -- do
   * not include the vendor name
   *
   * @param deviceName name of the device to create or null to not set a device name
   * @return instance of new device or null
   */
  public PluginDevice createDevice(String deviceName) {
    return createDevice(deviceName, false);
  }

  /**
   * Add a new device to the list of devices for this plugin.  If there is 
   * already a device with the same name or the passed device is 
   * invalid or it's from a different vendor, it's not added and false is returned.
   *
   * @param theDevice  device to add
   * @return true if device is unique and added in
   */
  public boolean addDevice(PluginDevice theDevice) {
    // Insure device is valid and not a dupe
    if ((theDevice == null) || (getDeviceNamed(theDevice.getDeviceName()) != null)) return false;
    
    // Insure the vendor matches
    if (!theDevice.getVendorName().equals(vendorName)) return false;
    
    // Find where to install the new item
    int deviceCount = getDeviceCount();
    String newDeviceName = theDevice.getDeviceName();
    int insertedAt = -1;
    for (int deviceIndex = 0; deviceIndex < deviceCount; deviceIndex++) {
      if (pluginDevices.get(deviceIndex).getDeviceName().compareTo(newDeviceName) < 0) continue;
      pluginDevices.add(deviceIndex, theDevice);
      insertedAt = deviceIndex;
      break;
    }
    
    // If not yet inserted, add it to the end
    if (insertedAt == -1) {
      insertedAt = pluginDevices.size();
      pluginDevices.add(theDevice);
    }

    // Add it in to tree
    if (!deviceList.contains(theDevice.getNodeElement())) deviceList.add((Object) theDevice.getNodeElement());

    theDevice.parent = this;
    
    markModified();
    fireEvent(theDevice, PluginEvent.EventCode.DEVICE_ADDED);
    return true;
  }

  /**
   * Given the passed index, remove the device at that index from the list
   * If the index is invalid, false is returned.  Otherwise the device
   * is removed and all resources for it are released.
   *
   * @param deviceIndex index of device to remove
   * @return true if device was removed and released
   */
  public boolean removeDeviceAt(int deviceIndex) {
    if ((deviceIndex < 0) || (deviceIndex >= deviceList.size())) return false;

    PluginDevice theVendorDevice = pluginDevices.get(deviceIndex);
    
    deviceList.remove(theVendorDevice.getNodeElement());
    pluginDevices.remove(deviceIndex);
    markModified();
    fireEvent(theVendorDevice, PluginEvent.EventCode.DEVICE_REMOVED, deviceIndex);

    // Release things
    theVendorDevice.releaseResources();
    return true;
  }

  public void copyInto(VendorPlugin theTarget) {
    boolean setCopiedFrom = (this.copiedFrom != theTarget);
    
    if (setCopiedFrom) theTarget.copiedFrom = this;
    
    if (theTarget.nodeElement == null) theTarget.nodeElement = new Element(nodeElement.getName());
    if (theTarget.pluginDocument == null) theTarget.pluginDocument = new Document(theTarget.nodeElement);
    if (theTarget.deviceList == null) theTarget.deviceList = theTarget.nodeElement.getChildren("device");
    
    // Copy attributes over
    theTarget.setVendorID(getVendorID());
    theTarget.setVendorName(getVendorName());
    theTarget.setPluginsFile(getPluginsFile());
    theTarget.setPluginURL(getPluginURL());
    theTarget.setPluginVersion(getPluginVersion());
    theTarget.setInfoURL(getInfoURL());
    
    // Add or update each device
    PluginDevice origDevice = null, targetDevice = null;
    for (int theIndex = 0; theIndex < getDeviceCount(); theIndex++) {
      origDevice = getDeviceAt(theIndex);
      
      if ((targetDevice = theTarget.getDeviceNamed(origDevice.getDeviceName())) == null) {
        targetDevice = (PluginDevice) origDevice.copyOf(theTarget);
        theTarget.addDevice(targetDevice);
      } else {
        origDevice.copyInto(targetDevice);
      }
    }
    
    // Purge obsolete devices
    for (int theIndex = theTarget.getDeviceCount() - 1; theIndex >= 0; theIndex--) {
      targetDevice = theTarget.getDeviceAt(theIndex);
      if (getDeviceNamed(targetDevice.getDeviceName()) != null) continue;
      theTarget.removeDeviceAt(theIndex);
    }

    // Copy status flags
    theTarget.isNewPlugin = this.isNewPlugin;
    theTarget.pluginsModified = this.pluginsModified;
    theTarget.vendorNameMismatchFound = this.vendorNameMismatchFound;
    theTarget.multipleVendorsFound = this.multipleVendorsFound;
  }
    
  public VendorPlugin copyOf() {
    VendorPlugin newCopy = new VendorPlugin();
    copyInto(newCopy);
    return newCopy;
  }

  /**
   * Release all resources for this plugin 
   */
  public void releaseResources() {
    super.releaseResources();

    pluginDocument = null;
    deviceList = null;
    vendorName = null;
    pluginFile = null;
    pluginsModified = false;
    
    if (pluginDevices != null) {
      for (PluginDevice theDevice: pluginDevices) {
        theDevice.releaseResources();
      }
      pluginDevices.clear();
      pluginDevices = null;
    }
  }
  
  public String toString() { return vendorName; }
}
