/*
 * LightingHelper.java
 *
 * Created on June 3, 2005, 11:20 PM
 */

package org.cdp1802.xpl.script;

import java.util.*;

import org.cdp1802.xpl.*;
import org.cdp1802.xpl.lighting.*;

/**
 *
 * @author gerry
 */
public class LightingHelper implements LightingGatewayListenerI, LightingDeviceListenerI, LightingSceneListenerI, Runnable {
  private static LightingGateway lightingGateway = null;
  private static String gatewaySemaphore = "SEMAPHORE";
  
  boolean helperEnabled = true;
  boolean reportOnce = true;
  HashMap<String, String> handlerTranslations = null;
  XPLScript theScript = null;
  
  LightingHelper(XPLScript theScript) {
    this.theScript = theScript;

    synchronized(gatewaySemaphore) {
      if (lightingGateway == null) {
        lightingGateway = new LightingGateway(XPLScriptEngine.getEngine().getModuleDevice());
      }
    }

    lightingGateway.addGatewayListener(this);
    
    // If we have not yet been discovered, start the discovery process
    if (!lightingGateway.isDiscoveryInProgress() && !lightingGateway.isDiscoveryInProgress()) {
      lightingGateway.startGatewayDiscovery();
    }

    checkGatewayStatus();
  }
  
  void releaseResources() {
    synchronized(gatewaySemaphore) {
      if (lightingGateway != null) {
        lightingGateway.removeAllDeviceListeners(this);
        lightingGateway.removeAllSceneListeners(this);
        lightingGateway.removeGatewayListener(this);
      }
    }
    
    if (handlerTranslations != null) handlerTranslations.clear();
    handlerTranslations = null;
    
    theScript = null;
  }

  void checkGatewayStatus() {
    // If discovery is done, then fire the read method
    if (lightingGateway.isDiscoveryComplete()) {
      if (helperEnabled) sendReport("theGateway", lightingGateway, new String[] { translate("GatewayReady") });
    }
  }
  
  void scriptReloaded() {
    checkGatewayStatus();
  }
  
  
  /**
   * This will not return until the gateway is ready (that is, discovery has
   * completed).  If the gateway has already done discovery, then it 
   * returns immediatly.  Otherwise, it'll pause until that happens
   */
  public void waitForGateway() {
    synchronized(gatewaySemaphore) {
      for(;;) {
        if (lightingGateway.isDiscoveryComplete()) return;
      
        try {
          gatewaySemaphore.wait();
        } catch (Exception anyError) { }
      }
    }
  }
  
  public void setEnabled(boolean isEnabled) {
    helperEnabled = isEnabled;
  }
  public boolean isEnabled() { return helperEnabled; }

  public void setReportOnce(boolean onlyOnce) {
    reportOnce = onlyOnce;
  }
  public boolean isReportOnce() { return reportOnce; }
  
  public LightingGateway getGateway() { return lightingGateway; }
  
  private String translate(String theMethodName) {
    if (handlerTranslations == null) return theMethodName;
    String translation = handlerTranslations.get(theMethodName);
    return (translation == null ? theMethodName : translation);
  }
  
  public void overrideHandler(String methodName, String translation) {
    if ((translation == null) && (handlerTranslations == null)) return;
    if (handlerTranslations == null) handlerTranslations = new HashMap<String, String>();
    if (translation == null)
      handlerTranslations.remove(methodName);
    else
      handlerTranslations.put(methodName, translation);
  }
  public void resetHandler(String methodName) {
    if (handlerTranslations == null) return;
    handlerTranslations.remove(methodName);
  }
  
  private void sendReport(String parameterNames[], Object theParameters[], String[] methodNames) {
    if ((parameterNames != null) != (theParameters != null)) return;
    
    StringBuffer theParameterList = new StringBuffer();
    if (parameterNames != null) {
      if (parameterNames.length != theParameters.length) return;
      
      for (int parmIndex = 0; parmIndex < parameterNames.length; parmIndex++) {
        theScript.setVar(parameterNames[parmIndex], theParameters[parmIndex]);
        
        if (theParameterList.length() != 0) theParameterList.append(", ");
        theParameterList.append(parameterNames[parmIndex]);
      }
    }
    
    for (String aMethod: methodNames) {
      if (theScript.hasMethod(aMethod)) {
        theScript.invokeMethod(aMethod, theParameterList.toString());
        if (reportOnce) break;
      }
    }
    
    if (parameterNames != null) {
      for (int parmIndex = 0; parmIndex < parameterNames.length; parmIndex++) {
        theScript.clearVar(parameterNames[parmIndex]);
      }
    }
  }

  private void sendReport(String parameterName, Object theParameter, String[] methodNames) {
    sendReport(new String[] { parameterName }, new Object[] { theParameter }, methodNames);
  }

  public void handleLightingGatewayEvent(LightingGatewayEvent theEvent) {
    switch(theEvent.getEventCode()) {
      case DISCOVERY_COMPLETE:
        lightingGateway.addDeviceListener(this);
        lightingGateway.addSceneListener(this);
 
        synchronized(gatewaySemaphore) {
          gatewaySemaphore.notifyAll();
        }
        if (helperEnabled) sendReport("theGateway", lightingGateway, new String[] { translate("GatewayReady") });
        
        break;
        
      case DISCOVERY_FAILED:
        // Restart discovery in 10 seconds
        LightingGateway.queueTask(this, "Lighting Gateway Discoverer");
        break;
    }
  }

  public void handleLightingDeviceEvent(LightingDeviceEvent theEvent) {
    if (!helperEnabled) return;

    if (theEvent.getEventCode() != LightingDeviceEvent.EventCode.DEVICE_STATE_CHANGED) return;
    
    String scriptName = translate("Light" + theEvent.getDevice().getDeviceID() + "Changed");
    String globalScriptName = translate("LightChanged");
    sendReport("theLight", theEvent.getDevice(), new String[] { scriptName, globalScriptName });
  }
  
  public void handleLightingSceneEvent(LightingSceneEvent theEvent) {
    if (!helperEnabled) return;
    
    if (theEvent.getEventCode() == LightingSceneEvent.EventCode.SCENE_ACTIVATED) {
      String scriptName = translate("Scene" + theEvent.getScene().getSceneID() + "Activated");
      String globalScriptName = translate("SceneActivated");
      sendReport(new String[] { "theScene", "theSource" }, new Object[] { theEvent.getScene(), theEvent.getSourceDevice() }, new String[] { scriptName, globalScriptName } );
      return;
    }
    
    if (theEvent.getEventCode() == LightingSceneEvent.EventCode.SCENE_DEACTIVATED) {
      String scriptName = translate("Scene" + theEvent.getScene().getSceneID() + "Deactivated");
      String globalScriptName = translate("SceneDeactivated");
      sendReport(new String[] { "theScene", "theSource" }, new Object[] { theEvent.getScene(), theEvent.getSourceDevice() }, new String[] { scriptName, globalScriptName } );
      return;
    }

    if (theEvent.getEventCode() == LightingSceneEvent.EventCode.SCENE_GOTO) {
      String scriptName = translate("Scene" + theEvent.getScene().getSceneID() + "Changed");
      String globalScriptName = translate("SceneChanged");
      sendReport(new String[] { "theScene", "theSource", "theLevel" }, new Object[] { theEvent.getScene(), theEvent.getSourceDevice(), LightingGateway.formatLevel(theEvent.getLevel()) }, new String[] { scriptName, globalScriptName } );
      return;
    }
  }
  
  public void run() {
    // Used to start the gateway discovery process after some time
    try {
      Thread.currentThread().sleep(10000);
    } catch (Exception anyError) { }
    
    if (!lightingGateway.isDiscoveryComplete() && !lightingGateway.isDiscoveryInProgress()) {
      lightingGateway.startGatewayDiscovery();
    }
  }
  
}
