/*
 * PeriodicTimer.java
 *
 * Created on June 1, 2005, 11:17 AM
 */

package org.cdp1802.xpl.script;

import bsh.EvalError;

/**
 * A timer that kicks over after a given interval
 *
 * @author gerry
 */
public class PeriodicTimer extends BaseTimer implements Runnable {
  public static final String DEFAULT_HANDLER = "handlePeriodicTimer";
  
  int timerInterval = 1;        // Interval in seconds
  boolean timerRepeats = false; // Is timer a one shot or repeating timer?
  int enableCount = 0;
  
  
  PeriodicTimer(XPLScript theScript) { 
    timerScript = theScript;
  }
  
  void releaseResources() {
    stopTimer();
    eventHandler = null;
    timerThread = null;
    timerScript = null;
  }

  void startTimer() {
    synchronized(timerScript) {
      enableCount++;
      if (timerRunning) {
        timerScript.notifyAll();
        return;
      }
      
      startIntTimer("Periodic Timer");
    }
  }
  
  void stopTimer() {
    stopIntTimer();
  }
  
  public boolean isEnabled() { return timerRunning; }
  public void setEnabled(boolean isEnabled) {
    if (timerScript == null)  return;
    
    if (isEnabled) 
      startTimer();
    else
      stopTimer();
  }

  public int getInterval() { return timerInterval; }
  public void setInterval(int newInterval) {
    if (timerRunning || (newInterval == timerInterval) || (newInterval <= 0)) return;
    timerInterval = newInterval;
  }
  
  public boolean isRepeating() { return timerRepeats; }
  public void setRepeating(boolean doesRepeat) {
    if (timerRunning || (timerRepeats == doesRepeat)) return;
    timerRepeats = doesRepeat;
  }
  
  public String getHandler() { return (eventHandler != null ? eventHandler : DEFAULT_HANDLER); }
  public void setHandler(String newHandler) {
    if (timerRunning) return;
    if ((newHandler != null) && (newHandler.length() == 0)) newHandler = null;
    eventHandler = newHandler;
  }
  
  public void run() {
    long startedAt = 0, sleepTime = 0;
    int lastEnableCount = 0;
    String theHandler = null;
    
    synchronized(timerScript) {
      // Flag us running
      timerThreadRunning = true;
      timerScript.notifyAll();
      
      // Now we actually runn
      mainLoop:
      while (timerRunning) {
        startedAt = System.currentTimeMillis();
        lastEnableCount = enableCount;
        
        while(timerRunning) {
          try {
            // Figure out how long to sleep
            sleepTime = (timerInterval - ((System.currentTimeMillis() - startedAt) / 1000)) * 1000;
            if (sleepTime <= 0) sleepTime = 100;
            
            // And sleep
            timerScript.wait(sleepTime);
          } catch (Exception ignoredException) { }
        
          // If we are no longer running, exit
          if (!timerRunning) break;
          
          // See if we've been restarted
          if (lastEnableCount != enableCount) continue mainLoop;
          
          // If we didn't sleep long enough, try again
          if (((System.currentTimeMillis() - startedAt) / 1000) < timerInterval) continue;
          
          // Install our timer
          timerScript.setVar("curTimer", this);
        
          // Invoke the handler
          theHandler = (eventHandler == null ?  DEFAULT_HANDLER : eventHandler);
          timerScript.invokeMethod(theHandler, "curTimer");

          // Clear things out
          timerScript.clearVar("curTimer");

          // And done
          break;
        }
        
        // If we are no longer running, exit
        if (!timerRunning) break;

        // See if this is a one shot or not
        if (!timerRepeats) break;
      }
      
      // Timer is no longer running
      timerThreadRunning = false;
      timerScript.notifyAll();
    }
  }
}
