/*
 * XPLScript.java
 *
 * Created on May 31, 2005, 12:37 PM
 */

package org.cdp1802.xpl.script;

import java.io.*;
import java.util.*;

import bsh.Interpreter;
import bsh.EvalError;

import org.cdp1802.xpl.*;
import org.cdp1802.xpl.device.*;

/**
 * Trackes one XPL Script
 * @author gerry
 */
public class XPLScript implements xPL_MessageListenerI, xPL_DeviceMessageListenerI {
  public static final Object ERROR_FLAG = new String("FAILED");   // Flag that an eval failed
  
  class DeviceInfo {
    xPL_DeviceI theDevice = null;
    String deviceHandler = null;
    
    DeviceInfo(xPL_DeviceI theDevice, String theHandler) {
      this.theDevice = theDevice;
      this.deviceHandler = theHandler;
    }
  }
  
  File scriptFile = null;
  long scriptLastChangedAt = 0;
  boolean scriptLoaded = false;
  boolean scriptHasErrors = false;
  Interpreter scriptInterpreter = new Interpreter();
  xPL_IdentifierI defaultSource = null;
  
  ArrayList<DeviceInfo> deviceList = null;
  ArrayList<MessageFilter> filterList = null;
  ArrayList<PeriodicTimer> periodicTimerList = null;
  ArrayList<ScheduledEvent> scheduledEventList = null;
  HVHelper hvHelper = null;
  LightingHelper lightingHelper = null;

  HashMap<String, String> dispatchedHandlerHistory = new HashMap<String, String>();
  String methodList[] = null;
  
  public XPLScript(String scriptFilename) {
    installScript(new File(scriptFilename));
  }
  
  public XPLScript(File scriptFile) {
    installScript(scriptFile);
  }
  
  public Interpreter getInterpreter() { return scriptInterpreter; }
  
  public File getScriptFile() { return scriptFile; }
  
  // True if a valid script file exists (even if it
  // may not be parsable).
  public boolean isScriptFileValid() {
    synchronized(this) {
      if (scriptFile == null) return false;
      if (!scriptFile.exists()) return false;
      if (!scriptFile.canRead()) return false;
      return true;
    }
  }

  // True if the script has been loaded and parsed without
  // any errors.
  public boolean isScriptValid() {
    synchronized(this) {
      if (!scriptLoaded) return false;
      if (scriptHasErrors) return false;
      return true;
    }
  }
  
  // True if the script as last parsed matches the
  // script on disk.  If the disk version has changed, then
  // this return false.
  public boolean isScriptUpToDate() {
  synchronized(this) {
      if (!isScriptFileValid()) return false;
      return (scriptFile.lastModified() == scriptLastChangedAt);
    }
  }
  
  // Install a script file.  Script files are 
  // only installed once ever per instance
  private void installScript(File theScriptFile) {
    scriptFile = theScriptFile;
  }
  
  private void clearInterpreter() {
    if (scriptInterpreter == null) return;
    
    // Give the script a chance to shutdown
    invokeOptionalMethod("cleanup");
    
    // Clear methods
    methodList = null;
    
    // Dump any devices we created
    if (deviceList != null) {
      for (DeviceInfo deviceInfo: deviceList) {
        deviceInfo.theDevice.removeMessageListener(this);
        deviceInfo.theDevice.setEnabled(false);
        deviceInfo.theDevice.doAutoPreserveConfig();
        DeviceManager.getManager().releaseDeviceResources(deviceInfo.theDevice);
        
        deviceInfo.deviceHandler = null;
        deviceInfo.theDevice = null;
      }
      
      deviceList.clear();
    }

    // Clear the filters
    if (filterList != null) {
      for (MessageFilter messageFilter: filterList) {
        messageFilter.releaseResources();
      }
      
      filterList.clear();
    }
    
    // Clear any periodic timers
    if (periodicTimerList != null) {
      for (PeriodicTimer periodicTimer: periodicTimerList) {
        periodicTimer.releaseResources();
      }
      
      periodicTimerList.clear();
    }
    
    // Clear any scheduled events
    if (scheduledEventList != null) {
      for (ScheduledEvent scheduledEvent: scheduledEventList) {
        scheduledEvent.releaseResources();
      }
      
      scheduledEventList.clear();
    }
    
    // Remove any message listeners we created
    xPL_Manager.getManager().removeMessageListener(this);
    
    // Release the interpreter space
    scriptInterpreter.getNameSpace().clear();
  }

  // Release all resources for this object.  After this, the
  // object cannot be used or reused
  public void releaseResources() {
    synchronized(this) {
      scriptFile = null;
      scriptLastChangedAt = 0;
      scriptLoaded = false;
      scriptHasErrors = false;

      defaultSource = null;

      if (hvHelper != null) hvHelper.releaseResources();
      hvHelper = null;
      
      if (lightingHelper != null) lightingHelper.releaseResources();
      lightingHelper = null;
      
      if (scriptInterpreter != null) clearInterpreter();
      scriptInterpreter = null;

      deviceList = null;
      filterList = null;
      periodicTimerList = null;
      scheduledEventList = null;
      methodList = null;
    }
  }
  
  public boolean reloadScript() {
    synchronized(this) {
      // See if this is loadable
      if (!isScriptFileValid()) {
        clearInterpreter();
        System.err.println("Script file not valid -- not loading");
        return false;
      }

      // Interpret it
      clearInterpreter();

      // Clear flags
      scriptLoaded = false;
      scriptHasErrors = false;

      // setup script environment
      initializeScriptContext(scriptInterpreter);

      // Source script
      try {
        scriptInterpreter.source(scriptFile.getAbsolutePath());
        methodList = scriptInterpreter.getNameSpace().getMethodNames();
        scriptInterpreter.eval("init()");
        scriptLoaded = true;
        scriptLastChangedAt = scriptFile.lastModified();
        if (hvHelper != null) hvHelper.scriptReloaded();
        if (lightingHelper != null) lightingHelper.scriptReloaded();
        
        System.out.println("SCRIPT:: (Re)Loaded scipt file " + scriptFile.getName());
        return true;
      } catch (FileNotFoundException ioError) {
        clearInterpreter();
        scriptLoaded = false;
        System.err.println("Unable to read script (no file) -- " + ioError.getMessage());
        return false;
      } catch (IOException ioError) {
        clearInterpreter();
        System.err.println("Unable to read script -- (io error)" + ioError.getMessage());
        scriptLoaded = false;
        return false;
      } catch (EvalError evalError) {
        System.err.println("Unable to read script (eval error) -- " + evalError.getMessage());
        clearInterpreter();
        evalError.printStackTrace();
        scriptLoaded = true;
        scriptHasErrors = true;
        return false;
      }
    }
  }

  public void setDefaultSource(xPL_IdentifierI theIdentifier) {
    defaultSource = theIdentifier;
  }
  public xPL_IdentifierI getDefaultSource() { return defaultSource; }
  
  public Object eval(String theCommand) {
    synchronized(this) {
      if (!isScriptValid()) return ERROR_FLAG;

      try {
        return scriptInterpreter.eval(theCommand);
      } catch (EvalError evalError) { 
        System.err.println("Error executing [" + theCommand + "] -- " + evalError.getMessage());
        evalError.printStackTrace();
        return ERROR_FLAG;
      }
    }
  }

  // Version for quiet failure
  public Object evalOptional(String theCommand) {
    synchronized(this) {
      if (!isScriptValid()) return ERROR_FLAG;

      try {
        return scriptInterpreter.eval(theCommand);
      } catch (EvalError evalError) { 
        return ERROR_FLAG;
      }
    }
  }
  
  public boolean hasMethod(String theMethod) {
    if (methodList == null) return false;
    for (String aMethod: methodList) {
      if (aMethod.equals(theMethod)) return true;
    }
    return false;
  }
  
  public Object invokeMethod(String methodName) {
    if (!hasMethod(methodName)) {
      System.out.println("Attempt to execute method " + methodName + " not found in script");
      return ERROR_FLAG;
    }
    
    return eval(methodName + "()");
  }
  
  public Object invokeMethod(String methodName, String parameters) {
    if (!hasMethod(methodName)) {
      System.out.println("Attempt to execute method " + methodName + " not found in script");
      return ERROR_FLAG;
    }

    return eval(methodName + "(" + parameters + ")");
  }

  public Object invokeOptionalMethod(String methodName) {
    if (!hasMethod(methodName)) return ERROR_FLAG;
    return eval(methodName + "()");
  }
  
  public Object invokeOptionalMethod(String methodName, String parameters) {
    if (!hasMethod(methodName)) return ERROR_FLAG;
    return eval(methodName + "(" + parameters + ")");
  }
  
  public boolean setVar(String varName, Object theValue) {
    // Set the current message in place
    try {
      scriptInterpreter.set(varName, theValue);
      return true;
    } catch (EvalError setError) { 
      System.err.println("Error installing value for variable " + varName + " -- " + setError.getMessage());
      return false; 
    }
  }
  public void clearVar(String varName) {
    setVar(varName, null);
  }
  
  private void initializeScriptContext(Interpreter theScript) {
    theScript.getNameSpace().importCommands(XPLScriptManager.BUILTIN_COMMAND_PACKAGE);
    theScript.setErr(System.err);
    theScript.setOut(System.out);
  }

  public HVHelper addHVHelper() {
    if (hvHelper != null) return hvHelper;
    return (hvHelper = new HVHelper(this));
  }
  
  public LightingHelper addLightingHelper() {
    if (lightingHelper != null) return lightingHelper;
    return (lightingHelper = new LightingHelper(this));
  }
  
  public MessageFilter addMessageFilter(String theFilter, String namedValues[][], String theHandler) {
    synchronized(this) {
      // Create the filter info
      if ((theFilter == null) || (theFilter.length() == 0)) theFilter = "*";
      if ((theHandler != null) && (theHandler.length() == 0)) theHandler = null;
      MessageFilter newFilter = new MessageFilter(this);
      
      // Install basics
      newFilter.setFilter(theFilter);
      newFilter.setHandler(theHandler);
      
      // Install named values, if any
      if (namedValues != null) {
        for (int nvIndex = 0; nvIndex < namedValues.length; nvIndex++) {
          newFilter.addNamedValueMatch(namedValues[nvIndex][0], namedValues[nvIndex][1]);
        }
      }

      // Install it
      if (filterList == null) filterList = new ArrayList<MessageFilter>();
      filterList.add(newFilter);

      // Make sure we are listening
      if (filterList.size() == 1) xPL_Manager.getManager().addMessageListener(this);
      
      // Enable the filter
      newFilter.setEnabled(true);
      return newFilter;
    }
  }

  public xPL_DeviceI addNewDevice(String deviceName, String theHandler) {
    synchronized(this) {
      xPL_IdentifierI theDeviceIdent = xPL_Manager.getManager().getIdentifierManager().parseNamedIdentifier(deviceName);
      if (theDeviceIdent == null) return null;

      xPL_DeviceI newDevice = xPL_Manager.getManager().getDeviceManager().createDevice(theDeviceIdent);
      if (newDevice == null) return null;

      if (deviceList == null) deviceList = new ArrayList<DeviceInfo>();
      if ((theHandler != null) && (theHandler.length() == 0)) theHandler = null;
      deviceList.add(new DeviceInfo(newDevice, theHandler));
      newDevice.addMessageListener(this);

      return newDevice;
    }
  }
  
  public PeriodicTimer addNewPeriodicTimer(int period, boolean repeating, String handler) {
    synchronized(this) {
      PeriodicTimer newTimer = new PeriodicTimer(this);
      newTimer.setInterval(period);
      newTimer.setRepeating(repeating);
      newTimer.setHandler(handler);

      if (periodicTimerList == null) periodicTimerList = new ArrayList<PeriodicTimer>();
      periodicTimerList.add(newTimer);
      return newTimer;
    }
  }
  
  public ScheduledEvent addNewScheduledEvent(String timeOfDay, String daysOfWeek, String handler) {
    synchronized(this) {
      ScheduledEvent newEvent = new ScheduledEvent(this);
      newEvent.setTimeOfDay(timeOfDay);
      newEvent.setDaysOfWeek(daysOfWeek);
      newEvent.setHandler(handler);

      if (scheduledEventList == null) scheduledEventList = new ArrayList<ScheduledEvent>();
      scheduledEventList.add(newEvent);
      return newEvent;
    }
  }
  
  public void handleXPLMessage(xPL_MessageI theMessage) {
    synchronized(this) {
      if (!XPLScriptManager.isScriptingEnabled() || (filterList == null) || (filterList.size() == 0)) return;

      // Set the current message in place
      setVar("curMessage", theMessage);

      // Displatch the message, making sure no handler sees the message more than once
      String theHandler = null;
      dispatchedHandlerHistory.clear();
      for (MessageFilter theFilter: filterList) {
        if (!theFilter.doesMessageMatchFilter(theMessage)) continue;

        // See if this handler already received this message
        theHandler = theFilter.getHandler();
        if (dispatchedHandlerHistory.containsKey(theHandler)) continue;

        // Dispatch the message and remember that we have used this handler already
        invokeMethod(theHandler, "curMessage");
        dispatchedHandlerHistory.put(theHandler, theHandler);
      }
      dispatchedHandlerHistory.clear();

      // And we are done
      clearVar("curMessage");
    }
  }
  
  public void handleXPLDeviceMessage(xPL_DeviceI theDevice, xPL_MessageI theMessage) {
    synchronized(this) {
      if (!XPLScriptManager.isScriptingEnabled() || (deviceList == null) || (deviceList.size() == 0)) return;

      // Set the current message & device in place
      setVar("curMessage", theMessage);
      setVar("curDevice", theDevice);

      // Displatch the message, making sure no handler sees the message more than once
      String theHandler = null;
      HashMap<String, String> handlerHistory = new HashMap<String, String>();
      for (DeviceInfo deviceInfo: deviceList) {
        if (deviceInfo.theDevice != theDevice) continue;

        theHandler = (deviceInfo.deviceHandler == null ?  "handleXPLDeviceMessage" : deviceInfo.deviceHandler);
        if (handlerHistory.containsKey(theHandler)) continue;
        invokeMethod(theHandler, "curDevice, curMessage");
        handlerHistory.put(theHandler, theHandler);
      }

      // And we are done
      clearVar("curMessage");
      clearVar("curDevice");
    }
  }
  
  public String toString() {
    if (scriptFile != null) return scriptFile.getName();
    return "Empty Script (" + super.toString() + ")";
  }
  
  
}
