/*
 * XPLScriptDirectory.java
 *
 * Created on May 31, 2005, 1:59 PM
 */

package org.cdp1802.xpl.script;

import java.io.*;
import java.util.*;

import bsh.Interpreter;

import org.cdp1802.xpl.*;

/**
 *
 * @author gerry
 */
public class XPLScriptDirectory {
  private static ScriptFileFilter scriptFilter = new ScriptFileFilter();
  
  File scriptDirectory = null;
  long directoryLastChangedAt = 0;
  HashMap<File, XPLScript> scriptList = new HashMap<File, XPLScript>();
  HashMap<Interpreter, XPLScript> scriptLookup = new HashMap<Interpreter, XPLScript>();

  public XPLScriptDirectory(String scriptDirectoryName) {
    initDirectory(new File(scriptDirectoryName));
  }
  
  public XPLScriptDirectory(File scriptDirectory) {
    initDirectory(scriptDirectory);
  }
  
  private void initDirectory(File theDirectory) {
    if (theDirectory == null) return;
    if (!theDirectory.exists()) return;
    if (!theDirectory.isDirectory()) return;
    if (!theDirectory.canRead()) return;
    
    scriptDirectory = theDirectory;
  }
  
  public File getDirectory() { return scriptDirectory; }
  
  public boolean isDirectoryValid() { 
    synchronized(this) {
      if (scriptDirectory == null) return false;
      if (!scriptDirectory.exists()) return false;
      if (!scriptDirectory.isDirectory()) return false;
      if (!scriptDirectory.canRead()) return false;
      return true;
    }
  }
  
  public boolean isDirectoryUpToDate() {
    synchronized(this) {
      if (!isDirectoryValid()) return false;
      return (directoryLastChangedAt == scriptDirectory.lastModified());
    }
  }
  
  public void releaseResources() {
    synchronized(this) {
      scriptDirectory = null;
      directoryLastChangedAt = 0;

      if (scriptList != null) {
        for (XPLScript theScript: scriptList.values()) {
          theScript.releaseResources();
        }
        scriptList.clear();
        scriptList = null;
      }
    }
  }
  
  private void createScript(File scriptFile) {
    XPLScript theScript = new XPLScript(scriptFile);
    scriptList.put(scriptFile, theScript);
    scriptLookup.put(theScript.getInterpreter(), theScript);
  }
  
  private void releaseScript(XPLScript theScript) {
    scriptLookup.remove(theScript.getInterpreter());
    theScript.releaseResources();
  }
  
  public boolean refreshDirectory(boolean forceUpdate) {    
    synchronized(this) {
      // Skip if not valid
      if (!isDirectoryValid()) return false;
      if (!forceUpdate && isDirectoryUpToDate()) return true;

      // Get time stamp
      directoryLastChangedAt = scriptDirectory.lastModified();

      // Get list of files
      File dirFileList[] = scriptDirectory.listFiles(scriptFilter);
      
      // If the directory is empty, then release any current scripts
      if ((dirFileList == null) || (dirFileList.length == 0)) {
        // Free up anything in the cache
        for (XPLScript theScript: scriptList.values()) {
          releaseScript(theScript);
        }
        scriptList.clear();
        scriptLookup.clear();
        
        return true;
      }

      // Check each script file to insure it's in the hashmap
      for (File scriptFile: dirFileList) {
        // See if the file already exists
        if (scriptList.containsKey(scriptFile)) continue;

        // If not, add it in
        createScript(scriptFile);
      }

      // Now review each script entry in our map and insure it
      // still exists on disk.  If so, check it for refresh.  If
      // not, delete it
      boolean scriptFound = false;
      XPLScript theScript = null;
      Iterator<XPLScript> theScripts = scriptList.values().iterator();
      while(theScripts.hasNext()) {
        theScript = theScripts.next();
        
        scriptFound = false;
        for (File theFile: dirFileList) {
          if (theScript.getScriptFile().equals(theFile)) {
            scriptFound = true;
            if (!theScript.isScriptUpToDate()) theScript.reloadScript();
            break;
          }
        }

        // If the script was not found, it not longer exists
        if (!scriptFound) {
          releaseScript(theScript);
          theScripts.remove();
        }
      }

      // Directory is now up to date
      return true;
    }
  }
  public boolean refreshDirectory() { return refreshDirectory(false); }
  
  public XPLScript getScriptByInterpreter(Interpreter theInterpreter) {
    synchronized(this) {
      return scriptLookup.get(theInterpreter);
    }
  }
}
