/*
 * XPLScriptManager.java
 *
 */

package org.cdp1802.xpl.script;

import java.io.*;
import java.util.*;

import bsh.Interpreter;

import org.cdp1802.xpl.*;

/**
 *
 * @author gerry
 */
public class XPLScriptManager implements Runnable {
  public static final int DEFAULT_REFRESH_INTERVAL = 10;
  public static final String BUILTIN_COMMAND_PACKAGE = "org.cdp1802.xpl.script.builtins";
  
  private static XPLScriptManager scriptManager = new XPLScriptManager();
  public static XPLScriptManager getManager() { return scriptManager; }
  public static boolean isScriptingEnabled() { return scriptManager.scriptEngineRunning; }
  
  HashMap<File, XPLScriptDirectory> directoryList = new HashMap<File, XPLScriptDirectory>();
  boolean scriptEngineRunning = false;
  Thread scriptUpdaterThread = null;
  int refreshInterval = DEFAULT_REFRESH_INTERVAL;
  
  private XPLScriptManager() {  }
  
  public void releaseResources() {
    stopScriptEngine();
    
    synchronized(directoryList) {
      // Clear all scripts and directories
      for (XPLScriptDirectory theDir: directoryList.values()) {
        theDir.releaseResources();
      }
      directoryList.clear();
    }
    
    scriptEngineRunning = false;
    scriptUpdaterThread = null;
  }
  
  public int getRefreshInterval() { return refreshInterval; }
  public void setRefreshInterval(int newRefreshInterval) {
    synchronized(this) {
      if (newRefreshInterval == refreshInterval) return;
      this.notifyAll();
    }
  }
  
  public XPLScriptDirectory getDirectory(File directory) {
    return directoryList.get(directory);
  }
  
  public XPLScriptDirectory getDirectory(String directoryName) {
    return getDirectory(new File(directoryName));
  }
  
  public Collection<XPLScriptDirectory> getAllDirectories() {
    if (directoryList == null) return null;
    return directoryList.values();
  }
  
  public boolean addDirectory(File theDirectory, boolean recursively) {
    // Insure this is a real directory
    if ((theDirectory == null) || !theDirectory.exists() || !theDirectory.isDirectory() || !theDirectory.canRead()) return false;

    synchronized(directoryList) {
      XPLScriptDirectory scriptDirectory = null;
      
      // Add us to the directory if not already there
      if (!directoryList.containsKey(theDirectory)) {
        scriptDirectory = new XPLScriptDirectory(theDirectory);
        directoryList.put(theDirectory, scriptDirectory);
      } else {
        if (!directoryList.get(theDirectory).isDirectoryUpToDate()) {
          scriptDirectory = directoryList.get(theDirectory);
        }
      }

      // Check on adding all sub directories too
      if (recursively) {
        File dirList[] = theDirectory.listFiles();
        if ((dirList != null) && (dirList.length != 0)) {
          for (int dirIndex = 0; dirIndex < dirList.length; dirIndex++) {
            if (!dirList[dirIndex].isDirectory()) continue;
            addDirectory(dirList[dirIndex], true);
          }
        }
      }
      
      // Refresh this directory
      if (scriptDirectory != null) scriptDirectory.refreshDirectory();
    }
    
    return true;
  }

  public boolean addDirectory(String theDirectoryName, boolean recursively) {
    return addDirectory(new File(theDirectoryName), recursively);
  }
  
  public boolean removeDirectory(XPLScriptDirectory theDirectory) {
    synchronized(directoryList) {
      // Remove from list
      directoryList.remove(theDirectory.getDirectory());
    
      // Release directory resources
      theDirectory.releaseResources();
      
      return true;
    }
  }
  public boolean removeDirectory(File directory) {
    synchronized(directoryList) {
      XPLScriptDirectory scriptDirectory = directoryList.get(directory);
      if (scriptDirectory == null) return false;
      return removeDirectory(scriptDirectory);
    }
  }
  public boolean removeDirectory(String directoryName) {
    return removeDirectory(new File(directoryName));
  }
  
  public void refreshScripts(boolean forceRefresh) {
    synchronized(directoryList) {
      for (XPLScriptDirectory theDirectory: directoryList.values()) {
        theDirectory.refreshDirectory(forceRefresh);
      }
    }
  }
  public void refreshScripts() { refreshScripts(false); }
  
  public XPLScript getScriptByInterpreter(Interpreter theInterpreter) {
    XPLScript theScript = null;
    for (XPLScriptDirectory theDirectory: directoryList.values()) {
      if ((theScript = theDirectory.getScriptByInterpreter(theInterpreter)) != null) {
        return theScript;
      }
    }
    
    return null;
  }
  
  public void startScriptEngine() {
    synchronized(this) {
      if (scriptEngineRunning) return;
      
      scriptEngineRunning = true;
      scriptUpdaterThread = new Thread(this, "Script Updator");
      scriptUpdaterThread.start();
    }
  }

  public void stopScriptEngine() {
    synchronized(this) {
      if (!scriptEngineRunning) return;
      scriptEngineRunning = false;
      this.notifyAll();
    }
    
    try {
      scriptUpdaterThread.join();
    } catch (Exception anything) { }
    
    scriptUpdaterThread = null;
  }

  public void run() {
    synchronized(this) {
      while(scriptEngineRunning) {
        // Refresh the scripts
        try {
          refreshScripts();
        } catch (Throwable anyError) { }

        // Sleep for a while
        try {
          this.wait(refreshInterval * 1000);
        } catch (Exception anyException) { }
      }
    }
  }
}
