/*
 * ModuleDefaultsParser.java
 * Copyright 2004, Gerald R Duprey Jr
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy of
 * the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0 
 *
 * Unless required by applicable law or agreed to in writing, software distributed
 * under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
 * CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License. 
 */

package org.cdp1802.xpl.server;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.PrintWriter;
import org.cdp1802.xpl.MutableNamedValuesI;
import org.cdp1802.xpl.NamedValueI;
import org.cdp1802.xpl.xPL_Util;

/**
 * This module takes care of reading a modules defaults
 *
 * @author gduprey
 */
class ModuleDefaultsParser {
  xPL_Server theContainer = xPL_Server.getServer();
  
  ModuleDefaultsParser(xPL_Server theContainer) {
    this.theContainer = theContainer;
  }

  void debug(String theMessage) {
    System.out.println("MOD_STORE:: " + theMessage);
  }
  
  MutableNamedValuesI parseNamedValues(BufferedReader from) {
    MutableNamedValuesI namedValues = xPL_Util.createMutableNamedValues();
    String theLine = null, theName = null, theValue = null;
    int delimPtr = -1;
    
    for(;;) {
      try {
        if ((theLine = from.readLine()) == null) break;
      } catch (IOException ioError) { return null; }

      // If this is an empty line, we're done
      if (theLine.length() == 0) break;
      
      // Find the delimiter -- skip the line if none
      if ((delimPtr = theLine.indexOf("=")) == -1) continue;
      theName = theLine.substring(0, delimPtr);
      theValue = theLine.substring(delimPtr + 1);
      
      // Validate the name
      if (theName.length() == 0) continue;
      
      // Parse out an install directory tag, if any
      if (theValue.indexOf("<INSTALLDIR>") != -1) {
        String installDir = xPL_Server.getServer().getInstallPath() + File.separator;
        StringBuffer workText = new StringBuffer(theValue);
        int tokenPtr = 0;
        
        while ((tokenPtr = workText.indexOf("<INSTALLDIR>")) != -1) {
          workText.replace(tokenPtr, tokenPtr + 12, installDir);
        }
        
        theValue = workText.toString();
      }
      
      // Add this in
      namedValues.addNamedValue(theName, theValue);
    }

    // Return the named values
    return namedValues;
  }
  
  // Parse the named/values for option/config/reconfig items, define
  // the configuration items in the module approrpriatly and remove
  // those items from the named values set.  If these values repesent
  // any changed, return true.  Otherwise, false.
  boolean parseConfigDefinitions(xPL_ModuleConfig theConfig, MutableNamedValuesI namedValues) {
    // Now we pick the config item definitions off
    String theName = null, theValue = null;
    boolean configChanged = false;
    MutableNamedValuesI defaultsList = theConfig.moduleDefaults;

    for(NamedValueI theNamedValue: namedValues.getAllNamedValues()) {
      theName = theNamedValue.getName();
      theValue = theNamedValue.getValue();
      
      // See if module is enabled
      if (theName.equalsIgnoreCase("ModuleEnabled")) {
        if (theConfig.isEnabled != theNamedValue.getBoolValue()) {
          theConfig.isEnabled = theNamedValue.getBoolValue();
          configChanged = true;
        }
        continue;
      }

      // Get the main class for this module
      if (theName.equalsIgnoreCase("ModuleClassName")) {
        if ((theConfig.moduleMainClass == null) || !theConfig.moduleMainClass.equals(theValue)) {
          theConfig.moduleMainClass = theValue;
          configChanged = true;
        }
        continue;
      }
        
      // Get the device Ident for this module
      if (theName.equalsIgnoreCase("ModuleDeviceIdent")) {
        if ((theConfig.moduleDeviceIdent == null) || !theConfig.moduleDeviceIdent.equals(theValue)) {
          theConfig.moduleDeviceIdent = theValue;
          configChanged = true;
        }
        continue;
      }
      
      // Get the version of this module
      if (theName.equalsIgnoreCase("ModuleVersion")) {
        if ((theConfig.moduleVersion == null) || !theConfig.moduleVersion.equals(theValue)) {
          theConfig.moduleVersion = theValue;
          configChanged = true;
        }
        continue;
      }

      // Get the device config file for this module
      if (theName.equalsIgnoreCase("ModuleDeviceConfigFilename")) {
        if ((theConfig.moduleDeviceConfigFilename == null) || !theConfig.moduleDeviceConfigFilename.equals(theValue)) {
          theConfig.moduleDeviceConfigFilename = theValue;
          configChanged = true;
        }
        continue;
      }

      // See if module is auto-load
      if (theName.equalsIgnoreCase("ModuleAutoLoad")) {
        if (theConfig.isAutoLoaded != theNamedValue.getBoolValue()) {
          theConfig.isAutoLoaded = theNamedValue.getBoolValue();
          configChanged = true;
        }
        continue;
      }

      // See if module is auto-started
      if (theName.equalsIgnoreCase("ModuleAutoStart")) {
        if (theConfig.isAutoStarted != theNamedValue.getBoolValue()) {
          theConfig.isAutoLoaded = theNamedValue.getBoolValue();
          configChanged = true;
        }
        continue;
      }
      
      // Add as new attr
      defaultsList.addNamedValue(theName, theValue);
      configChanged = true;
    }
     
    return configChanged;
  }
  
  /**
   * Read in a module configuration from the passed stream.  If 
   * "allowOverrides" is true and the module is already loaded, the
   * information is read into the module and the module is re-initialized.
   * If "allowOverrides" is false and the module already exists, it
   * is skipped
   *
   * @param fromStream stream to read module definitions from
   * @param allowOverrides true to update/replace existing modules,
   *                       false to skip existing modules
   * @return true if any changes to the config was found.
   */
  boolean readModuleDefaults(InputStream fromStream, boolean allowOverrides) {
    // Make sure we even have a chance
    if (fromStream == null) return false;
    
    BufferedReader moduleConfigReader = new BufferedReader(new InputStreamReader(fromStream));
    String moduleName = null;
    xPL_ModuleConfig theConfig = null;
    xPL_ModuleConfigI theConfigI = null;
    String theLine = null;
    boolean configChanged = false;
    boolean newModule = false;
    
    // Look for a header
    for(;;) {
      // Read a line, bailing out if we hit end of the stream
      try {
        if ((theLine = moduleConfigReader.readLine()) == null) break;
      } catch (IOException ioError) { 
        System.err.println("CONFIG:: Error reading module configuration -- " + ioError.getMessage());
        return false;
      }

      // See if this line is a header
      if (!theLine.startsWith("[")) continue;
      
      // Extract the module name
      if ((moduleName = theLine.substring(1, theLine.indexOf("]"))).length() == 0) continue;

      // See if we already have this module on file
      if ((theConfigI = theContainer.getNamedModuleConfig(moduleName)) == null) {
        theConfig = new xPL_ModuleConfig();
        theConfig.moduleName = moduleName;
        newModule = true;
      } else {
        // Skip module if it already exists and we're not allowing overrides
        if (!allowOverrides) continue;
        
        // Lookup the module
        newModule = false;
        if (theConfigI instanceof xPL_ModuleConfig) 
          theConfig = (xPL_ModuleConfig) theConfigI;
        else
          theConfig = null;
      }
      
      // Parse the named values
      MutableNamedValuesI namedValues = parseNamedValues(moduleConfigReader);
      if (namedValues == null) continue;
    
      // Parse out configuration item definitions
      if (theConfig != null) {
        if (parseConfigDefinitions(theConfig, namedValues)) {
          configChanged = true;          
        } 
      }

      // If this is a new module, install it
      if (newModule) {
        theContainer.moduleConfigs.add(theConfig);
        if (theContainer.managementModule != null) theContainer.managementModule.sendModuleStatus(theConfig, true);
      }
    }
    
    // And we are done
    return configChanged;
  }
  
  /**
   * Load the module configurations from the passed input stream. 
   *
   * @param fromStream stream to read state from
   * @return true if any changes to the config were found, false if not or there was an error
   */
  boolean restoreModuleConfigs(InputStream fromStream) {
    return readModuleDefaults(fromStream, true);
  }
  
  /**
   * Restore the passed module config from the passed input file
   *
   * @param fromFile file name to read from
   * @return true if any module config was added/changed, false if not
   */
  boolean restoreModuleConfigs(File fromFile) {
    // Make sure we have a chance
    if (!fromFile.exists() || !fromFile.canRead()) return false;
    
    FileInputStream fromStream = null;
    try {
      fromStream = new FileInputStream(fromFile);
      return restoreModuleConfigs(fromStream);
    } catch (Exception anyError) { 
      anyError.printStackTrace();
      return false;
    } finally {
      if (fromStream != null) try { fromStream.close(); } catch (IOException ignoredError) { }
    }
  }
}
