/*
 * ModuleTracker.java
 *
 * ModuleTracker checks the list of module directories for changes
 * and when they are detected, attempts to load any new modules 
 * into the server automatically
 *
 * The process goes like this:
 * 1) Once ever 10 seconds, the jars subdirectory is checked for a change
 * 2) If there is a change, the directory is enumerated and each .jar file
 *    is added to the classpath (if any)
 * 3) After all jars are added to the class path, each jar is scanned for
 *    a META-INF/xPL_Modules.cfg file.  If not found, the jar is skipped
 *    (though it remains in the classpath)
 * 4) If there is a defaults file, it's read.  If a module in the file is
 *    already loaded, then it's skipped.  If not, the config is loaded for
 *    the module and then, depending on the config parms, the module itself
 *    may be loaded and even started.
 */

package org.cdp1802.xpl.server;

import java.io.*;
import java.net.*;
import java.util.*;
import java.util.jar.*;
import java.util.zip.*;

/**
 *
 * @author gerry
 */
public class ModuleTracker implements Runnable {
  public static final String DEFAULT_MODULE_CONFIG_FILE = "META-INF/xPL_Modules.cfg";
  public static final int DEFAULT_FOLDER_SCAN_INTERVAL = 30;  // In seconds

  class ModuleDirInfo {
    File moduleDir = null;
    long lastUpdate = -1;
    
    ModuleDirInfo(File theDir) {
      moduleDir = theDir;
    }
    
    public boolean hasDirChanged() {
      if (moduleDir == null) return false;

      long modifiedAt = moduleDir.lastModified();
      if (modifiedAt == lastUpdate) return false;
      
      lastUpdate = modifiedAt;
      return true;
    }
  }

  
  class JARFileFilter implements FileFilter {
    public boolean accept(File f) {
      // Skip directories
      if (f.isDirectory()) return false;

      String extension = getExtension(f);
      if (extension.equalsIgnoreCase("jar")) return true;

      return false;
    }

    /**
      Method to get the extension of the file, in lowercase
     */
    private String getExtension(File f) {
      String s = f.getName();
      int i = s.lastIndexOf('.');
      if (i > 0 &&  i < s.length() - 1) return s.substring(i+1).toLowerCase();
      return "";
    }
  }
  
  HashMap<File, ModuleDirInfo> moduleDirList = new HashMap<File, ModuleDirInfo>();
  JARFileFilter jarFileFilter = new JARFileFilter();
  Thread updaterThread = null;
  boolean moduleUpdatorRunning = false;
  boolean moduleUpdatesPaused = false;
  int folderScanInterval = DEFAULT_FOLDER_SCAN_INTERVAL;
  
  public ModuleTracker() { }

  void debug(String theMessage) {
    System.out.println("MOD_TRACK:: " + theMessage);
  }

  public boolean isUpdatesPaused() { return moduleUpdatesPaused; }
  public void setUpdatesPaused(boolean arePaused) {
    moduleUpdatesPaused = arePaused;
  }
  
  
  public void initializeModuleTracker() {
    // Stop the updater, if needed
    stopUpdater();

    // Clear all entries out
    moduleDirList.clear();
  }
  
  public void addModuleDir(File newModuleDir) {
    synchronized(moduleDirList) {
      if (newModuleDir == null) return;
      if (!newModuleDir.exists() || !newModuleDir.isDirectory() || !newModuleDir.canRead()) return;
      if (moduleDirList.containsKey(newModuleDir)) return;
      
      moduleDirList.put(newModuleDir, new ModuleDirInfo(newModuleDir));
      moduleDirList.notifyAll();
    }
  }
  
  public void addModuleDir(String newModuleDirName) {
    addModuleDir(new File(newModuleDirName));
  }

  public void removeModuleDir(File moduleDir) {
    if (moduleDir == null) return;
    
    synchronized(moduleDirList) {
      moduleDirList.remove(moduleDir);
    }
  }
  
  public int getModuleDirCount() { return moduleDirList.size(); }
  
  public Collection<ModuleDirInfo> getModuleDirs() { return moduleDirList.values(); }
  
  public ModuleDirInfo getModuleDir(File forPath) {
    return moduleDirList.get(forPath);
  }
  
  // Load any modules from the passed jar.  If there are no
  // modules loaded (or errors), false is returned.  If 
  // at least one new module is loaded, true is returned
  boolean testAndLoadModulesFromJAR(File theJAR) {
    JarFile theJARFile = null;
    ZipEntry moduleConfigFile = null;
    InputStream configStream = null;
    boolean moduleStateDirty = false;
    ModuleDefaultsParser moduleDefaultsParser = xPL_Server.getServer().getModuleDefaultsParser();
    
    try {
      // Get access to the JAR file
      theJARFile = new JarFile(theJAR);
    
      // See if it has a default entry
      moduleConfigFile = theJARFile.getEntry(DEFAULT_MODULE_CONFIG_FILE);
      if (moduleConfigFile == null) {
        return false;
      }
      
      // Get a stream to the file
      configStream = theJARFile.getInputStream(moduleConfigFile);
      
      // Read in the module config
      return moduleDefaultsParser.readModuleDefaults(configStream, false);
    } catch (Exception jarError) {
      System.err.println("MOD_TRACK:: Error reading JAR " + theJAR.getName() + " for file " + DEFAULT_MODULE_CONFIG_FILE + " -- " + jarError.getMessage());
      jarError.printStackTrace();
      return false;
    } finally {
      if (configStream != null) {
        try { configStream.close(); } catch (IOException closeError) { }
      }
      if (theJARFile != null) {
        try { theJARFile.close(); } catch (IOException closeError) { }
      }
    }
  }
  
  private File urlToFile(URL theURL) {
    try {
      return new File(URLDecoder.decode(theURL.getFile(), "UTF-8"));
    } catch (Exception anything) { 
      return null;
    }
  }
  
  // Scan all .jar files on the class path looking for any
  // module definitions.  If any are found, load them.  
  // If at least one module was loaded, true is returned
  boolean scanClassPathJarsForModules() {
    // Check each JAR to see if it contains modules
    boolean moduleStateDirty = false;
    
    String jarPath = null;
    File jarFile = null;
    for (URL theURL: ClassPathManager.enumerateClassPaths()) {
      // See if this is (potentially) a jar
      if (theURL.getPath().toLowerCase().endsWith(".jar")) {
        if ((jarFile = urlToFile(theURL)) == null) continue;
        moduleStateDirty |= testAndLoadModulesFromJAR(jarFile);
      }
    }

    // If the module state is dirty, then ensure they are all correct
    if (moduleStateDirty) {
      xPL_Server.getServer().ensureConsistantModuleStates();
    }
    
    return moduleStateDirty;
  }
  
  // Check the passed directory for jar files.  If
  // any are found, see if they are already in the 
  // class path.  If not, add them in.  If any jars
  // are added to the class path, return true.  If
  // there are no new jars, then return false.
  boolean checkAndInstallJARs(ModuleDirInfo theDir) {
    // Get list of jars, if any
    File jarList[] = theDir.moduleDir.listFiles(jarFileFilter);
    if ((jarList == null) || (jarList.length == 0)) return false;
    
    // Process each .jar into the class path
    boolean newJARsFound = false;
    for (File theJAR: jarList) {
      if (ClassPathManager.existsInClassPath(theJAR.getAbsolutePath())) continue;
      if (ClassPathManager.ensurePathInClassPath(theJAR.getAbsolutePath())) {
        newJARsFound = true;
      }
    }
    
    return newJARsFound;
  }

  // Scan the module dirs for new jars and install them.  If there
  // are any new jars, true is returned.  Otherwise, false.
  boolean scanModuleDirs() {
    synchronized(moduleDirList) {
      boolean moduleStateDirty = false;
      
      for (ModuleDirInfo theDir: moduleDirList.values()) {
        // See if no change
        if (!theDir.hasDirChanged()) continue;
        
        // If there are any new jars
        moduleStateDirty = true;
        checkAndInstallJARs(theDir);
      }
      
      // If no changes, we're done
      return moduleStateDirty;
    }
  }
  
  public void startUpdater() {
    synchronized(moduleDirList) {
      if (moduleUpdatorRunning) return;

      // Create/start the thread
      updaterThread = new Thread(this, "Module Tracker Updater");
      updaterThread.start();
      
      try {
        moduleDirList.wait();
      } catch (InterruptedException intError) { }
      
      return;
    }
  }
  
  public void stopUpdater() {
    synchronized(moduleDirList) {
      if (!moduleUpdatorRunning) return;
      
      // Clear the flag, notify the runner and wait for shutdown
      moduleUpdatorRunning = false;
      moduleDirList.notifyAll();
    }
    
    try {
      updaterThread.join();
    } catch (Exception anything) { }
  }
  
  public void run() {
    synchronized(moduleDirList) {
      // At startup, let the starter know we are going
      moduleUpdatorRunning = true;
      moduleDirList.notifyAll();
      
      // Check for any new modules in bootstrap class path
      scanModuleDirs();
      scanClassPathJarsForModules();
      
      // Now loop until we are shutdown
      while (moduleUpdatorRunning) {
        // Check for module updates
        if (!moduleUpdatesPaused) {
          if (scanModuleDirs()) scanClassPathJarsForModules();
        }
        
        // And then sleep a bit
        try {
          moduleDirList.wait(folderScanInterval * 1000);
        } catch (Exception anyError) { }
      }
    }
  }

}
