/*
 * NetworkModule.java
 * Copyright 2004, Gerald R Duprey Jr
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy of
 * the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0 
 *
 * Unless required by applicable law or agreed to in writing, software distributed
 * under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
 * CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License. 
 */

package org.cdp1802.xpl.server;

import java.net.InetAddress;
import org.cdp1802.xpl.*;
import org.cdp1802.xpl.ethernet.*;
import org.cdp1802.xpl.device.xPL_DeviceConfigItemI;
import org.cdp1802.xpl.device.xPL_DeviceChangeEventI;
import org.cdp1802.xpl.device.xPL_DeviceChangeListenerI;

/**
 * This module handles the xPL Network interfaces
 *
 * @author gduprey
 */
class NetworkModule extends xPL_Module implements xPL_DeviceChangeListenerI {
  xPL_Manager xplManager = xPL_Manager.getManager();
  
  public NetworkModule() { }

  // Insure we have at least one valid network connection.  Normally, modules
  // would not bother doing this, but as the entire server hangs on having
  // a network and without one, nothing else in the server can run, this 
  // makes some sense
  void installDefaultNetwork() {
    // Skip if no chance of hope
    if (moduleDevice == null) return;
    
    // Create the network configurable, if it doesn't exist
    xPL_DeviceConfigItemI defaultNetwork = moduleDevice.getConfigurable("Network");
    if (defaultNetwork == null) defaultNetwork = moduleDevice.addConfigurable("Network", 255, true, true);
    
    // Add in the default network, if it doesn't exist
    if (defaultNetwork.getValueCount() == 0) {
      defaultNetwork.addValue("0.0.0.0");
    }
  }
  
  boolean isExistingNetworkHandler(String theTestAddress) {
    EthernetHandler netHandler = null;
    
    for (xPL_MediaHandlerI theHandler: xPL_Manager.getManager().getMediaHandlers()) {
      // Skip if not a network handler
      if (!(theHandler instanceof EthernetHandler)) continue;
      netHandler = (EthernetHandler) theHandler;
      
      InetAddress curAddress = netHandler.getConnectedNetworkInterface();
      if (curAddress == null) curAddress = netHandler.getNetworkInterface();
      if (curAddress == null) curAddress = netHandler.getAdvertisedInterface();
      if (curAddress == null) continue;
      
      if (curAddress.getHostAddress().equals(theTestAddress)) return true;
    }
    
    return false;
  }
  
  // This compares the list of current network media handlers and the 
  // configuration for this module and figures of what changes, if any,
  // need to be done.
  void updateNetworkHandlers() {
    // Bind to each network specified
    String theValue = null;
    InetAddress theAddress = null;
    xPL_DeviceConfigItemI networkList = moduleDevice.getConfigurable("Network");
    
    for(String theNetwork: networkList.getValues()){
      // Attempt to parse into network address
      try {
        theAddress = InetAddress.getByName(theNetwork);
      } catch (Exception anError) {
        continue;
      }

      // See if this address is already running
      if (isExistingNetworkHandler(theAddress.getHostAddress())) {
        continue;
      }
      
      // Do a smart start for this network
      try {
        xplManager.createAndStartNetworkHandler(theAddress);
      } catch (xPL_MediaHandlerException startError) { 
        System.err.println("ENET_HANDLER:: Error starting network interface for " + theAddress.getHostAddress() + " -- " + startError.getMessage());
        startError.printStackTrace();
        continue; 
      }
    }
    
    // Now see if there are any network handlers running that shouldn't be anymore
    synchronized(xPL_Manager.getManager().getMediaHandlers()) {
      xPL_MediaHandlerI theHandler = null;
      EthernetHandler netHandler = null;
      int handlerCount = xPL_Manager.getManager().getMediaHandlerCount();
      for (int handlerIndex = handlerCount - 1; handlerIndex >= 0; handlerIndex--) {
        theHandler = xPL_Manager.getManager().getMediaHandlerAt(handlerIndex);

        // Skip if not a network handler
        if (!(theHandler instanceof EthernetHandler)) return;
        netHandler = (EthernetHandler) theHandler;

        InetAddress curAddress = netHandler.getConnectedNetworkInterface();
        if (curAddress == null) curAddress = netHandler.getNetworkInterface();
        if (curAddress == null) curAddress = netHandler.getAdvertisedInterface();
        if (curAddress == null) continue;

        // See if there is a network with that value
        boolean foundNetwork = false;
        String testNetwork = curAddress.getHostAddress();
        for (String theNetwork: networkList.getValues()) {
          if (theNetwork.equals(testNetwork)) {
            foundNetwork = true;
            break;
          }
        }

        // If not found, we have to remove this 
        if (!foundNetwork) {
          xPL_Manager.getManager().removeMediaHandler(netHandler);
        }
      }
    }
  }
  
  public void startModule() {
    // Skip out if already started
    if (isModuleStarted()) return;
    
    // Add listener
    moduleDevice.addDeviceChangeListener(this);
    
    // Install the default network, if needed
    installDefaultNetwork();
    
    // Start existing defined networks
    updateNetworkHandlers();
    
    // Do module startup
    super.startModule();
  }

  public void stopModule() {
    moduleDevice.removeDeviceChangeListener(this);
    xplManager.doSmartShutdown();
    super.stopModule();
  }
  
  public void handleXPLDeviceChange(xPL_DeviceChangeEventI theEvent) {
    if (theEvent.getDevice() != moduleDevice) return;
    if (theEvent.getChangeReason() == xPL_DeviceChangeEventI.ChangeReason.CONFIGURATION_CHANGED) {
      updateNetworkHandlers();
    }
  }
}
