/*
 * ServerManagerModule.java
 * Copyright 2004, Gerald R Duprey Jr
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy of
 * the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0 
 *
 * Unless required by applicable law or agreed to in writing, software distributed
 * under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
 * CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License. 
 */

package org.cdp1802.xpl.server;

import java.io.File;
import java.util.*;

import org.cdp1802.xpl.*;
import org.cdp1802.xpl.device.*;
import org.cdp1802.xpl.tracker.*;

/**
 * This module provides the management interface to the xPL_Server.  If you
 * do not want the management console, set ModuleEnabled=0 to disable it.
 * 
 * By default, the management device will identify itself as cdp1802-xpl4java.X
 * where X is a fairly unique identifier.  If you want to change that identification,
 * you can supply a line in the server configuration like this:
 * 
 *   ModuleDeviceIdent=myVendor-myDevice.myInstance
 * 
 * NOTE: If you set myInstance to *, a fairly unique identifier will be 
 * assigned as the instance ID in place of this.
 * 
 * Messages need to be directed at a schema class/type of server.manage and
 * have the mandatory named value COMMAND=.  Commands accepted are:
 * 
 * SHUTDOWN_SERVER - shutdown the server.  All started modules are given a chance
 * to logically stop and then the server itself is stopped.
 * 
 * START_MODULE - start the module named as specified in the MODULE_NAME= named value.  If
 * the module has not yet been loaded, it is first loaded.  If the module is not
 * enabled, it will not be started
 * 
 * STOP_MODULE - stop the module named as specified in the MODULE_NAME= named value.  If
 * the module has not yet been loaded, nothing happens.
 * 
 * UNLOAD_MODULE - unload the module named as specified in the MODULE_NAME= named value.  If
 * the module has not be loaded, nothing happens.  If the module was loaded and
 * started, it is stopped before being unloaded.
 * 
 * RELOAD_MODULE - unload and reload the named module as specified in MODULE_NAME=.  If the
 * module was started or loaded, the reload unloads it and attempts to bring it
 * back to it's previous state.  If the module was not loaded or started, it is
 * unloaded (if possible) and then the autoload/start module settings are 
 * honored.
 * 
 * MODULE_STATUS - send the state of the named module specified by the 
 * MODULE_NAME= .  If the module does not exist, nothing happens.  You can
 * ask for the status of all messages by using a special module name of ALL.  
 * Be aware that this can flood the network and may result in lost packets.  One
 * message will be sent out for each known module.
 * 
 * LIST_MODULES - send a list of all known modules out.  In order to work inside
 * the UDP packet size limit, there are at most 10 modules sent per message
 * and "n" messages where "n" is # of modules / 10, rounded up.  To make 
 * receiving messages simpler, the first message will have a LIST=START tag
 * in it and the last message will have a LIST=END.  Each module is listed
 * using a MODULE_NAME tag.
 * 
 * You can control how many modules are listed in each packet by include 
 * a COUNT= in the request (valid values are 1 to 999).
 * 
 * NOTE: Modules are returned in the order they were created which, from the
 * receiving end, may not seem like any order at all.
 * 
 * Status Updates:
 * ========================================================================
 * Whenever the state of a module changes, a module status message is sent 
 * out.  It has a schema type of server.module and contains the following
 * items in the message body:
 * 
 * MODULE_NAME=<name of the module>
 * MODULE_ENABLED=[1|0]
 * MODULE_STATE=[DISABLED|UNLOADED|LOADED|STARTED|STOPPING|UNLOADING|DELETED|RENAMED]
 * MODULE_CLASS_NAME=<class name of the module>
 * MODULE_AUTO_LOAD=[1|0]
 * MODULE_AUTO_START=[1|0] *
 * NOTE: You only receive a state of STOPPING once, when the module is actually 
 * being stopped.  Subsequently, the state will be LOADED.  Similary, you only
 * receive the state UNLOADING when the module is being unloaded (after which,
 * that module will not have any status until it may someday be reloaded).  If you
 * receive a RENAMED status, there will be a MODULE_OLD_NAME attribute with the
 * modules previous name.
 * 
 * SERVER_STATUS - sends back information about the server including how
 * many modules are defined, loaded and stated, memory usage and uptime.
 * Status message consists of:
 *  SERVER_UPTIME=<# of minutes since server started>
 *  MODULE_COUNT=<# of known modules>
 *  LOADED_MODULE_COUNT=<# of the known modules that have been loaded>
 *  STARTED_MODULE_COUNT=<# of loaded modules that are started/running>
 *  MAX_SYS_MEMORY= Maximum amount of system memory available for Java (MBs)
 *  USED_SYS_MEMORY=Amount of system memory actually allocated for Java (tMBs)
 *  FREE_JAVA_MEMORY=Amount of the allocated system memory still available for Java to use (MBs)
 *  USED_JAVA_MEMORY=Amount of the allocated system memory currently in use by Java (MBs)
 *  
 * RESCAN_MODULES - forces an immediate scan for new modules on the module
 * paths
 * 
 * @author gduprey
 */
class ServerManagerModule extends xPL_Module implements xPL_DeviceMessageListenerI, xPL_DeviceChangeListenerI, xPL_ManagerListenerI  {
  public static final int MEGABYTE = 1048576;
  
  xPL_Server theServer = xPL_Server.getServer();
  xPL_MutableMessageI statusMessage = null;
  ModuleTracker moduleTracker = null;
  boolean pathUpdatesAllowed = false;
  long serverStartTime = System.currentTimeMillis();
  boolean discoveryRequestSent = false;
  TrackedDeviceManager trackingManager = null;
  
  public ServerManagerModule() { }
  
  public void loadModule(xPL_ModuleConfigI ourConfig, xPL_DeviceI ourDevice) {
    theServer.managementModule = this;
    moduleTracker = theServer.moduleTracker;

    // Create a new message
    statusMessage = ourDevice.createBroadcastMessage(xPL_MessageI.MessageType.STATUS, "server", "status");

    super.loadModule(ourConfig, ourDevice);
  }
  
  public void startModule() {
    // Skip out if already started
    if (isModuleStarted()) return;
    
    // Start the device
    moduleDevice.addMessageListener(this);
    moduleDevice.addDeviceChangeListener(this);

    // Get hook on manager
    if (trackingManager == null) {
      trackingManager = xPL_Manager.getManager().getTrackedDeviceManager();
      trackingManager.setTrackingDevice(moduleDevice);
    }
    
    // Listen to manager events and change messages
    xPL_Manager.getManager().addManagerListener(this);
  
    // Start tracker
    discoveryRequestSent = false;
    trackingManager.startTracking();
    if (xPL_Manager.getManager().areHubsConfirmed()) {
      discoveryRequestSent = true;
      trackingManager.sendDiscoveryRequest();
    }
    
    // And we are running
    super.startModule();
  }
  
  public void stopModule() {
    // Stop device, if any
    if (moduleDevice != null) {
      moduleDevice.removeMessageListener(this);
      moduleDevice.removeDeviceChangeListener(this);
    }

    // Stop tracker
    xPL_Manager.getManager().getTrackedDeviceManager().stopTracking();

    xPL_Manager.getManager().removeManagerListener(this);
    super.stopModule();
  }

  public void unloadModule() {
    stopModule();

    // Release resources
    theServer = null;
    if (statusMessage != null) statusMessage.releaseResources();
    statusMessage = null;
    moduleTracker = null;

    super.unloadModule();
  }
  
  private void updateModuleConfig() {
    if (!pathUpdatesAllowed) return;
    xPL_DeviceConfigItemI searchPaths = moduleDevice.getConfigurable("SearchPath");
    if (searchPaths == null) return;
    
    boolean pathsChanged = false;
    
    // Suspend automatic scanning until we're done
    moduleTracker.setUpdatesPaused(true);

    // Update any refresh interval
    moduleTracker.folderScanInterval = moduleDevice.getIntConfigValue("ScanInterval", moduleTracker.folderScanInterval);
    
    // First, add in any new paths
    File searchPath = null;
    for (String aSearchPath: searchPaths.getValues()) {
      // Skip empties
      if ((aSearchPath == null) || (aSearchPath.length() == 0)) continue;
      
      // Skip existing paths
      searchPath = new File(aSearchPath);
      if (moduleTracker.getModuleDir(searchPath) != null) continue;
      
      // Add new paths
      moduleTracker.addModuleDir(searchPath);
      pathsChanged = true;
    }

    // Identify any paths that should no longer be there
    ArrayList<File> purgeList = null;
    for (ModuleTracker.ModuleDirInfo theModule: moduleTracker.getModuleDirs()) {
      // See if the directory is in the list of paths
      boolean pathFound = false;
      for (String aSearchPath: searchPaths.getValues()) {
        // Skip empties
        if ((aSearchPath == null) || (aSearchPath.length() == 0)) continue;

        // Skip existing paths
        searchPath = new File(aSearchPath);
        if (searchPath.equals(theModule.moduleDir)) {
          pathFound = true;
          break;
        }
      }
      
      // If the path was not found, add it to the purge list
      if (!pathFound) {
        if (purgeList == null) purgeList = new ArrayList<File>();
        purgeList.add(theModule.moduleDir);
      }
    }

    // Now purge no longer relevant paths
    if ((purgeList != null) && (purgeList.size() != 0)) {
      for (File purgePath: purgeList) {
        moduleTracker.removeModuleDir(purgePath);
        pathsChanged = true;
      }
    }
    
    // If there were any path changes, rescan for them
    if (pathsChanged) {
      if (moduleTracker.scanModuleDirs()) moduleTracker.scanClassPathJarsForModules();
    }

    // Resume automatic scanning
    moduleTracker.setUpdatesPaused(false);
  }
  
  void startModulePathUpdates() {
    if (pathUpdatesAllowed) return;
    pathUpdatesAllowed = true;
    
    updateModuleConfig();
    moduleTracker.startUpdater();
  }
  
  private void listModules(xPL_MessageI theMessage) {
    int perMessageCount = theMessage.getIntNamedValue("COUNT", 10);
    if (perMessageCount < 0) perMessageCount = 10;
    
    xPL_MutableMessageI listMessage = moduleDevice.createBroadcastMessage(xPL_MessageI.MessageType.STATUS, "server", "list");
    listMessage.addNamedValue("LIST", "START");
    
    int moduleCount = theServer.moduleConfigs.size();
    int writtenCount = 0;
    for (int moduleIndex = 0; moduleIndex < moduleCount; moduleIndex++) {
      listMessage.addNamedValue("MODULE_NAME", theServer.moduleConfigs.get(moduleIndex).getModuleName());
      writtenCount++;
      
      if ((writtenCount % perMessageCount) == 0) {
        if (writtenCount == moduleCount) listMessage.addNamedValue("LIST", "END");
        moduleDevice.sendMessage(listMessage);
        listMessage.clearMessageBody();
      }
    }
    
    // If there are any left overs, wrap it up
    if (listMessage.getMessageBody().getNamedValuesCount() != 0) {
      listMessage.addNamedValue("LIST", "END");
      moduleDevice.sendMessage(listMessage);
    }

    // Free resources
    listMessage.releaseResources();
  }

  public void handleXPLDeviceChange(xPL_DeviceChangeEventI theEvent) {
    if (theEvent.getDevice() != moduleDevice) return;
    if (theEvent.getChangeReason() == xPL_DeviceChangeEventI.ChangeReason.CONFIGURATION_CHANGED) {
      updateModuleConfig();
    }
  }
  
  private xPL_ModuleConfigI parseModule(xPL_MessageI theMessage) {
    String moduleName = theMessage.getNamedValue("MODULE_NAME");
    if (moduleName == null) return null;
    return theServer.getNamedModuleConfig(moduleName);
  }
  
  public void handleXPLDeviceMessage(xPL_DeviceI theDevice, xPL_MessageI theMessage) {
    // Skip unless this is for us
    if (!theMessage.getSchemaClass().equalsIgnoreCase("server") || !theMessage.getSchemaType().equalsIgnoreCase("manage")) return;
    
    // Decode commands
    xPL_ModuleConfigI moduleConfig = null;
    String theCommand = theMessage.getNamedValue("COMMAND");
    if (theCommand == null) return;
    
    if (theCommand.equalsIgnoreCase("SHUTDOWN_SERVER")) {
      theServer.stopXPLServer();
      System.exit(0);
      return;
    }
    
    if (theCommand.equalsIgnoreCase("START_MODULE")) {
      if ((moduleConfig = parseModule(theMessage)) == null) {
        debug("Attempt to start module when module cannot be found");
        return;
      }
      
      if (!theServer.startModule(moduleConfig.getModuleName())) debug("Unable to start module");
      return;
    }
    
    if (theCommand.equalsIgnoreCase("STOP_MODULE")) {
      if ((moduleConfig = parseModule(theMessage)) == null) return;
      theServer.stopModule(moduleConfig.getModuleName());
      return;
    }
    
    if (theCommand.equalsIgnoreCase("UNLOAD_MODULE")) {
      if ((moduleConfig = parseModule(theMessage)) == null) return;
      theServer.unloadModule(moduleConfig.getModuleName());
      return;
    }
    
    if (theCommand.equalsIgnoreCase("RELOAD_MODULE")) {
      if ((moduleConfig = parseModule(theMessage)) == null) return;
      theServer.reloadModule(moduleConfig.getModuleName());
      return;
    }

    if (theCommand.equalsIgnoreCase("MODULE_STATUS")) {
      if ((moduleConfig = parseModule(theMessage)) != null) {
        sendModuleStatus(moduleConfig, false);
        return;
      }
      
      // Check for a special "All Modules" mode
      String moduleName = theMessage.getNamedValue("MODULE_NAME");
      if ((moduleName != null) && moduleName.equalsIgnoreCase("ALL")) {
        int moduleCount = theServer.moduleConfigs.size();
        for (int moduleIndex = 0; moduleIndex < moduleCount; moduleIndex++) {
          sendModuleStatus(theServer.moduleConfigs.get(moduleIndex), false);
        }
      }
        
      return;
    }
    
    if (theCommand.equalsIgnoreCase("LIST_MODULES")) {
      listModules(theMessage);
      return;
    }
    
    if (theCommand.equalsIgnoreCase("RESCAN_MODULES")) {
      if (pathUpdatesAllowed) {
        if (moduleTracker.scanModuleDirs()) moduleTracker.scanClassPathJarsForModules();
      }
      return;
    }
    
    if (theCommand.equalsIgnoreCase("SERVER_STATUS")) {
      sendServerStatus();
      return;
    }
    
  }

  public void sendModuleStatus(xPL_ModuleConfigI theModule, boolean stateChanged, String moduleState, String oldName) {
    // Safety trap
    if (moduleDevice == null) return;
    
    // Set message type
    if (stateChanged) 
      statusMessage.setType(xPL_MessageI.MessageType.TRIGGER);
    else
      statusMessage.setType(xPL_MessageI.MessageType.STATUS);

    // Build the message body
    statusMessage.setSchemaType("module");
    statusMessage.clearMessageBody();
    statusMessage.addNamedValue("MODULE_NAME", theModule.getModuleName());
    statusMessage.addNamedValue("MODULE_ENABLED", theModule.isEnabled());
    statusMessage.addNamedValue("MODULE_STATE", moduleState);
    statusMessage.addNamedValue("MODULE_CLASS_NAME", theModule.getMainClass());
    statusMessage.addNamedValue("MODULE_VERSION", theModule.getModuleVersion());
    statusMessage.addNamedValue("MODULE_DEVICE_IDENT", theModule.getModuleDeviceIdent());
    statusMessage.addNamedValue("MODULE_AUTO_LOAD", theModule.isAutoLoaded());
    statusMessage.addNamedValue("MODULE_AUTO_START", theModule.isAutoStarted());
    
    // Send the message
    moduleDevice.sendMessage(statusMessage);
  }
  
  public void sendModuleStatus(xPL_ModuleConfigI theModule, boolean stateChanged) {
    String moduleState = null;
    
    if (!theModule.isEnabled()) 
      moduleState = "DISABLED";
    else if (theModule.isLoaded()) {
      if (theModule.isStarted()) {
        moduleState = "STARTED";
      } else {
        moduleState = "LOADED";
      }
    } else {
      moduleState = "UNLOADED";
    }

    sendModuleStatus(theModule, stateChanged, moduleState, null);
  }
  
  public void sendServerStatus() {
    // Safety trap
    if (moduleDevice == null) return;
    
    // Set message type
    statusMessage.setType(xPL_MessageI.MessageType.STATUS);
    statusMessage.setSchemaType("status");

    // Build the message body
    statusMessage.clearMessageBody();
    statusMessage.addNamedValue("SERVER_UPTIME", (int) ((System.currentTimeMillis() - serverStartTime) / 60000));
    statusMessage.addNamedValue("MODULE_COUNT", theServer.moduleConfigs.size());
    statusMessage.addNamedValue("LOADED_MODULE_COUNT", theServer.loadedModules.size());
    
    int startedModules = 0;
    for (xPL_ModuleConfigI modConfig: theServer.moduleConfigs) {
      if (modConfig.isStarted()) startedModules++;
    }
    statusMessage.addNamedValue("STARTED_MODULE_COUNT", startedModules);
    statusMessage.addNamedValue("MAX_SYS_MEMORY", (int) (Runtime.getRuntime().maxMemory() / MEGABYTE));
    statusMessage.addNamedValue("USED_SYS_MEMORY", (int) (Runtime.getRuntime().totalMemory() / MEGABYTE));
    statusMessage.addNamedValue("FREE_JAVA_MEMORY", (int) (Runtime.getRuntime().freeMemory() / MEGABYTE));
    statusMessage.addNamedValue("USED_JAVA_MEMORY", (int) ((Runtime.getRuntime().totalMemory() - Runtime.getRuntime().freeMemory()) / MEGABYTE));
  
    // Send the message
    moduleDevice.sendMessage(statusMessage);
  }
  
  public void handleXPLManagerEvent(xPL_ManagerEvent theEvent) {
    if (theEvent.getEventCode() == xPL_ManagerEvent.EventCode.ALL_HUBS_DETECTED) {
      if (!discoveryRequestSent) {
        discoveryRequestSent = true;
        trackingManager.sendDiscoveryRequest();
      }
    }
  }
}
