/*
 * xPL_Module.java
 * Copyright 2004, Gerald R Duprey Jr
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy of
 * the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0 
 *
 * Unless required by applicable law or agreed to in writing, software distributed
 * under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
 * CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License. 
 */

package org.cdp1802.xpl.server;
import org.cdp1802.xpl.device.xPL_DeviceI;
import org.cdp1802.xpl.device.DeviceManager;

/**
 * Abstract base module with some standard features.  Just
 * override the methods you careabout/need.
 *
 * @author gduprey
 */
public class xPL_Module implements xPL_ModuleI {
  protected xPL_ModuleConfigI moduleConfig = null;
  protected boolean moduleIsStarted = false;
  protected xPL_DeviceI moduleDevice = null;
  
  public xPL_Module() {  }

  protected void debug(String theMessage) {
    if (moduleConfig == null)
      System.err.println("XPL_MODULE[UNKNOWN]::" + theMessage);
    else
      System.err.println("XPL_MODULE[" + moduleConfig.getModuleName() + "]:: " + theMessage);
  }

  protected void warn(String theMessage) {
    if (moduleConfig == null)
      System.err.println("XPL_MODULE[UNKNOWN] (Warning)::" + theMessage);
    else
      System.err.println("XPL_MODULE[" + moduleConfig.getModuleName() + "] (Warning) :: " + theMessage);
  }
  
  protected void error(String theMessage) {
    if (moduleConfig == null)
      System.err.println("XPL_MODULE[UNKNOWN] (ERROR)::" + theMessage);
    else
      System.err.println("XPL_MODULE[" + moduleConfig.getModuleName() + "] (ERROR) :: " + theMessage);
  }
  
  public xPL_ModuleConfigI getModuleConfig() { return moduleConfig; }
  public xPL_DeviceI getModuleDevice() { return moduleDevice; }
  public boolean isModuleStarted() { return moduleIsStarted; }
  
  /**
   * This should always be invoked by descendent classes (super.loadModule(x,x))
   * 
   * If you do not supply a ModuleDeviceIdent= in the xPL_Modules.cfg for this
   * module, no device is created and null is passed as the device.  If you
   * do provide a serivce ident, a device will be created and passed in.  The
   * device ident can have an instance ID of * which will result in a unique
   * instance ID being assigned the modules device.
   * 
   * If you supply a ModuleDeviceConfigFilename= in the xPL_Modules.cfg file,
   * the device created will be a configurable device and will attempt to load
   * previous values (if any) from the named file.  In addition, it will save 
   * any values it is configured for into the same file.  If there are not
   * saved values, the default values for the module (as defined, if any, in
   * xPL_Modules.cfg) will be installed as configurable items.
   * 
   * If you do NOT want to use the supplied device, just ignore it (though you
   * should then omit any reference to ModuleDeviceIdent= in the xPL_Modules.cfg
   * file).  After the module is loaded, the manager will compare what it passed
   * to loadModule() and what you have nstalled (which can be null) and if they 
   * do not match, it assumes you are  not going to use the default device and
   * so it destroys it.
   * 
   * If the manager notices that you are not using the same device it provided
   * but you are using a device (getModuleDevice() != null), it will attempt to
   * install default configuration values, if needed, into the device.  If you
   * do not want this to happen (it's usually a very good thing), just leave the
   * results of getModuleDevice() returning null and store the device you are 
   * using in a different place (note: this is generally NOT a good idea)
   * 
   * @param moduleConfig configuration for this module
   * @param moduleDevice default device for this module or null if module does not use a device
   */
  public void loadModule(xPL_ModuleConfigI moduleConfig, xPL_DeviceI moduleDevice) {
    this.moduleConfig = moduleConfig;
    this.moduleDevice = moduleDevice;
  }

  /**
   * This should always be invoked by descendent classes (super.startModule))
   *
   * If not already started, this will start/enable the modules device
   *
   */
  public void startModule() {
    moduleIsStarted = true;
    if (moduleDevice != null) {
      moduleDevice.setEnabled(true);
    }
  }

  /**
   * This should always be invoked by descent classes (super.stopModule)
   *
   * If not already stopped, this will stop the modules device and release
   * it's resources.  
   */
  public void stopModule() {
    moduleIsStarted = false;
    if (moduleDevice != null) moduleDevice.setEnabled(false);
  }

  /**
   * This should always be involed by descendent classes (super.unloadModule)
   *
   * If not already released, this will release the devices resources
   */
  public void unloadModule() {
    stopModule();
    this.moduleConfig = null;

    if (moduleDevice != null) {
      DeviceManager.getManager().releaseDeviceResources(moduleDevice);
      moduleDevice = null;
    }
  }
}
