/*
 * xPL_ModuleConfig.java
 * Copyright 2004, Gerald R Duprey Jr
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy of
 * the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0 
 *
 * Unless required by applicable law or agreed to in writing, software distributed
 * under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
 * CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License. 
 */

package org.cdp1802.xpl.server;

import java.lang.reflect.Method;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.*;
import java.io.File;

import org.cdp1802.xpl.*;
import org.cdp1802.xpl.device.xPL_DeviceConfigItemI;
import org.cdp1802.xpl.device.xPL_DeviceI;
import org.cdp1802.xpl.device.DeviceManager;
import org.cdp1802.xpl.device.xPL_Device;

/**
 *
 * @author gduprey
 */
class xPL_ModuleConfig implements xPL_ModuleConfigI {
  String moduleName = null;
  String moduleVersion = null;
  String moduleMainClass = null;
  String moduleDeviceIdent = null;
  String moduleDeviceConfigFilename = null;

  boolean isEnabled = true;
  boolean isAutoLoaded = true;
  boolean isAutoStarted = true;
  xPL_ModuleI loadedModule = null;
  MutableNamedValuesI moduleDefaults = xPL_Util.createMutableNamedValues();

  xPL_ModuleConfig() { }
  
  private void debug(String theMessage) {
    System.out.println("MOD_CONF:: " + theMessage);
  }
  
  private void error(String theError) {
    System.err.println("MOD_CFG[" + (moduleName == null ? "<UNKNOWN>" : moduleName) + "] *ERROR* " + theError);
  }
  
  public String getModuleName() { return moduleName; }
  public String getModuleVersion() { return moduleVersion; }
  public String getModuleDeviceIdent() { return moduleDeviceIdent; }
  public String getModuleDeviceConfigFilename() { return moduleDeviceConfigFilename; }

  public String getMainClass() { return moduleMainClass; }
  
  public boolean isEnabled() { return isEnabled; }
  public boolean isAutoLoaded() { return isAutoLoaded; }
  public boolean isAutoStarted() { return isAutoStarted; }

  public boolean isLoaded() { return loadedModule != null; }
  public boolean isStarted() { return ((loadedModule != null) && loadedModule.isModuleStarted()); }
  
  public xPL_ModuleI getModule() {
    return loadedModule;
  }
  
  public MutableNamedValuesI getModuleDefaults() { return moduleDefaults; }

  // This will compare the list of module defaults (loaded from the modules
  // configuration section of xPL_Modules.cfg) and what values the module
  // has in it's device.  The install only those defaults that are not already set
  // (from the device restoring it's last set values).  
  //
  // This allows the module to retain it's settings and yet still allows new module
  // options to "show through".
  private void installModuleDefaults(xPL_DeviceI theDevice) {
    // Make sure this is even necessary
    if (theDevice == null) return;
    if ((moduleDefaults == null) || (moduleDefaults.getNamedValuesCount() == 0)) return;
    
    // Get copy of the defaults that we can modify
    MutableNamedValuesI defaultValues = moduleDefaults.getMutableCopy();
    
    // Purge anything already loaded as a configurable.  That means if the
    // item has been defined before, even if it has no values, it's going to
    // be removed from the list.  
    for (xPL_DeviceConfigItemI configItem: theDevice.getConfigurables()) {
      defaultValues.removeNamedValues(configItem.getName());
    }

    // This will define the actual configurable elements and their limits (like
    // indicating if a configuration element can be reconfigured, what the maximum
    // number of values for a config item (or filters or group) is, etc.
    DeviceManager.getManager().getDevicePersisterr().parseConfigDefinitions(theDevice, defaultValues);
    
    // What is left in the defaultValues is anything that wasn't definitional -- in short,
    // configuration values.  If this is a first time, we'll install all values (interval,
    // filters, groups, etc).  If not, then only update configurables that were not
    // part of the device before (those that were have already been purged).
    ((xPL_Device) theDevice).installConfigValues(defaultValues,  true, theDevice.wasRestored());
  }
  
  boolean loadModule() {
    // If already loaded, great, we are done
    if (loadedModule != null) return true;
    
    // If module is not enabled, it cannot be loaded
    if (!isEnabled) return false;
    
    // Insure we are configured enough for this to work
    if ((moduleName == null) || (moduleMainClass == null)) {
      error("Unable to load module -- no name or class specified");
      return false;
    }
    
    // Attempt to get this class from the stock class loader
    try {
      Class theClass = ClassPathManager.findClass(moduleMainClass);
      loadedModule = (xPL_ModuleI) theClass.newInstance();
    } catch (Throwable anyError) { 
      error("Unable to instantiate class [" + moduleMainClass + "] -- " + anyError.getMessage());
      anyError.printStackTrace();
      return false; 
    }

    // Create a device for it
    xPL_DeviceI theDevice = null;
    if (moduleDeviceIdent != null) {
      xPL_IdentifierI deviceIdent = xPL_Manager.getManager().getIdentifierManager().parseNamedIdentifier(moduleDeviceIdent);
      if (deviceIdent != null) {
        // Make sure and empty string is as good as no string
        if ((moduleDeviceConfigFilename != null) && (moduleDeviceConfigFilename.length() == 0)) 
          moduleDeviceConfigFilename = null;
        
        // Create a configurable device or not
        if (moduleDeviceConfigFilename != null) {
          theDevice = DeviceManager.getManager().createConfigurableDevice(deviceIdent, new File(moduleDeviceConfigFilename));
        } else {
          theDevice = DeviceManager.getManager().createDevice(deviceIdent);
        }
        
        // Install the version
        theDevice.setVersion(moduleVersion);
      }
      
      // Install defaults into the device
      installModuleDefaults(theDevice);
    }
    
    // Now we try to load it
    try {
      loadedModule.loadModule(this, theDevice);
    } catch (Throwable anyException) {
      error("Error loading module -- " + anyException.getMessage());
      anyException.printStackTrace();
      try { loadedModule.unloadModule(); } catch (Exception ignoredError) { }
      loadedModule = null;
      return false;
    }

    // If the passed device and the installed device do not match, the module
    // does not want to use the default device -- release it
    if ((theDevice != null) && (loadedModule.getModuleDevice() != theDevice)) {
      DeviceManager.getManager().releaseDeviceResources(theDevice);
      
      // If there is a device for the module, load defaults, if needed
      if (loadedModule.getModuleDevice() != null) installModuleDefaults(loadedModule.getModuleDevice());
    }
    
    // We have loaded it
    return true;
  }
}
