/*
 * xPL_ModuleConfigI.java
 * Copyright 2004, Gerald R Duprey Jr
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy of
 * the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0 
 *
 * Unless required by applicable law or agreed to in writing, software distributed
 * under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
 * CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License. 
 */

package org.cdp1802.xpl.server;

import java.util.Collection;
import org.cdp1802.xpl.MutableNamedValuesI;
import org.cdp1802.xpl.NamedValueI;

/**
 * Describes the configuration of a module.  The configuration is seperate
 * from the module because a module may be loaded and unloaded (resulting
 * in multiple instance of the modules class) many times over a servers 
 * lifetime.
 * @author gduprey
 */
public interface xPL_ModuleConfigI {
  /**
   * Get the modules name
   * @return modules name
   */
  public String getModuleName();

  /**
   * Get the main class for this module.  This is the name used to create instance
   * of the module when a module is actually loaded.
   * @return full name of the modules main class
   */
  public String getMainClass();
  
  /**
   * Get the default device ident for this module.  This is what will be used
   * if the module does not otherwise create it's own device.  
   *
   * NOTE: The instance portion may be a * meaning that an appropriate instance
   * ID will be created when the device is created
   *
   * @return name of the device for this module (i.e. vendor-device.instance)
   */
  public String getModuleDeviceIdent();

  /**
   * Get the module version, as defined for this module at configuration
   *
   * NOTE: It's possible there will be no version and you could receive
   * null back
   *
   * @return module version string or null
   */
  public String getModuleVersion();
  
  /**
   * Get the default device configuration file.  This file is used to save and
   * restore configuration values for the modules device.  If there is no
   * config file, the device will not be configurable
   *
   * @return name of the configuration file or null if none
   */
  public String getModuleDeviceConfigFilename();
  
  /**
   * Is this module currently enabled?
   * @return true if enabled
   */
  public boolean isEnabled();
  /**
   * Does this module automatically get started when the server discovers it 
   * (usually at startup)?  Note: If this is true, the isAutoLoaded() should also
   * be true.
   * @return true if the module is started when the server starts up
   */
  public boolean isAutoStarted();
  /**
   * Is the modules class loaded/instantiated when the server starts or when
   * the contain discovers the module (i.e. after a reload of the configuration)
   * @return true if module will be instantiated when the server starts
   */
  public boolean isAutoLoaded();

  /**
   * If the module is loaded, get the current instance of the modules class.
   * 
   * NOTE: Over the lifetime of the server, the actual instance of this module
   * may change, so you should not cache this value.
   * @return instance of the loaded module or null if the module has not been loaded yet
   */
  public xPL_ModuleI getModule();
  /**
   * Is the module actually loaded?
   * @return true if module has been loaded/instantiated
   */
  public boolean isLoaded();
  /**
   * Is the module actually started?
   * @return true if the module has successfully been started
   */
  public boolean isStarted();
  
  /**
   * Get the modules defaults as a set of named/values
   * @return modules defaults as a set of named values
   */
  public MutableNamedValuesI getModuleDefaults();
}
