/*
 * xPL_ModuleI.java
 * Copyright 2004, Gerald R Duprey Jr
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy of
 * the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0 
 *
 * Unless required by applicable law or agreed to in writing, software distributed
 * under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
 * CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License. 
 */

package org.cdp1802.xpl.server;
import org.cdp1802.xpl.device.xPL_DeviceI;

/**
 * An xPL Module is one that can be loaded and managed by the
 * xPL loader.  The xPL Loader allows multiple xPL apps to run
 * concurrently in the same Java JVM instance and share xPL
 * configuration and setup. 
 * 
 * Contract by which an xPL_ModuleI must play by:
 * 
 * <LI> No resource allocation in the module constructor
 * <LI> Module must never execute the System.exit() call
 * <LI> The module should not allocate extra resources when loaded, only when started
 * <LI> A module should create it's device when loaded and release it when unloaded
 * <LI> A module should not enable the device until it's started and disable it when stopped
 * <LI> All resources allocated when starting must be released when stopped.
 * <LI> If a module is done/complete, it must invoke xPL_Loader.unloadModule(this)
 * <LI> It must be able to be started and stopped multiple times
 * @author gduprey
 */
public interface xPL_ModuleI {
  /**
   * Do module "load-time" initialization and setup.  This method is invoked
   * when the module is first loaded, but before it's ever started.  This 
   * is a good place to initialize state, but would not be a good place to 
   * allocate any resources (database connection, serial ports, network
   * listeners, etc).  
   * 
   * Keep in mind an indefinite period of time may pass between when
   * loadModule() is invoked and the first time startModule() is invoked (it's
   * possible startModule() may never be invoked).
   * 
   * loadModule is called only once in a modules lifetime.  If it fails, the
   * module will be unloaded (and unloadModule() called) and removed from the
   * list of available modules in the loader.
   *
   * If you do not supply a ModuleDeviceIdent= in the xPL_Modules.cfg for this
   * module, no device is created and null is passed as the device.  If you
   * do provide a serivce ident, a device will be created and passed in.  The
   * device ident can have an instance ID of * which will result in a unique
   * instance ID being assigned the modules device.
   *
   * If you supply a ModuleDeviceConfigFilename= in the xPL_Modules.cfg file,
   * the device created will be a configurable device and will attempt to load
   * previous values (if any) from the named file.  In addition, it will save 
   * any values it is configured for into the same file.  If there are not
   * saved values, the default values for the module (as defined, if any, in
   * xPL_Modules.cfg) will be installed as configurable items.
   *
   * If you do NOT want to use the supplied device, just ignore it (though you
   * should then omit any reference to ModuleDeviceIdent= in the xPL_Modules.cfg
   * file).  After the module is loaded, the manager will compare what it passed
   * to loadModule() and what you have nstalled (which can be null) and if they 
   * do not match, it assumes you are  not going to use the default device and
   * so it destroys it.
   *
   * If the manager notices that you are not using the same device it provided
   * but you are using a device (getModuleDevice() != null), it will attempt to
   * install default configuration values, if needed, into the device.  If you
   * do not want this to happen (it's usually a very good thing), just leave the
   * results of getModuleDevice() returning null and store the device you are 
   * using in a different place (note: this is generally NOT a good idea)
   *
   * @param moduleConfig configuration for this module
   * @param moduleDevice default device for this module or null if module does not use a device
   */
  public void loadModule(xPL_ModuleConfigI moduleConfig, xPL_DeviceI moduleDevice);

  /**
   * Unload a module and free any resources.  This method is invoked 
   * when the module is about to be unloaded.  After this method is invoked,
   * this instance of the module will never be used again.
   * 
   * Keep in mind that the module may not be in a logical state when this
   * method is invoked -- inparticular, if loadModule() fails for any reason,
   * this method will be called and your internal state may not be consistant.
   * In addition, the management console may choose to unload a module at 
   * any time (though if the module was started, it will have first been
   * stopped before being unloaded).
   *
   * unloadModule is called only once in a modules lifetime.  It cannot fail
   * and must not throw any exceptions or otherwise not complete.  Write it
   * as defensively as possible and cleanup every resource possible.
   *
   * unloadModule MUST clear the module configuration out.  It is a requirement
   * of the contract that after unloadModule is called, getModuleConfig() must
   * return null.
   */
  public void unloadModule();

  /**
   * Start a module.  This is called to enable/start the modules processing.
   * Any resources/threads/etc needed by the module should be allocated 
   * here.
   * 
   * The module should be able to handle being started if it is already 
   * started (i.e. know it's running and just silently exit from the 
   * startModule() invocation).
   * 
   * If a module fails to start, it should reset it state to that of a
   * properly stopped module (that is, any resources allocated before the
   * failure are released).  Properly designed, you should be able to use
   * your own stopModule() as cleanup after a failure.
   * 
   * The module must be able to be started and stopped multiple times without failing.
   */
  public void startModule();

  /**
   * Stop a previously started module.  This is called stop stop/disable the 
   * modules processing.  Any resources/threads/etc should be released when
   * the module is stopped.
   * 
   * The module should be able to handle being stopped even if it is not 
   * currently started (i.e. know it's not running and just silently exit
   * from the stopModule() invocation).  A defensive approach is best here
   * (don't assume startModule() was called or completed -- it may have
   * failed).
   * 
   * The module must be able to be stopped and started multiple times without
   * failing.
   * 
   * This method must NOT fail.  It must always work and never throw an exception
   * or leave the module in an undefined/illegal/inconsistant state.
   */
  public void stopModule();

  /**
   * Determine if the module is currently running.  If it has been successfully
   * started via startModule() and not yet stopped via stopModule(), this should
   * return true.
   *
   * @return true if module is currently started/running
   */
  public boolean isModuleStarted();
  
  /**
   * Return the device associated with this module.  All modules must have a
   * device associated with them to be valid.
   *
   * @return the device for this module (or null if not loaded)
   */
  public xPL_DeviceI getModuleDevice();
  
  
  /**
   * Return the instance of the module configuration.  This should never
   * return null (except after the module is unloaded).
   *
   * @return configuration for this module
   */
  public xPL_ModuleConfigI getModuleConfig();
}
