/*
 * xPL_Server.java
 * Copyright 2004, Gerald R Duprey Jr
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy of
 * the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0 
 *
 * Unless required by applicable law or agreed to in writing, software distributed
 * under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
 * CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License. 
 */

package org.cdp1802.xpl.server;

import java.io.*;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;

import org.cdp1802.xpl.*;

/**
 * The xPL Server is designed to allow multiple xPL4Java apps to share the
 * same JVM, to have xPL devices configured for them and available and to allow
 * remote managment of the apps.
 * 
 * <P>Each app is called a module.  The modules the server will run are loaded
 * from the configuration file (xPL_Server.cfg, by default).  Each module
 * has a unique name, can be disabled or enabled, specifies a class path
 * and module class name, can be automatically loaded and started and have an
 * optional set of attributes the module can access that are stored in the 
 * configuration.
 * 
 * Each module can be in one of these states:
 *    UNLOADED - the module is known, but the actual modules main class is not
 *               yet loaded.
 *    LOADED - the module is known and an instance of it's main class is loaded
 *             and initialized, but it's not yet started.  
 *    STARTED - the module is loaded and started.  It's doing whatever you think 
 *              it should do.
 *              
 *    STOPPING - The module is stopping logically.  When done, the module state
 *               will change to LOADED.  The module could be restarted at that time.
 *               
 *    UNLOADING - the module is being unloaded. Subsequent attempts to start the
 *                module will load a new instance of the module
 *                
 *    DISABLED - the module is disabled and cannot be loaded or started.
 * 
 * @author gduprey
 */
public class xPL_Server implements xPL_ManagerListenerI {
  private static xPL_Server theXPLServer = new xPL_Server();
  /**
   * Get the single instance to the xPL server
   * @return instance of the xPL server
   */
  public static xPL_Server getServer() { return theXPLServer; }

  boolean serverRunning = false;
  boolean trackerRunning = false;
  String installationPath = null;
  xPL_ModuleConfigI managementModuleConfig = null;
  ServerManagerModule managementModule = null;
  
  ModuleDefaultsParser moduleDefaultsParser = new ModuleDefaultsParser(this);
  ModuleTracker moduleTracker = new ModuleTracker();
  ArrayList<xPL_ModuleI> loadedModules = new ArrayList<xPL_ModuleI>();
  ArrayList<xPL_ModuleConfigI> moduleConfigs = new ArrayList<xPL_ModuleConfigI>();
  
  xPL_Server() {  }
  
  private void debug(String theMessage) {
    System.out.println("xPL_CONT:: " + theMessage);
  }
  
  private void error(String theError) {
    System.err.println("xPL_CNTNR *ERROR*:: " + theError);
  }

  ModuleTracker getModuleTracker() { return moduleTracker; }
  
  /**
   * Get the number of known modules in this server
   * @return number of known modules
   */
  public int getModuleCount() { return loadedModules.size(); }

  private int getNamedModuleIndex(String moduleName) {
    synchronized(loadedModules) {
      int moduleCount = loadedModules.size();
      for (int moduleIndex = 0; moduleIndex < moduleCount; moduleIndex++) {
        if (loadedModules.get(moduleIndex).getModuleConfig().getModuleName().equalsIgnoreCase(moduleName))
          return moduleIndex;
      }
    }
    
    return -1;
  }
  
  /**
   * Given a passed name, return an instance of the module that matches that name.
   * If the module has not been loaded yet or does not exist, you'll receive null
   * back.
   *
   * @param moduleName name of loaded module to lookup
   * @return instance of the loaded module or null if module not loaded
   */
  public xPL_ModuleI getNamedModule(String moduleName) {
    synchronized(loadedModules) {
      int moduleIndex = getNamedModuleIndex(moduleName);
      if (moduleIndex == -1) return null;
      return loadedModules.get(moduleIndex);
    }
  }

  private int getNamedModuleConfigIndex(String moduleName) {
    synchronized(loadedModules) {
      int moduleCount = moduleConfigs.size();
      for (int moduleIndex = 0; moduleIndex < moduleCount; moduleIndex++) {
        if (moduleConfigs.get(moduleIndex).getModuleName().equalsIgnoreCase(moduleName))
          return moduleIndex;
      }
    }
    
    return -1;
  }
  
  /**
   * Get the configuration definition for the named module.  If there is no 
   * such named module, null is returned
   *
   * @param moduleName Name of module configuration to return
   * @return instance of the named modules configuration or null if there is no such module
   */
  public xPL_ModuleConfigI getNamedModuleConfig(String moduleName) {
    synchronized(loadedModules) {
      int moduleIndex = getNamedModuleConfigIndex(moduleName);
      if (moduleIndex == -1) return null;
      return moduleConfigs.get(moduleIndex);
    }
  }

  /**
   * Get all currently loaded modules in this server
   *
   * @return collection of all loaded modules
   */
  public Collection<xPL_ModuleI> getLoadedModules() { return loadedModules; }

  /**
   * Load the module described by the passed configuration.  If it fails,
   * we return false.  If successful, true is returned.
   *
   * <P>There are a number of reasons loading can fail -- can't find
   * the JAR file, can't load the class, not enabled, etc
   *
   * @param theModuleConfig config to load module from
   * @return true if module has been loaded
   */
  public boolean loadModule(xPL_ModuleConfigI theModuleConfig) {
    if (theModuleConfig == null) return false;
    if (theModuleConfig.isLoaded()) return true;
    if (!theModuleConfig.isEnabled()) return false;
  
    // Make sure we can actually load it
    if (!(theModuleConfig instanceof xPL_ModuleConfig)) return false;
    
    // Attempt to load this
    synchronized(loadedModules) {
      if (((xPL_ModuleConfig) theModuleConfig).loadModule()) {
        int moduleIndex = getNamedModuleIndex(theModuleConfig.getModuleName());
        if (moduleIndex != -1)
          loadedModules.set(moduleIndex, theModuleConfig.getModule());
        else
          loadedModules.add(theModuleConfig.getModule());

        if (managementModule != null) managementModule.sendModuleStatus(theModuleConfig, true);
        return true;
      } 
    }
    
    return false;
  }

  /**
   * Load the named module.  If the named module is not known or there is an
   * error loading the module itself, false is returned.
   *
   * <P>If the module is an auto-start module, it will be started after
   * being successfully loaded.
   *
   * <P>There are a number of reasons loading can fail -- can't find
   * the JAR file, can't load the class, etc.  
   *
   * @param moduleName name of known module to load
   * @return true if module has been loaded
   */
  public boolean loadModule(String moduleName) {
    // Insure we have a chance
    if ((moduleName == null) || (moduleName.length() == 0)) return false;
    
    synchronized(loadedModules) {
      xPL_ModuleConfigI theConfig = getNamedModuleConfig(moduleName);
      if (theConfig == null) return false;
      return loadModule(theConfig);
    }
  }

  /**
   * Attempt to start the passed module. 
   *
   * @param theModule module to start
   * @return true if module started, false if unable to start module
   */
  public boolean startModule(xPL_ModuleI theModule) {
    if (theModule == null) return false;
    if (theModule.isModuleStarted()) return true;
    if (!theModule.getModuleConfig().isEnabled()) return false;

    try {
      theModule.startModule();
      if (theModule.isModuleStarted()) {
        if (managementModule != null) managementModule.sendModuleStatus(theModule.getModuleConfig(), true);
        return true;
      } else {
        return false;
      }

    } catch (Throwable anyError) { 
      try {
        theModule.stopModule();
      } catch (Throwable stopError) { }
      
      System.err.println("XPL_CONT:: ERROR attempting to start module " + theModule.getModuleConfig().getModuleName() + " -- " + anyError.getMessage());
      anyError.printStackTrace();
      
      return false;
    }
  }

  /**
   * Find the named module and attempt to start it.  If the module is
   * known but not yet loaded, we first attempt to load it, then start
   * it.
   *
   * @param moduleName name of module to start
   * @return true if module is now started, false if unable to start
   */
  public boolean startModule(String moduleName) {
    // Insure we have a chance
    if ((moduleName == null) || (moduleName.length() == 0)) return false;
    
    // See if the module is loaded and if not, load it
    xPL_ModuleI theModule = getNamedModule(moduleName);
    if (theModule == null) {
      if (!loadModule(moduleName)) return false;
      if ((theModule = getNamedModule(moduleName)) == null) return false;
    }
    
    // Attempt to start it
    return startModule(theModule);
  }

  /**
   * Attempt to stop the passed module.  Modules should always be
   * stoppable, so this should never fail (unless you pass a null)
   *
   * @param theModule module to stop
   * @return true if module was stopped, false if there was an error
   */
  public boolean stopModule(xPL_ModuleI theModule) {
    if (theModule == null) return false;
    if (!theModule.isModuleStarted()) return true;
    
    try {
      theModule.stopModule();
      if (managementModule != null) {
        managementModule.sendModuleStatus(theModule.getModuleConfig(), true, "STOPPING", null);
        managementModule.sendModuleStatus(theModule.getModuleConfig(), true);
      }
    } catch (Throwable anyException) { return false; }
    
    return true;
  }
  
  /**
   * Attempt to stop the passed named module.  Modules should always
   * be stoppable, so this should only fail if the name you pass
   * does not identify an existing module or is invalid.
   *
   * @param moduleName the module name to stop
   * @return true if module was stopped, false if there was an error
   */
  public boolean stopModule(String moduleName) {
    // Insure we have a chance
    if ((moduleName == null) || (moduleName.length() == 0)) return false;
    
    // See if the module is loaded and if not, fail
    xPL_ModuleI theModule = getNamedModule(moduleName);
    if (theModule == null) return true;
    
    return stopModule(theModule);
  }

  /**
   * Attempt to unload the passed module.  If the module was started,
   * it will first be stopped, then unloaded.  Modules should always
   * be unloadable, so this should never fail (unless you pass a null)
   *
   * @param theModule module to unload
   * @return true if module is unloaded
   */
  public boolean unloadModule(xPL_ModuleI theModule) {
    if (theModule == null) return false;
    
    synchronized(loadedModules) {
      // Stop & Unload the module
      xPL_ModuleConfigI theConfig = theModule.getModuleConfig();
      if (theConfig != null) {
        // See if it needs to be stopped first
        if (theModule.isModuleStarted()) {
          try {
            theModule.stopModule();
          } catch (Throwable stopException) { }
        }

        // Now unload the module
        try {
          if (managementModule != null) managementModule.sendModuleStatus(theConfig, true, "UNLOADING", null);
          theModule.unloadModule();
        } catch (Throwable unloadException) {
          error("Got an error while unloading module [" + theConfig.getModuleName() + "] -- " + unloadException.getMessage());
          unloadException.printStackTrace();
          return false;
        } finally {
          // Clear the loaded instance from the config
          if (theConfig instanceof xPL_ModuleConfig) ((xPL_ModuleConfig) theConfig).loadedModule = null;

          // Send out final state
          try {
            if (managementModule != null) managementModule.sendModuleStatus(theConfig, true);
          } catch (Throwable anythingAtAll) { }

          // Look the module up & remove it from our list
          loadedModules.remove(theModule);
        }
      }
    }
    
    // And we are done
    return true;
  }

  /**
   * Attempt to unloaded the passed named module.  If the module was started,
   * it will first be stopped, then unloaded.  Modules should always be 
   * unloadable, so this should never fail unless you pass a bad module 
   * name
   *
   * @param moduleName name of module to unload
   * @return true if module was unloaded, false if there was an error
   */
  public boolean unloadModule(String moduleName) {
    // Insure we have a chance
    if ((moduleName == null) || (moduleName.length() == 0)) return false;
    
    // See if the module is loaded and if not, fail
    synchronized(loadedModules) {
      xPL_ModuleI theModule = getNamedModule(moduleName);
      if (theModule == null) return true;
      return unloadModule(theModule);
    }
  }

  /**
   * Reload the named module.  The module, if loaded or started, is stopped
   * and unloaded, then reloaded.  
   *
   * @param moduleName module name to reload
   * @return true if reloaded, false if module unknown, there was an error, etc
   */
  public boolean reloadModule(String moduleName) {
    // See if we should reload module
    xPL_ModuleConfigI theModule = getNamedModuleConfig(moduleName);
    if (theModule == null) return false;

    boolean moduleWasLoaded = theModule.isLoaded();
    boolean moduleWasStarted = theModule.isStarted();
    
    // Unload module, if it exists
    if (!unloadModule(moduleName)) return false;

    // See if we should re-load it
    if (theModule.isAutoLoaded() || moduleWasLoaded || moduleWasStarted) {
      if (!loadModule(theModule)) return false;
    }
    
    // See if we should restart it
    if (theModule.isAutoStarted() || moduleWasStarted) {
      if (!startModule(theModule.getModule())) return false;
    }
    
    // We are done
    return true;
  }
  
  /** 
   * Stop all known modules and unload them.  THis is generally done only as a
   * prelude to shutdown.
   *
   */
  public void unloadAllModules() {
    for (xPL_ModuleConfigI theConfig: moduleConfigs) {
      unloadModule(theConfig.getModule());
    }
  }
  
  ModuleDefaultsParser getModuleDefaultsParser() { return moduleDefaultsParser; }

  // Ensure the passed modules state matches the configuration
  // state.  If not (in particular, if it's enabling has changed), alter
  // the module to match that state
  void ensureConsistantModuleState(xPL_ModuleConfigI moduleConfig) {
    // See if this module is disabled but was running
    if (!moduleConfig.isEnabled()) {
      // If this module had been the management console, clear it out
      if (moduleConfig == managementModuleConfig) managementModule = null;

      // If this is started or loaded, stop it and unload it
      if (moduleConfig.isLoaded()) {
        stopModule(moduleConfig.getModule());
        unloadModule(moduleConfig.getModule());
      }

      return;
    }

    // Insure that this is in fact loaded
    if (moduleConfig.isAutoLoaded() && !moduleConfig.isLoaded()) {
      if (!loadModule(moduleConfig)) return;
    }

    // Insure that this is in fact started

    // See if we should auto-start it
    if (moduleConfig.isAutoLoaded() && moduleConfig.isAutoStarted() && !moduleConfig.isStarted()) {
      if (!startModule(moduleConfig.getModule())) return;
    } 
  
    // Install management hook, if possible
    if (moduleConfig == managementModuleConfig) managementModule = (ServerManagerModule) moduleConfig.getModule();
  }


  /** Check all modules that are known to exist and insure that
   * their actual state matches their defined state
   */
  public void ensureConsistantModuleStates() {
    for (xPL_ModuleConfigI moduleConfig: moduleConfigs) {
      ensureConsistantModuleState(moduleConfig);
    }
  }
  
  // Scan for the directory the passed targetFile is in from 
  // the list of paths passed.  If found, the path to the directory
  // (but not to the file) is returned.  If not found on any of
  // of the passed paths, null is returned
  private File scanForFileInPaths(String pathList[], String targetFileName) {
    targetFileName = targetFileName.toLowerCase();
    
    for (int pathIndex = 0; pathIndex < pathList.length; pathIndex++) {
      // Convert into a file and skip if it doesn't exist
      File theFile = new File(pathList[pathIndex]);
      if (!theFile.exists()) continue;
      
      // If this is a file, see if it's us
      if (theFile.isFile()) {
        if (pathList[pathIndex].toLowerCase().indexOf(targetFileName) != -1) {
          try {
            return new File(theFile.getAbsoluteFile().getParentFile().getAbsolutePath());
          } catch (Exception ignoredError) { }
        }
        
        continue;
      }
      
      // If this isn't a directory, we don't want to see it
      if (!theFile.isDirectory()) continue;
      
      // Scan the directory for this file
      try {
        final String matchFileName = targetFileName;
        File[] matchingFiles = theFile.listFiles(new FileFilter() {
          public boolean accept(File thePath) {
            return (thePath.getName().equalsIgnoreCase(matchFileName));
          }
        });
        
        // See if we got anything
        if ((matchingFiles == null) || (matchingFiles.length == 0)) continue;
        
        // Okay, return the first directory we have
        return theFile;
      } catch (Exception anyError) { continue; }
    }
    
    return null;
  }

  private String trimTrailingChar(String theString, String trailingChar) {
    while (theString.endsWith(trailingChar)) theString = theString.substring(0, theString.length() - 1);
    return theString;
  }

  public String getInstallPath() {
    File theInstallPath = null;
    
    // Handle cached install
    if (installationPath != null) return installationPath;
    
    // Try the class path
    String theClassPaths[] = System.getProperty("java.class.path").split(File.pathSeparator);
    if ((theClassPaths != null) && (theClassPaths.length != 0)) {
      if ((theInstallPath = scanForFileInPaths(theClassPaths, "xPL4Java.jar")) != null) {
        installationPath = theInstallPath.getAbsolutePath();
        installationPath = trimTrailingChar(installationPath, ".");
        installationPath = trimTrailingChar(installationPath, File.separator);
        return installationPath;
      } 
    }
     
    // Try the current directory
    if ((theInstallPath = scanForFileInPaths(new String[] { System.getProperty("user.dir") }, "xPL4Java.jar")) != null) {
      installationPath = theInstallPath.getAbsolutePath();
      installationPath = trimTrailingChar(installationPath, ".");
      installationPath = trimTrailingChar(installationPath, File.separator);
      return installationPath;
    }

    // Try the users home directory
    if ((theInstallPath = scanForFileInPaths(new String[] { System.getProperty("user.home") }, "xPL4Java.jar")) != null) {
      installationPath = theInstallPath.getAbsolutePath();
      installationPath = trimTrailingChar(installationPath, ".");
      installationPath = trimTrailingChar(installationPath, File.separator);
      return installationPath;
    }
    
    // If we get here, we could not find it - fail!
    System.err.println("XPL_SERVER:: Unable to workout where xPL4Java.jar is installed");
    return null;
  }

  /**
   * Start XPL server.  In succession, this does the following:
   * 
   * <LI> Restores the configuration for all known modules (if any)
   * <LI> Loads all in the jars/ subdirectory
   * <LI> Starts the underlying xPL network for each defined interface
   * <LI> Loads all modules that are set to auto-load
   * <LI> Starts all modules that are set to auto-start
   *
   * @return true if server is starting, false if there was a failure
   */
  public boolean startXPLServer() {
    // Skip if already running
    if (serverRunning) return true;
    
    System.out.println("xPL4Java server V" + xPL_Manager.XPL4JAVA_VERSION + " copyright (c) 2005, Gerald R Duprey Jr.");
    System.out.println("Starting up with Java version " + System.getProperty("java.version"));
    System.out.println("More info at http://www.xPL4Java.org");
    
    // Flag us as running
    serverRunning = true;
    
    // Create the module tracker
    moduleTracker.initializeModuleTracker();
    moduleTracker.scanClassPathJarsForModules();
    
    // Load all intially known modules
    ensureConsistantModuleStates();
    
    // Add a listener for hub updates.  In the unlikely event
    // that all media handlers do not have hubs (and thus respond
    // that hubs are ready immediatly), check the status after
    // installing a listener
    xPL_Manager.getManager().addManagerListener(this);
    if (xPL_Manager.getManager().areHubsConfirmed()) {
      startModuleTracking();
    }
    
    // And we are done
    return true;
  }
  
  // We dont' start this until all hubs are responding
  synchronized void startModuleTracking() {
    if (trackerRunning) return;
    xPL_Manager.getManager().removeManagerListener(this);
    managementModule.startModulePathUpdates();
    trackerRunning = true;
  }
  
  public void handleXPLManagerEvent(xPL_ManagerEvent theEvent) {
    if (theEvent.getEventCode() == xPL_ManagerEvent.EventCode.ALL_HUBS_DETECTED) startModuleTracking();
  }
  
  
  /**
   * Shutdown the XPL server.  In succession, this does the following
   *
   * <LI> Stops any running modules
   * <LI> Unloads any loaded modules
   * <LI> Shutdown the networking modules/xPL
   * <LI> Releases resources
   *
   * the XPL_Server can be restarted after it has been stopped, if 
   * you need to (i.e. it's restartable)
   */
  public void stopXPLServer() {
    if (!serverRunning) return;
    
    // Now stop the modules (in reverse order)
    xPL_ModuleI theModule = null;
    for (int moduleIndex = loadedModules.size() - 1; moduleIndex >= 0; moduleIndex--) {
      theModule = loadedModules.get(moduleIndex);
      
      // Stop this
      if (theModule.isModuleStarted()) stopModule(theModule);
      
      // Unload it
      unloadModule(theModule);
    }
    
    serverRunning = false;
  }

  /**
   * Main entry point for the server.  
   * 
   * You can pass the config file to use via the first parameter to override the default
   * @param args the command line arguments
   */
  public static void main(String[] args) {
    // Install config file
    
    // Start things up
    theXPLServer.startXPLServer();
  }
}
