/*
 * TrackedDevice.java
 *
 * Created on September 21, 2005, 4:22 PM
 *
 */

package org.cdp1802.xpl.tracker;

import java.util.*;
import org.cdp1802.xpl.*;
import org.cdp1802.xpl.device.xPL_DeviceConfigItem;
import org.cdp1802.xpl.device.xPL_DeviceConfigItemI;

/**
 * Contain everything we know about a device we are tracking.  This also allows
 * us to send configuration data out for this device as well.
 *
 * @author gerry
 */
public class TrackedDevice { 
  static String deviceSemaphore = "SEMAPHORE";
  static IdentifierManager identifierManager = xPL_Manager.getManager().getIdentifierManager();
  static TrackedDeviceManager trackedDeviceManager = TrackedDeviceManager.getTrackedDeviceManager();
  static xPL_MutableMessageI configureMessage = null;
  
  private TrackedDevice originalDevice = null;
  private xPL_IdentifierI deviceIdent = null;
  private long lastHeartBeatAt = 0;
  private long discoveredAt = 0;
  
  private boolean deviceConfigurable = false;
  private boolean deviceConfigured = true;
  private boolean hasConfigItemsDefined = false;
  private boolean hasConfigValuesDefined = false;

  private TreeMap<String, xPL_DeviceConfigItemI> configItems = new TreeMap<String, xPL_DeviceConfigItemI>();

  private xPL_DeviceConfigItemI filterItem = null;
  private xPL_DeviceConfigItemI groupItem = null;
  
  private int heartBeatInterval = 0;
  private int remotePort = 0;
  private String remoteIPAddr = null;
  private String deviceVersion = "";

  public TrackedDevice(xPL_IdentifierI theIdent) {
    deviceIdent = theIdent;
    discoveredAt = System.currentTimeMillis();
    
    configItems.put("FILTER", new xPL_DeviceConfigItem("filter", 16, false, true));
    configItems.put("GROUP", new xPL_DeviceConfigItem("group", 16, false, true));
  }

  public TrackedDevice() { }
  
  /**
   * Return the identifer for this tracked device
   *
   * @return identifier of this device
   */
  public xPL_IdentifierI getIdent() { return deviceIdent; }

  /**
   * Return the time, as represented in milliseconds, we last heard from the
   * device.  Note that if we have never heard from the device, this is 0.
   *
   * @return time last heard from the device
   */
  public long getLastHeartBeat() { return lastHeartBeatAt; }
  
  /**
   * Return the time, as represented in milliseconds, that we first heard of
   * this device.  Note that if we have never heard from the device, this is 0
   *
   * @return time first heard from the device
   */
  public long getDiscoveryTime() { return discoveredAt; }

  /**
   * Return the heartbeat interval rate, in minutes, as published by the device
   *
   * @return heartbeat interval, in minutes, of this device
   */
  public int getHeartBeatInterval() { return heartBeatInterval; }

  /**
   * Return the remote port # for this device.  If the device is not a network
   * device, this will be 0.
   *
   * NOTE: This is informational only -- you should never attempt to contact
   * the device directly on this port.
   *
   * @return remote port # used by a networked device
   */
  public int getRemotePort() { return remotePort; }
  
  /**
   * Get the remote IP address that is associated with this device.  If the
   * device is not a network device, this will be null
   *
   * NOTE: This is informational only -- you should never attempt to contact
   * the device directly on this port
   *
   * @return remote IP address of the networked device or null
   */
  public String getRemoteIPAddr() { return remoteIPAddr; }
  
  /**
   * Get the version string for this device, if any.  Versions are optional
   * and not all devices will have them
   *
   * @return device version string or an empty string if no version info available
   */
  public String getDeviceVersion() { return deviceVersion; }
  
  /**
   * True if device is configurable or not.  This can change as 
   * we learn more about the device
   *
   * @return true if configurable, false if not
   */
  public boolean isConfigurable() { return deviceConfigurable; }
  
  /**
   * True if the device is either not configurable or is configurable
   * and we have received a configuration for it.
   *
   * @return true if configured, false if not
   */
  public boolean isConfigured() { return deviceConfigurable && deviceConfigured; }
  
  /**
   * True if configuration items definitions (though maybe not values) 
   * have been received.
   *
   * NOTE: This only makes sense on configured, configurable 
   * devices.
   *
   * @return true if there are at least one config item, false if not
   */
  public boolean hasConfigItems() { return hasConfigItemsDefined; }
  
  /**
   * True if configuration item definitions and values have been
   * received
   *
   * NOTE: This only makes sense on configured, configurable 
   * devices.
   *
   * @return true if we have config values, false if not
   */
  public boolean hasConfigValues() { return hasConfigValuesDefined; }
 
  /**
   * Get the maximum number of allowed filters for this device
   *
   * NOTE: This only makes sense on configured, configurable 
   * devices.
   *
   * @return max number of allowed filters
   */
  public int getMaxFilterCount() { 
    if (filterItem == null) return 0;
    return filterItem.getMaxValueCount();
  }
  
  /**
   * Get the current number of defined filters for this 
   * device.  
   *
   * NOTE: This only makes sense on configured, configurable 
   * devices.
   *
   * NOTE: To get actual filters, get the config item named "filter"
   *
   * @return number of currently defined filters
   */
  public int getFilterCount() { 
    if (filterItem == null) return 0;
    return filterItem.getValueCount();
  }
  
  /**
   * Get the maximum number of allowed groups for this device
   *
   * NOTE: This only makes sense on configured, configurable 
   * devices.
   *
   * @return max number of allowed groups
   */
  public int getMaxGroupCount() { 
    if (groupItem == null) return 0;
    return groupItem.getMaxValueCount();
  }

  /**
   * Get the current number of defined groupss for this 
   * device.  
   *
   * NOTE: This only makes sense on configured, configurable 
   * devices.
   *
   * NOTE: To get actual groups, get the config item named "group"
   *
   * @return number of currently defined groups
   */
  public int getGroupCount() {
    if (groupItem == null) return 0;
    return groupItem.getValueCount();
  }

  /**
   * Return the number of available configuration items for this
   * device.  
   *
   * NOTE: This only makes sense on configured, configurable 
   * devices.
   *
   * @return number of configuable items for the device
   */
  public int getConfigItemCount() { return configItems.size(); }

  /**
   * Get a configurable item by name.  If the item does not exist
   * then return null.  Lookups are case insensitive.
   *
   * NOTE: This only makes sense on configured, configurable 
   * devices.
   *
   * @param configItemName name of config item to lookup
   * @return config item of that name or null
   */
  public xPL_DeviceConfigItemI getConfigItem(String configItemName) {
    return configItems.get(configItemName.toLowerCase());
  }
  
  /** 
   * Return a collection of configuration items
   *
   * NOTE: This only makes sense on configured, configurable 
   * devices.
   *
   * @return collection of known configuration items
   */
  public Collection<xPL_DeviceConfigItemI> getConfigItems() { return configItems.values(); }
  
  // Parse the passed message as a heartbeat
  void parseHeartBeat(xPL_MessageI theMessage) {
    // Skip unless it's a heartbeat
    if (!theMessage.isHeartbeat()) return;
    
    // Update last time we heard about this
    lastHeartBeatAt = System.currentTimeMillis();

    // Update interval, if supplied
    int newValue = theMessage.getIntNamedValue("interval");
    if (newValue != -1) heartBeatInterval = newValue;
      
    // Update remote port, if supplied
    if ((newValue = theMessage.getIntNamedValue("port")) != -1) {
      remotePort = newValue;
    }
      
    // Get the remote IP address, if any
    String newAddr = theMessage.getNamedValue("remote-ip");
    if ((newAddr != null) && (newAddr.length() != 0)) remoteIPAddr = newAddr;

    // Parse version info
    String newVersion = theMessage.getNamedValue("version");
    if ((newVersion != null) && (newVersion.length() != 0)) 
      deviceVersion = newVersion;
    else
      deviceVersion = "";
    
    // See if it needs to be configured
    if (theMessage.getSchemaClass().equalsIgnoreCase("config")) {
      deviceConfigured = false;  // If the schema class is config, it clearly needs to be configured
      deviceConfigurable = true; // And it's configurable
    } else {
      deviceConfigured = true;   // Regular heartbeat means it does not need to be configured anymore
    }
  }
  
  // Parse the passed message as a defintion of configurables
  void parseConfigDefinitions(xPL_MessageI theMessage) {
    // Now we pick the config item definitions off
    String theName = null, theValue = null, theParam = null;
    int maxValues = -1, startDelim = -1, endDelim = -1;
    boolean isMandatory = false, isReconfigurable = false;
    xPL_DeviceConfigItemI configItem = null;

    // Clear all existing configuration
    Iterator<xPL_DeviceConfigItemI> itemList =  configItems.values().iterator();
    xPL_DeviceConfigItemI theItem = null;
    while (itemList.hasNext()) {
      theItem = itemList.next();
      
      if (theItem.getName().equalsIgnoreCase("filter") || theItem.getName().equalsIgnoreCase("group")) continue;

      // Release anything else and remove it
      theItem.releaseResources();
      itemList.remove();
    }
    
    Iterator<NamedValueI> allNamedValues = theMessage.getMessageBody().getAllNamedValues().iterator();
    NamedValueI theNamedValue = null;
    while(allNamedValues.hasNext()) {
      theNamedValue = allNamedValues.next();
      theName = theNamedValue.getName();
      theValue = theNamedValue.getValue();
      
      // See if these are configuration item definitions and skip if not
      if (theName.equalsIgnoreCase("config")) {
        isMandatory = true;
        isReconfigurable = false;
      } else if (theName.equalsIgnoreCase("reconf")) {
        isMandatory = true;
        isReconfigurable = true;
      } else if (theName.equalsIgnoreCase("option")) {
        isMandatory = false;
        isReconfigurable = true;
      } else {
        // Not one of our tags
        continue;
      }
        
      // Take our tag out of the list of tags
      allNamedValues.remove();
      
      // Get the config item name and parse for a max values indicator
      maxValues = 1;
      if ((startDelim = theValue.indexOf("[")) != -1) {
        // Skip if no end delimiter
        if ((endDelim = theValue.indexOf("]", startDelim + 1)) == -1) continue;
        theParam = theValue.substring(startDelim + 1, endDelim);
        theValue = theValue.substring(0, startDelim);
        if (theParam.length() != 0) {
          try {
            maxValues = Integer.parseInt(theParam);
          } catch (Exception anError) { continue; }
        } else {
          maxValues = 1;
        }
      }

      // Get existing config item, if any, or create one
      if ((configItem = getConfigItem(theValue)) == null) {
        configItem = new xPL_DeviceConfigItem(theValue, maxValues, isMandatory, isReconfigurable);
        configItems.put(configItem.getName().toLowerCase(), configItem);
      } else {
        configItem.setMaxValueCount(maxValues);
        configItem.setMandatory(isMandatory);
        configItem.setReconfigurable(isReconfigurable);
      }
    }  
    
    // Set flags
    deviceConfigurable = true;       // We are assuredly configurable
    hasConfigItemsDefined = true;    // We have defitions
    hasConfigValuesDefined = false;  // But no values
  }

  // Parse the passed message as config values
  void parseConfigValues(xPL_MessageI theMessage) {
    NamedValuesI configValues = theMessage.getMessageBody();
    
    // See if there is a new instance name
    String instanceID = configValues.getNamedValue("newconf");
    if ((instanceID != null) && !instanceID.equalsIgnoreCase(deviceIdent.getInstanceID())) {
      // Attempt to update/reconfigure the device and if we do, flag us needing a new heartbeat
      xPL_IdentifierI newDevice = identifierManager.reconfigureInstanceID(deviceIdent, instanceID);
      if (newDevice != null) deviceIdent = newDevice;
    }      

    // See if there is a new interval
    int newInterval = configValues.getIntNamedValue("interval", -1);
    if (newInterval != -1) heartBeatInterval = newInterval;

    // Iterate over all the remaining values and install them as config
    // items, if they in fact match a config item
    HashMap<String,String> uniqueItemList = new HashMap<String,String>();
    xPL_DeviceConfigItemI theConfigItem = null;
    String theName = null;
      
    for(NamedValueI namedValue: configValues.getAllNamedValues()) {
      // Get the config item and cache the name
      theName = namedValue.getName();

      // Lookup the config item.  If it's unknown, skip it
      if ((theConfigItem = configItems.get(theName.toLowerCase())) == null) continue;

      // If this is the first instance of this name, purge any values
      if (!uniqueItemList.containsKey(theName)) {
        uniqueItemList.put(theName, theName);
        theConfigItem.clearValues();
      }
      
      // Add the value on
      if ((namedValue.getValue() != null) && (namedValue.getValue().length() != 0)) {
        theConfigItem.addValue(namedValue.getValue());
      }
    }

    // Get filter/group hooks
    filterItem = configItems.get("FILTER");
    groupItem = configItems.get("GROUP");
    
    // Set Flags
    deviceConfigurable = true;      // We are assuredly configurable
    hasConfigValuesDefined = true;  // And we have config values
  }

  /**
   * This will send all current configuration values for a device to that
   * device and then immediatly request a reload of configuration values
   * to check on which were installed and which were not
   *
   * NOTE: The TrackedDeviceManager must have a device installed, the device
   * must be enabled and the tracked device manager must be active or this is
   * ignored.
   */
  public void sendConfigValues() {
    synchronized(deviceSemaphore) {
      // Insure there is a device and it's enabled
      if (!trackedDeviceManager.isTrackingActive() || (trackedDeviceManager.getTrackingDevice() == null) || !trackedDeviceManager.getTrackingDevice().isEnabled()) return;
      
      // Create a message, if needed
      if (configureMessage ==  null) {
        configureMessage = trackedDeviceManager.getTrackingDevice().createMessage(xPL_MessageI.MessageType.COMMAND, "config", "response");
      }
      
      // Clear the message and set the target
      configureMessage.clearMessageBody();
      configureMessage.setTarget(deviceIdent);
      
      // Now include everything else
      String configName = null;
      for (xPL_DeviceConfigItemI configItem: configItems.values()) {
        configName = configItem.getName();

        if (configItem.getValueCount() == 0) {
          configureMessage.addNamedValue(configName, "");
        } else {
          for (String aValue: configItem.getValues()) {
            configureMessage.addNamedValue(configName, aValue);
          }
        }
      }
    
      // Send the message
      trackedDeviceManager.getTrackingDevice().sendMessage(configureMessage);
      
      // Now we request a re-send of the device values
      trackedDeviceManager.sendConfigValueRequest(this);
    }
  }

  /** 
   * Copy the values of this TrackedDevice into the passed tracked 
   * device instance.  Any information in the passed tracked device is
   * lost.  You should only use tracked device instances you created 
   * via copyOf() or "new TrackedDevice" invocations.  Using any other
   * instance of a TrackedDevice as the target is sure to break things
   * badly
   *
   * @param newDevice device to copy values from this device into
   */
  public void copyInto(TrackedDevice newDevice) {
    // Insure config items are clean
    if (newDevice.configItems == null)
      newDevice.configItems = new TreeMap<String, xPL_DeviceConfigItemI>();
    else 
      newDevice.configItems.clear();
        
    // Copy simple/shallow things over  
    newDevice.deviceIdent = deviceIdent;
    newDevice.lastHeartBeatAt = lastHeartBeatAt;
    newDevice.discoveredAt = discoveredAt;
    newDevice.deviceConfigurable = deviceConfigurable;
    newDevice.deviceConfigured = deviceConfigured;
    newDevice.hasConfigItemsDefined = hasConfigItemsDefined;
    newDevice.hasConfigValuesDefined = hasConfigValuesDefined;
    newDevice.filterItem = null;
    newDevice.groupItem = null;
    newDevice.heartBeatInterval = heartBeatInterval;
    newDevice.remotePort = remotePort;
    newDevice.remoteIPAddr = remoteIPAddr;
    newDevice.deviceVersion = deviceVersion;
    
    // Only set original device if we are copying from an original device
    newDevice.originalDevice = (this.originalDevice == null ? this : null);
  
    // Copy config items
    if (configItems != null) {
      xPL_DeviceConfigItem newItem = null;

      for (xPL_DeviceConfigItemI existingItem: configItems.values()) {
        newItem = new xPL_DeviceConfigItem(existingItem.getName(), existingItem.getMaxValueCount(), existingItem.isMandatory(), existingItem.isReconfigurable());
        for (String aValue: existingItem.getValues()) {
          newItem.addValue(aValue);
        }
      
        // Add to the list
        newDevice.configItems.put(newItem.getName().toLowerCase(), newItem);
      }
    }
    
    // Get hooks on filter/group items
    newDevice.filterItem = newDevice.configItems.get("FILTER");
    newDevice.groupItem = newDevice.configItems.get("GROUP");
  }

  /**
   * Create a copy of this tracked device and return it.  The copy is not
   * part of any device tracking, so subsequent changes to the original 
   * device will NOT be reflected here.
   *
   * You should use the releaseResources() method on the copy when you
   * are done with it
   *
   * @return copy of this TrackedDevice
   */
  public TrackedDevice copyOf() {
    TrackedDevice newDevice = new TrackedDevice();
    copyInto(newDevice);
    return newDevice;
  }

  /**
   * Return the original device if this instance is a copy of another 
   * device (created via copyInfo or copyOf() methods).  If this is
   * the original device, null is returned
   *
   * @return original version of this device or null
   */
  public TrackedDevice getOrignalDevice() { return originalDevice; }

  
  /**
   * Release all resources related to this device.  You should only call
   * this if *YOU* created the device (i.e. had a new TrackedDevice() or
   * used copyOf()).  If you call it for any other reason, you'll almost
   * assuredly bust the entire TrackedDeviceManager.  You have been warned.
   */
  public void releaseResources() {
    deviceIdent = null;
    remoteIPAddr = null;
    deviceVersion = null;
    
    filterItem = null;
    groupItem = null;
    
    if (configItems != null) {
      for (xPL_DeviceConfigItemI theItem: configItems.values()) {
        theItem.releaseResources();
      }
      configItems.clear();
    }
    configItems = null;
    originalDevice = null;
  }
  
  /**
   * Return true if the device passed is exactly us (same instance)
   *
   * @param anyObject oject to compare us with
   * @return true if this is the same object, false otherwise
   */
  public boolean equals(Object anyObject) {
    if (!(anyObject instanceof TrackedDevice)) return false;
    return (anyObject == this);
  }

  /**
   * Return a string describing this tracked devices vendor
   *
   * @return vendorID of this tracked device, if any
   */
  public String toString() {
    if (deviceIdent != null) return deviceIdent.getVendorID();
    return "Unknown";
  }
}
