/*
 * TrackedDeviceManager.java
 *
 * Created on September 21, 2005, 8:27 PM
 *
 */
package org.cdp1802.xpl.tracker;

import java.util.*;

import org.cdp1802.xpl.*;
import org.cdp1802.xpl.device.xPL_DeviceI;

/**
 * TrackedDeviceManager takes care of discovering new devices, tracking devices and
 * generally keeping on top of the xPL network
 * @author gerry
 */
public class TrackedDeviceManager implements xPL_MessageListenerI, Runnable {
  private static TrackedDeviceManager trackedDeviceManager = new TrackedDeviceManager();
  public static TrackedDeviceManager getTrackedDeviceManager() { return trackedDeviceManager; }
  
  boolean trackingActive = false;
  boolean autoCreatedDevice = false;
  xPL_DeviceI trackerDevice = null;
  xPL_MutableMessageI commandRequestMessage = null;
  xPL_MutableMessageI discoveryMessage = null;
  HashMap<String, TrackedDevice> deviceList = new HashMap<String, TrackedDevice>();
  ArrayList<TrackedDeviceListenerI> managerListenerList = new ArrayList<TrackedDeviceListenerI>();

  Thread trackerThread = null;
  boolean trackerThreadRunning = false;
  
  private TrackedDeviceManager() { }
  
  /**
   * Return a synchronizable() semaphor for the tracker.  Synchronizing on this
   * will alow you to consult the tracked devices without them changing for
   * the duration.  Obviously, you should not hang on to this moment longer
   * than you need to.
   *
   * @return object to synchronize tracker device list operations
   */
  public Object getTrackerSemaphore() { return deviceList; }
  
  /**
   * Start the device tracking process.  If there is no device installed, 
   * we'll create one and enable it (if there is one installed, it should be
   * enabled before tracking is started).
   */
  public void startTracking() {
    if (trackingActive) return;

    // Create a device if we need one
    if (trackerDevice == null) {
      trackerDevice = xPL_Manager.getManager().getDeviceManager().createDevice("cdp1802", "devtrack", null);
      trackerDevice.setEnabled(true);
      autoCreatedDevice = true;
    }
    
    // Create a device if we need one
    commandRequestMessage = trackerDevice.createMessage(xPL_MessageI.MessageType.COMMAND, "config", "list");
    commandRequestMessage.addNamedValue("command", "request");
    
    // Install message listener
    xPL_Manager.getManager().addMessageListener(this);
    
    // And we're running
    trackingActive = true;
    
    // Create the thread, if needed, and start if
    trackerThread = new Thread(this, "Timedout Device Tracker");
    trackerThreadRunning = true;
    trackerThread.start();
  }
  
  /**
   * Send a message out asking all devices to identify themselves immediatly for
   * device discovery.
   *
   * NOTE: This is NOT automatically done when the tracker is started as the 
   * tracker may be started and stopped multiple times.  You should execute
   * this if you want immediate (vs eventual) device discovery when your
   * app is ready for it.
   *
   * NOTE: This can put a load on the network.  Please don't use it unless you
   * really need to.
   */
  public void sendDiscoveryRequest() {
    if (discoveryMessage == null) {
      discoveryMessage = trackerDevice.createBroadcastMessage(xPL_MessageI.MessageType.COMMAND, "hbeat", "request");
      discoveryMessage.addNamedValue("command", "request");
    }
    
    // Send the message out
    trackerDevice.sendMessage(discoveryMessage);
  }
  
  /**
   * Stop the tracking process.  The device, if we created it, is disabled and
   * we stop listening for new trackable devices
   */
  public void stopTracking() {
    // Test/clear the active flag
    if (!trackingActive) return;
    trackingActive = false;
  
    // Stop listening
    xPL_Manager.getManager().removeMessageListener(this);
    
    // Stop the device, if we created it
    if (autoCreatedDevice) trackerDevice.setEnabled(false);

    // Stop the thread
    synchronized(deviceList) {
      deviceList.notifyAll();
    }
    
    // Wait for the thread toend
    try {
      trackerThread.join();
    } catch (Exception anything) { }
    
    // Release resources
    commandRequestMessage.releaseResources();
    commandRequestMessage = null;
  }
  
  /**
   * Determine if device tracking is active or not
   *
   * @return true if we are tracking devices, false otherwise
   */
  public boolean isTrackingActive() { return trackingActive; }

  /**
   * Install a device to use for sending out messages involved with
   * device tracking.  If you pass null here, then a device will be
   * automatically created (which is the default).  If you pass a 
   * device, you must insure it's enabled BEFORE you call startTracker
   *
   * NOTE: Any attempt to change the device while tracking is active
   * is rejected.
   *
   * @param deviceToTrackWith device to use or null to auto-create device as needed
   */
  public void setTrackingDevice(xPL_DeviceI deviceToTrackWith) {
    if (trackingActive || (deviceToTrackWith == trackerDevice)) return;
    
    // Install it
    trackerDevice = deviceToTrackWith;
    autoCreatedDevice = false;
  }
  
  /**
   * Return the device being used for the tracker.  If you call this before
   * you call startTracker(), you may get a null back as if the tracker will
   * need to create an automatic device, it's not done until that time.
   *
   * @return device tracker is using, possibly null if tracker not started
   */
  public xPL_DeviceI getTrackingDevice() { return trackerDevice; }
  
  /**
   * Return the number of devices that we are currently tracking.
   *
   * @return number of devices currently known/tracked
   */
  public int getDeviceCount() { return deviceList.size(); }

  /** 
   * Return the number of devices that we are tracking that are known
   * to be configured.  Configured means either that the device is 
   * configurable and has received values or that the device is NOT
   * configurable (and thus is as configured as it's going to get)
   *
   * @return number of configured, tracked devices
   */
  public int getConfiguredDeviceCount() {
    int theCount = 0;
    synchronized(deviceList) {
      for (TrackedDevice theDevice: deviceList.values()) {
        if (!theDevice.isConfigurable() || theDevice.isConfigured()) theCount++;
      }
    }
    return theCount;
  }

  /**
   * Return the number of devices that we are tracking that are known
   * to not yet be configured.  That means we know the device is 
   * configurable, but as of yet, it has not received it's first set of
   * configuration values
   *
   * @return number of unconfigured, tracked devices
   */
  public int getUnconfiguredDeviceCount() {
    int theCount = 0;
    synchronized(deviceList) {
      for (TrackedDevice theDevice: deviceList.values()) {
        if (theDevice.isConfigurable() && !theDevice.isConfigured()) theCount++;
      }
    }
    return theCount;
  }

  /**
   * Look for a device by the passed device name and return it or null if
   * we are not currently tracking a device by that name. Lookups are case
   * insensitive.
   *
   * @param deviceID - xPL Device ID to look for (in the vendor-device.instance format)
   * @return TrackedDevice of the given name or null if no known device
   */
  public TrackedDevice getDeviceNamed(String deviceID) { return deviceList.get(deviceID); }

  /**
   * Look for a device by the passed device ident and return it or null if
   * we are not currently tracking a device by that identification.
   *
   * @param deviceID - xPL Device ID to look for
   * @return TrackedDevice of the given name or null if no known device
   */
  public TrackedDevice getDeviceNamed(xPL_IdentifierI deviceID) { return deviceList.get(deviceID.toString()); }
  
  /**
   * Return a collection of tracked devices
   *
   * @return collection of known, tracked devices
   */
  public Collection<TrackedDevice> getDevices() { return deviceList.values(); }

  // Send out a request for config definitions
  void sendConfigListRequest(TrackedDevice theDevice) {
    commandRequestMessage.setTarget(theDevice.getIdent());
    commandRequestMessage.setSchemaType("list");
    trackerDevice.sendMessage(commandRequestMessage);
  }
  
  // Send out a request for config values
  void sendConfigValueRequest(TrackedDevice theDevice) {
    commandRequestMessage.setTarget(theDevice.getIdent());
    commandRequestMessage.setSchemaType("current");
    trackerDevice.sendMessage(commandRequestMessage);
  }

  /**
   * Add a new, TrackedDeviceManager listener.  This listener will any changes to 
   * managed devices
   * 
   * @param theListener listener to receive TrackedDeviceManager events
   */
  public void addManagerListener(TrackedDeviceListenerI theListener) {
    synchronized(managerListenerList) {
      if (managerListenerList.contains(theListener)) return;
      managerListenerList.add(theListener);
    }
  }

  /**
   * Remove an existing device manager listener.
   *
   * @param theListener listener to remove
   */
  public void removeManagerListener(TrackedDeviceListenerI theListener) {
    synchronized(managerListenerList) {
      managerListenerList.remove(theListener);
    }
  }
  
  // Fire messages to tracked device listeners
  private void fireDeviceEvent(TrackedDeviceEvent theEvent) {
    int listenerCount = managerListenerList.size();
    for (int listenerIndex = listenerCount - 1; listenerIndex >= 0; listenerIndex--) {
      try {
        managerListenerList.get(listenerIndex).handleTrackedDeviceEvent(theEvent);
      } catch (Throwable ignoredException) { }
    }
    
    // Free event resources
    theEvent.releaseResources();
  }
  
  // Install a new device in the cache
  private TrackedDevice addNewDevice(xPL_MessageI fromMessage) {
    TrackedDevice newDevice = null;
        
    // See if device already exists
    synchronized(deviceList) {
      String deviceIdent = fromMessage.getSource().toString();
      if ((newDevice = deviceList.get(deviceIdent)) != null) return newDevice;

      // Create new device
      newDevice = new TrackedDevice(fromMessage.getSource());
      newDevice.parseHeartBeat(fromMessage);
      deviceList.put(deviceIdent, newDevice);
    }
    
    // Send a request for configuration elements
    sendConfigListRequest(newDevice);
    
    // Fire device added message
    fireDeviceEvent(new TrackedDeviceEvent(TrackedDeviceEvent.EventCode.DEVICE_DISCOVERED, newDevice));
    return newDevice;
  }

  // Remove an existing device from the cache
  private void removeDevice(TrackedDevice theDevice) {
    if (theDevice.getIdent() == null) return;
    
    synchronized(deviceList) {
      // Purge device from list (if not in the list, someone else just purged it)
      if (deviceList.remove(theDevice.getIdent().toString()) == null) return;
    }

    // Fire device removed message
    fireDeviceEvent(new TrackedDeviceEvent(TrackedDeviceEvent.EventCode.DEVICE_SHUTDOWN, theDevice));
    
    // Free resources
    theDevice.releaseResources();
  }
  
  // Update heartbeat info for a known device
  private void recordHeartBeat(TrackedDevice theDevice, xPL_MessageI fromMessage) {
    // Update the device
    theDevice.parseHeartBeat(fromMessage);
    
    // Fire heartbeat detected 
    fireDeviceEvent(new TrackedDeviceEvent(TrackedDeviceEvent.EventCode.HEARTBEAT_RECEIVED, theDevice));
  }

  // Install configuration definitions for a known device
  private void installConfigDefinitions(TrackedDevice theDevice, xPL_MessageI fromMessage) {
    // Install the config definitions
    theDevice.parseConfigDefinitions(fromMessage);
   
    // Fire define config definitions received message
    fireDeviceEvent(new TrackedDeviceEvent(TrackedDeviceEvent.EventCode.CONFIG_DEFINITIONS_RECEIVED, theDevice));
    
    // After definitions, request values
    sendConfigValueRequest(theDevice);
  }
  
  // Install configuaration values for a known device
  private void installConfigValues(TrackedDevice theDevice, xPL_MessageI fromMessage) {
    // Install the config values
    theDevice.parseConfigValues(fromMessage);
    
    // Fire defined config values message
    fireDeviceEvent(new TrackedDeviceEvent(TrackedDeviceEvent.EventCode.CONFIG_VALUES_RECEIVED, theDevice));
  }

  // Check for device timeouts -- that is, devices that we know about but have
  // not heard from in more than twice their stated heartbeat interval
  private void doTimeoutChecks() {
    long rightNow = System.currentTimeMillis();
    int elapsedMinutes = 0;
    
    ArrayList<TrackedDevice> purgeList = null;
    
    // Identify devices we haven't heard from in a while
    synchronized(deviceList) {
      for (TrackedDevice theDevice: deviceList.values()) {
        elapsedMinutes = (int) ((rightNow - theDevice.getLastHeartBeat()) / 60000);
        if (elapsedMinutes > (theDevice.getHeartBeatInterval() * 2)) {
          if (purgeList == null) purgeList = new ArrayList<TrackedDevice>();
          purgeList.add(theDevice);
        }
      }
    }
    
    // Purge the timed out devices (if any)
    if (purgeList != null) {
      for (TrackedDevice theDevice: purgeList) {
        fireDeviceEvent(new TrackedDeviceEvent(TrackedDeviceEvent.EventCode.DEVICE_TIMEDOUT, theDevice));
        removeDevice(theDevice);
      }
    }
  }
  
  // Dispatch a newly received xPL message
  public void handleXPLMessage(xPL_MessageI theMessage) {
    // We only care about status messages
    if (theMessage.getType() != xPL_MessageI.MessageType.STATUS) return;

    // Insure this is only a heartbeat or configuration message -- ignore all others
    if (!theMessage.isHeartbeat() && !theMessage.isConfigurationMessage()) return;
    
    // See if this is a device we know about.  If not, we add it
    TrackedDevice theDevice = deviceList.get(theMessage.getSource().toString());
    if (theDevice == null) {
      // If this is a shutdown heartbeat, no sense adding the device now
      if (theMessage.isHeartbeatEnd()) return;
      if ((theDevice = addNewDevice(theMessage)) == null) return;
    }
    
    // Handle end-of-life
    if (theMessage.isHeartbeatEnd()) {
      removeDevice(theDevice);
      return;
    }
    
    // Handle normal heartbeat
    if (theMessage.isHeartbeat()) {
      recordHeartBeat(theDevice, theMessage);
      return;
    }
    
    // If we get this far, this is a configuration message
    if (theMessage.isConfigurationMessage()) {
      String schemaType = theMessage.getSchemaType().toLowerCase();
      if (schemaType.equals("list")) {
        installConfigDefinitions(theDevice, theMessage);
        return;
      }

      if (schemaType.equals("current")) {
        installConfigValues(theDevice, theMessage);
        return;
      }
    }
  }
  
  public void run() {
    // Insure thread running flag is set
    trackerThreadRunning = true;

    // Do this as long as the tracking is runningt
    for(;;) {
      synchronized(deviceList) {
        // Wait a minute
        try {
          deviceList.wait(60000);
        } catch (Exception anything) { }
      }

      // Insure we are still active 
      if (!trackingActive) break;

      // Check for timedout devices
      try {
        doTimeoutChecks();
      } catch (Throwable theError) {
        System.err.println("TRACKED_DEVMGR:: Error while doing timedout device tracking -- " + theError.getMessage());
        theError.printStackTrace();
      }
    }
    
    // Clear thread running flag
    trackerThreadRunning = false;
  }
}
