/*
 * xPL_Filter.java
 * Copyright 2004, Gerald R Duprey Jr
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy of
 * the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0 
 *
 * Unless required by applicable law or agreed to in writing, software distributed
 * under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
 * CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License. 
 */

package org.cdp1802.xpl;

import java.util.NoSuchElementException;
import java.util.StringTokenizer;


/**
 * Implementation of an xPL filter
 * @author gerry
 */
public class xPL_Filter implements xPL_FilterI {
  xPL_MessageI.MessageType filterOnMessageType = xPL_MessageI.MessageType.UNKNOWN;
  String vendorID = null;
  String deviceID = null;
  String instanceID = null;
  String schemaClass = null;
  String schemaType = null;
  String filterSpec = null;
  
  /**
   * Cretea a new filter with all values set to wildcard (filter matches everything)
   */
  public xPL_Filter() {
    resetFilters();
  }

  /**
   * Create a new filter and set values based on the passed filter spec, per
   * the xPL docs.  See the setFilter() method for the format of this string
   *
   * @param filterSpec filterSpec to install into this new filter
   */
  public xPL_Filter(String filterSpec) {
    resetFilters();
    setFilter(filterSpec);
  }

  /**
   * Create a new filter and set all filter values explicitly.
   * @param messageType Message type to filter on or UNKNOWN for wildcard
   * @param theVendor source vendor to filter on or null for wildcard
   * @param theDeviceID source device to filter on or null for wildcard
   * @param theInstance source instance to filter on or null for wildcard
   * @param theSchemaClass schema class to filter on or null for wildcard
   * @param theSchemaType schema type to filter on or null for wildcard
   */
  public xPL_Filter(xPL_MessageI.MessageType messageType, String theVendor, String theDeviceID, String theInstance, String theSchemaClass, String theSchemaType) {
    resetFilters();
    setFilter(messageType, theVendor, theDeviceID, theInstance, theSchemaClass, theSchemaType);
  }

  public void resetFilters() {
    filterOnMessageType = xPL_MessageI.MessageType.UNKNOWN;
    vendorID = null;
    deviceID = null;
    instanceID = null;
    schemaClass = null;
    schemaType = null;
    filterSpec = null;
  }

  /**
   * This will parse and install values for this filter from the passed 
   * string.  If the string is empty or null or does not define all the
   * elements of the filter, then false is returned.  If all elements are
   * there, true is returned.
   *
   * Even when false is returned, the elements parsed so far are installed
   * and the filter is always reset first (even for empty/null strings).
   *
   * @param filterSpec string version of this filter
   * @return true if filter is complete and valid, false if there are any problems
   */
  public boolean parseAndValidateFilter(String filterSpec) {
    // Unconditionally clear the filter
    resetFilters();

    // Handle empty/null specs
    if ((filterSpec == null) || (filterSpec.length() == 0)) return false;
    
    StringTokenizer theTokenizer = new StringTokenizer(filterSpec, ".");
    
    try {
      // Parse Message Type
      String messageType = theTokenizer.nextToken();
      if (messageType.equals("*"))
        setMessageTypeFilter(xPL_MessageI.MessageType.UNKNOWN);
      else if (messageType.equalsIgnoreCase("xpl-cmnd"))
        setMessageTypeFilter(xPL_MessageI.MessageType.COMMAND);
      else if (messageType.equalsIgnoreCase("xpl-stat"))
        setMessageTypeFilter(xPL_MessageI.MessageType.STATUS);
      else if (messageType.equalsIgnoreCase("xpl-trig"))
        setMessageTypeFilter(xPL_MessageI.MessageType.TRIGGER);
      
      // Parse out vendor
      setVendorIDFilter(theTokenizer.nextToken());
      setDeviceIDFilter(theTokenizer.nextToken());
      setInstanceIDFilter(theTokenizer.nextToken());
      
      // Parse out Schema
      setSchemaClassFilter(theTokenizer.nextToken());
      setSchemaTypeFilter(theTokenizer.nextToken());
      
      // There should not be anything else
      if (theTokenizer.hasMoreTokens()) return false;
      return true;
    } catch (NoSuchElementException shortError) {
      return false;
    }
  }
  
  public void setFilter(xPL_MessageI.MessageType messageType, String theVendor, String theDeviceID, String theInstance, String theSchemaClass, String theSchemaType) {
    setMessageTypeFilter(messageType);
    setDeviceFilter(theVendor, theDeviceID, theInstance);
    setSchemaFilter(theSchemaClass, theSchemaType);
  }
  
  public void setFilter(String filterSpec) {
    parseAndValidateFilter(filterSpec);
  }
  
  public xPL_MessageI.MessageType getMessageTypeFilter() { return filterOnMessageType; }

  public String getDeviceIDFilter() { return (deviceID == null ? "*" : deviceID); }
  public String getVendorIDFilter() { return (vendorID == null ? "*" : vendorID); }
  public String getInstanceIDFilter() { return (instanceID == null ? "*" : instanceID); }

  public String getSchemaClassFilter() { return (schemaClass == null ? "*" : schemaClass); }
  public String getSchemaTypeFilter() { return (schemaType == null ? "*" : schemaType);  }

  public void setDeviceFilter(xPL_IdentifierI theIdentifier) {
    if (theIdentifier == null) {
      setVendorIDFilter(null);
      setDeviceIDFilter(null);
      setInstanceIDFilter(null);
    } else {
      setVendorIDFilter(theIdentifier.getVendorID());
      setDeviceIDFilter(theIdentifier.getDeviceID());
      setInstanceIDFilter(theIdentifier.getInstanceID());
    }
  }

  public void setDeviceFilter(String theVendor, String theDeviceID, String theInstance) {
    setVendorIDFilter(theVendor);
    setDeviceIDFilter(theDeviceID);
    setInstanceIDFilter(theInstance);
  }

  public void setVendorIDFilter(String theVendor) {
    if ((theVendor == null) || (theVendor.length() == 0) || theVendor.equals("*"))
      vendorID = null;
    else
      vendorID = theVendor;

    filterSpec = null;
  }

  public void setDeviceIDFilter(String theDeviceID) {
    if ((theDeviceID == null) || (theDeviceID.length() == 0) || theDeviceID.equals("*"))
      deviceID = null;
    else
      deviceID = theDeviceID;

    filterSpec = null;
  }

  public void setInstanceIDFilter(String theInstanceID) {
    if ((theInstanceID == null) || (theInstanceID.length() == 0) || theInstanceID.equals("*"))
      instanceID = null;
    else
      instanceID = theInstanceID;

    filterSpec = null;
  }

  public void setMessageTypeFilter(xPL_MessageI.MessageType messageType) {
    filterOnMessageType = messageType;
    filterSpec = null;
  }

  public void setSchemaClassFilter(String theSchemaClass) {
    if ((theSchemaClass == null) || (theSchemaClass.length() == 0) || theSchemaClass.equals("*"))
      schemaClass = null;
    else
      schemaClass = theSchemaClass;

    filterSpec = null;
  }

  public void setSchemaTypeFilter(String theSchemaType) {
    if ((theSchemaType == null) || (theSchemaType.length() == 0) || theSchemaType.equals("*"))
      schemaType = null;
    else
      schemaType = schemaType;

    filterSpec = null;
  }

  public void setSchemaFilter(String schemaClass, String schemaType) {
    setSchemaClassFilter(schemaClass);
    setSchemaTypeFilter(schemaType);
  }
  
  public boolean doesMessageMatchFilter(xPL_MessageI theMessage) {
    // If we have a specific message type, insure this matches
    if ((filterOnMessageType != xPL_MessageI.MessageType.UNKNOWN) && (filterOnMessageType != theMessage.getType())) return false;

    // Get the device for parsing and filter with it
    xPL_IdentifierI sourceDevice = theMessage.getSource();
    if (sourceDevice == null) return false;
    if ((vendorID != null) && !sourceDevice.getVendorID().equalsIgnoreCase(vendorID)) return false;
    if ((deviceID != null) && !sourceDevice.getDeviceID().equalsIgnoreCase(deviceID)) return false;
    if ((instanceID != null) && !sourceDevice.getInstanceID().equalsIgnoreCase(instanceID)) return false;
    
    // Check/Filter on schema
    if ((schemaClass != null) && !theMessage.getSchemaClass().equalsIgnoreCase(schemaClass)) return false;
    if ((schemaType != null) && !theMessage.getSchemaType().equalsIgnoreCase(schemaType)) return false;
    
    // Filter matches
    return true;
  }
  
  public String toString() {
    if (filterSpec != null) return filterSpec;
    
    StringBuffer theSpec = new StringBuffer();

    switch(filterOnMessageType) {
      case STATUS:
        theSpec.append("xpl-stat");
        break;
      case COMMAND:
        theSpec.append("xpl-cmnd");
        break;
      case TRIGGER:
        theSpec.append("xpl-trig");
        break;
      default:
        theSpec.append("*");
        break;
    }
    theSpec.append('.');
    
    theSpec.append(vendorID == null ? '*' : vendorID);
    theSpec.append('.');
    theSpec.append(deviceID == null ? '*' : deviceID);
    theSpec.append('.');
    theSpec.append(instanceID == null ? '*' : instanceID);
    theSpec.append('.');
    
    theSpec.append(schemaClass == null ? '*' : schemaClass);
    theSpec.append('.');
    theSpec.append(schemaType == null ? '*' : schemaType);
    
    return (filterSpec = theSpec.toString());
  }
}
