/*
 * xPL_FilterI.java
 * Copyright 2004, Gerald R Duprey Jr
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy of
 * the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0 
 *
 * Unless required by applicable law or agreed to in writing, software distributed
 * under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
 * CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License. 
 */

package org.cdp1802.xpl;



/**
 * Describes a single filter item to be used for filtering broadcast
 * messages.
 *
 * @author  gerry
 */
public interface xPL_FilterI {
  /**
   * Clear all filters (reset all filter items to wildcards) 
   */
  public void resetFilters();

  /**
   * This will parse and install values for this filter from the passed 
   * string.  If the string is empty or null or does not define all the
   * elements of the filter, then false is returned.  If all elements are
   * there, true is returned.
   *
   * Even when false is returned, the elements parsed so far are installed
   * and the filter is always reset first (even for empty/null strings).
   *
   * @param filterSpec string version of this filter
   * @return true if filter is complete and valid, false if there are any problems
   */
  public boolean parseAndValidateFilter(String filterSpec);
  
  /**
   * Install a filter by parsing a filter string spec.  The spec looks like
   * a filter from the xPL protocol docs.  Like this:
   *
   *   type.vendor.device.instance.schema_class.schema_type
   *
   * For example:
   *    *.*.*.*.*.*  -- match everything
   *    STATUS.cdp1802.*.*.config.* -- match only config status updates for vendor cdp1802
   *
   * @param filterSpec string specification of the filter
   */
  public void setFilter(String filterSpec);

  /**
   * Install all individual components of a filter at the same time.  Passing
   * a null indicates that item should be a wild card (for message type, wildcard
   * is specified with a type of UNKNOWN)
   *
   * @param messageType    type of UNKNOWN for wildcard
   * @param theVendor      vendor name or null for wildcard
   * @param theDeviceID    device name or null for wildcard
   * @param theInstance    instance ID or null for wildcard
   * @param theSchemaClass schema class or null for wildcard
   * @param theSchemaType  schema type or null for wildcard
   */
  public void setFilter(xPL_MessageI.MessageType messageType, String theVendor, String theDeviceID, String theInstance, String theSchemaClass, String theSchemaType);
  
  /** 
   * Set the message type filter.  UNKNOWN is used for wildcard match
   *
   * @param messageType message type or UNKNOWN for wildcard
   */
  public void setMessageTypeFilter(xPL_MessageI.MessageType messageType);

  /**
   * Return the current message type filter.  UNKNOWN means wildcard match
   *
   * @return message type filter or UNKNOWN for wildcard
   */
  public xPL_MessageI.MessageType getMessageTypeFilter();
  
  /**
   * Set the source vendor filter value
   * @param theVendor vendor to filter on or null for wildcard
   */
  public void setVendorIDFilter(String theVendor);
  /**
   * Get the current source vendor filter value
   * @return source vendor filter value or null (wildcard)
   */
  public String getVendorIDFilter();
  
  /**
   * Set the source device filter value
   * @param theDeviceID Device to filter on or null for wildcard
   */
  public void setDeviceIDFilter(String theDeviceID);
  /**
   * Get the current source device filter value
   * @return source device value or null for wildcard
   */
  public String getDeviceIDFilter();
  
  /**
   * Set source instance filter value
   * @param theInstanceID Instsance to filter on or null for wildcard
   */
  public void setInstanceIDFilter(String theInstanceID);
  /**
   * Get the current source instance filter value
   * @return source instance value or null for wildcard
   */
  public String getInstanceIDFilter();
  
  /**
   * Set source filter values
   * @param theVendor vendor to filter on or null for wildcard
   * @param theDeviceID device to filter on or null for wildcard
   * @param theInstance instance to filter on or null for wildcard
   */
  public void setDeviceFilter(String theVendor, String theDeviceID, String theInstance);
  
  /**
   * Set source filter values
   * @param theIdentifier device to extract vendor, device and instance filter values or null to set all three to wildcards
   */
  public void setDeviceFilter(xPL_IdentifierI theIdentifier);
  
  /**
   * Set Schema Class to filter on
   * @param theSchemaClass schema class to filter on or null for wildcard
   */
  public void setSchemaClassFilter(String theSchemaClass);
  /**
   * Get current schema class filter value
   * @return schema class filter value or null for wildcard
   */
  public String getSchemaClassFilter();
  
  /**
   * Set schema type to filter on
   * @param theSchemaType schema type to filter on or null for wildcard
   */
  public void setSchemaTypeFilter(String theSchemaType);
  /**
   * Get current schema type filter value
   * @return current schema type value or null for wildcard
   */
  public String getSchemaTypeFilter();
  
  /**
   * Set schema class and type filter values
   * @param schemaClass schema class to filter on or null for wildcard
   * @param schemaType schema type to filter on or null for wildcard
   */
  public void setSchemaFilter(String schemaClass, String schemaType);
  
  /**
   * Compare passed message to the filters installed values and determine if the
   * message match/pass this filter
   * @param theMessage Message to test against.  An invalid message or null value always fails
   * @return true if message matches/passes this filter, false if one or more items do not match
   */
  public boolean doesMessageMatchFilter(xPL_MessageI theMessage);
  
  /**
   * Return the filter as a string in filter spec format
   *
   * @return string representing this filter
   */
  public String toString();
}
