/*
 * xPL_Manager.java
 * Copyright 2004, Gerald R Duprey Jr
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy of
 * the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0 
 *
 * Unless required by applicable law or agreed to in writing, software distributed
 * under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
 * CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License. 
 */

package org.cdp1802.xpl;

import java.net.InetAddress;
import java.util.ArrayList;
import java.util.Collection;
import org.cdp1802.xpl.ethernet.EthernetHandler;
import org.cdp1802.xpl.device.*;
import org.cdp1802.xpl.tracker.TrackedDeviceManager;
import org.cdp1802.xpl.plugins.PluginManager;

/**
 * XPL Manger -- "go-to" object for xPL.
 *
 * @author  gerry
 */
public class xPL_Manager implements xPL_MediaHandlerListenerI {
  public static final String XPL4JAVA_VERSION = "1.3h";
  private static xPL_Manager theXPLManager = new xPL_Manager();

  /**
   * Return an instance to the XPL Manager.  There is only one
   * XPL Manager for any instance of a JVM
   *
   * @return the XPL Manager Instance
   */
  public static xPL_Manager getManager() { return theXPLManager; }
  
  IdentifierManager identifierManager = IdentifierManager.getManager();
  DeviceManager deviceManager = DeviceManager.getManager();
  TrackedDeviceManager trackedDeviceManager = null;
  PluginManager pluginManager = null;
  Thread shutdownThread = null;
  ArrayList<xPL_MessageListenerI> messageListenerList = new ArrayList<xPL_MessageListenerI>();
  ArrayList<xPL_ManagerListenerI> managerListenerList = new ArrayList<xPL_ManagerListenerI>();
  ArrayList<xPL_MediaHandlerI> mediaHandlerList = new ArrayList<xPL_MediaHandlerI>();

  boolean hubConfirmed = false;
  
  private xPL_Manager() {  }

  private void debug(String theMessage) {
    System.out.println("XPL_MGR:: " + theMessage);
  }

  /** 
   * See if all media handlers have seen their hub yet. 
   *
   * @return true if all media handlers have confirmed there is a hub, false if one or more are still waiting
   */
  public boolean areHubsConfirmed() { return hubConfirmed; }
  
  /**
   * Return an instance of the XPL identifier manager.  This is normally not
   * needed and is used to cache common copies of XPL identifier identifiers
   *
   * @return instance of the identifier manager
   */
  public IdentifierManager getIdentifierManager() { return identifierManager; }
  
  /**
   * Return an instance of the device tracker.  Note that the tracker must
   * be started before it will begin listening for devices and tracking 
   * them
   *
   * @return instance of the device tracking manager
   */
  public TrackedDeviceManager getTrackedDeviceManager() { 
    if (trackedDeviceManager == null) trackedDeviceManager = TrackedDeviceManager.getTrackedDeviceManager();
    return trackedDeviceManager;
  }
  
  /**
   * Return an instance of plugins manager
   *
   * @return instance of the plugins manager
   */
  public PluginManager getPluginManager() {
    if (pluginManager == null) pluginManager = PluginManager.getManager();
    return pluginManager;
  }
  
  /**
   * Return an instance of the xPL Device Manager.  This is very useful
   * for creating new devices and managing them.
   *
   * @return instance of the device manager
   */
  public DeviceManager getDeviceManager() { return deviceManager; }
  
  
  /**
   * Add a new, raw XPL message listener.  This listener will see ever XPL
   * message received, whether it is targeted at a device on this machine
   * or not.  
   *
   * If you are using a managed XPL device, you would likely be better 
   * off using a listener targeted at that device
   *
   * @param theListener listener to receive all XPL messages
   */
  public void addMessageListener(xPL_MessageListenerI theListener) {
    synchronized(messageListenerList) {
      if (messageListenerList.contains(theListener)) return;
      messageListenerList.add(theListener);
    }
  }

  /**
   * Remove an existing XPL message listener.
   *
   * @param theListener listener to remove
   */
  public void removeMessageListener(xPL_MessageListenerI theListener) {
    synchronized(messageListenerList) {
      messageListenerList.remove(theListener);
    }
  }
  
  /**
   * Send the passed message out via all known media handlers
   *
   * @param outgoingMessage message to send out
   */
  public void sendMessage(xPL_MessageI outgoingMessage) {
    for (xPL_MediaHandlerI theHandler: mediaHandlerList) {
      theHandler.sendMessage(outgoingMessage);
    }
  }

  /**
   * Add a new, xPL_Manager listener.  This listener will see major changes
   * and updates to the xPL manager
   *
   * @param theListener listener to receive all xPL Manager messages
   */
  public void addManagerListener(xPL_ManagerListenerI theListener) {
    synchronized(managerListenerList) {
      if (managerListenerList.contains(theListener)) return;
      managerListenerList.add(theListener);
    }
  }

  /**
   * Remove an existing XPL manager listener.
   *
   * @param theListener listener to remove
   */
  public void removeManagerListener(xPL_ManagerListenerI theListener) {
    synchronized(managerListenerList) {
      managerListenerList.remove(theListener);
    }
  }

  /**
   * Distribute the passed event to all listeners.  When the event has
   * been distributed, the events resources are auatomatically released
   *
   * @param theEvent Event to send to all this managers listeners
   */
  protected void fireManagerEvent(xPL_ManagerEvent theEvent) {
    int listenerCount = managerListenerList.size();
    for (int listenerIndex = listenerCount - 1; listenerIndex >= 0; listenerIndex--) {
      try {
        managerListenerList.get(listenerIndex).handleXPLManagerEvent(theEvent);
      } catch (Throwable ignoredException) { }
    }
    
    // Free event resources
    theEvent.releaseResources();
  }
  
  /**
   * Send a heartbeat message for the passed device
   *
   * @param forDevice device to send heartbeat message for
   * @param timelyHeartbeat true if this is a normal heartbeat, false if forced off the scheduled time
   */
  public void sendHeartbeat(xPL_DeviceI forDevice, boolean timelyHeartbeat) {
    for (xPL_MediaHandlerI theHandler: mediaHandlerList) {
      theHandler.sendHeartbeat(forDevice, timelyHeartbeat);
    }
  }

  /**
   * Shutdown xPL in preparation for termination.  Any hub(s) are shutdown, 
   * devices are shutdown (and notice sent to everyone) and media handlers
   * are all shutdown.
   *
   * This is the best way to shutdown prior to application exit
   */
  public void doSmartShutdown() {
    // Let listeners know
    fireManagerEvent(new xPL_ManagerEvent(xPL_ManagerEvent.EventCode.XPL_SHUTTING_DOWN));
    
    // Stop all devices
    DeviceManager.getManager().shutdownAllManagedDevices();
    
    // Stop all media handlers
    stopAllMediaHandlers();
  }
  
  // Insure the shutdown hook is installed
  private void ensureShutdownEnabled() {
    if (shutdownThread != null) return;

    // Create the shutdown thread
    shutdownThread = new Thread() {
      public void run() {
        doSmartShutdown();
      }
    };
    
    // Install the thread
    Runtime.getRuntime().addShutdownHook(shutdownThread);
  }
  
  /**
   * Add a new media handler to the internally managed list
   * of handlers.  In most cases, you do not need to use 
   * this call (unless you have a unique, custom media handler)
   *
   * @param theHandler media handler to manage
   */
  public void addMediaHandler(xPL_MediaHandlerI theHandler) {
    // Insure this is real
    if (theHandler == null) return;
    
    synchronized(mediaHandlerList) {
      if (mediaHandlerList.contains(theHandler)) return;
      mediaHandlerList.add(theHandler);
      theHandler.addListener(this);
      ensureShutdownEnabled();
    }

    fireManagerEvent(new xPL_ManagerEvent(xPL_ManagerEvent.EventCode.MEDIA_HANDLER_ADDED, theHandler));

    // See if this hub is one that is always OK (serial ports and such)
    checkForHubCompletion();
  }

  /**
   * Remove an existing media handler from the internally managed
   * list of handlers.  In mosts cases, you do not need to use this
   * method (unless you have a unqiue, custom media handler)
   *
   * If the media handler is started, it will be stopped before being
   * removed.
   *
   * @param theHandler handler to remove from list of managed handlers
   */
  public void removeMediaHandler(xPL_MediaHandlerI theHandler) {
    // Insure we have something to remove
    if (theHandler == null) return;

    // Insure it's stopped
    theHandler.stopHandler();

    // Actually remove it
    boolean wasRemoved = false;
    synchronized(mediaHandlerList) {
      theHandler.removeListener(this);
      wasRemoved = mediaHandlerList.remove(theHandler);
    }
      
    if (wasRemoved) fireManagerEvent(new xPL_ManagerEvent(xPL_ManagerEvent.EventCode.MEDIA_HANDLER_REMOVED, theHandler));
  }

  /**
   * Get an iteration of media handlers
   *
   * @return list of media handlers
   */
  public Collection<xPL_MediaHandlerI> getMediaHandlers() {
    return mediaHandlerList;
  }

  /**
   * Get the number of media handlers currently defined
   *
   * @return number of media handlers currently defined
   */
  public int getMediaHandlerCount() { return mediaHandlerList.size(); }
  
  /**
   * Return the media handler at the specified index.  If the index
   * is invalid, null is returned
   *
   * @param theIndex index of media handler to return
   * @return mediaHandler at that index or null if index is invalid
   */
  public xPL_MediaHandlerI getMediaHandlerAt(int theIndex) {
    return mediaHandlerList.get(theIndex);
  }
  
  /**
   * Start All managed media handlers. If a handler is already started, then
   * it is passed over.  if not, it's started.  If there is any error starting
   * a media handler, that error is returned and the process stops.
   * 
   * Note: you do not normally need to call this if you use the various
   * standard media creators.  However, it won't hurt to use it.
   *
   * @throws xPL_MediaHandlerException thrown for any problems encountered while starting a handler
   */
  public void startAllMediaHandlers() throws xPL_MediaHandlerException {
    synchronized(mediaHandlerList) {
      for (xPL_MediaHandlerI mediaHandler: mediaHandlerList) {
        mediaHandler.startHandler();
      }
    }
  }

  /**
   * Stop all managed media handlers.  If a handler is already stopped, 
   * it is silently skipped
   */
  public void stopAllMediaHandlers() {
    synchronized(mediaHandlerList) {
      for (xPL_MediaHandlerI mediaHandler: mediaHandlerList) {
        mediaHandler.stopHandler();
      }
    }
  }
  
  /**
   * Create a basic network media handler.  The handler is not started and
   * you will have to configure it to some extent (at least minimally
   * setting the connection style).  This is very flexible, but requires
   * you to:
   *
   * 1) Set the connection mode (standalone)
   * 2) Set the network interface to use (optional - default: all networks)
   * 3) Make any other settings
   * 4) Start the handler
   * 5) Add the media handler to the list of media handlers (i.e. addMediaHandler())
   *
   * @return network media handler, ready for configuring
   */
  public EthernetHandler createNetworkHandler() {
    return new EthernetHandler();
  }

  /** 
   * Create and start a standard network handler.  This is for a hub-based
   * network connect (i.e. a hub must already be running).  Once successfully
   * started, the handler is automatically added tothe list of media handlers.
   *
   * The passed network address specifies which interface this handler
   * should bind too.  If you pass a null, then all interfaces are 
   * bound to (which is the most common usage). 
   *
   * Note: If you pass null, while all network interfaces will be listened to,
   * the first non-loopback interface is as identification for output packets
   * (specially, xPL heartbeat messages).   This is usually fine, but if you need
   * to override this, you can changed the advertised address in the returned
   * EthernetHandler (you'll have to then stop and restart the handler to make
   * that take effect).
   *
   * @param bindTo network address to bind to or null to bind to all
   * @return instance of the network handler
   * @throws xPL_MediaHandlerException thrown for an error starting the handler
   */
  public EthernetHandler createAndStartNetworkHandler(InetAddress bindTo) throws xPL_MediaHandlerException {
    // Creat the handler and set it up
    EthernetHandler ethernetHandler = createNetworkHandler();
    if (bindTo != null) ethernetHandler.setNetworkInterface(bindTo);
    ethernetHandler.setConnectMode(EthernetHandler.ConnectMode.VIA_HUB);

    // Start it
    ethernetHandler.startHandler();
    
    // Add it too our managed list
    addMediaHandler(ethernetHandler);
    return ethernetHandler;
  }
  
  /** 
   * Create and start a standard network handler.  This is for a hub-based
   * network connect (i.e. a hub must already be running).  Once successfully
   * started, the handler is automatically added tothe list of media handlers.
   *
   * This will bind to all network interfaces (both listening and sending).
   *
   * Note: While all network interfaces will be listened to, the first
   * non-loopback interface is as identification for output packets
   * (specially, xPL heartbeat messages).  This is usually fine, but if you need
   * to override this, you can changed the advertised address in the returned
   * EthernetHandler (you'll have to then stop and restart the handler to make
   * that take effect).
   *
   * @return instance of the network handler
   * @throws xPL_MediaHandlerException thrown for an error starting the handler
   */
  public EthernetHandler createAndStartNetworkHandler() throws xPL_MediaHandlerException {
    return createAndStartNetworkHandler(null);
  }
  
  /**
   * This is used by the various media handlers to distribute incoming messages.
   * It should not be used anywhere else.
   *
   * @param incomingMessage parsed, ready to distrubute incoming message
   */
  public void distributeReceivedMessage(xPL_MessageI incomingMessage) {
    int listenerCount = messageListenerList.size();
    for (int listenerIndex = listenerCount - 1; listenerIndex >= 0; listenerIndex--) {
      try {
        messageListenerList.get(listenerIndex).handleXPLMessage(incomingMessage);
      } catch (Throwable ignoredException) { }
    }
  }

  private void checkForHubCompletion() {
    // Determine if all handlers have confirmed hearing a hub
    boolean waitingForHub = false;
    for (xPL_MediaHandlerI theHandler: mediaHandlerList) {
      if (!theHandler.isHubConfirmed()) waitingForHub = true;
    }

    // If no handlers are waiting for the hub, set confirmation flag
    boolean originalHubConfirmed = hubConfirmed;
    hubConfirmed = !waitingForHub;

    // If we confirmed all handlers found a hub, send a general note
    if (hubConfirmed && (originalHubConfirmed != hubConfirmed)) {
      fireManagerEvent(new xPL_ManagerEvent(xPL_ManagerEvent.EventCode.ALL_HUBS_DETECTED));
    }
  }
  
  /**
   * Keep track of the goings on in the media handlers we have.
   *
   * @param theEvent event describing what happened to a handler 
   */
  public void handleMediaHandlerEvent(xPL_MediaHandlerEvent theEvent) {
    // Log a handler starting
    if (theEvent.getEvent() == xPL_MediaHandlerEvent.EventCode.HANDLER_ACTIVE)
      fireManagerEvent(new xPL_ManagerEvent(xPL_ManagerEvent.EventCode.MEDIA_HANDLER_STARTED, theEvent.getHandler()));
    
    // Log a handler stopping
    if (theEvent.getEvent() == xPL_MediaHandlerEvent.EventCode.HANDLER_INACTIVE) 
      fireManagerEvent(new xPL_ManagerEvent(xPL_ManagerEvent.EventCode.MEDIA_HANDLER_STOPPED, theEvent.getHandler()));

    // Log a handler detecting a hub
    if (theEvent.getEvent() == xPL_MediaHandlerEvent.EventCode.HANDLER_CONFIRMED_HUB) 
      fireManagerEvent(new xPL_ManagerEvent(xPL_ManagerEvent.EventCode.MEDIA_HANDLER_DETECTED_HUB, theEvent.getHandler()));
      
    // When a handler is started
    if ((theEvent.getEvent() == xPL_MediaHandlerEvent.EventCode.HANDLER_CONFIRMED_HUB)
      ||(theEvent.getEvent() == xPL_MediaHandlerEvent.EventCode.HANDLER_ACTIVE)) {
      checkForHubCompletion();
    }
  }
}
