/*
 * xPL_MediaHandler.java
 * Copyright 2004, Gerald R Duprey Jr
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy of
 * the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0 
 *
 * Unless required by applicable law or agreed to in writing, software distributed
 * under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
 * CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License. 
 */

package org.cdp1802.xpl;

import java.util.*;

/**
 * Base class for media handlers
 *
 * @author  gerry
 */
public abstract class xPL_MediaHandler implements xPL_MediaHandlerI {
  protected String handlerDescription = "";
  protected boolean handlerStarted = false;
  protected boolean hubConfirmed = false;
  xPL_Manager xplManager = xPL_Manager.getManager();
  ArrayList<xPL_MediaHandlerListenerI> handlerListenerList = new ArrayList<xPL_MediaHandlerListenerI>();
  
  public xPL_MediaHandler() { }

  /**
   * Add a new listener to this media handler.  Listeners get reports
   * of the state of and changes to media handlers
   *
   * @param theListener listener to add, if not already added
   */
  public void addListener(xPL_MediaHandlerListenerI theListener) {
    synchronized(handlerListenerList) {
      if (handlerListenerList.contains(theListener)) return;
      handlerListenerList.add(theListener);
    }
  }
  
  /**
   * Remove a previously added listener from this media handler .
   *
   * @param theListener listener to remove from handler
   */
  public void removeListener(xPL_MediaHandlerListenerI theListener) {
    synchronized(handlerListenerList) {
      handlerListenerList.remove(theListener);
    }
  }

  /**
   * Distribute the passed event to all listeners.  When the event has
   * been distributed, the events resources are auatomatically released
   *
   * @param theEvent Event to send to all this handlers listeners
   */
  protected void fireHandlerEvent(xPL_MediaHandlerEvent theEvent) {
    int listenerCount = handlerListenerList.size();
    for (int listenerIndex = listenerCount - 1; listenerIndex >= 0; listenerIndex--) {
      try {
        handlerListenerList.get(listenerIndex).handleMediaHandlerEvent(theEvent);
      } catch (Throwable ignoredException) { }
    }
    
    // Free event resources
    theEvent.releaseResources();
  }
  
  public String getDescription() { return handlerDescription; }
  protected void setDescription(String newDescription) {
    handlerDescription = newDescription;
  }

  public boolean isHubConfirmed() { return hubConfirmed; }
  
  public boolean isStarted() { return handlerStarted; }
  protected void setStarted(boolean isStarted) {
    handlerStarted = isStarted;
  }

  // Common logic for preparing a message to be sent.  This insures
  // the message is sendable, valid and handles converting it into
  // a text form (including inter-media hop incrementing).  If there
  // is any problem, null is returned.  Otherwise, the ready to send
  // text message is returned.
  protected String prepareMessage(xPL_MessageI theMessage) {
    // Make sure we can actually send a message and we have a message to send.
    if (!handlerStarted || (theMessage == null) || !theMessage.isValid()) return null;

    // Determine if this was a message received from one media handler
    // and forwarded to another, then bump the hop count. 
    boolean interMediaBump = ((theMessage.getReceivedFrom() != null) && (theMessage.getReceivedFrom() != this));
    
    // Format the message
    return xPL_Util.formatMessage(theMessage, this, interMediaBump);
  }
  
  public String toString() { return handlerDescription; }
}
