/*
 * xPL_MediaHandlerI.java
 * Copyright 2004, Gerald R Duprey Jr
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy of
 * the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0 
 *
 * Unless required by applicable law or agreed to in writing, software distributed
 * under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
 * CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License. 
 */

package org.cdp1802.xpl;

import org.cdp1802.xpl.device.xPL_DeviceI;

/**
 * Media Handler defined an interface that a media adapter (serial,
 * ethernet, etc) must implement to be supported by the xPL lib.
 *
 * A Media Handler should be able to be started and stopped multiple
 * times without failure.  All media handlers are in a stopped state
 * until specifically started.
 *
 * If the media handler cannot start, it returns false.  The media handler
 * should not bind to any resource (network port, serial port, etc) until
 * the handler is started and should release any resources when stopped
 *
 * If the media handler does not support a hub (like a serial port), it's 
 * still required to return true to isHubConfirmed() method and to fire
 * a xPL_MediaHandlerEvent.EventCode.HANDLER_CONFIRMED_HUB.  Failure to
 * do this will keep xPL4Java in a hobbled state and may result in applications
 * not correctly starting (or not starting at all, if they are waiting for 
 * notification).
 *
 * @author  gerry
 */
public interface xPL_MediaHandlerI {
  /**
   * Add a new listener to this media handler.  Listeners get reports
   * of the state of and changes to media handlers
   *
   * @param theListener listener to add, if not already added
   */
  public void addListener(xPL_MediaHandlerListenerI theListener);
  
  /**
   * Remove a previously added listener from this media handler .
   *
   * @param theListener listener to remove from handler
   */
  public void removeListener(xPL_MediaHandlerListenerI theListener);
  
  /**
   * Broadcast the passed text as a single message over the
   * media.  If the message is sent successfully, true is
   * returned.
   *
   * @param theMessage text to send over the media
   * @return true if message sent OK, false if there was a failure
   */
  public boolean sendMessage(xPL_MessageI theMessage);
  
  /**
   * Send a heartbeat message out for the passed device.  Heartbeat
   * messages need to be tailored for each media handler as the contents
   * of the message vary based on handler specifics (short vs long, network
   * port/IP address info, etc)
   *
   * The timelyHeartbeat is used primarily to let the handler know the heartbeat 
   * message is being sent out of cadence for the device.  This often is done
   * when the handler has not detected a hub yet and as a result, some handlers 
   * will IGNORE messages that are sent off the clock.  FYI
   *
   * @param forDevice device to send heartbeat message for
   * @param timelyHeartbeat true if the heartbeat is a normal heartbeat, false if a forced, off-the-clock one
   * @return true is message was sent, false if not (handler not running, error, etc)
   */
  public boolean sendHeartbeat(xPL_DeviceI forDevice, boolean timelyHeartbeat);
  
  /**
   * Start receiving messages on this media handler.
   *
   */
  public void startHandler() throws xPL_MediaHandlerException;
  
  /** 
   * Stop receiving/sending messages on this media handler.
   *
   */
  public void stopHandler();
  
  /**
   * Return true if the media handler is started.
   *
   * @return true if stated, false if stopped
   */
  public boolean isStarted();
  
  /**
   * Return true if the media handler has confirmed it knows there is a hub.
   * For media that does not have hubs, this should always return true.
   *
   *@return true if the media handler has seen a hub, false if still waiting
   */
  public boolean isHubConfirmed();
  
  /**
   * Return a description of the handler.  For example, a 
   * serial port handler might return "Serial Port (/dev/ttyS1)"
   *
   * @return shortish summary description of the handler
   */
  public String getDescription();
}
