/*
 * xPL_Message.java
 * Copyright 2004, Gerald R Duprey Jr
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy of
 * the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0 
 *
 * Unless required by applicable law or agreed to in writing, software distributed
 * under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
 * CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License. 
 */

package org.cdp1802.xpl;

import java.util.HashMap;
import java.util.Collection;

/**
 *
 * @author  gerry
 */
public class xPL_Message implements xPL_MessageI {
  int hopCount = 1;
  MessageType msgType = MessageType.UNKNOWN;
  String schemaClass = "", schemaType = "";
  xPL_IdentifierI msgSource = null, msgTarget = null;
  NamedValuesI namedValues = null;
  xPL_MediaHandlerI receivedFrom = null;
  
  xPL_Message() {  }

  public boolean isValid() {
    return ((msgType != MessageType.UNKNOWN) 
         && (schemaClass.length() != 0) && (schemaType.length() != 0)
         && (msgSource != null) && (msgTarget != null)
         && (namedValues != null));
  }
  
  public MessageType getType() { return msgType; }
  public int getHopCount() { return hopCount; }

  public String getSchemaClass() { return schemaClass; }
  public String getSchemaType() { return schemaType; }

  public xPL_IdentifierI getSource() { return msgSource; }
  public xPL_IdentifierI getTarget() { return msgTarget; }

  public boolean isBroadcast() {
    if ((msgTarget != null) && msgTarget.isBroadcastIdentifier()) return true;
    return false;
  }
  
  public boolean isGroupMessage() {
    if ((msgTarget != null) && msgTarget.isGroupIdentifier()) return true;
    return false;
  }
  
  public boolean isHeartbeat() {
    // Make sure this is the right type of message
    if (msgType != MessageType.STATUS) return false;

    // Make sure we are in the right class
    if (!schemaClass.equalsIgnoreCase("hbeat") && !schemaClass.equalsIgnoreCase("config")) return false;
    
    // Make sure we have the right type
    return (schemaType.equalsIgnoreCase("app") || schemaType.equalsIgnoreCase("basic") || schemaType.equalsIgnoreCase("end"));
  }
  
  public boolean isHeartbeatEnd() {
    // Make sure this is the right type of message
    if (msgType != MessageType.STATUS) return false;

    // Make sure we are in the right class
    if (!schemaClass.equalsIgnoreCase("hbeat") && !schemaClass.equalsIgnoreCase("config")) return false;
    
    // Make sure we have the right type
    return schemaType.equalsIgnoreCase("end");
  }
  
  public boolean isHeartbeatRequest() {
    // Make sure this is the right type of message
    if (msgType != MessageType.COMMAND) return false;

    // Make sure we are in the right class
    if (!schemaClass.equalsIgnoreCase("hbeat") && !schemaClass.equalsIgnoreCase("config")) return false;
    
    // Make sure we have the right type
    return schemaType.equalsIgnoreCase("request");
  }
  
  public boolean isConfigurationMessage() {
    // Has to be a config message
    if (!schemaClass.equalsIgnoreCase("config")) return false;
    
    // If it's a config heartbeat, bail
    if (isHeartbeat()) return false;
    
    // Anything else is a configuration message
    return true;
  }
  
  public NamedValuesI getMessageBody() {
    return namedValues;
  }

  public xPL_MediaHandlerI getReceivedFrom() { return receivedFrom; }

  public boolean hasNamedValue(String theName) {
    return namedValues.hasNamedValue(theName);
  }

  public boolean hasNamedValue(String theName, String theValue) {
    return namedValues.hasNamedValue(theName, theValue);
  }
  
  public boolean hasNamedValue(String theName, int theValue) {
    return namedValues.hasNamedValue(theName, theValue);
  }

  public boolean hasNamedValue(String theName, boolean theValue) {
    return namedValues.hasNamedValue(theName, theValue);
  }
  
  public boolean getBoolNamedValue(String theName) {
    return namedValues.getBoolNamedValue(theName);
  }

  public boolean getBoolNamedValue(String theName, boolean defaultValue) {
    return namedValues.getBoolNamedValue(theName, defaultValue);
  }

  public int getIntNamedValue(String theName) {
    return namedValues.getIntNamedValue(theName);
  }

  public int getIntNamedValue(String theName, int defaultValue) {
    return namedValues.getIntNamedValue(theName, defaultValue);
  }

  public String getNamedValue(String theName) {
    return namedValues.getNamedValue(theName);
  }

  public String getNamedValue(String theName, String defaultValue) {
    return namedValues.getNamedValue(theName, defaultValue);
  }

  public void releaseResources() {
    hopCount = 0;
    msgType = MessageType.UNKNOWN;
    schemaClass = "";
    schemaType = "";
    msgSource = null;
    msgTarget = null;
    if (namedValues != null) namedValues.releaseResources();
    namedValues = null;
    receivedFrom = null;
  }

  public String toString() {
    StringBuffer messageText = new StringBuffer();
    messageText.append("MSGTYPE=");
    messageText.append(msgType);
    messageText.append(", FROM=");
    messageText.append((msgSource == null) ? "<UNKNOWN>" : msgSource.toString());
    messageText.append(", TO=");
    messageText.append((msgTarget == null) ? "<UNKNOWN>" : msgTarget.toString());
    messageText.append(", HOPS=");
    messageText.append(hopCount);
    messageText.append(", SCHEMA=");
    messageText.append((schemaClass.length() == 0) ? "<UNKNOWN>" : schemaClass);
    messageText.append('.');
    messageText.append((schemaType.length() == 0) ? "<UNKNOWN>" : schemaType);
    messageText.append(", ");
    messageText.append(namedValues.getNamedValuesCount());
    messageText.append(" named values");
    return messageText.toString();
  }
  
}
