/*
 * xPL_MessageI.java
 * Copyright 2004, Gerald R Duprey Jr
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy of
 * the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0 
 *
 * Unless required by applicable law or agreed to in writing, software distributed
 * under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
 * CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License. 
 */

package org.cdp1802.xpl;

import java.util.Collection;

/**
 * Description of a xPL message
 *
 * @author  gerry
 */
public interface xPL_MessageI {
  public enum MessageType { UNKNOWN, STATUS, COMMAND, TRIGGER };

  /**
   * Return the number of hops, to date, this message has crossed.
   *
   * @return hop count for this message
   */
  public int getHopCount();

  /**
   * Return the type of message this is
   *
   * @return message type
   */
  public MessageType getType();
  
  /**
   * Return the schema class for this messge.  This will
   * never return null.
   *
   * @return schema class for message or "" if no class set
   */
  public String getSchemaClass();
  
  /**
   * Return the schema type for this message.  This will
   * never return null.
   *
   * @return schema type for message or "" if no type set
   */
  public String getSchemaType();
  
  /**
   * Return true if this message was a broadcast (the target
   * was *)
   *
   * @return true if this is a broadcast message
   */
  public boolean isBroadcast();

  /**
   * Return true if this message is a group targeted message)
   *
   * @return true if this is a group targeted message
   */
  public boolean isGroupMessage();
  
  /**
   * Determine if this message is a heartbeat message or not.  Heartbeat messages
   * are those with a schema class of "hbeat" or "config" and a schema type of
   * "app", "basic" or "end".  Both schema class and type must match to be considered
   * a heartbeat message (so a config.list message is NOT a heartbeat)
   *
   * @return true if this is a heartbeat, false if any other sort of message
   */
  public boolean isHeartbeat();
  
  /**
   * Determine if this message is a heartbeat "end"/shutdown message.  Such messages
   * are sent when a device is going offline and should no longer be contacted/used
   * or expected to say anything else.
   *
   * @return true if this is a heartbeat "end" message, false if any other sort
   */
  public boolean isHeartbeatEnd();
  
  /**
   * Determine if this message is a request for heartbeats by clients who receive
   * it.  If so, return true
   * 
   * @return true if this is a request for devices to send their heartbeats
   */
  public boolean isHeartbeatRequest();
  
  /**
   * Determine if this message is a configuration message or not.  Configuration
   * messages always have a schema class of "config" and have a schema type
   * of list, current, response or any other schema message with a schema class
   * of "config" and a schema type NOT equal to "app", "basic" or "end" (those schema
   * types are for heartbeats)
   *
   * @return true if this is a configuration message, false if any other sort of message
   */
  public boolean isConfigurationMessage();

  /**
   * Return the identifier describing the sender of this message
   *
   * @return sender of this message or null if no sender installed
   */
  public xPL_IdentifierI getSource();
  
  /**
   * Return the target for this message.
   *
   * Note: For group based messages, the identifier will have a
   * vendor of XPL, a device of GROUP and the group name in the
   * instance ID (though check the .isGroupIdentifier() to confirm)
   *
   * Note: for broadcast messages, check is isBroadcastIdentifier()
   *
   * @return target of this message or null if no sender installed
   */
  public xPL_IdentifierI getTarget();

  /**
   * Return true if this is a valid message (all elements are filled in)
   *
   * @return true if this is a valid message
   */
  public boolean isValid();

  /**
   * Return the media handler, if any, that this message was received
   * from.  If the message was not received, this will be null
   *
   * @return media handler message was received from or null
   */
  public xPL_MediaHandlerI getReceivedFrom();
  

  /**
   * Return the name/values that comprise the body of this message
   * or null if no message body has been installed.  Note: there may be
   * a non-null message body that has no values in it.
   *
   * @return name/values body of this message or possibly null
   */
  public NamedValuesI getMessageBody();
  
  /**
   * Check the message body and determine if the message body has a value of the passed name
   * @param theName name to check message body for
   * @return true if the named value exists in the messages body
   */
  public boolean hasNamedValue(String theName);
  
  /**
   * Returns true if there is at least one named value by the 
   * passed name that has the passed value.
   *
   * @param theName the name to look for
   * @param theValue the value to match
   * @return true if there is at least named value with the passed name and value
   */
  public boolean hasNamedValue(String theName, String theValue);
  
  /**
   * Returns true if there is at least one named value by the 
   * passed name that has the passed value.
   *
   * @param theName the name to look for
   * @param theValue the value to match
   * @return true if there is at least named value with the passed name and value
   */
  public boolean hasNamedValue(String theName, int theValue);
  
  /**
   * Returns true if there is at least one named value by the 
   * passed name that has the passed value.
   *
   * @param theName the name to look for
   * @param theValue the value to match
   * @return true if there is at least named value with the passed name and value
   */
  public boolean hasNamedValue(String theName, boolean theValue);

  /**
   * Given the passed name, return the matching value from the message
   * body.  If the named value does not exist, null is returned.
   *
   * If there is more than one value for this name, only the first
   * set/installed value is returned.
   *
   * @param theName name of the value to get the value for
   * @return value associated with this name or null if the name does not exist
   */
  public String getNamedValue(String theName);
  
  /**
   * Given the passed name, return the matching value from the message
   * body.  If the named value does not exist, the default value is returned.
   *
   * If there is more than one value for this name, only the first
   * set/installed value is returned.
   *
   * @param theName name of the value to return the value for
   * @param defaultValue value to return if the named value does not exist in this message
   * @return actual value of for this named value in the message or the default value if the named value is not found
   */
  public String getNamedValue(String theName, String defaultValue);
  
  /**
   * Given the passed name, return the numeric value from the message that matches
   * that name.  If the name does not exist in the message body or the value is not
   * numeric, -1 is returned.
   *
   * If there is more than one value for this name, only the first
   * set/installed value is returned.
   *
   * @param theName named value to return from the message body
   * @return numeric value of that named value or -1 if value is not found or is not numeric
   */
  public int getIntNamedValue(String theName);

  /**
   * Given the passed name, return the numeric value from the message that matches
   * that name.  If the name does not exist in the message body or the value is not
   * numeric, the default value is returned.  If the value is not numeric, then -1
   * is returned.
   *
   * If there is more than one value for this name, only the first
   * set/installed value is returned.
   *
   * @param theName named value to return from the message body
   * @param defaultValue value to return if named value does not exist
   * @return numeric value of that named value or -1 if value is not found or is not numeric
   */
  public int getIntNamedValue(String theName, int defaultValue);
  
  /**
   * Given the passed name, return the boolean value from the message body that
   * matches that named.
   * 
   * If there is more than one value for this name, only the first
   * set/installed value is returned.
   *
   * If there is no instance of a value with this name, false is returned.
   *
   * The following values are considered true: 'true', 'yes', 't', 'y' or '1'
   * The check is case-insensitive
   *
   * If there value for the name is empty, false is returned
   *
   * @param theName name of the value to fetch
   * @return value of the first matching named value or false if not found
   */
  public boolean getBoolNamedValue(String theName);
  
  /**
   * Given the passed name, return the boolean value from the message body that
   * matches that named.
   * 
   * If there is more than one value for this name, only the first
   * set/installed value is returned.
   *
   * If there is no instance of a value with this name, defaultValue is returned.
   *
   * The following values are considered true: 'true', 'yes', 't', 'y' or '1'
   * The check is case-insensitive
   *
   * If there value for the name is empty, false is returned
   *
   * @param theName name of the value to fetch
   * @param defaultValue value to return if named value does not exist
   * @return value of the first matching named value or false if not found
   */
  public boolean getBoolNamedValue(String theName, boolean defaultValue);
  
  /** 
   * Release the resources for the message.  The message will be invalid
   * after this
   */
  public void releaseResources();
}
