/*
 * xPL_MutableMessageI.java
 * Copyright 2004, Gerald R Duprey Jr
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy of
 * the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0 
 *
 * Unless required by applicable law or agreed to in writing, software distributed
 * under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
 * CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License. 
 */

package org.cdp1802.xpl;

/**
 * A message whose parameters can be changed
 *
 * @author  gerry
 */
public interface xPL_MutableMessageI extends xPL_MessageI {
  /**
   * Change the type of the message
   * @param newMessageType new message type to install for this message (must not be UNKNOWN)
   */
  public void setType(MessageType newMessageType);

  /**
   * Change the type of the message
   * @param typeName  name of message type to install (xpl-trig, xpl-stat or xpl-cmnd)
   */
  public void setType(String typeName);
  
  /**
   * Set the Schema Class of the message
   * @param newSchemaClass new schema class to install for this message (must not be null)
   */
  public void setSchemaClass(String newSchemaClass);

  /**
   * Set the Schema Type of the message
   * @param newSchemaType New schema type for this message (must not be null)
   */
  public void setSchemaType(String newSchemaType);
  
  /**
   * Set the Schema Class and Type of message at once
   * @param newSchemaClass new schema class to install for this message (must not be null)
   * @param newSchemaType new schema type to install for this message (must not be null)
   */
  public void setSchema(String newSchemaClass, String newSchemaType);
  
  /**
   * Set the schema class and type of message at once from a string in
   * the format of class.type.
   *
   * @param newSchemaInfo class and type in a single string, delimited by a period
   */
  public void setSchema(String newSchemaInfo);
  
  /**
   * Install the source identifier for the message
   *
   * @param newSourceIdentifier new source identifier to install (must not be null)
   */
  public void setSource(xPL_IdentifierI newSourceIdentifier);

  /**
   * Install the source identifier for the message
   *
   * @param newSourceIdentifier new source identifier to install (must not be null)
   */
  public void setSource(String newSourceIdentifier);
  
  /**
   * Install the target identifier for the message
   *
   * @param newTargetIdentifier new target identifier to install (must not be null)
   */
  public void setTarget(xPL_IdentifierI newTargetIdentifier);

  /**
   * Install the target identifier for the message
   *
   * @param newTargetIdentifier new target identifier to install (must not be null)
   */
  public void setTarget(String newTargetIdentifier);

  /**
   * 
   * Install a new set of named values.  The passed NamedValuesI is
   * installed (not copied) so if the passed list is mutable, further
   * changes to the passed list will appear in the message.
   * 
   * Any existing NamedValues for the message are lost
   * 
   * If you install a non-mutable namedvalue list, further attempts
   * to modify it (clearNamedValues/removeNamedValues/setNamedValue)
   * will silently fail/be ignored.
   * @param newNamedValues New set of namedvalues to install/replace this message body
   */
  public void setMessageBody(NamedValuesI newNamedValues);
  
  /**
   * Returns the message body pre-cast into a mutable set of name/values
   * for the message.  This is really just a simpler version of using
   * getMessageBody(), then insuring the result was an instanceof MutableNameValuesI
   * and then casting it.
   *
   * If the underlying NameValues is not mutable (which is only possible if a
   * previous call to setMessageBody installed a non-mutable NamedValueI), then
   * this will return null.
   *
   * @return the mutable name/values for this message body or null if body is not mutable
   */
  public MutableNamedValuesI getMutableMessageBody();
  
  
  /**
   * Clear all named values in the message body.
   *
   * NOTE: If you have installed a non-mutable NameValues based message body with
   * a previous setMessageBody(), then this method will silently fail/be ignored
   *
   */
  public void clearMessageBody();

  /**
   * Add the passed named/value to the message body.  If you add a value using a name which already
   * exists, the original named value is left alone and this new one is added to
   * the end.
   * 
   * If you attempt to set a value of null, the value is converted into an empty
   * string ("").  An attempt to add a null or empty name is ignored/rejected.
   * 
   * NOTE: If you have installed a non-mutable NameValues based message body with
   * a previous setMessageBody(), then this method will silently fail/be ignored
   * @param theName name of value to add
   * @param theValue value to add with this name or null if message body is immutable
   * @return Instance of the named value created to represent this
   */
  public MutableNamedValueI addNamedValue(String theName, String theValue);

  /**
   * Add the passed named/value to the message body .  If you add a value using a name which already
   * exists, the original named value is left alone and this new one is added to
   * the end.
   * 
   * An attempt to add a null or empty name is ignored/rejected.
   * 
   * NOTE: If you have installed a non-mutable NameValues based message body with
   * a previous setMessageBody(), then this method will silently fail/be ignored
   * @param theName name of value to add
   * @param theValue value to add with this name or null if message body is immutable
   * @return Instance of the named value created to represent this
   */
  public MutableNamedValueI addNamedValue(String theName, int theValue);

  /**
   * Add the passed named/value to the message body.  If you add a value using a name which already
   * exists, the original named value is left alone and this new one is added to
   * the end.
   * 
   * An attempt to add a null or empty name is ignored/rejected.
   * 
   * NOTE: If you have installed a non-mutable NameValues based message body with
   * a previous setMessageBody(), then this method will silently fail/be ignored
   * @param theName name of value to add
   * @param theValue value to add with this name or null if message body is immutable
   * @return Instance of the named value created to represent this
   */
  public MutableNamedValueI addNamedValue(String theName, boolean theValue);
  
  /**
   * Set a named value in the message.  If there is an existing value OR values for the
   * name, they are purged and the single name/value installed.  THis is
   * a convience method and does the equivilent of a removeNamedValues()
   * followed by an addNamedValue() call.
   *
   * An attempt to add a null or empty name is ignored/rejected.
   *
   * @param theName name of value to add
   * @param theValue value to add with this name
   */
  public MutableNamedValueI setNamedValue(String theName, String theValue);
  
  /**
   * Set a named value in the message.  If there is an existing value OR values for the
   * name, they are purged and the single name/value installed.  THis is
   * a convience method and does the equivilent of a removeNamedValues()
   * followed by an addNamedValue() call.
   *
   * An attempt to add a null or empty name is ignored/rejected.
   *
   * @param theName name of value to add
   * @param theValue value to add with this name
   */
  public MutableNamedValueI setNamedValue(String theName, int theValue);

  /**
   * Set a named value in the message.  If there is an existing value OR values for the
   * name, they are purged and the single name/value installed.  THis is
   * a convience method and does the equivilent of a removeNamedValues()
   * followed by an addNamedValue() call.
   *
   * An attempt to add a null or empty name is ignored/rejected.
   *
   * @param theName name of value to add
   * @param theValue value to add with this name
   */
  public MutableNamedValueI setNamedValue(String theName, boolean theValue);

  /**
   * Remove all occurances of the values with the passed name from the message body.  The number of
   * values actually removed is returned.
   *
   * NOTE: If you have installed a non-mutable NameValues based message body with
   * a previous setMessageBody(), then this method will silently fail/be ignored
   *
   * @param theName name of the values to remove
   * @return the number of values actually removed (0 if message body is immutable)
   */
  public int removeNamedValues(String theName);
}