/*
 * xPL_Util.java
 * Copyright 2004, Gerald R Duprey Jr
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy of
 * the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0 
 *
 * Unless required by applicable law or agreed to in writing, software distributed
 * under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
 * CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License. 
 */

package org.cdp1802.xpl;

import org.cdp1802.xpl.device.xPL_DeviceI;

/**
 * Various accessors/creators and utilities for xPL
 * @author gerry
 */
public class xPL_Util {
  private static xPL_Manager xplManager = xPL_Manager.getManager();
  private static IdentifierManager identifierManager = xplManager.getIdentifierManager();
  private static MessageFormatter messageFormatter = new MessageFormatter();
  private static MessageParser messageParser = new MessageParser();
  
  
  /**
   * Parse the passed textual xPL message into an xPL_Message instance.
   *
   * This is not normally needed as xPLLib will parse the message as needed
   *
   * @param textMessage text format of message to parse
   * @param receivedFrom media handler message was received from (can be null)
   * @return xPL_MessageI instance that represents the text message
   * @throws xPL_ParseException thrown if the message is malformed
   */
  public static xPL_MessageI parseMessage(String textMessage, xPL_MediaHandlerI receivedFrom) throws xPL_ParseException {
    return messageParser.parseMessage(textMessage, receivedFrom);
  }
  
  /**
   * Parse the passed textual xPL message into an xPL_Message instance.
   *
   * This is not normally needed as xPLLib will parse the message as needed
   *
   * @param textMessage text format of message to parse
   * @return xPL_MessageI instance that represents the text message
   * @throws xPL_ParseException thrown if the message is malformed
   */
  public static xPL_MessageI parseMessage(String textMessage) throws xPL_ParseException {
    return messageParser.parseMessage(textMessage, null);
  }

  /**
   * Format the passed xPL_Message instance into it's textual equivilent.
   * If the message is not properly setup, null is returned
   *
   * @param theMessage valid xPL_MessageI instance to format into text
   * @param forMediaHandler mediaHandler this message is destined to be written to (can be null)
   * @param bumpHopCount if true, the hop count is written as hop+1 to the text file
   * @return the textual equivilent of that message or null if there was an error
   */
  public static String formatMessage(xPL_MessageI theMessage, xPL_MediaHandlerI forMediaHandler, boolean bumpHopCount) {
    return messageFormatter.formatMessage(theMessage, forMediaHandler, bumpHopCount);
  }
 
  /**
   * Format the passed xPL_Message instance into it's textual equivilent.
   * If the message is not properly setup, null is returned
   *
   * @param theMessage valid xPL_MessageI instance to format into text
   * @return the textual equivilent of that message or null if there was an error
   */
  public static String formatMessage(xPL_MessageI theMessage) {
    return messageFormatter.formatMessage(theMessage, null, false);
  }
  
  /** Create a new set of mutable named values
   *
   * @return new isntance of named/values
   */
  public static MutableNamedValuesI createMutableNamedValues() {
    return new MutableNamedValues();
  }
  
  /**
   * Create a basic, mutable message
   *
   * @return a new empty message
   */
  public static xPL_MutableMessageI createMessage() {
    return new xPL_MutableMessage();
  }

  /**
   * Create a new message from the passed device targeted for the destination device 
   * with a specific message type and schema class/type
   *
   * @param from xPL idenifier to serve as the source of this message
   * @param to xPL identifier to serve as the target of this message
   * @param messageType type of message to create (status, trigger, command)
   * @param schemaClass schema class to install in this message
   * @param schemaType schema type to install in this message
   * @return a new message initialized with the passsed source/target/type/schema of null if bad parameters
   */
  public static xPL_MutableMessageI createMessage(xPL_IdentifierI from, xPL_IdentifierI to, xPL_MessageI.MessageType messageType, String schemaClass, String schemaType) {
    // Validate parameters
    if ((from == null) || from.isBroadcastIdentifier() || from.isGroupIdentifier()) return null;
    if (to == null) return null;
    if (messageType == xPL_MessageI.MessageType.UNKNOWN) return null;
    if ((schemaClass == null) || (schemaClass.length() == 0) || (schemaType == null) || (schemaType.length() == 0)) return null;
    
    xPL_MutableMessageI theMessage = createMessage();
    theMessage.setSource(from);
    theMessage.setTarget(to);
    theMessage.setType(messageType);
    theMessage.setSchema(schemaClass, schemaType);
    return theMessage;
  }
  
  /**
   * Create a new group message from the passed device targeted for the destination group
   * with a specific message type and schema class/type
   *
   * @param from xPL idenifier to serve as the source of this message
   * @param toGroup xPL group identifier to serve as the target of this message
   * @param messageType type of message to create (status, trigger, command)
   * @param schemaClass schema class to install in this message
   * @param schemaType schema type to install in this message
   * @return a new message initialized with the passsed source/target/type/schema
   */
  public static xPL_MutableMessageI createGroupMessage(xPL_IdentifierI from, xPL_IdentifierI toGroup, xPL_MessageI.MessageType messageType, String schemaClass, String schemaType) {
    // Validate parameters
    if ((from == null) || from.isBroadcastIdentifier() || from.isGroupIdentifier()) return null;
    if ((toGroup == null) || toGroup.isBroadcastIdentifier() || !toGroup.isGroupIdentifier()) return null;
    if (messageType == xPL_MessageI.MessageType.UNKNOWN) return null;
    if ((schemaClass == null) || (schemaClass.length() == 0) || (schemaType == null) || (schemaType.length() == 0)) return null;

    xPL_MutableMessageI theMessage = createMessage();
    theMessage.setSource(from);
    theMessage.setTarget(toGroup);
    theMessage.setType(messageType);
    theMessage.setSchema(schemaClass, schemaType);
    return theMessage;
  }

  /**
   * Create a new group message from the passed device targeted for the destination group
   * with a specific message type and schema class/type
   *
   * @param from xPL idenifier to serve as the source of this message
   * @param toGroupName group name to direct message toward
   * @param messageType type of message to create (status, trigger, command)
   * @param schemaClass schema class to install in this message
   * @param schemaType schema type to install in this message
   * @return a new message initialized with the passsed source/target/type/schema
   */
  public static xPL_MutableMessageI createGroupMessage(xPL_IdentifierI from, String toGroupName, xPL_MessageI.MessageType messageType, String schemaClass, String schemaType) {
    // Get group ref
    xPL_IdentifierI toGroup = identifierManager.getGroupIdentifier(toGroupName);
    if (toGroup == null) return null;
    return createGroupMessage(from, toGroup, messageType, schemaClass, schemaType);
  }

  /** 
   * Create a new broadcast message from the passed device with a specific 
   * message type and schema class/type
   *
   * @param from xPL device identifier to serve as source of this message
   * @param messageType type of message to create (status, trigger, command)
   * @param schemaClass schema class to install in this message
   * @param schemaType schema type to install in this message
   * @return a new message set to be a broadcast message from the passed source/type/schema
   */
  public static xPL_MutableMessageI createBroadcastMessage(xPL_IdentifierI from, xPL_MessageI.MessageType messageType, String schemaClass, String schemaType) {
    // Validate parameters
    if ((from == null) || from.isBroadcastIdentifier() || from.isGroupIdentifier()) return null;
    if (messageType == xPL_MessageI.MessageType.UNKNOWN) return null;
    if ((schemaClass == null) || (schemaClass.length() == 0) || (schemaType == null) || (schemaType.length() == 0)) return null;

    xPL_MutableMessageI theMessage = createMessage();
    theMessage.setSource(from);
    theMessage.setTarget(identifierManager.getBroadcastIdentifier());
    theMessage.setType(messageType);
    theMessage.setSchema(schemaClass, schemaType);
    return theMessage;
  }
  
}
